/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose
 * such Confidential Information and shall use it only in accordance
 * with the terms of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY
 * DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */
package examples.acrs;

import com.cafesoft.cams.Config;
import com.cafesoft.cams.ConfigException;

import com.cafesoft.cams.access.AbstractAccessControlRule;
import com.cafesoft.cams.access.AccessControlRequest;
import com.cafesoft.cams.access.AccessControlResponse;
import com.cafesoft.cams.access.EvaluationException;
import com.cafesoft.cams.access.InternalAccessControlRequest;
import com.cafesoft.cams.access.InternalAccessControlResponse;

import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.access.SyntaxException;

import com.cafesoft.core.jdbc.JdbcUtils;

import com.cafesoft.core.log.Logger;

import javax.security.auth.Subject;

import java.util.Iterator;
import java.util.Properties;
import java.util.Set;

/**
 * <code>UserAttributeAcr</code> demonstrates creation of a custom Cams
 * AccessControlRule that only grants access to a resource to users with
 * specified user attributes. The user attribute is fetched from a Cams
 * ManagedSession.
 *
 * @version $Revision: 1.9 $ $Date: 2009/08/14 23:52:22 $
 * @author Gary Gwin
 *
 * @see com.cafesoft.cams.access.AccessControlRule
 * @see com.cafesoft.cams.access.AbstractAccessControlRule
 * @see com.cafesoft.cams.access.EvaluationException
 * @see com.cafesoft.cams.access.SyntaxException
 */
public class UserAttributeAcr extends AbstractAccessControlRule
{
	/** The configured account type */
	private String resourceAttribute;

	/** The configured session attribute namespace */
	private String nameSpace;

	/** The configured session attribute name */
	private String attributeName;

	/**
	 * Create a new UserAttributeAcr.
	 */
	public UserAttributeAcr()
	{
		super();
		this.resourceAttribute = null;
		this.nameSpace = null;
		this.attributeName = null;
	}

	/**
	 * Create a new UserAttributeAcr with the specified id.
	 *
	 * @param id the AccessControlRule id.
	 */
	public UserAttributeAcr(String id)
	{
		this();
		setId(id);
	}

	//
	// extending AbstractAccessControlRule
	//

	/**
	 * Initialize the AccessControlRule.
	 *
	 * @param config a Config object that provides access to configuration
	 * 	parameters and a runtime Context.
	 */
	public void initialize(Config config) throws ConfigException
	{
		super.initialize(config);

		// Get the supplied logger for the security domain.
		logger = config.getContext().getLogger();
	}

	/**
	 * Set the resource attribute.
	 *
	 * @param resourceAttribute the resource attribute.
	 * @exception SyntaxException if the resource attribute is null or empty.
	 */
	public void setResourceAttribute(String resourceAttribute) throws SyntaxException
	{
		if (resourceAttribute == null || "".equals(resourceAttribute))
			throw new SyntaxException(
				"Resource attribute cannot be null or empty");

		this.resourceAttribute = resourceAttribute;
	}

	/**
	 * Get the resource attribute.
	 *
	 * @return the resource attribute.
	 */
	public String getResourceAttribute()
	{
		return resourceAttribute;
	}

	/**
	 * Set the session attribute namespace.
	 *
	 * @param nameSpace the session attribute namespace to use.
	 * @exception SyntaxException if the namespace is null or empty
	 */
	public void setNameSpace(String nameSpace) throws SyntaxException
	{
		if ((nameSpace == null) || ("".equals(nameSpace)))
			throw new SyntaxException(
				"The session attribute namespace must be specified.");

		this.nameSpace = nameSpace;
	}

	/**
	 * Get the session attribute namespace.
	 *
	 * @return the session attribute namespace.
	 */
	public String getNameSpace()
	{
		return nameSpace;
	}

	/**
	 * Set the session attribute name.
	 *
	 * @param attributeName the session attribute name to use.
	 * @exception SyntaxException if the attribute name is null or empty
	 */
	public void setAttributeName(String attributeName) throws SyntaxException
	{
		if ((attributeName == null) || ("".equals(attributeName)))
			throw new SyntaxException(
				"The session attribute namespace must be specified.");

		this.attributeName = attributeName;
	}

	/**
	 * Get the session attribute name.
	 *
	 * @return the session attribute name.
	 */
	public String getAttributeName()
	{
		return attributeName;
	}

	/**
	 * Evaluate the access control rule.
	 *
	 * @param accessControlRequest the access request.
	 * @param accessControlResponse the access response.
	 * @return true in all cases.
	 * @exception EvaluationException if an error when attempting to
	 *	evaluate the rule.
	 * @see AccessControlRequest
	 * @see AccessControlResponse
	 */
	public boolean evaluate(InternalAccessControlRequest accessControlRequest,
							InternalAccessControlResponse accessControlResponse) 
		throws EvaluationException
	{
		if (accessControlRequest == null)
			throw new EvaluationException("null AccessControlRequest");

		if (accessControlResponse == null)
			throw new EvaluationException("null AccessControlResponse");

		if (resourceAttribute == null)
			throw new EvaluationException(
				"the resource attribute has not been set");

		// Get the attribute for comparison
		String userAttribute =
			accessControlRequest.getSession().getAttribute(
			nameSpace, attributeName).toString();

		// Is the account type valid for this resource
		if (userAttribute == null || !userAttribute.equals(resourceAttribute))
		{
			// User account type not valid for the resource
			accessControlResponse.setReason(
				AccessControlResponse.RC_ACCESS_DENIED_CONDITIONALLY);
			accessControlResponse.setMessage(
				"attribute '" + userAttribute + "' not valid for this resource");
			return false;
		}

		return true;
	}

}  // End of class: UserAttributeAcr
