/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose
 * such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package examples.acrs;

import examples.acrs.UserAttributeAcr;

import com.cafesoft.cams.access.AbstractAccessControlRulePersistenceManager;
import com.cafesoft.cams.access.AccessControlRule;
import com.cafesoft.cams.access.SyntaxException;

import com.cafesoft.cams.Config;
import com.cafesoft.cams.ConfigException;
import com.cafesoft.cams.PersistenceException;

import com.cafesoft.core.xml.dom.XmlDomUtils;

import com.cafesoft.core.log.Logger;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import java.io.PrintWriter;
import java.io.IOException;

/**
 * <code>XmlUserAttributeAcrPm</code> implements methods to create, load, store,
 * and remove instances of <code>UserAttributeAcr</code>.
 * <P>
 * XML representation of UserAttributeAcr in "access-control-policy.xml":
 * <PRE>
 * <example:user-attribute-acr
 *         xmlns:example="http://cafesoft.com/example-user-attribute-acr_1_0.dtd"
 *         id="account type gold" desc="Restrict access to 'gold' accounts">
 *         <example:user-attribute
 *            name="gold"
 *            session-attr-namespace="userinfo"
 *            session-attr-name="account_type"/>
 * </example:user-attribute-acr>
 * </PRE>
 *
 * @version $Revision: 1.6 $ $Date: 2009/08/14 23:52:22 $
 * @author Gary Gwin
 */
public class XmlUserAttributeAcrPm
	extends AbstractAccessControlRulePersistenceManager
{
	/**
	 * Create a new <code>XmlUserAttributeAcrPm</code> instance.
	 */
	public XmlUserAttributeAcrPm()
	{
	}

	//
	// extending AbstractAccessControlRulePersistenceManager
	//

	/**
	 * Create a new/empty UserAttributeAcr.
	 *
	 * @return a new/empty UserAttributeAcr instance.
	 */
	public AccessControlRule acrCreate()
	{
		return new UserAttributeAcr();
	}

	/**
	 * Load a UserAttributeAcr from persistent storage.
	 *
	 * @param object the XML DOM element associated with the
	 *	UserAttributeAcr to be created.
	 *
	 * @return the UserAttributeAcr instance.
	 * @exception PersistenceException if an error creating the new
	 *	AccessControlRule.
	 * @exception IllegalArgumentException if the parameter is null or
	 *	it's type is not supported by the factory implementation.
	 */
	public AccessControlRule acrLoad(Object object)
		throws PersistenceException
	{
		if (object == null)
			throw new IllegalArgumentException(
				"null XML DOM Element");

		if (! (object instanceof Element))
			throw new IllegalArgumentException(
				"parameter is not an XML DOM Element");

		Element element = (Element)object;

		// The Element must have local name "user-attribute-acr"
		if (!element.getLocalName().equalsIgnoreCase("user-attribute-acr"))
			throw new PersistenceException(
				"unexpected top-level tag name: " +
				element.getLocalName() +
				", expected: user-attribute-acr");

		// Create the AccessControlRule name.
		UserAttributeAcr acr = new UserAttributeAcr(
			element.getAttribute("id"));

		// Traverse all Child Nodes to populate the rule
		NodeList nodeList = element.getChildNodes();
		for (int i = 0; i < nodeList.getLength(); i++)
		{
			Node node = nodeList.item(i);

			// Branch to handle different Child node types.
			switch (node.getNodeType())
			{
				case Node.ELEMENT_NODE:
					Element child = (Element)node;

					// Make sure the child Element is in the
					// same namespace as the business-hours-acr
					if (XmlDomUtils.nameSpaceURIMatches(element, child))
					{
						// Set the attribute name
						try
						{
							acr.setResourceAttribute(child.getAttribute("name"));
						}
						catch (SyntaxException e)
						{
							logger.error(this, "invalid name", e);
							throw new PersistenceException(e);
						}

						// Set the session attribute namespace
						try
						{
							acr.setNameSpace(child.getAttribute("session-attr-namespace"));
						}
						catch (SyntaxException e)
						{
							logger.error(this, "invalid namespace", e);
							throw new PersistenceException(e);
						}

						// Set the session attribute name
						try
						{
							acr.setAttributeName(child.getAttribute("session-attr-name"));
						}
						catch (SyntaxException e)
						{
							logger.error(this, "invalid attribute name", e);
							throw new PersistenceException(e);
						}
					}
					break;

				case Node.TEXT_NODE:
					// Ignore white space text
					break;

				default:
					// Ignore all other node types.
			}
		}

		// Done populating the AccessControlRule
		return acr;
	}

	/**
	 * Store an AccessControlRule to persistent storage.
	 *
	 * @param acr the AccessControlRule to be stored.
	 * @param dest an object that may be needed for some AccessControlRule
	 *	storage destination types. For example, if the AccessControlRule
	 *	is a file, then a PrintStream might be provided. Some storage
	 *	destinations (e.g. a relational database), may not use this
	 *	parameter.
	 *
	 * @exception PersistenceException if an error storing the
	 *	AccessControlRule.
	 * @exception IllegalArgumentException if a required parameter is null,
	 *	or it's type is not supported by the implementation.
	 */
	public void acrStore(AccessControlRule acr, Object dest)
		throws PersistenceException
	{
		// Make sure the AccessControl rule was specified and is the
		// expected type.
		if (acr == null)
			throw new PersistenceException("null access control rule");

		if (!(acr instanceof UserAttributeAcr))
			throw new PersistenceException("unsupported AccessControlRule type: " +
				acr.getClass().getName() + ", expected: " +
				UserAttributeAcr.class.getName());

		// Since storage is XML, the destination Object is expected
		// to be a PrintWriter.
		if (dest == null)
			throw new PersistenceException("null destination. Expected PrintWriter");

		if (!(dest instanceof PrintWriter))
			throw new PersistenceException("unsupported destination type: " +
				dest.getClass().getName() + ", expected PrintWriter");

		UserAttributeAcr rule = (UserAttributeAcr)acr;
		PrintWriter out = (PrintWriter)dest;

		// Write the BusinessHoursAcr to the PrintWriter as XML
		out.print("\t\t<example:user-attribute-acr ");
		out.print("xmlns=\"xmlns:example=");
		out.print("\"http://cafesoft.com/example-user-attribute-acr_1_0.dtd\" ");
		out.println("id=\"" + rule.getId() + "\">");

		out.print("\t\t\t<example:set-user-attribute ");
		out.print("name=\"" + rule.getResourceAttribute() + "\" ");
		out.print("session-attr-namespace=\"" + rule.getNameSpace() + "\" ");
		out.print("session-attr-name=\"" + rule.getAttributeName() + "\" ");

		out.println("\t\t</example:user-attribute-acr>");
	}

	/**
	 * Remove an AccessControlRule from persistent storage.
	 *
	 * @param acr the AccessControlRule to be removed from persistent
	 *	storage.
	 *
	 * @return the AccessControlRule instance.
	 * @exception PersistenceException if an error creating the new
	 *	AccessControlRule.
	 * @exception IllegalArgumentException if the parameter is null or
	 *	it's type is not supported by the factory implementation.
	 */
	public void acrRemove(AccessControlRule acr)
		throws PersistenceException
	{
		// Do nothing. The AccessControlRule instance is simply not
		// written to the XML file. (Note: Whatever class is editing
		// the AccessControlPolicy containing the referenced rule,
		// will simply remove it from it's AccessControlRuleLibrary.
		// When that is persisted, this rule will not be written).
	}

}  // End of class: XmlUserAttributeAcrPm
