/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package examples.client;

import com.cafesoft.cams.Config;
import com.cafesoft.cams.ConfigException;
import com.cafesoft.cams.StandardConfig;

import com.cafesoft.cams.access.AccessControlException;
import com.cafesoft.cams.access.AccessControlRequest;
import com.cafesoft.cams.access.AccessControlResponse;
import com.cafesoft.cams.access.AccessControlService;
import com.cafesoft.cams.access.InvalidResourceException;
import com.cafesoft.cams.access.ResourceRequest;
import com.cafesoft.cams.access.ResourceRequestFactory;
import com.cafesoft.cams.access.ResourceType;

import com.cafesoft.cams.access.url.UrlResourceRequestFactory;

import com.cafesoft.cams.auth.AuthException;
import com.cafesoft.cams.auth.AuthRequest;
import com.cafesoft.cams.auth.AuthResponse;
import com.cafesoft.cams.auth.AuthenticationService;

import com.cafesoft.cams.client.CamsClient;
import com.cafesoft.cams.client.CamsClientFactory;
import com.cafesoft.cams.client.CamsTransportException;
import com.cafesoft.cams.client.ConnectionException;
import com.cafesoft.cams.client.ConnectionExceptionListener;

import com.cafesoft.cams.oblig.Obligation;

import com.cafesoft.cams.ping.PingException;
import com.cafesoft.cams.ping.PingRequest;
import com.cafesoft.cams.ping.PingResponse;
import com.cafesoft.cams.ping.PingService;

import com.cafesoft.cams.session.RemoteSession;
import com.cafesoft.cams.session.SessionId;

import com.cafesoft.cams.session.access.SessionAccessException;
import com.cafesoft.cams.session.access.SessionAccessRequest;
import com.cafesoft.cams.session.access.SessionAccessResponse;
import com.cafesoft.cams.session.access.SessionAccessService;

import com.cafesoft.cams.session.control.SessionControlAction;
import com.cafesoft.cams.session.control.SessionControlException;
import com.cafesoft.cams.session.control.SessionControlRequest;
import com.cafesoft.cams.session.control.SessionControlResponse;
import com.cafesoft.cams.session.control.SessionControlService;

import com.cafesoft.security.common.access.StandardResourceType;

import com.cafesoft.security.common.client.StandardCamsClientFactory;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerException;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.service.ServiceException;
import com.cafesoft.core.service.ServiceFinder;

import com.cafesoft.core.util.ClassInstantiationException;
import com.cafesoft.core.util.ClassUtils;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.Date;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

/**
 * <code>CamsClientExample</code> provides sample code using the CamsClient
 * API. This API is used to implement existing Cams web agents and may be
 * used by licensed customers to create custom agents.
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.7 $ $Date: 2009/08/14 23:52:22 $ $Author: norb $
 * @since 2/4/2008
 */
public class CamsClientExample implements ConnectionExceptionListener
{
	//
	// Static constants
	//

	/**
	 * The ResourceType name for sample access control checks on resources.
	 */
	private static String RESOURCE_TYPE_NAME = "http";

	/**
	 * The ResourceType description for sample resources.
	 */
	private static String RESOURCE_TYPE_DESC = "HTTP Resource Type";

	/**
	 * The ResourceType actions for sample resources.
	 * <P>
	 * NOTE: THE ORDER OF THESE ELEMENTS MUST MATCH THE ORDERING OF THE
	 * THE SAME ELEMENTS IN THE cams.conf FILE DEPLOYED ON CAMS POLICY 
	 * SERVER(s). FOR HISTORIC REASONS, EACH RESOURCE ACTION WITHIN A
	 * GIVEN TYPE IS ASSIGNED A UNIQUE BIT VALUE BASED ON ITS POSITION IN 
	 * THIS LIST, WHICH AIDS WITH EFFICIENT ACTION MATCHING.
	 */
	private static String[] RESOURCE_TYPE_ACTIONS =
	{ "GET","POST","PUT","DELETE","HEAD","OPTIONS","TRACE","PROPFIND",
		"PROPPATCH","MKCOL","COPY","MOVE","LOCK","UNLOCK","DEBUG"
	};

	/**
	 * The "http" ResourceType object used when constructing a "resource
	 * request".
	 * <p>
	 * NOTE: This variable is static because it represents a ResourceType
	 * supported by this example agent. There is no need to have more than
	 * one instance. Existing Cams web agents support this ResourceType and
	 * no other types.
	 */
	private static ResourceType httpResourceType = new StandardResourceType(
		RESOURCE_TYPE_NAME, RESOURCE_TYPE_DESC, RESOURCE_TYPE_ACTIONS);

	/**
	 * Default configuration file path.
	 */
	private static final String DEFAULT_CONFIG_FILE_PATH =
		"./cams-client-example.conf";

	//
	// Instance Variables
	//

	/**
	 * The CamsClient used by this example.
	 */
	private CamsClient camsClient = null;

	/**
	 * The ResourceRequestFactory.
	 */
	private ResourceRequestFactory resReqFactory = null;

	/**
	 * The Logger used by DEBUG, INFO, WARN, ERROR, and FATAL messages.
	 */
	private Logger logger;

	/**
	 * A flag used to enable/disable DEBUG-level information for this example.
	 */
	private boolean debug = false;

	/**
	 * A timestamp used to keep track of the last report access control
	 * policy modification time, which is return with all access control
	 * responses.
	 */
	private long lastModified = 0L;

	//
	// Constructors
	//

	/**
	 * Create a new CamsClientExample instance.
	 *
	 * @param camsClient the CamsClient to be used by this example.
	 * @param resReqFactory the ResourceRequestFactory.
	 * @param logger the Logger.
	 * @param debug a flag used to enable/disable DEBUG-level log messages.
	 */
	public CamsClientExample(CamsClient camsClient, Logger logger,
		ResourceRequestFactory resReqFactory, boolean debug)
	{
		this.camsClient = camsClient;
		this.resReqFactory = resReqFactory;
		this.logger = logger;
		this.debug = debug;
	}

    //
	// Main Program
	//

	/**
	 * Main program. Creates a new CamsClientExample instance and runs it.
	 *
	 * @param args command line arguments.
	 */
	public static void main(String[] args)
	{
		String configFilePath = DEFAULT_CONFIG_FILE_PATH;
		CamsClientExample example = null;

		// Handle command line arguments
		int i = 0;
		while (i < args.length)
		{
			if ("-f".equals(args[i]))
			{
				// Get the config file path.
				if ((i + 1) < args.length)
				{
					i++;
					configFilePath = args[i];
				}
				else
				{
					System.err.println("Command line parameter: -f " +
						"is not followed by the configuration file path");
					printUsage();
					System.exit(1);
				}
			}
			else if ("-help".equals(args[i]))
			{
				printUsage();
				System.exit(0);
			}
			else
			{
				System.err.println("Unrecognized command line parameter: " + 
					args[i]);
				printUsage();
				System.exit(1);
			}

			i++;
		}

		// Make sure the config file exists.
		File configFile = new File(configFilePath);
		if (configFile.exists() == false)
		{
			System.err.println("Configuation file: " + configFilePath +
				" does not exist.");
			printUsage();
			System.exit(1);
		}

		try
		{
			// Create the Config object.
			Config config = createConfig(configFile);

			// Create/initialize the CamsClient.
			CamsClient camsClient = createCamsClient(config);

			// Set the global debug flag.
			String debugString = config.getInitParameter("cams.debug", "false");
			boolean debugFlag = "true".equals(debugString);
			com.cafesoft.security.common.Constants.DEBUG = debugFlag;

			// Create/initialize a factory for ResourceRequest instances.
			ResourceRequestFactory resReqFactory = 
				new UrlResourceRequestFactory();
			Properties p = new Properties();
			p.setProperty("defaultPort.http", "80");
			p.setProperty("defaultPort.https", "443");
			resReqFactory.initialize(p);

			// Create a new CamsClientExample instance.
			example = new CamsClientExample(camsClient, 
				config.getContext().getLogger(),
				resReqFactory,
				debugFlag);

			// Connect to configured Cams Policy Server(s).
			boolean connected = example.connect();
			if (connected)
			{
				// Test the "ping" service
				example.testPingService();

				// Test the "authentication" service.
				SessionId sessionId = example.testAuthenticationService();

				// If a session exists: test the "session access" service.
				if (sessionId != null)
					example.testSessionAccessService(sessionId);

				// Test the "access control" service.
				example.testAccessControlService(sessionId);

				// If a session exists: test the "session control" (logout)
				// service.
				if (sessionId != null)
					example.testSessionControlService(sessionId);

				// Disconnect the CamsClient.
				example.disconnect();
			}
		}
		catch (ConfigException e)
		{
			e.printStackTrace();

			// Exit with an error status.
			System.exit(1);
		}
		catch (Exception e)
		{
			e.printStackTrace();

			// Exit with an error status.
			System.exit(1);
		}
		finally
		{
			// Destroy the CamsClientExample.
			if (example != null)
				example.destroy();
		}

		// Exit with a success status.
		System.exit(0);
	}

	/**
	 * Create a Config object from a File.
	 *
	 * @param configFile the configuration file.
	 * @return an initialized CamsClient instance.
	 * @exception ConfigException if a CamsClient configuration error.
	 */
	private static Config createConfig(File configFile)
		throws ConfigException
	{
		// Load configuration properties from a file.
		Properties properties = loadPropertiesFromFile(configFile);

		// Create the Logger.
		Logger logger = createLogger(properties);

		logger.info(CamsClientExample.class, "Creating Config object");
		Config config = new StandardConfig(properties, "", null, logger);

		return config;
	}

	/**
	 * Create and initialize a CamsClient instance.
	 *
	 * @param config the Config object.
	 * @return an initialized CamsClient instance.
	 * @exception ConfigException if a CamsClient configuration error.
	 */
	private static CamsClient createCamsClient(Config config)
		throws ConfigException
	{
		CamsClient camsClient = null;

		CamsClientFactory clientFactory = new StandardCamsClientFactory();
		camsClient = clientFactory.create(config);
		camsClient.initialize(config);

		return camsClient;
	}

	/**
	 * Print command line usage for this example to System.out.
	 */
	private static void printUsage()
	{
		System.out.println("usage: java examples.client.CamsClientExample " +
			"[-f configFilePath] [-help]");
	}

	//
	// Implements ConnectionExceptionListener interface.
	//

	/**
	 * Called when a Connection exception occurrs.
	 *
	 * @param e the exception that occurred
	 */
	public void handleException(ConnectionException e)
	{
		logger.error(this, "Connection error occurred.", e.getCause());
	}

	//
	// Public Methods
	//

	/**
	 * Connect to the configured Cams Policy Server(s).
	 *
	 * @return true if the CamsClient successfully connects, else false.
	 */
	public boolean connect()
	{
		// Tell the CamsClient to connect to configured policy server(s).
		logger.info(this, "CamsClient connecting to configured Cams Policy Server(s)");
		camsClient.connect();

		// Is the CamsClient to connected?
		boolean connected = camsClient.isConnected();
		logger.info(this, "CamsClient connected = " + connected);

		return connected;
	}

	/**
	 * Disconnect from the configured Cams Policy Server(s).
	 *
	 * @return true if the CamsClient successfully disconnects, else false.
	 */
	public void disconnect()
	{
		// Remove the ConnectionExceptionListener.
		logger.info(this, "Removing ConnectionExceptionListener from CamsClient");
		camsClient.removeConnectionExceptionListener(this);

		// Disconnect from Cams Policy Servers.
		if (camsClient.isConnected())
		{
			logger.info(this, 
				"CamsClient disconnecting from configured Cams Policy Server(s)");
			camsClient.disconnect();

			boolean connected = camsClient.isConnected();
			logger.info(this, "CamsClient disconnected = " + !connected);
		}
		else
		{
			logger.info(this, "CamsClient not connected: no need to disconnect");
		}
	}

	/**
	 * Destroy the CamsClientExample.
	 */
	public void destroy()
	{
		camsClient.destroy();
		camsClient = null;
		debug = false;
		resReqFactory = null;
		logger = null;
	}

	/**
	 * Test the Cams access control service.
	 *
	 * @param sessionId a Cams session identifier object representing
	 * 	an authenticated Cams user or null if no authenticated user.
	 */
	public void testAccessControlService(SessionId sessionId)
	{
		AccessControlService acService = null;
		AccessControlRequest acRequest = null;
		AccessControlResponse acResponse = null;

		logger.info(this, "Starting access control service test");

		try
		{
			// Get the access control service from the CamsClient.
			acService = (AccessControlService)camsClient.find(
				"access-control", AccessControlService.class);

			// Get an AccessControlRequest object from the service.
			acRequest = acService.createAccessControlRequest();

			// Create an HTTP ResourceRequest.
			ResourceRequest resourceRequest =
				resReqFactory.createResourceRequest(httpResourceType,
					"http://localhost:8080/cams/camstest.jsp", "GET");
			acRequest.setResourceRequest(resourceRequest);
			acRequest.setRemoteAddr("127.0.0.1");
			acRequest.setRemoteHost("localhost");
			acRequest.setLoginConfigEntry("http");
			acRequest.setConfidential(false);

			if (sessionId != null)
				acRequest.addSessionId(sessionId);

			if (debug)
				acRequest.log(logger, true);

			try
			{
				acResponse = acService.checkAccess(acRequest);
			}
			catch(CamsTransportException cte)
			{
				if(!acRequest.hasSessionIds())
				{
					logger.error(this, "Error occurred checking access", cte);
					return;
				}

				acRequest.clearSessionIds();

				try
				{
					acResponse = acService.checkAccess(acRequest);
				}
				catch(CamsTransportException cte2)
				{
					throw new AccessControlException(
						"Error occurred sending request", cte2);
				}
			}

			if (debug)
				acResponse.log(logger);

			// Your agent should handle the access control response as needed
			// for the agent environment. Sample code is provided in private
			// method: handleAccessControlResponse();
			handleAccessControlResponse(acRequest, acResponse);
		}
		catch(ServiceException se)
		{
			logger.error(this, "Error occurred checking access", se);
		}
		catch(AccessControlException ace)
		{
			logger.error(this, "Error occurred checking access", ace);
		}
		catch(InvalidResourceException ire)
		{
			logger.error(this, "Error occurred checking access", ire);
		}
		finally
		{
			if (acService != null)
			{
				if (acRequest != null)
					acService.destroy(acRequest);
				if (acResponse != null)
					acService.destroy(acResponse);
				acService = null;
			}
		}

		logger.info(this, "Completed access control service test");
	}

	/**
	 * Test the Cams ping service.
	 */
	private void testPingService()
	{
		PingService pingService = null;
		PingRequest pingRequest = null;
		PingResponse pingResponse = null;

		logger.info(this, "Starting ping service test");

		try
		{
			// Get the ping service from the CamsClient.
			pingService = (PingService)camsClient.find("ping", PingService.class);

			// Create a PingRequest object.
			pingRequest = pingService.createPingRequest();

			// Set the IP address for this agent.
			pingRequest.setRemoteAddr("127.0.0.1");

			// Set the Cams Policy Server name to be pinged.
			pingRequest.setServerName("MyCamsServer");

			// Ping the Cams Policy Server.
			pingResponse = pingService.ping(pingRequest);

			// Handle the PingResponse.
			int pingStatus = pingResponse.getStatus();
			if (pingStatus == PingResponse.SC_SUCCESS)
			{
				logger.info(this, "Ping status is: SUCCESS");
				logger.info(this, "Ping message is: " + pingResponse.getMessage());
				logger.info(this, "Ping server name is: " + pingResponse.getServerName());
				logger.info(this, "Ping server address is: " + pingResponse.getServerAddr());
				logger.info(this, "Ping server port is: " + pingResponse.getServerPort());
			}
			else if (pingStatus == PingResponse.SC_FAILED)
			{
				logger.info(this, "Ping status is: FAILED");
				logger.info(this, "Ping message is: " + pingResponse.getMessage());
				logger.info(this, "Ping server name is: " + pingResponse.getServerName());
				logger.info(this, "Ping server address is: " + pingResponse.getServerAddr());
				logger.info(this, "Ping server port is: " + pingResponse.getServerPort());
			}
			else if (pingStatus == PingResponse.SC_PENDING)
			{
				logger.warning(this, "Ping status is: PENDING (request timed out)");
			}
			else
			{
				logger.error(this, "Ping status is unrecognized: status=" + pingStatus);
			}
		}
		catch(ServiceException se)
		{
			logger.error(this, "Error occurred pinging server", se);
		}
		catch(CamsTransportException cte)
		{
			logger.error(this, "Error occurred pinging server", cte);
		}
		catch(PingException pe)
		{
			logger.error(this, "Error occurred pinging server", pe);
		}
		finally
		{
			if (pingService != null)
			{
				if (pingRequest != null)
					pingService.destroy(pingRequest);
				if (pingResponse != null)
					pingService.destroy(pingResponse);
				pingService = null;
			}
		}
	}

	/**
	 * Test the Cams authentication service.
	 *
	 * @return a Cams SessionId object if authentication was successful.
	 */
	public SessionId testAuthenticationService()
	{
		SessionId sessionId = null;
		AuthenticationService authService = null;
		AuthRequest authRequest = null;
		AuthResponse authResponse = null;

		try
		{
			// Get the authentication service from the CamsClient.
			authService = (AuthenticationService)camsClient.find(
				"authentication", AuthenticationService.class);

			// Populate the authentication request.
			authRequest = authService.createAuthRequest();
			authRequest.setSecurityDomainName("system");
			authRequest.setLoginConfigEntryName("http");
			authRequest.setRemoteAddr("127.0.0.1");
			authRequest.setRemoteHost("localhost");
			authRequest.addCallbackValue("username", "guest");
			authRequest.addCallbackValue("password", "password");

			// NOTE: If your agent needs to implement "re-authentication" or
			// request additional authentication tokens, you can add an 
			// existing Cams SessionId object here.
			// For example:
			//
			// authRequest.addSessionId(sessionId);

			if (debug)
				authRequest.log(logger);

			try
			{
				// Authenticate the user.
				authResponse = authService.authenticate(authRequest);
			}
			catch(CamsTransportException cte)
			{
				// Transport exception occurred, which means that the request failed
				// to a specific Cams Policy Server chosen by the CamsClient. 

				// If request had no sessions it was anonymous and we must fail
				// because the CamsClient has already attempted authentication
				// with all configured policy servers.
				if (!authRequest.hasSessionIds())
					throw new AuthException("Failed to send request", cte);

				// If the request has existing Cams sessions, then we won't be
				// able to "re-authenticate" or get additional authentication 
				// credentials with the specific policy server for which the 
				// session was created. That's because Cams sessions are "sticky"
				// to a policy server.
				//
				// In this case, clear the existing Cams session and try a fresh
				// (anonymous) authentication, which might be handled by another
				// available Cams Policy Server. 
				authRequest.clearSessionIds();

				try
				{
					// Attempt another authentication.
					authResponse = authService.authenticate(authRequest);
				}
				catch(CamsTransportException cte2)
				{
					// Authentication failure: all avenues have been exhausted.
					throw new AuthException("Failed to send request", cte2);
				}
			}

			// An authentication response was received.

			if (debug)
				authResponse.log(logger);

			// If authentication succeeded a session id will be returned.
			sessionId = authResponse.getSessionId();
			if (sessionId != null)
			{
				logger.info(this, "User authentication succeeded: session id=" +
					sessionId.toString());
			}
			else
			{
				logger.info(this, "User authentication failed: " +
					"status code=" + authResponse.getStatus() +
					", reason code=" + authResponse.getReason() +
					", message=" + authResponse.getMessage());
			}

			// Your agent should handle the authentication response as needed
			// for the agent environment. Sample code is provided in private
			// method: handleAuthResponse();
			handleAuthResponse(authRequest, authResponse);
		}
		catch(ServiceException se)
		{
			logger.error(this, "authentication service error", se);
		}
		catch(AuthException ae)
		{
			logger.error(this, "authentication error", ae);
		}
		finally
		{
			if (authService != null)
			{
				if (authRequest != null)
					authService.destroy(authRequest);
				if (authResponse != null)
					authService.destroy(authResponse);
				authService = null;
			}
		}

		return sessionId;
	}

	/**
	 * Test the Cams session access service.
	 *
	 * @param sessionId the identifier for the Cams session to be accessed.
	 */
	private void testSessionAccessService(SessionId sessionId)
	{
		SessionAccessService saService = null;
		SessionAccessRequest saRequest = null;
		SessionAccessResponse saResponse = null;

		logger.info(this, "Starting session access service test");

		try
		{
			// Get the session access service from the CamsClient.
			saService = (SessionAccessService)camsClient.find(
				"session-access", SessionAccessService.class);

			// Get a session access request object from the service.
			saRequest = saService.createSessionAccessRequest();

			// Set the session id and the current user IP address.
			saRequest.setSessionId(sessionId);
			saRequest.setRemoteAddr("127.0.0.1");

			if (debug)
				saRequest.log(logger);

			try
			{
				saResponse = saService.accessSession(saRequest);
			}
			catch(CamsTransportException cte)
			{
				throw new SessionAccessException(
					"Error occurred sending session access request", cte);
			}

			if (debug)
				saResponse.log(logger);

			// Handle the session access response.
			int saStatus = saResponse.getStatus();
			if (saStatus == SessionAccessResponse.SC_SUCCESS)
			{
				// SessionAccess succeeded
				RemoteSession session = saResponse.getSession();
				logger.info(this, "Session access status is: SUCCESS");
				logger.info(this, "Session id=" + session.getId());
				logger.info(this, "Session status=" + session.getStatus());
				logger.info(this, "Security Domain Name=" + session.getSecurityDomainName());
				logger.info(this, "Creation Time=" + new Date(session.getCreationTime()));
				logger.info(this, "Last Touch Time=" + new Date(session.getLastTouchTime()));
				logger.info(this, "Idle Time=" + session.getIdleTime());

				// Get the subject name
				logger.info(this, "Subject (user) name=" + session.getSubjectName());

				// Get the principal names (roles)
				String[] principalName = session.getPrincipalNames();
				for (int i = 0; i < principalName.length; i++)
				{
					logger.info(this, "Principal (role) name=" + principalName[i]);
				}

				// Get all the session attribute namespaces
				String[] namespace = session.getNamespaces();
				for (int i = 0; i < namespace.length; i++)
				{
					// Get all attribute names within a namespace
					String[] attrName = session.getAttributeNames(namespace[i]);
					for (int j = 0; j < attrName.length; j++)
					{
						// Dump each name/value pair within the namespace
						logger.info(this, "Session Attribute: " + 
							namespace[i] + "." + attrName[j] + "=" + 
							session.getAttribute(namespace[i], attrName[j]));
					}
				}
			}
			else if (saStatus == SessionAccessResponse.SC_FAILED)
			{
				logger.info(this, "Session access status is: FAILED, reason code=" +
					saResponse.getReason());

				// SessionAccess failed. Take action based on the reason.
				switch (saResponse.getReason())
				{
					case SessionAccessResponse.RC_SESSION_DOES_NOT_EXIST:
						// The session id probably expired, cleanup cached session info
						// if necessary
						break;

					case SessionAccessResponse.RC_GENERAL_SERVER_ERROR:
					case SessionAccessResponse.RC_INVALID_REMOTE_IP_ADDRESS:
					case SessionAccessResponse.RC_UNAUTHORIZED_AGENT:
					case SessionAccessResponse.RC_UNKNOWN_SECURITY_DOMAIN:
						// Misconfiguration of agent or security domain
						logger.error(this, "session access failed: reason(" +
							saResponse.getReason() + "): " + saResponse.getMessage());
						break;

					case SessionAccessResponse.RC_INVALID_SESSION_ID:
					case SessionAccessResponse.RC_POISONED_SESSION:
						// The session id is invalid or corrupted. The format of the
						// session id is not understood or an internal checksum 
						// indicates the session id was tampered with or is NOT
						// associated with the RemoteAddress.
						break;

					default:
						// Unhandled reason code
						logger.error(this, "Unknown session access reason code: " +
							saResponse.getReason() + ", message=" + saResponse.getMessage());
						break;
				}
			}
			else if (saStatus == SessionAccessResponse.SC_PENDING)
			{
				logger.warning(this, "Session access status is: PENDING (request timed out)");
			}
			else
			{
				logger.error(this, "Session access status is unrecognized: status=" + saStatus);
			}
		}
		catch(ServiceException se)
		{
			logger.error(this, "Error occurred accessing session", se);
		}
		catch(SessionAccessException sae)
		{
			logger.error(this, "Error occurred accessing session", sae);
		}
		finally
		{
			if (saService != null)
			{
				if (saRequest != null)
					saService.destroy(saRequest);
				if (saResponse != null)
					saService.destroy(saResponse);
				saService = null;
			}
		}

		logger.info(this, "Completed session access service test");
	}

	/**
	 * Test the Cams session control service.
	 *
	 * @param sessionId the Cams session identifier.
	 */
	private void testSessionControlService(SessionId sessionId)
	{
		SessionControlService scService = null;
		SessionControlRequest scRequest = null;
		SessionControlResponse scResponse = null;

		logger.info(this, "Starting session control service test");

		try
		{
			// Get the access control service from the CamsClient.
			scService = (SessionControlService)camsClient.find(
				"session-control", SessionControlService.class);

			scRequest = scService.createSessionControlRequest();
			scRequest.setSessionId(sessionId);
			scRequest.setRemoteAddr("127.0.0.1");
			scRequest.setAction(SessionControlAction.CLOSE);

			if (debug)
				scRequest.log(logger);

			try
			{
				scResponse = scService.controlSession(scRequest);
			}
			catch(CamsTransportException cte)
			{
				throw new SessionControlException(
					"Error occurred sending session control request", cte);
			}

			if (debug)
				scResponse.log(logger);

			// Handle the session control response.
			int scStatus = scResponse.getStatus();
			if (scStatus == SessionControlResponse.SC_SUCCESS)
			{
				logger.info(this, "Session Control status is: SUCCESS");
			}
			else if (scStatus == SessionControlResponse.SC_FAILED)
			{
				logger.info(this, "Session Control status is: FAILED, reason code=" +
					scResponse.getReason());

				switch (scResponse.getReason())
				{
					case SessionControlResponse.RC_NOT_APPLICABLE:
						logger.info(this, "reason=RC_NOT_APPLICABLE");
						break;

					case SessionControlResponse.RC_GENERAL_SERVER_ERROR:
						logger.info(this, "reason=RC_GENERAL_SERVER_ERROR");
						break;

					case SessionControlResponse.RC_UNKNOWN_SECURITY_DOMAIN:
						logger.info(this, "reason=RC_GENERAL_SERVER_ERROR");
						break;

					case SessionControlResponse.RC_INVALID_SESSION_ID:
						logger.info(this, "reason=RC_GENERAL_SERVER_ERROR");
						break;

					case SessionControlResponse.RC_POISONED_SESSION:
						logger.info(this, "reason=RC_GENERAL_SERVER_ERROR");
						break;

					case SessionControlResponse.RC_SESSION_DOES_NOT_EXIST:
						logger.info(this, "reason=RC_GENERAL_SERVER_ERROR");
						break;

					default:
						logger.info(this, "reason is unrecognized");
						break;
				}
			}
			else if (scStatus == SessionControlResponse.SC_PENDING)
			{
				logger.warning(this, "Session Control status is: PENDING (request timed out)");
			}
			else
			{
				logger.error(this, "Session Control status is unrecognized: status=" + scStatus);
			}
		}
		catch(ServiceException se)
		{
			logger.error(this, "Session Control service error on close", se);
		}
		catch(SessionControlException sce)
		{
			logger.error(this, "Session Control exception on close", sce);
		}
		finally
		{
			if (scService != null)
			{
				if (scRequest != null)
					scService.destroy(scRequest);
				if (scResponse != null)
					scService.destroy(scResponse);
				scService = null;
			}
		}

		logger.info(this, "Completed session control service test");
	}

	/**
	 * Load the configuration Properties from a File.
	 *
	 * @param configFile the configuration File object.
	 * @return a new Properties object with all of the appropriate properties.
	 */
	private static Properties loadPropertiesFromFile(File configFile)
		throws ConfigException
	{
		Properties properties = new Properties();
		FileInputStream fis = null;

		try
		{
			fis = new FileInputStream(configFile);
			properties.load(fis);
		}
		catch (FileNotFoundException e)
		{
			throw new ConfigException("Error loading configuration file: " + 
				configFile.getAbsolutePath(), e);
		}
		catch (IOException e)
		{
			throw new ConfigException("Error loading configuration file: " + 
				configFile.getAbsolutePath(), e);
		}
		finally
		{
			try
			{
				if (fis != null)
					fis.close();
			}
			catch (Exception e)
			{
			}
		}

		return properties;
	}

	/**
	 * Create the Logger for use by the CamsClient.
	 *
	 * @param properties the configuration properties.
	 * @return a new Logger instance.
	 */
	private static Logger createLogger(Properties properties)
	{
		Logger logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		String className = properties.getProperty("logger.class");

		try
		{
			properties.put("logger.name", "cams-client-example");
			logger = (Logger)ClassUtils.createInstance(className);
			logger.initialize(properties);
		}
		catch (ClassInstantiationException cie)
		{
			StringBuffer buffer = new StringBuffer();
			buffer.append("Could not create Logger object, reason=");
			buffer.append(cie.getMessage());
			buffer.append(" creating StderrLogger");
			logger = StderrLogger.DEFAULT_STDERR_LOGGER;
			logger.warning(CamsClientExample.class, buffer.toString());
		}
		catch (LoggerException le)
		{
			StringBuffer sb = new StringBuffer();
			sb.append("Could not initialize Logger object, ");
			sb.append("reason=");
			sb.append(le.getMessage());
			sb.append(", Creating StderrLogger");
			logger = StderrLogger.DEFAULT_STDERR_LOGGER;
			logger.warning(CamsClientExample.class, sb.toString());
		}

		return logger;
	}

	/**
	 * Handle the access control response
	 *
	 * @param acRequest the access control response associated with the response.
	 * @param acResponse the access control response associated with the request.
	 */
	private void handleAccessControlResponse(
		AccessControlRequest acRequest, AccessControlResponse acResponse)
	{
		// Handle the access control response.
		int acStatus = acResponse.getStatus();
		if (acStatus == AccessControlResponse.SC_GRANTED)
		{
			if (debug)
			{
				SessionId sessionId = acResponse.getSessionId();
				logger.debug(this, "access GRANTED to resource id=" +
					acRequest.getResourceId().toString() + "'" +
					", security domain=" + acResponse.getSecurityDomainName() +
					", session id=" + 
					(sessionId != null ? sessionId.toString() : "null"));
			}

			// The server configuration "last modified time" is useful for clearing
			// cached state.
			if (lastModified != acResponse.getLastModificationTime())
			{
				// Clear out cached state

				// Update the lastModified time
				lastModified = acResponse.getLastModificationTime();
			}

			// NOTE: Handle successful access control obligations here.
			// Obligations are actions that an agent is obligated (required)
			// to perform as a Policy Enforcement Point (PEP) based on a decision
			// by the Policy Decision Point (PDP), which in this case is the 
			// AuthenticationService at the Cams Policy Server.
			//
			// Obligations are specific to the type of agent. For a web agent,
			// an obligation may specify: to redirect the User-Agent to a 
			// specific URL, add an HTTP response header, respond with a 
			// speciic HTTP error code, etc. Using the Obligations API on
			// the Cams Policy Server, you can create/return obligations that
			// are very specific to the types of actions needed by your agent's
			// environment.
			//
			// See the Cams Obligations API for for more details.
			Iterator obligationItr = acResponse.getObligations();
			while (obligationItr.hasNext())
			{
				Obligation obligation = (Obligation)obligationItr.next();
				logger.info(this, "Access Control Obligation: id=" + 
					obligation.getId() + ", description=" + 
					obligation.getDescription());

				// Handle Attributes returned by the Obligation here.
			}

			// Do whatever is necessary on successful authentication here.
			// For example: create a session cookie, return the session id
			// to a web service, save the session id in a Java client, etc.

			// Access to the requested resource is granted. Various other values are
			// returned, which may be useful to an agent. 
			// Take action based on the reason.
			switch (acResponse.getReason())
			{
				case AccessControlResponse.RC_ACCESS_GRANTED_UNCONDITIONALLY:
				case AccessControlResponse.RC_DEFAULT_BIAS_APPLIED:
					// Cache unconditionally granted access control checks?
					break;

				case AccessControlResponse.RC_ACCESS_GRANTED_CONDITIONALLY:
					// Access was granted by one or more conditional 
					// access control rules
				break;

				default:
					// Ignore all other reason codes
					break;
			}

			// Fall through, or do whatever else is necessary to return the resource
		}
		else if (acStatus == AccessControlResponse.SC_DENIED)
		{
			if (debug)
			{
				SessionId sessionId = acResponse.getSessionId();
				logger.debug(this, "access DENIED to resource id=" + 
					acRequest.getResourceId().toString() + "'" +
					", message=" + acResponse.getMessage() +
					", reason code=" + acResponse.getReason() +
					", security domain=" + acResponse.getSecurityDomainName() +
					", session id=" + (sessionId != null ? sessionId.toString() : "null"));
			}

			// The server configuration "last modified time" might be useful for
			// clearing cached state.
			if (lastModified != acResponse.getLastModificationTime())
			{
				// Clear out cached state?

				// Update the lastModified time
				lastModified = acResponse.getLastModificationTime();
			}

			// NOTE: Handle denied access control obligations here.
			// Obligations are actions that an agent is obligated (required)
			// to perform as a Policy Enforcement Point (PEP) based on a decision
			// by the Policy Decision Point (PDP), which in this case is the 
			// AuthenticationService at the Cams Policy Server.
			//
			// Obligations are specific to the type of agent. For a web agent,
			// an obligation may specify: to redirect the User-Agent to a 
			// specific URL, add an HTTP response header, respond with a 
			// speciic HTTP error code, etc. Using the Obligations API on
			// the Cams Policy Server, you can create/return obligations that
			// are very specific to the types of actions needed by your agent's
			// environment.
			//
			// See the Cams Obligations API for for more details.
			Iterator obligationItr = acResponse.getObligations();
			while (obligationItr.hasNext())
			{
				Obligation obligation = (Obligation)obligationItr.next();
				logger.info(this, "Access Control Obligation: id=" + 
					obligation.getId() + ", description=" + 
					obligation.getDescription());

				// Handle Attributes returned by the Obligation here.
			}

			// Access to the requested resource was denied.
			// Take action based on the reason.
			switch (acResponse.getReason())
			{
				case AccessControlResponse.RC_ACCESS_DENIED_UNCONDITIONALLY:
				case AccessControlResponse.RC_ACCESS_DENIED_CONDITIONALLY:
				case AccessControlResponse.RC_DEFAULT_BIAS_APPLIED:

					// Perform "access-denied" action.
					break;

				case AccessControlResponse.RC_ACCESS_DENIED_AUTHENTICATION_REQUIRED:
				case AccessControlResponse.RC_ACCESS_DENIED_SESSION_EXPIRED:

					// Get the "login parameters" returned with the response.
					// NOTE: These values are configured in the login-config.xml
					// file on the Cams Policy Server for a given security domain
					// and login-config-entry
    
					Map loginParamMap = acResponse.getLoginParameters();

					// Prompt user to authenticate? (agent-specific action)
					break;

				case AccessControlResponse.RC_ACCESS_DENIED_CONFIDENTIALITY_REQUIRED:
					// Notify that access to resource requires confidentiality.
					// e.g. Redirect to "https" URL?
					break;

				case AccessControlResponse.RC_ACCESS_DENIED_EVALUATION_ERROR:
				case AccessControlResponse.RC_GENERAL_SERVER_ERROR:
				case AccessControlResponse.RC_INVALID_REMOTE_HOST_NAME:
				case AccessControlResponse.RC_INVALID_REMOTE_IP_ADDRESS:
				case AccessControlResponse.RC_INVALID_RESOURCE_IDENTIFIER:
				case AccessControlResponse.RC_UNKNOWN_RESOURCE_ACTION:
				case AccessControlResponse.RC_UNKNOWN_RESOURCE_TYPE:
				case AccessControlResponse.RC_UNKNOWN_SECURITY_DOMAIN:
				case AccessControlResponse.RC_UNKNOWN_LOGIN_CONFIG:
				case AccessControlResponse.RC_UNAUTHORIZED_AGENT:

					// Configuration or runtime error.
					logger.error(this, "access control error: " + 
						acResponse.getMessage());

					// Take appropriate "error" action here.

					break;

				default:

					// Unhandled access denied reason code.
					logger.error(this, "Unknown access denied reason code: " +
						acResponse.getReason() + ", message=" + 
						acResponse.getMessage());

					// Take appropriate "error" action here.
	
					break;
			}
		}
		else if (acStatus == AccessControlResponse.SC_PENDING)
		{
			logger.warning(this, "Access control descision is: PENDING (request timed out)");

			// Try again or take appropriate "error" action here.
		}
		else
		{
			logger.error(this, "Access control descision is unrecognized: status=" + acStatus);

			// Take appropriate "error" action here.
		}
	}

	/**
	 * Handle the authentication response.
	 *
	 * @param authRequest the authentication request associated with the response.
	 * @param authResponse the authentication response.
	 */
	private void handleAuthResponse(AuthRequest authRequest, AuthResponse authResponse)
	{
		if (authResponse == null)
			throw new IllegalArgumentException(
				"AuthResponse is null, cannot handle null response");

		int statusCode = authResponse.getStatus();
		if (statusCode == AuthResponse.SC_SUCCESS)
		{
			// Authentication was successful.

			// Get the SessionId
			SessionId sessionId = authResponse.getSessionId();

			// Do whatever is necessary on successful authentication here.
			// For example: create a session cookie, return the session id
			// to a web service, save the session id in a Java client, etc.

			logger.info(this, "Authentication was successful: session id=" +
				sessionId.toString());

			// NOTE: Handle successful authentication obligations here.
			// Obligations are actions that an agent is obligated (required)
			// to perform as a Policy Enforcement Point (PEP) based on a decision
			// by the Policy Decision Point (PDP), which in this case is the 
			// AuthenticationService at the Cams Policy Server.
			//
			// Obligations are specific to the type of agent. For a web agent,
			// an obligation may specify: to redirect the User-Agent to a 
			// specific URL, add an HTTP response header, respond with a 
			// speciic HTTP error code, etc. Using the Obligations API on
			// the Cams Policy Server, you can create/return obligations that
			// are very specific to the types of actions needed by your agent's
			// environment.
			//
			// See the Cams Obligations API for for more details.
			Iterator obligationItr = authResponse.getObligations();
			while (obligationItr.hasNext())
			{
				Obligation obligation = (Obligation)obligationItr.next();
				logger.info(this, "Authentication Obligation: id=" + 
					obligation.getId() + ", description=" + obligation.getDescription());

				// Handle Attributes returned by the Obligation here.
			}
		}
		else if (statusCode == AuthResponse.SC_FAILED)
		{
			logger.info(this, "Authentication failed");

			// NOTE: Handle failed authentication obligations here.
			// Obligations are actions that an agent is obligated (required)
			// to perform as a Policy Enforcement Point (PEP) based on a decision
			// by the Policy Decision Point (PDP), which in this case is the 
			// AuthenticationService at the Cams Policy Server.
			//
			// Obligations are specific to the type of agent. For a web agent,
			// an obligation may specify: to redirect the User-Agent to a 
			// specific URL, add an HTTP response header, respond with a 
			// speciic HTTP error code, etc. Using the Obligations API on
			// the Cams Policy Server, you can create/return obligations that
			// are very specific to the types of actions needed by your agent's
			// environment.
			//
			// See the Cams Obligations API for for more details.
			Iterator obligationItr = authResponse.getObligations();
			while (obligationItr.hasNext())
			{
				Obligation obligation = (Obligation)obligationItr.next();
				logger.info(this, "Authentication Obligation: id=" + 
					obligation.getId() + ", description=" + obligation.getDescription());

				// Handle Attributes returned by the Obligation here.
			}

			// Handle based on the failed authentication reason.
			int reasonCode = authResponse.getReason();
			switch (reasonCode)
			{
				case AuthResponse.RC_LOGIN_FAILED:
				case AuthResponse.RC_EXPIRED_ACCOUNT:
				case AuthResponse.RC_EXPIRED_CREDENTIAL:

					// Login has failed due to invalid credentials, an expired
					// account, or expired credentials. Take appropriate action
					// here.

					logger.info(this, "Login failed: invalid credentials, expired account " +
						"or expired credentials: reason code=" + reasonCode);

					// Dump any login parameters that might have been returned.
					Map loginParametersMap = authResponse.getLoginParameters();
					if ((loginParametersMap != null) && (loginParametersMap.size() > 0))
					{
						Iterator iter = loginParametersMap.keySet().iterator();
						while (iter.hasNext())
						{
							String name = (String)iter.next();
							String value = (String)loginParametersMap.get(name);
							logger.info(this, "   Login parameter: " + name + "=" + value);
						}
					}

					break;

				case AuthResponse.RC_CALLBACK_HANDLER_ERROR:
				case AuthResponse.RC_GENERAL_SERVER_ERROR:
				case AuthResponse.RC_INVALID_REMOTE_HOST_NAME:
				case AuthResponse.RC_INVALID_REMOTE_IP_ADDRESS:
				case AuthResponse.RC_NOT_APPLICABLE:
				case AuthResponse.RC_UNAUTHORIZED_AGENT:
				case AuthResponse.RC_UNKNOWN_SECURITY_DOMAIN:
				case AuthResponse.RC_UNKNOWN_LOGIN_CONFIG:
				case AuthResponse.RC_LOGIN_ERROR:

					// An error occured.
					logger.error(this, "login error: reason code=" + reasonCode);

					// Take appropriate error action here.

				default:

					// An error occured.
					logger.error(this, "login error: unknown reason code=" + reasonCode);

					// Take appropriate error action here.

					break;
			}
		}
		else
		{
			logger.error(this, "login error: unknown status code=" + statusCode);
			
			// Take appropriate error action here.
		}
	}

} // End of class : CamsClientExample
