/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose
 * such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package com.cafesoft.cams.auth.callback;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import java.io.IOException;

import java.util.Map;
import java.util.HashMap;

/**
 * <code>MapCallbackHandler</code> extends the standard JAAS
 * <code>CallbackHandler</code> interface with methods to set/get 
 * callback values by name. This enables Cams to provide a generic, 
 * general purpose CallbackHandler that does not need a new set/get 
 * method for every callback value.
 *
 * @author Norbert K. Kuhnert
 * @version $Revision: 1.2 $ $Date: 2009/08/14 23:49:50 $ $Author: norb $
 * @since 9/22/2008
 */
public class MapCallbackHandler implements CallbackHandler
{
	//
	// private instances
	//
	private Map cbMap = null;

	/**
	 * Create a new MapCallbackHandler instance.
	 */
	public MapCallbackHandler()
	{
		this.cbMap = new HashMap();
	}
	
	/**
	 * Set a callback value.
	 * 
	 * @param name the callback value name.
	 * @param value the callback value.
	 */
	public void setCallbackValue(String name, String value)
	{
		if ((name == null) || (value == null))
			return;
		
		this.cbMap.put(name, value);
	}

	/**
	 * Set callback values from an existing Map.
	 *
	 * @param srcMap the source Map from wich callback values are to be copied.
	 */
	public void setCallbackValues(Map srcMap)
	{
		if (srcMap == null)
			return;

		this.cbMap.putAll(srcMap);
	}

	/**
	 * Retrieve or display the information requested in the provided Callbacks.
	 * The handle method implementation checks the instance(s) of the Callback
	 * object(s) passed in to retrieve or display the requested information.
	 *
	 * @param callbacks an array of Callback objects provided by an underlying
	 * security service which contains the information requested to be retrieved
	 * or displayed.
	 *
	 * @exception IOException if an input or output error ocurrs
	 * @exception UnsupportedCallbackException if the implementation of this
	 * method does not support one or more of the Callbacks specified in the
	 * callbacks parameter
	 */
	public void handle(Callback[] callbacks)
		throws IOException, UnsupportedCallbackException
	{
		// Loop over all Callbacks
		for (int i = 0; i < callbacks.length; i++)
		{
			Callback cb = callbacks[i];

			if (cb instanceof MapCallback)
			{
				((MapCallback)cb).setCallbackValues(cbMap);
			}
			else
			{
				throw new UnsupportedCallbackException(
					cb, "Unrecognized Callback type: " + 
					cb.getClass().getName());
			}
		}
	}

} // End of class: MapCallbackHandler

