/**
 * Copyright (c) 1996-2012 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.service.ServiceClient;
import com.cafesoft.core.service.ServiceException;
import com.cafesoft.core.service.ServiceFinder;

import com.cafesoft.cams.service.ActiveDirectoryGroupNameService;

import com.novell.ldap.LDAPAuthHandler;
import com.novell.ldap.LDAPAuthProvider;
import com.novell.ldap.LDAPAttribute;
import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPConstraints;
import com.novell.ldap.LDAPDN;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;
import com.novell.ldap.LDAPReferralException;
import com.novell.ldap.LDAPReferralHandler;
import com.novell.ldap.LDAPSearchResults;
import com.novell.ldap.LDAPJSSESecureSocketFactory;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;
import java.io.UnsupportedEncodingException;

import java.security.Principal;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.StringTokenizer;

/**
 * <p><code>AESOActiveDirectoryLoginModule</code> is a JAAS LoginModule used
 * with Cams "Automatic Enterprise Sign-On" (AESO) and Microsoft Active
 * Directory.</p>
 * <p>This LoginModule is designed to trust a user identity authenticated at the 
 * web layer, most likely via Kerberos, NTLM, HTTP Basic or HTTP Digest
 * authentication methods. The user is authenticated within the Cams security
 * context login succeeds within this LoginModule.<p>
 * <p>Responsibilities of this LoginModule include:
 * <ul>
 * <li>using LDAP to bind to the configured ActiveDirectory as a privileged
 * administrative user that has permission to query for user information.
 * <li>querying for existence of the user based on callback value "username".
 * <li>querying for group memberships and assigning role principals.
 * <li>populating a JAAS Subject with the user Principal and role Principals.
 * <li>adding the Authentication Method to the Subject public credential
 * cache based on the value of callback value "auth_type". Supported values
 * from "weakest" to "strongest" include: "PASSWORD", "NTLM", "Kerberos".
 * NOTE: the "auth_method" value is populated by proper configuration of 
 * the corresponding AESO callback value at the Cams AESO agent.
 * </ul>
 * Supported configuration options are listed below:
 * <ul>
 * <li><b>debug</b> - Enable/disable debugging [true|false].
 * <li><b>host</b> - The DNS name or IP address of the Active Directory
 * host.</li>
 * <li><b>port</b> - The host's Active Directory port to use.</li>
 * <li><b>useSSL</b> - Enables/disables use of SSL.</li>
 * <li><b>connectTimeout</b> - (Optional) The number of milliseconds to wait for
 * 	a connection to the ActiveDirectory before timing out.</li>
 * <li><b>bindUsername</b> - The username for a Windows account with permissions
 * to query Active Directory for user and group information.</li>
 * <li><b>bindPassword</b> - The password for the "bindUser" account.</li>
 * <li><b>agentSecret</b> - The expected value for the "agent_secret" callback
 * value, which must be present in the authentication request. This value is 
 * used to authenticated Cams agents, which assert the identity of a user
 * already authenticated at the web layer. The agent secret is used to protect 
 * against impersonation attacks via carefully created HTTP requests 
 * attempting to subvert AESO component trust relationships.</li>
 * <li><b>defaultRoles</b> - (Optional) A list of one or more default roles
 * separated by a comma to be added to the user session.</li>
 * <li><b>roleRegexPattern</b> - (optional) Limits roles saved in the user 
 * session to those that match a regular expression pattern. This option is 
 * useful when the role search returns many roles that are not used by the 
 * access control policy. 
 * <li><b>domain</b> - (Optional) The Active Directory domain to use. If this
 * value is present and the user name does not end with @domain, then append
 * "@" and the value of domain to the user name before attempting to the 
 * userPrincipalName search.</li>
 * <li><b>useRoleSearch</b> - (Optional) A value of "true" (default) 
 * indicates that a role search will be attempted [true|false].</li>
 * <li><b>useDomainInRoleSearch</b> - (Optional) A value of "true" (default) 
 * indicates that an optionally supplied "domain" used for authentication
 * will be included in the role search filter's {username} substitution value 
 * [true|false].</li>
 * <li><b>baseDN</b> - (Conditional) The base distinguished name for the User entries.
 * <li><b>scope</b> - (Conditional) The scope for the role search. Valid values
 *   are "BASE", "ONE", and "SUB".</li>
 * <li><b>filter</b> - (Conditional) The role search filter.</li>
 * <li><b>roleAttr</b> - (Conditional) The user entity attribute where the roles
 * are listed.</li>
 *
 * WARNING: The values in this object can be accessed by other Cams components 
 * to enable SSO to other other systems, so administrator must take care that only 
 * trusted components have are used!</li>
 * </ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.14 $ $Date: 2012/08/22 22:05:11 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 10/02/2008
 */
public class AESOActiveDirectoryLoginModule
	implements LoginModule, LoggerClient, ServiceClient
{
	//
	// Static Variables.
	//
	
	/**
	 * The LoginModule name (used in log messages throughout the code).
	 */
	private static final String LMN = "[AESOActiveDirectoryLoginModule]";
	
	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag indicating that SSL must be used.
	 */
	private boolean useSSL;
	
	/**
	 * Flag indicating that a role search will be tried.
	 */
	private boolean useRoleSearch;

	/**
	 * Flag indicating that a domain value, if supplied, will be used in the 
	 * user name substitution value for a role search.
	 */
	private boolean useDomainInRoleSearch;
	
	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating user.
	 */
	private String username;

	/**
	 * The fully-qualified username (including domain) if needed 
	 * when searching ActiveDirectory.
	 */
	private String fqUsername;

	/**
	 * The agent secret (used to protect against impersonation attacks).
	 */
	private String agentSecret;
	
	/**
	 * A default list of user roles to associate with an authenticated user.
	 */
	private List roleList;

	/**
	 * Used to lookup required Cams services.
	 */
	private ServiceFinder serviceFinder;

	//
	// Constructor
	//

	/**
	 * Default AESOActiveDirectoryLoginModule Constructor.
	 */
	public AESOActiveDirectoryLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.useSSL = false;
		this.useRoleSearch = true;
		this.useDomainInRoleSearch = true;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.subject = null;
		this.username = null;
		this.agentSecret = null;
		this.roleList = new ArrayList();
		this.serviceFinder = null;
	}

	//
	// Implements the ServiceClient interface
	//

	/**
	 * Set the ServiceFinder.
	 *
	 * @param finder the class used by find Services by the ServiceClient.
	 */
	public void setServiceFinder(ServiceFinder finder)
	{
		this.serviceFinder = finder;
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *			with the end user (prompting for usernames and
	 *			passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.debug = "true".equalsIgnoreCase((String)options.get("debug"));
		this.useSSL = "true".equalsIgnoreCase((String)options.get("useSSL"));
		this.useRoleSearch = 
			! "false".equalsIgnoreCase((String)options.get("useRoleSearch"));
		this.useDomainInRoleSearch = 
			! "false".equalsIgnoreCase((String)options.get("useDomainInRoleSearch"));
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt  succeeded (checked
	 * by retrieving the private state saved by the <code>login</code> method),
	 * then this method cleans up any state that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 * <p>
	 * This method executes the following steps:
	 * <ol>
	 * <li>Checks preconditions
	 * <li>Handles callback objects
	 * <li>Get a connection to the Active Directory
	 * <li>Authenticates the users using bind (only method available
	 * with Active Directory)
	 * <li>Close the connection
	 * <li>Reset state if authentication failed
	 * <li>Return results of authentication attempt
	 * </ol>
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @exception LoginException thrown if any error occurs while trying to
	 * authenticate the user.
	 * @exception FailedLoginException if authentication fails.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException,
		AccountExpiredException, CredentialExpiredException
	{
		// Check preconditions
		if (callbackHandler == null)
			throw new LoginException(LMN + " Error: no CallbackHandler available");

		String host = (String)options.get("host");
		if ((host == null) || (host.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required 'host' property missing");

		String port = (String)options.get("port");
		if ((port == null) || (port.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required 'port' property missing");

		try
		{
			int portInt = Integer.parseInt(port);
			if (portInt < 1)
				throw new LoginException(LMN + " Error: " +
					"'port' property must be an integer value greater than zero");
		}
		catch (NumberFormatException e)
		{
			throw new LoginException(LMN + " Error: " +
				"'port' property is not an integer value");
		}

		String bindUsername = (String)options.get("bindUsername");
		if ((bindUsername == null) || (bindUsername.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'bindUsername' property missing");

		String bindPassword = (String)options.get("bindPassword");
		if ((bindPassword == null) || (bindPassword.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'bindPassword' property missing");
		
		String expectedAgentSecret = (String)options.get("agentSecret");
		if ((expectedAgentSecret == null) || (expectedAgentSecret.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'agentSecret' property missing");

		String domain = (String)options.get("domain");

		// Handle callback objects.
		handleCallbacks();

		// An agent secret must always be supplied as a callback value and
		// its value must match the expected value.
		if (agentSecret == null || agentSecret.length() == 0)
			throw new LoginException(LMN + " Error: Required callback value 'agent_secret' " +
				"does not exist or is empty");
		else if (agentSecret.equals(expectedAgentSecret) == false)
			throw new LoginException(LMN + " Error: Callback value 'agent_secret' " +
				"does not match expected value in option 'agentSecret'");
		
		// A user name must always be supplied
		if (username == null || username.length() == 0)
			throw new FailedLoginException(LMN + " Authentication failed");

		/**
		 * If the username callback value has a backslash, then assume it
		 * is a UNC (Universal Naming Convention) name. In this case, chop off 
		 * the windows domain name because the UNC form does not work well 
		 * with Active Directory.
		 */
		int index = username.indexOf('\\');
		if (index > 0)
			username = username.substring(index + 1);
		
		// Optionally append the domain to the user name if supplied. Do
		// not modify username as it is required "as is" for the role filter.
		fqUsername = username;
		if ((domain != null) && ! ("".equals(domain.trim())))
			fqUsername = username + "@" + domain;

		if (debug)
			logger.debug(this, LMN + " Will attempt to bind " +
				 "with user name = '" + bindUsername + "'");

		// Get a connection to Active Directory
		LDAPConnection connection = null;
		connection = getConnection();

		if (debug)
			logger.debug(this, LMN + " Connected to Active Directory: '" + 
				connection.isConnected() + "'");

		try
		{
			// Bind to Active Directory using credentials with "admin" permissions.
			connection.bind(3, bindUsername, bindPassword.getBytes());
		}
		catch (LDAPException e)
		{
			if (debug)
				logger.debug(this, LMN + " An error occured while trying to bind user: '" + 
					bindUsername + "'" +
					"\n\tMessage: " + e.getMessage() +
					"\n\tLDAPErrorMessage: " + e.getLDAPErrorMessage() +
					"\n\tResultCode: " + e.getResultCode(), e);

			// Login failure during LDAP bind.
			throw new LoginException(LMN + " Error while trying to " +
					"bind user: " + e.getLDAPErrorMessage() + "\n");
		}

		if (debug)
			logger.debug(this, LMN + " Successfully " +
				"bound to Active Directory as user '" + bindUsername + "'");

		try
		{
			String baseDN = (String)options.get("baseDN");
			if ((baseDN == null) || "".equals(baseDN.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'baseDN'");

			String filter = (String)options.get("filter");
			if ((filter == null) || "".equals(filter.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'filter'");

			String scope = (String)options.get("scope");
			if ((!"BASE".equalsIgnoreCase(scope)) &&
					(!"ONE".equalsIgnoreCase(scope)) &&
					(!"SUB".equalsIgnoreCase(scope)))
				throw new LoginException(LMN + " " +
					"Error: Required option 'scope' must be 'BASE', 'ONE', or 'SUB'");

			// Set the search scope, the default of 0 is "BASE"
			int scopeInt = 0;
			if ("ONE".equalsIgnoreCase(scope))
				scopeInt = 1;
			else if ("SUB".equalsIgnoreCase(scope))
				scopeInt = 2;

			// Set the search constraints to automatically follow referrals
			LDAPConstraints ldapConstraints = new LDAPConstraints();
			ldapConstraints.setReferralFollowing(true);

			// Create and set the referral handler
			LDAPReferralHandler referralHandler = new AuthImpl(bindUsername, bindPassword);
			ldapConstraints.setReferralHandler(referralHandler);

			// Set the search constraints
			connection.setConstraints(ldapConstraints);

			// Specify the attributes that the search will return
			String[] attrs = {"userPrincipalName"};

			// Substitute the user name in the role search filter. Optionally,
			// don't use the supplied domain in the user name to enable searches
			// using attributes other than the userPrincipalName, such as 
			// sAMAccountName, uid, or cn, which don't use the fully-
			// qualified UPN.
			if (useDomainInRoleSearch)
				filter = replaceUsername(filter, "username", fqUsername);
			else
				filter = replaceUsername(filter, "username", username);
			
			if (debug)
				logger.debug(this, LMN + " Searching for user entry attributes using filter: " +
					filter );

			// Search for the user entry attributes
			LDAPSearchResults searchResults = connection.search(baseDN,
				scopeInt, filter, attrs, false);

			// Process the search results to get the user dn
			if (searchResults.hasMore())
			{
				LDAPEntry entry = searchResults.next();
				LDAPAttribute attribute = entry.getAttribute(attrs[0]);
				if (attribute != null)
				{
					if (debug)
						logger.debug(this, LMN + " login() succeeded");
					
					// Found the userPrincipalName, so login succeeded.
					loginSucceeded = true;
				}
				else
				{
					if (debug)
						logger.debug(this, LMN + " login() failed");
								
					throw new FailedLoginException(
						LMN + " Authentication failed");
				}
			}
			else
			{
				if (debug)
					logger.debug(this, LMN + " Attribute 'userPrincipalName' " +
						"not found for user '" + username + "'");
			
				throw new LoginException ("Attribute 'userPrincipalName' " +
					"not found for user '" + username + "'");
			}
		}
		catch (LDAPException e)
		{
			if (debug)
				logger.debug(this, LMN + " An error '" +
					"occured while trying to bind user: '" + bindUsername + "'" +
					"\n\tMessage: " + e.getMessage() +
					"\n\tLDAPErrorMessage: " + e.getLDAPErrorMessage() +
					"\n\tResultCode: " + e.getResultCode(), e);

			// Login failure during LDAP bind.
			throw new LoginException(
				LMN + " Error while trying to " +
					"bind user: " + e.getLDAPErrorMessage() + "\n");
		}

		// Associate roles from group memberships.

		// Add default roles (if any) to list of roles
		String defaultRoles = (String)options.get("defaultRoles");
		if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
		{
			StringTokenizer st = new StringTokenizer(defaultRoles, ",");
			while(st.hasMoreTokens())
				roleList.add(st.nextElement().toString().trim());
		}

		// Check the required options when using a role search
		if (useRoleSearch)
		{
			String baseDN = (String)options.get("baseDN");
			if ((baseDN == null) || "".equals(baseDN.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'baseDN'");

			String filter = (String)options.get("filter");
			if ((filter == null) || "".equals(filter.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'filter'");

			String scope = (String)options.get("scope");
			if ((!"BASE".equalsIgnoreCase(scope)) &&
					(!"ONE".equalsIgnoreCase(scope)) &&
					(!"SUB".equalsIgnoreCase(scope)))
				throw new LoginException(LMN +
					" Error: Required option 'scope' must be 'BASE', 'ONE', or 'SUB'");

			// Set the search scope, the default of 0 is "BASE"
			int scopeInt = 0;
			if ("ONE".equalsIgnoreCase(scope))
				scopeInt = 1;
			else if ("SUB".equalsIgnoreCase(scope))
				scopeInt = 2;

			// Get and validate the attribute where roles are stored (the
			// "memberOf" groups attribute is often used but we'll enable
			// external configuration to supply an alternative attribute.
			String roleAttr = (String)options.get("roleAttr");
			if ((roleAttr == null) || "".equals(roleAttr.trim()))
				throw new LoginException(LMN + 
					" Error: Missing required option 'roleAttr'");

			try
			{
				// Set the search constraints to automatically follow referrals
				LDAPConstraints ldapConstraints = new LDAPConstraints();
				ldapConstraints.setReferralFollowing(true);

				// Create and set the referral handler
				LDAPReferralHandler referralHandler = new AuthImpl(bindUsername, bindPassword);
				ldapConstraints.setReferralHandler(referralHandler);

				// Set the search constraints
				connection.setConstraints(ldapConstraints);

				// Specify the attributes that the search will return
				String[] attrs = {roleAttr};

				// Substitute the user name in the role search filter. Optionally,
				// don't use the supplied domain in the user name to enable searches
				// using attributes other than the userPrincipalName, such as 
				// sAMAccountName, uid, or cn, which don't use the fully-
				// qualified UPN.
				if (useDomainInRoleSearch)
					filter = replaceUsername(filter, "username", fqUsername);
				else
					filter = replaceUsername(filter, "username", username);

				// If supplied, only use roles that match roleRegexPattern
				String roleRegexPattern = (String)options.get("roleRegexPattern");
				
				// The "tokenGroups" attribute contains all groups of which a
				// a user is member (including primary and nested groups). However,
				// it is only returned if you search with the user's
				// distinguishedName. So, we first search to get the user's
				// distinguishedName, then we do a read to get the "tokenGroups".
				// Because the "tokenGroups" are returned as SIDs, we then
				// use a Cams service to efficiently obtain their names.
				if ("tokenGroups".equals(attrs[0]))
				{
					// Get the user's distinguishedName
					attrs[0] = "distinguishedName";

					if (debug)
					{
						logger.debug(this, LMN + " Search " +
							"for the user's 'distinguishedName' using:" +
							"\n\tbaseDN: " + baseDN +
							"\n\tfilter: " + filter +
							"\n\t attrs: " + attrs[0]);
					}

					// Search for the user entry attributes
					LDAPSearchResults searchResults = connection.search(baseDN,
						scopeInt, filter, attrs, false);

					// Process the search results to get the user dn
					if (searchResults.hasMore())
					{
						LDAPEntry entry = searchResults.next();
						LDAPAttribute attribute = entry.getAttribute(attrs[0]);
						if (attribute != null)
							baseDN = attribute.getStringValue();
					}
					else
					{
						throw new LoginException ("Attribute 'distinguishedName' " +
							"not found for user '" + bindUsername + "'");
					}

					// Now that we have the user's dn, looking for tokenGroups
					attrs[0] = "tokenGroups";

					if (debug)
						logger.debug(this, LMN + " Reading " +
							"'tokenGroups' attribute for user '" + baseDN + "'");

					// Read the user's tokenGroups values
					LDAPEntry entry = connection.read(baseDN, attrs);

					// Save the Active Directory groups as roles for commit method
					if (entry != null)
					{
						LDAPAttribute attribute = entry.getAttribute(attrs[0]);

						// Get an enumeration of the attribute values
						Enumeration allValues = attribute.getByteValues();
						String role = null;
						String sid = null;

						if ((roleRegexPattern != null) && !("".equals(roleRegexPattern)))
						{
							Pattern pattern = Pattern.compile(roleRegexPattern);
							while (allValues.hasMoreElements())
							{
								// Get the sid
								sid = Sid2String((byte[]) allValues.nextElement());

								if (debug)
									logger.debug(this, LMN + " Search result SID: '" + sid + "'");
									
								role = getGroupName(sid);

								if (debug)
									logger.debug(this, LMN + " SID role name: '" + role + "'");

								if (pattern.matcher(role).matches())
								{
									this.roleList.add(role);
																			if (debug)
									logger.debug(this, LMN + 
										" Role: '" + role + 
										"' matched roleRegexPattern: '" + 
										roleRegexPattern + "'");
								}
								else
								{
									logger.debug(this, LMN + 
										" Role: '" + role + 
										"' did not match roleRegexPattern: '" + 
										roleRegexPattern + "'");
								}
							}
						}
						else
						{
							while (allValues.hasMoreElements())
							{
								// Get the sid
								sid = Sid2String((byte[]) allValues.nextElement());

								if (debug)
									logger.debug(this, LMN + " Search result SID: '" + sid + "'");

								role = getGroupName(sid);

								if (debug)
									logger.debug(this, LMN + " SID role name: '" + role + "'");

								this.roleList.add(role);
							}
						}
					}
				}

				// The role attribute is "memberOf", so we need to get the
				// get string values and explode them to get the group names
				else if ("memberOf".equals(attrs[0]))
				{
					if (debug)
					{
						logger.debug(this, LMN + " Search " +
							"for the user's Active Directory groups using:" +
							"\n\tbaseDN: " + baseDN +
							"\n\tfilter: " + filter +
							"\n\t attrs: " + attrs[0]);
					}					

					// Search for the user entry attributes
					LDAPSearchResults searchResults = connection.search(baseDN,
						scopeInt, filter, attrs, false);

					if (debug)
					{
						logger.debug(this, LMN + " After LDAP Search, searchResults=" + searchResults);
						
						if (searchResults != null)
							logger.debug(this, LMN + " After LDAP Search, searchResults.hasMore()=" +
									searchResults.hasMore());
					}

					// Get the roles and save them for commits
					if (searchResults.hasMore())
					{
						LDAPEntry entry = searchResults.next();
						LDAPAttribute attribute = entry.getAttribute(attrs[0]);
						if (attribute != null)
						{
							// Get an enumeration of the attribute values
							Enumeration allValues = attribute.getStringValues();
							String role = null;

							if ((roleRegexPattern != null) && !("".equals(roleRegexPattern)))
							{
								Pattern pattern = Pattern.compile(roleRegexPattern);
								while (allValues.hasMoreElements())
								{
									// Get the role dn to add
									role = (String)allValues.nextElement();

									if (debug)
										logger.debug(this, LMN + 
											" Exploding fully-qualifed role: '" + 
											role + "'");

									// Active Directory returns memberOf attributes as
									// fully qualified DNs, get the common name for each
									// by exploding the DN into parts
									String[] explodedDN = LDAPDN.explodeDN(role, true);

									role = explodedDN[0].trim();

									if (pattern.matcher(role).matches())
									{
										this.roleList.add(role);

										if (debug)
											logger.debug(this, LMN + 
												" Role: '" + role + 
												"' matched roleRegexPattern: '" + 
												roleRegexPattern + "'");
									}
									else
									{
										if (debug)
											logger.debug(this, LMN + 
												" Role: '" + role + 
												"' did not match roleRegexPattern: '" + 
												roleRegexPattern + "'");
									}
								}
							}
							else
							{
								while (allValues.hasMoreElements())
								{
									// Get the role dn to add
									role = (String)allValues.nextElement();

									if (debug)
										logger.debug(this, LMN + 
											" Exploding fully-qualifed role: '" + 
											role + "'");

									// Active Directory returns memberOf attributes as
									// fully qualified DNs, get the common name for each
									// by exploding the DN into parts
									String[] explodedDN = LDAPDN.explodeDN(role, true);

									role = explodedDN[0].trim();

									if (debug)
										logger.debug(this, LMN + 
											" Using role name: '" + role + "'");

									this.roleList.add(role);
								}
							}
						}
					}
					else
					{
						logger.debug(this, LMN + " " +
						" No search results returned");
					}
				}
				else
				{
					logger.error(this, LMN + " " +
						"Unsupported role search attribute, must be " +
						"'memberOf' or 'tokenGroups'");
				}
			}
			catch (LDAPReferralException e)
			{
				// Log any failed referrals.
				if (debug)
				{
					String[] referrals = e.getReferrals();
					for (int i = 0; i < referrals.length; i++)
					{
						logger.debug(this, LMN + " referral '" +
							referrals[i] + "' not found");
					}
				}
			}
			catch (LDAPException e)
			{
				if (debug)
					logger.debug(this, LMN + " An error '" +
						"occured while searching for groups for user '" +
						this.username + "'", e);

				// Search filter failed
				throw new LoginException(
					LMN + " Error: " +
						e.getMessage() + "\n");
			}
			finally
			{
				// Close the connection
				if (connection != null)
				{
					try
					{
						connection.disconnect();
						connection = null;
					}
					catch(LDAPException e)
					{
						// Only log the exception
						logger.error(this,
							LMN + " Error while disconnecting",
							e);
					}
				}
			}
		}

		if (debug)
			logger.debug(this, LMN + " Returning from authenticate(): loginSucceeded='" + loginSucceeded);
					
		// Return authentication results
		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @throws LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		if (debug)
			logger.debug(this, LMN + " Starting associateRules(): loginSucceeded=" + loginSucceeded);
		
		// The login failed, clean up state
		if (!loginSucceeded)
			return false;

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList();

		// Repository operations to get default role principals
		Iterator roleListItr = roleList.iterator();
		while(roleListItr.hasNext())
		{
			String rolename = (String)roleListItr.next().toString().trim();
			
			if (debug)
				logger.debug(this, LMN + " Creating CSRolePrincipal: rolename='" + rolename + "'");
		
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				if (debug)
					logger.debug(this,
						LMN + " Adding CSRolePrincipal " +
						rolename + " to principalList for: '" +
						username + "'");

				principalList.add(rolePrincipal);
			}
			else
			{
				if (debug)
					logger.debug(this,
						LMN + " CSRolePrincipal '" +
						rolename + "' already in principalList for '" +
						username + "', not added");
			}
		}

		if (debug)
			logger.debug(this, LMN + " Creating CSUserPrincipal: username='" + username + "'");
			
		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			if (debug)
				logger.debug(this,
					LMN + " Adding CSUserPrincipal: '" + username +
					"' to principalList");

			principalList.add(userPrincipal);
		}
		else if (debug)
		{
			logger.debug(this,
				LMN + " CSUserPrincipal: '" + username +
				"' already in principalList, not added");
		}

		logger.debug(this, LMN + " Adding principalList to Subject");

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		if(debug)
		{
			logger.debug(this, LMN + " Adding " +
				AuthenticationMethod.CAMS_AESO_LOGIN.getName() +
				" authentication method");
		}

		subject.getPublicCredentials().add(AuthenticationMethod.CAMS_AESO_LOGIN);

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.agentSecret = null;

		if (debug)
			logger.debug(this, LMN + " Returning from associateRules(): commitSucceeded='" + commitSucceeded);
		
		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @throws LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.
			this.loginSucceeded = false;
			this.username = null;
			this.agentSecret = null;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @throws LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if(this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if (this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.logger = null;
		this.callbackHandler = null;
		this.username = null;
		this.agentSecret = null;
		this.options = null;
		this.loginSucceeded = false;
		this.commitSucceeded = false;
		this.debug = false;

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username and
	 * password instance variables with their appropriate value.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// This LoginModule requires a username and an agent_secret
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Get the callback values.
			callbackHandler.handle(callbacks);

			// Get the user name
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");
			
			// Get the agent secret
			this.agentSecret = ((MapCallback)callbacks[0]).getCallbackValue("agent_secret");
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available",
				e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (debug)
		{
			logger.debug(this, LMN + " User entered username = " + username);
			logger.debug(this, LMN + " AESO agent secret = " + agentSecret);
		}
	}

	/**
	 * Get an Active Directory connection.
	 *
	 * @return a LDAPConnection.
	 * @exception LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getConnection()
		throws LoginException
	{
		// Get the connection values to use
		String host = (String)options.get("host");
		int port = Integer.parseInt((String)options.get("port"));
		int timeout = Integer.parseInt((String)options.get("connectTimeout"));

		// Initialize the connection
		LDAPConnection connection = null;

		try
		{
			if (debug)
				logger.debug(this, LMN + 
					" Instantiating LDAPConnection: host=" + host +
					", port=" + port + ", useSSL=" + useSSL +
					", connectTimeout=" + timeout);

			if (useSSL)
			{
				// Warn if required SSL System property for trust store is not set
				String trustStoreProp = System.getProperty("javax.net.ssl.trustStore");
				if (trustStoreProp == null)
					logger.warning(this, "Missing required SSL system property: " +
						"javax.net.ssl.trustStore");

				connection = new LDAPConnection((new LDAPJSSESecureSocketFactory())); 
			}
			else
			{
				connection = new LDAPConnection(); 
			}

			// Set the connection timeout
			connection.setSocketTimeOut(timeout);

			// Debug to help customers determine LDAP response time 
			long timer = System.currentTimeMillis(); 

			// Attempt connection
			connection.connect(host, port);

			if (debug)
			{
				logger.debug(this, LMN + "Time to connect: '" +
					(System.currentTimeMillis() - timer) + "'");

				if (connection.isConnected() && connection.isConnectionAlive())
					logger.debug(this, LMN + " LDAP connection succeeded");
				else
					logger.debug(this, LMN + " LDAP connection failed");

				logger.debug(this, LMN + " Returning LDAPConnection");
			}

			return connection;
		}
		catch (LDAPException e)
		{
			logger.error(this, "LDAP error", e);

			// Dump the nested Throwables to get at the root cause
			int i = 1;
			Throwable cause = e.getCause();
			do
			{
				if (cause != null)
					logger.error(this, "LDAP error: CAUSE " + i + ": " +
						cause.getMessage());
			}
			while ((cause = cause.getCause()) != null);

			throw new LoginException("[ActiveDirectoryLoginModule] Error: " +
				e.getMessage());
		}
	}

	/**
	 * Return the input pattern replaced with the match value.
	 *
	 * @param pattern The input pattern to use
	 * @param matchName The variable name to match
	 * @param matchValue The value to replace
	 * @return The input pattern with matched value
	 */
	private String replaceUsername(String pattern, String matchName,
		String matchValue)
	{
		StringBuffer sb = new StringBuffer();
		StringTokenizer st = new StringTokenizer(pattern, "{}");
		while (st.hasMoreTokens())
		{
			String tmp = st.nextToken();
			if (tmp.trim().equals(matchName))
				sb.append(matchValue);
			else
				sb.append(tmp);
		}

		return sb.toString();
	}

	/**
	 * Gets the ActiveDirectoryGroupNameService for the LoginModule.
	 *
	 * @throws LoginException if an error occurs getting the
	 * ActiveDirectoryGroupNameService object.
	 */
	private String getGroupName(String sid)
		throws LoginException
	{
		String groupName = null;
		try
		{
			if (serviceFinder == null)
				throw new LoginException(
					LMN + " Error: ServiceFinder is null");

			String serviceId = (String)options.get("serviceId");
			if ((serviceId == null) || (serviceId.trim().length() == 0))
				throw new LoginException(LMN + " Error: " +
					"Required 'serviceId' property missing");

			if (debug)
			{
				logger.debug(this, "Using serviceId '" + serviceId + "'");
				logger.debug(this, "Getting group id '" + sid + "'");
			}

			ActiveDirectoryGroupNameService adgns = (ActiveDirectoryGroupNameService)
				serviceFinder.find(serviceId, ActiveDirectoryGroupNameService.class);
			groupName = adgns.getGroupNameBySid(sid);

			if (debug)
				logger.debug(this, "Returned group name '" + groupName + "'");
		}
		catch(ServiceException e)
		{
			logger.error(this, "Error finding ActiveDirectoryGroupNameService", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}

		return groupName;
	}

	/**
	 * Convert the native byte representation of a SID into string format
	 *
	 * @param sid a byte representation of a SID
	 * @return a string formatted representation of the SID
	 */
	private static String Sid2String(byte[] SID)
	{
		long version;
		long authority;
		long count;
		long rid;
		String strSID;
		strSID = "S";
		version = SID[0];
		strSID = strSID + "-" + Long.toString(version);
		authority = SID[4];
		for (int i = 0;i<4;i++)
		{
			authority <<= 8;
			authority += SID[4+i] & 0xFF;
		}
		strSID = strSID + "-" + Long.toString(authority);
		count = SID[2];
		count <<= 8;
		count += SID[1] & 0xFF;
		for (int j=0;j<count;j++)
		{
			rid = SID[11 + (j*4)] & 0xFF;
			for (int k=1;k<4;k++)
			{
				rid <<= 8;
				rid += SID[11-k + (j*4)] & 0xFF;
			}
			strSID = strSID + "-" + Long.toString(rid);
		}
		return strSID;
	}

	/**
	 * A simple LDAPAuthHandler for following referrals
	 */
	private class AuthImpl implements LDAPAuthHandler
	{
		// The dn used to authenticate
		private String dn;

		// The password used to authenticate
		private byte[] passwd;

		/**
		 * Construct this instance using data
		 */
		private AuthImpl(String username, String password)
		{

			// Create an LDAPAuthProvider class to encapsulate the
			// authentication credentials.
			dn = username;
			try
			{
				passwd = password.getBytes("UTF8");
			}
			catch (UnsupportedEncodingException e)
			{
				// Should never get this
				logger.error(this,
					"Unable to encode password to UTF-8 for user: " + dn);
			}

			return;
		}

		/**
		 * Choose the login dn and password based on the host and port
		 * (this implementation simply reuses the original credentials)
		 */
		public LDAPAuthProvider getAuthProvider(String host, int port)
		{
			if (debug)
				logger.debug(this, "Following directory referral, rebinding " +
					"to: " + host + ":" + port);

			return new LDAPAuthProvider(dn, passwd);
		}
	}
} // End of class : AESOActiveDirectoryLoginModule

