/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 * OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
 * BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
 * OR ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.core.jdbc.JdbcUtils;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.service.ServiceClient;
import com.cafesoft.core.service.ServiceFinder;

import java.security.Principal;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * <code>AESOAESOJdbcLoginModule</code> provides login support for Cams
 * Automatic Enterprise Sign-On (AESO) against relational databases
 * accessible via JDBC.
 * <P>
 * All configuration parameters are listed below:
 * <Ul>
 * <LI><B>jdbcDriver</B> - the JDBC Driver
 * <LI><B>jdbcUrl</B> - the database JDBC URL
 * <LI><B>jdbcUsername</B> - the database username
 * <LI><B>jdbcPassword</B> - the user's password on the database server
 * <li><b>agentSecret</b> - The expected value for the "agent_secret" callback
 * value, which must be present in the authentication request. This value is 
 * used to authenticated Cams agents, which assert the identity of a user
 * already authenticated at the web layer. The agent secret is used to protect 
 * against impersonation attacks via carefully created HTTP requests 
 * attempting to subvert AESO component trust relationships.</li>
 * <LI><B>userPreparedStatement</B> - the prepared statement to execute to
 * lookup/validate the user
 * <LI><B>usernameColumn</B> - the database column name for the user's
 * username field
 * <LI><B>accountExpiredPreparedStatement</B> - (optional) the prepared
 * statement to determine if the account has expired. Prepared statement
 * must return a count of "1" if the account has expired, "0" if it has not.
 * This prepared statement is executed conditionally after the password has
 * been validated.
 * <LI><B>defaultRoles</B> - (Optional) A list of one or more default roles
 * separated by a comma to be added to the user session.
 * <LI><B>rolePreparedStatement</B> - the prepared statement to execute to
 * obtain the user's roles
 * <LI><B>roleColumn</B> - the database column name for the role field
 * </UL>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.4 $ $Date: 2009/08/14 23:49:51 $ $Author: norb $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 11/20/2008
 */
public class AESOJdbcLoginModule
	implements LoginModule, LoggerClient
{
	/**
	 * The LoginModule name (used in log messages throughout the code).
	 */
	private static final String LMN = "[AESOJdbcLoginModule]";

	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating User.
	 */
	private String username;

	/**
	 * The agent secret (used to protect against impersonation attacks).
	 */
	private String agentSecret;

	//
	// Constructor
	//

	/**
	 * Default AESOJdbcLoginModule Constructor.
	 */
	public AESOJdbcLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.subject = null;
		this.username = null;
		this.agentSecret = null;
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *			with the end user (prompting for usernames and
	 *			passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.debug = "true".equals((String)options.get("debug"));
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @throws LoginException thrown if any error occurs while trying to
	 * authenticate the user.
	 * @throws FailedLoginException if authentication fails.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException,
		AccountExpiredException, CredentialExpiredException
	{
		String expectedAgentSecret = (String)options.get("agentSecret");
		if ((expectedAgentSecret == null) || (expectedAgentSecret.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'agentSecret' property missing");

		handleCallbacks();

		// An agent secret must always be supplied as a callback value and
		// its value must match the expected value.
		if (agentSecret == null || agentSecret.length() == 0)
			throw new LoginException(LMN + " Error: Required callback value 'agent_secret' " +
				"does not exist or is empty");
		else if (agentSecret.equals(expectedAgentSecret) == false)
			throw new LoginException(LMN + " Error: Callback value 'agent_secret' " +
				"does not match expected value in option 'agentSecret'");

		// A user name must always be supplied
		if (username == null || username.length() == 0)
			throw new FailedLoginException(
				LMN + " Authentication failed");

		// Get JdbcDriver
		String jdbcDriver = (String)options.get("jdbcDriver");
		if ((jdbcDriver == null) || "".equals(jdbcDriver.trim()))
			throw new LoginException(LMN +
				" Error: Missing required option 'jdbcDriver'");

		// Get the JdbcUrl
		String jdbcUrl = (String)options.get("jdbcUrl");
		if ((jdbcUrl == null) || "".equals(jdbcUrl.trim()))
			throw new LoginException(LMN +
				" Error: Missing required option 'jdbcUrl'");

		// Get the JdbcUsername and JdbcPassword (empty may be valid)
		String jdbcUsername  = (String)options.get("jdbcUsername");
		String jdbcPassword  = (String)options.get("jdbcPassword");

		// Get UserPreparedState used to query the username
		String userPreparedStatement =
			(String)options.get("userPreparedStatement");

		if ((userPreparedStatement == null) ||
			"".equals(userPreparedStatement.trim()))
		{
			throw new LoginException(LMN +
				" Error: Missing required option 'userPreparedStatement'");
		}

		// Get PasswordColumn (the name of the password result set column)
		String usernameColumn =	(String)options.get("usernameColumn");
		if ((usernameColumn == null) || "".equals(usernameColumn.trim()))
			throw new LoginException(LMN +
				" Error: Missing required option 'usernameColumn'");

		Connection c = null;
		PreparedStatement s = null;
		ResultSet r = null;

		try
		{
			// Load the database driver
			JdbcUtils.loadDriver(jdbcDriver);

			// Get a JDBC Connection
			c = DriverManager.getConnection(
				jdbcUrl, jdbcUsername, jdbcPassword);

			// Query the username/password
			s = c.prepareStatement(userPreparedStatement);
			s.setString(1, username);
			r = s.executeQuery();

			// Check the password
			if (r != null)
			{
				// there should only be one row in the result set as a valid
				// username is assumed to be unique
				if (r.next())
				{
					String queriedUsername = r.getString(usernameColumn);

					// Fail the login for null or empty password.
					if ((queriedUsername == null) ||
						(queriedUsername.length() < 1))
					{
						if (debug)
						{
							logger.debug(this, LMN + " '" +
								username + "' does not exist");
						}
					}
					else
					{
						// Set the login succeeded flag by comparing the usernames.
						// NOTE: Some databases may return a case-insensitive username.
						// This test ensures the username case matches.
						loginSucceeded = queriedUsername.equals(username);
					}
				}

				// No user by that name, FailedLoginException thrown later
				else
				{
					if (debug)
					{
						logger.debug(this, LMN + " User '" +
							username + "' does not exist");
					}
				}
			}
			else
			{
				if (debug)
				{
					// Null result set, a FailedLoginException will be
					// thrown later
					logger.debug(this,
						LMN + " Result set null for query '" +
						s.toString() + "'");
				}
			}

			// Optionally check for account or credential expired states
			if (loginSucceeded)
			{
				// Throw an AccountExpiredException if account has expired
				String accountExpiredPreparedStatement =
					(String)options.get("accountExpiredPreparedStatement");

				if ((accountExpiredPreparedStatement != null) &&
					!("".equals(accountExpiredPreparedStatement.trim())))
				{
					s = c.prepareStatement(accountExpiredPreparedStatement);
					s.setString(1, username);
					r = s.executeQuery();

					if (r != null)
					{
						// Only one row should be returned
						if (r.next())
						{
							// Expecting "1" (expired) or "0" (not expired)
							if (r.getInt(1) == 1)
							{
								throw new AccountExpiredException(
									LMN + " Error: Authentication " +
									"failed, account expired");
							}
						}
					}
				}
			}

		}
		catch (SQLException e)
		{
			// Log the SQL error for debug
			if (debug)
				JdbcUtils.err(e);

			throw new LoginException(
				LMN + " Error: " + e.getMessage());
		}
		finally
		{
			JdbcUtils.close(c,s,r);
		}

		// Authentication failed, clean up state and throw a FailedLoginException
		if (!loginSucceeded)
		{
			this.username = null;
			this.agentSecret = null;
			throw new FailedLoginException(
				LMN + " Authentication failed");
		}

		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @throws LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login method failed, clean up
		if (!loginSucceeded)
			return false;

		// Add all principals to an ArrayList, then commit in single transaction
		List principalList = new ArrayList(5);

		// Add default roles (if any) to list of roles
		String defaultRoles = (String)options.get("defaultRoles");
		if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
		{
			StringTokenizer st = new StringTokenizer(defaultRoles, ",");
			while(st.hasMoreTokens())
				principalList.add(
					new CSRolePrincipal(st.nextElement().toString().trim()));
		}

		// Get the JDBC connection parameters (values validated in login method)
		String jdbcDriver = (String)options.get("jdbcDriver");
		String jdbcUrl = (String)options.get("jdbcUrl");
		String jdbcUsername  = (String)options.get("jdbcUsername");
		String jdbcPassword  = (String)options.get("jdbcPassword");

		// Get roles prepared statement used to query the user's role names
		String rolePreparedStatement =
			(String)options.get("rolePreparedStatement");

		if ((rolePreparedStatement == null) ||
			"".equals(rolePreparedStatement.trim()))
		{
			throw new LoginException(
				LMN + " Missing required option: rolePreparedStatement");
		}

		// Get role column (the column name in the result set for the password)
		String roleColumn = (String)options.get("roleColumn");
		if ((roleColumn == null) || "".equals(roleColumn.trim()))
			throw new LoginException(
				LMN + " Missing required option: roleColumn");

		// Get the roles for the authenticated user

		Connection c = null;
		PreparedStatement s = null;
		ResultSet r = null;

		try
		{
			// Load the database driver
			JdbcUtils.loadDriver(jdbcDriver);

			// Get the JDBC connection
			c = DriverManager.getConnection(
				jdbcUrl, jdbcUsername, jdbcPassword);

			// Query the user's roles
			s = c.prepareStatement(rolePreparedStatement);
			s.setString(1, username);
			r = s.executeQuery();

			// Get the roles from the result set
			if (r != null)
			{
				while (r.next())
				{
					String rolename = r.getString(roleColumn);
					if ((rolename != null) && ("".equals(rolename) == false))
						rolename = rolename.trim();
					else
					{
						if (debug)
							logger.debug(this,
								LMN + " Role query failed: " +
								"null returned from result set column '" +
								roleColumn + "'");

						throw new LoginException("Role assignment failed: " +
							"null or empty role name in column: " + roleColumn);
					}

					// Create the role Principal
					CSRolePrincipal rolePrincipal = new CSRolePrincipal(rolename);

					// Add the CSRolePrincipal if it does not already exist
					if (!subject.getPrincipals(rolePrincipal.getClass()).contains(rolePrincipal))
					{
						principalList.add(rolePrincipal);

						// Log that the CSRolePrincipal that was added
						if (debug)
						{
							logger.debug(this,
								LMN + " Added CSRolePrincipal " +
								rolename + " to principal ArrayList for user '" +
								username + "'");
						}
					}
					else
					{
						if(debug)
						{
							// Log the CSRolePrincipal was not added because it
							// already exists
							logger.debug(this,
								LMN + " CSRolePrincipal '" + username +
								"' already in Subject for '" + username +
								"', not added to " + "principal List");
						}
					}
				}
			}
		}
		catch (SQLException e)
		{
			// Log the SQL error
			if (debug)
				JdbcUtils.err(e);

			throw new LoginException(e.getMessage());
		}
		finally
		{
			JdbcUtils.close(c,s,r);
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		CSUserPrincipal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(userPrincipal.getClass()).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if(debug)
			{
				logger.debug(this, LMN + " Added CSUserPrincipal '" +
					username + "' to principal List");
			}
		}
		else
		{
			if(debug)
			{
				logger.debug(
					this, LMN + " CSUserPrincipal '" + username +
					"' already in Subject, not added to principal List");
			}
		}

		// Add all new principals to Subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		if (debug)
		{
			logger.debug(this, LMN + " Adding " +
				AuthenticationMethod.PASSWORD.getName() +
				" AuthenticationMethod");
		}

		subject.getPublicCredentials().add(
			AuthenticationMethod.PASSWORD);

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.agentSecret = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @throws LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// If authentication failed return false. If commit
		// succeeded log the user out return true. Otherwise,
		// Cleanup the login module and and return true.

		if (!loginSucceeded)
			return false;

		if (commitSucceeded)
			logout();
		else
			cleanup();

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @throws LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if(subject != null)
			subject.getPrincipals().clear();

		cleanup();

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username and
	 * password instance variables with their appropriate value.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// Check for a callbackHandler
		if (callbackHandler == null)
			throw new LoginException(
				LMN + " Error: no CallbackHandler available");

		// This LoginModule requires a username and a password
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Prompt the user for a username and password
			callbackHandler.handle(callbacks);

			// Save the user name
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");

			// Get the agent secret
			this.agentSecret = ((MapCallback)callbacks[0]).getCallbackValue("agent_secret");
		}
		catch(IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch(UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available",
				e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (debug)
		{
			logger.debug(this, LMN + " User entered username = " + username);
			logger.debug(this, LMN + " AESO agent secret = " + agentSecret);
		}
	}

	/**
	 * Clean up the resources of the LoginModule
	 */
	private void cleanup()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.logger = null;
		this.loginSucceeded = false;
		this.options = null;
		this.subject = null;
		this.username = null;
		this.agentSecret = null;
	}

} // End of class : AESOJdbcLoginModule
