/**
 * Copyright (c) 1996-2012 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-
 * INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR
 * ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.novell.ldap.LDAPAttribute;
import com.novell.ldap.LDAPAuthHandler;
import com.novell.ldap.LDAPAuthProvider;
import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPConstraints;
import com.novell.ldap.LDAPReferralHandler;
import com.novell.ldap.LDAPReferralException;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;
import com.novell.ldap.LDAPSearchResults;
import com.novell.ldap.LDAPSocketFactory;
import com.novell.ldap.LDAPJSSESecureSocketFactory;

import java.security.Principal;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;
import java.io.UnsupportedEncodingException;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * <p><code>AESOLdapLoginModule</code> is a JAAS LoginModule used
 * with Cams "Automatic Enterprise Sign-On" (AESO) and LDAP v3 compliant
 * directory servers.</p>
 * <p>This LoginModule is designed to trust a user identity authenticated at the 
 * web layer, most likely via Kerberos, NTLM, HTTP Basic or HTTP Digest
 * authentication methods. The user is authenticated within the Cams security
 * context login succeeds within this LoginModule.<p>
 * <UL>
 * <li><b>debug</b> - Enable/disable debugging [true|false].
 * <li><b>host</b> - The DNS name or IP address of the Active Directory
 * host.</li>
 * <li><b>port</b> - The host's Active Directory port to use.</li>
 * <li><b>useSSL</b> - Enables/disables use of SSL.</li>
 * <li><b>connectTimeout</b> - (Optional) The number of milliseconds to wait for
 * 	a connection to the ActiveDirectory before timing out.</li>
 * <li><b>bindUsername</b> - The username for a Windows account with permissions
 * to query Active Directory for user and group information.</li>
 * <li><b>bindPassword</b> - The password for the "bindUser" account.</li>
 * <li><b>agentSecret</b> - The expected value for the "agent_secret" callback
 * value, which must be present in the authentication request. This value is 
 * used to authenticated Cams agents, which assert the identity of a user
 * already authenticated at the web layer. The agent secret is used to protect 
 * against impersonation attacks via carefully created HTTP requests 
 * attempting to subvert AESO component trust relationships.</li>
 * <li><b>defaultRoles</b> - (Optional) A list of one or more default roles
 * separated by a comma to be added to the user session.</li>
 * <LI><B>userBaseDN</B> - The distinguished name pattern to match to find a
 * user. The username entered by the the user at login will be substituted
 * for the value "{username}". The LDAP login module will not successfully
 * find users in schemas that use a distinguished name that does not use
 * the username to distinguish users.
 * <LI><B>defaultRoles</B> - (Optional) A list of one or more default roles
 * separated by a comma to be added to the user session.
 * <LI><B>roleBaseDN</B> - The base distinguished name to use for your
 * LDAP schema role searches.
 * <LI><B>roleScope</B> - Specifies the LDAP search scope to use for
 * role searches. Valid values are BASE (search is done only at the base
 * dn tree node), ONE (search is done only at the first child tree node
 * of the base dn), and SUB (search is done at the base dn node and all
 * child nodes).
 * <LI><B>roleAttrName</B> - The name of the attribute used to store the
 * role names (usually one name is used but can be a comma-delimited list).
 * <LI><B>roleFilter</B> - The filter to use to search for a users roles.
 * The username entered by the the user at login will be substituted for
 * the value "{username}".
 * </Ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.6 $ $Date: 2012/08/22 22:05:11 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 12/01/2008
 */
public class AESOLdapLoginModule
	implements LoginModule, LoggerClient
{
	//
	// static constants
	//

	/**
	 * The LoginModule name (used in log messages throughout the code).
	 */
	private static final String LMN = "[AESOLdapLoginModule]";
	
	/**
	 * A tag used to identify a generic LDAP server type.
	 */
	private static final String GENERIC_LDAP_SERVER_TYPE = "GENERIC";
	
	/**
	 * A tag used to identify an Active Directory LDAP server type.
	 */
	private static final String ACTIVE_DIRECTORY_LDAP_SERVER_TYPE = "ACTIVE_DIRECTORY";
	
	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag indicating that SSL must be used.
	 */
	private boolean useSSL;

	/**
	 * Flag indicating that the LDAP directory is to be searched for roles.
	 */
	private boolean useRoleSearch;
	/**
	 * The LDAP server type, which is useful for customizing error handling.
	 */
	private String ldapServerType = GENERIC_LDAP_SERVER_TYPE;
	
	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating User.
	 */
	private String username;

	/**
	 * A list of user roles. This list is populated in the login method and
	 * used in the commit method.
	 */
	private List roleList;

	/**
	 * The agent secret (used to protect against impersonation attacks).
	 */
	private String agentSecret;

	//
	// Constructor
	//

	/**
	 * Default AESOLdapLoginModule Constructor.
	 */
	public AESOLdapLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.useSSL = false;
		this.useRoleSearch = true;
		this.ldapServerType = GENERIC_LDAP_SERVER_TYPE;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.agentSecret = null;
		this.subject = null;
		this.username = null;
		this.roleList = new ArrayList();
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for accessing
	 * 	callback values (e.g. username).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.debug = "true".equals((String)options.get("debug"));
		this.useSSL = "true".equals((String)options.get("useSSL"));
		this.useRoleSearch = 
			! "false".equalsIgnoreCase((String)options.get("useRoleSearch"));
		
		// Set the LDAP server type
		if (ACTIVE_DIRECTORY_LDAP_SERVER_TYPE.equals(options.get("ldapServerType")))
			this.ldapServerType = ACTIVE_DIRECTORY_LDAP_SERVER_TYPE;
		else
			this.ldapServerType = GENERIC_LDAP_SERVER_TYPE;
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 * <p>
	 * This method executes the following steps:
	 * <ol>
	 * <li>Checks preconditions
	 * <li>Handles callback objects
	 * <li>Get a connection to the LDAP repository
	 * <li>Authenticates the users by attepting and LDAP bind with the
	 * username and password
	 * <li>Return the connection to its ObjectPool
	 * <li>Reset state if authentication failed
	 * <li>Return results of authentication attempt
	 * </ol>
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @exception LoginException thrown if any error occurs while trying to
	 *	authenticate the user.
	 * @exception FailedLoginException if authentication fails.
	 * @exception AccountExpiredException if the user account is expired.
	 * @exception CredentialExpiredException if the user credential is expired.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException,
			AccountExpiredException, CredentialExpiredException
	{
		// Check preconditions
		if (callbackHandler == null)
			throw new LoginException(LMN + " Error: no CallbackHandler available");

		String host = (String)options.get("host");
		if ((host == null) || (host.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required 'host' property missing");

		String port = (String)options.get("port");
		if ((port == null) || (port.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required 'port' property missing");

		try
		{
			int portInt = Integer.parseInt(port);
			if (portInt < 1)
				throw new LoginException(LMN + " Error: " +
					"'port' property must be an integer value greater than zero");
		}
		catch (NumberFormatException e)
		{
			throw new LoginException(LMN + " Error: " +
				"'port' property is not an integer value");
		}

		String bindUsername = (String)options.get("bindUsername");
		if ((bindUsername == null) || (bindUsername.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'bindUsername' property missing");

		String bindPassword = (String)options.get("bindPassword");
		if ((bindPassword == null) || (bindPassword.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'bindPassword' property missing");

		String expectedAgentSecret = (String)options.get("agentSecret");
		if ((expectedAgentSecret == null) || (expectedAgentSecret.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'agentSecret' property missing");

		// Handle callback objects.
		handleCallbacks();

		// An agent secret must always be supplied as a callback value and
		// its value must match the expected value.
		if (agentSecret == null || agentSecret.length() == 0)
			throw new LoginException(LMN + " Error: Required callback value 'agent_secret' " +
				"does not exist or is empty");
		else if (agentSecret.equals(expectedAgentSecret) == false)
			throw new LoginException(LMN + " Error: Callback value 'agent_secret' " +
				"does not match expected value in option 'agentSecret'");

		// A username must always be supplied
		if (username == null || username.length() == 0)
			throw new FailedLoginException(
				LMN + " Authentication failed");

		// Get a connection to the LDAP repository
		LDAPConnection connection = getConnection(host, port);

		if (debug)
			logger.debug(this, LMN + " Successfully connected to LDAP Server");

		try
		{
			// Bind to LDAP Server using credentials with "admin" permissions.
			connection.bind(3, bindUsername, bindPassword.getBytes());
		}
		catch (LDAPException e)
		{
			if (debug)
				logger.debug(this, LMN + " An error occured while trying to bind user: '" + 
					bindUsername + "'" +
					"\n\tMessage: " + e.getMessage() +
					"\n\tLDAPErrorMessage: " + e.getLDAPErrorMessage() +
					"\n\tResultCode: " + e.getResultCode(), e);

			// Login failure during LDAP bind.
			throw new LoginException(LMN + " Error while trying to " +
					"bind user: " + e.getLDAPErrorMessage() + "\n");
		}

		if (debug)
			logger.debug(this, LMN + " Successful " +
				"bind to LDAP Server as user '" + bindUsername + "'");

		try
		{
			String userBaseDN = (String)options.get("userBaseDN");
			if ((userBaseDN == null) || "".equals(userBaseDN.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'userBaseDN'");
			
			String userFilter = (String)options.get("userFilter");
			if ((userFilter == null) || "".equals(userFilter.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'userFilter'");

			// Get the user search base DN, with possible substitution of username.
			String baseDN = replaceUsername(userBaseDN, "username", username);
			
			// Get the user search filter, with possible substitution of username.
			String filter = replaceUsername(userFilter, "username", username);

			String userScope = (String)options.get("userScope");
			if ((!"BASE".equalsIgnoreCase(userScope)) &&
					(!"ONE".equalsIgnoreCase(userScope)) &&
					(!"SUB".equalsIgnoreCase(userScope)))
				throw new LoginException(LMN + " " +
					"Error: Required option 'userScope' must be 'BASE', 'ONE', or 'SUB'");

			// Set the search scope, the default of 0 is "BASE"
			int scopeInt = 0;
			if ("ONE".equalsIgnoreCase(userScope))
				scopeInt = 1;
			else if ("SUB".equalsIgnoreCase(userScope))
				scopeInt = 2;
			
			String userAttrName = (String)options.get("userAttrName");
			if ((userAttrName == null) || "".equals(userAttrName.trim()))
			throw new LoginException(LMN + " " +
				"Error: Required option 'userAttrName' is null or empty");

			// Set the search constraints to automatically follow referrals
			LDAPConstraints ldapConstraints = new LDAPConstraints();
			ldapConstraints.setReferralFollowing(true);

			// Create and set the referral handler
			LDAPReferralHandler referralHandler = new AuthImpl(bindUsername, bindPassword);
			ldapConstraints.setReferralHandler(referralHandler);

			// Set the search constraints
			connection.setConstraints(ldapConstraints);

			// Specify the attributes that the search will return
			String[] attrs = {userAttrName};

			if (debug)
				logger.debug(this, LMN + " Searching for user DN entry using filter: " +
					baseDN);

			// Search for the user entry attributes
			LDAPSearchResults searchResults = connection.search(baseDN,
				scopeInt, filter, attrs, false);

			// Process the search results to get the user dn
			if (searchResults.hasMore())
			{
				LDAPEntry entry = searchResults.next();
				if (entry != null)
				{
					if (debug)
						logger.debug(this, LMN + " login() succeeded");
					
					// Found the userPrincipalName, so login succeeded.
					loginSucceeded = true;
				}
				else
				{
					if (debug)
						logger.debug(this, LMN + " login() failed");
								
					throw new FailedLoginException(
						LMN + " Authentication failed");
				}
			}
			else
			{
				if (debug)
					logger.debug(this, LMN + " LDAPEntry '" + filter + "' " +
						"not found for user '" + username + "'");
			
				throw new LoginException ("Attribute '" + filter + "' " +
					"not found for user '" + username + "'");
			}
		}
		catch (LDAPException e)
		{
			if (debug)
				logger.debug(this, LMN + " An error '" +
					"occured while trying to search for userBaseDN for user: '" + bindUsername + "'" +
					"\n\tMessage: " + e.getMessage() +
					"\n\tLDAPErrorMessage: " + e.getLDAPErrorMessage() +
					"\n\tResultCode: " + e.getResultCode(), e);

			// Login failure during LDAP bind.
			throw new LoginException(
				LMN + " Error while trying to " +
					"authenticate user: " + e.getLDAPErrorMessage() + "\n");
		}


		// Add default roles (if any) to list of roles
		String defaultRoles = (String)options.get("defaultRoles");
		if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
		{
			StringTokenizer st = new StringTokenizer(defaultRoles, ",");
			while(st.hasMoreTokens())
				roleList.add(st.nextElement().toString().trim());
		}

		// If configured, search for roles stored in the LDAP directory.
		if (useRoleSearch)
		{
			String roleBaseDN = (String)options.get("roleBaseDN");
			if ((roleBaseDN == null) || "".equals(roleBaseDN.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'roleBaseDN'");

			String roleFilter = (String)options.get("roleFilter");
			if ((roleFilter == null) || "".equals(roleFilter.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'roleFilter'");

			String roleAttrName = (String)options.get("roleAttrName");
			if ((roleAttrName == null) || "".equals(roleAttrName.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'roleAttrName'");

			String roleScope = (String)options.get("roleScope");
			if ((!"BASE".equalsIgnoreCase(roleScope)) &&
					(!"ONE".equalsIgnoreCase(roleScope)) &&
					(!"SUB".equalsIgnoreCase(roleScope)))
				throw new LoginException(LMN +
					" Error: Required option 'roleScope' must be 'BASE', 'ONE', or 'SUB'");

			// Set the search scope, the default of 0 is "BASE"
			int scopeInt = 0;
			if ("ONE".equalsIgnoreCase(roleScope))
				scopeInt = 1;
			else if ("SUB".equalsIgnoreCase(roleScope))
				scopeInt = 2;

			try
			{
				// Set the search constraints to automatically follow referrals
				LDAPConstraints ldapConstraints = new LDAPConstraints();
				ldapConstraints.setReferralFollowing(true);

				// Create and set the referral handler
				LDAPReferralHandler referralHandler = new AuthImpl(bindUsername, bindPassword);
				ldapConstraints.setReferralHandler(referralHandler);

				// Set the search constraints
				connection.setConstraints(ldapConstraints);

				// Substitute the user name in the role search filter. Optionally,
				// don't use the supplied domain in the user name to enable searches
				// using attributes other than the userPrincipalName, such as 
				// sAMAccountName, uid, or cn, which don't use the fully-
				// qualified UPN.
				String filter = replaceUsername(roleFilter, "username", username);

				String[] attrs = getStringArray(roleAttrName, ",");
				
				if (debug)
				{
					logger.debug(this, LMN + " Search " +
						"for the user's LDAP server groups using:" +
						"\n\tbaseDN: " + roleBaseDN +
						"\n\tfilter: " + filter +
						"\n\t attrs: " + attrs[0]);
				}

				// Search for the user entry attributes
				LDAPSearchResults searchResults = connection.search(roleBaseDN,
					scopeInt, filter, attrs, false);
					
				if (debug)
				{
					logger.debug(this, LMN + " After LDAP Search, searchResults=" + searchResults);
					
					if (searchResults != null)
						logger.debug(this, LMN + " After LDAP Search, searchResults.hasMore()=" +
							searchResults.hasMore());
				}

				// Get the roles and save them for commits
				if (searchResults.hasMore())
				{
					LDAPEntry entry = searchResults.next();
					LDAPAttribute attribute = entry.getAttribute(attrs[0]);
					if (attribute != null)
					{
						// Get an enumeration of the attribute values
						Enumeration allValues = attribute.getStringValues();

						while (allValues.hasMoreElements())
						{
							// Get the role dn to add
							String role = (String)allValues.nextElement();

							if (debug)
								logger.debug(this, LMN + " " +
									"unexploded role '" + role +
									"' will be exploded");

							// Add the left most element of the role dn
							this.roleList.add(role);

							if (debug)
								logger.debug(this, LMN + " " +
									 "role '" + role + "' added to list");
						}
					}
				}
				else
				{
					logger.debug(this, LMN + " " +
					"Search results were null");
				}
			}
			catch (LDAPReferralException e)
			{
				// Log any failed referrals.
				if (debug)
				{
					String[] referrals = e.getReferrals();
					for (int i = 0; i < referrals.length; i++)
					{
						logger.debug(this, LMN + " referral '" +
							referrals[i] + "' not found");
					}
				}
			}
			catch (LDAPException e)
			{
				if (debug)
					logger.debug(this, LMN + " An error '" +
						"occured while searching for groups for user '" +
						this.username + "'", e);

				// Search filter failed
				throw new LoginException(
					LMN + " Error: " +
						e.getMessage() + "\n");
			}
			finally
			{
				// Close the connection
				if (connection != null)
				{
					try
					{
						connection.disconnect();
						connection = null;
					}
					catch(LDAPException e)
					{
						// Only log the exception
						logger.error(this,
							LMN + " Error while disconnecting",
							e);
					}
				}
			}
		}

		if (debug)
			logger.debug(this, LMN + " Returning from authenticate(): loginSucceeded='" + loginSucceeded);
					
		// Return authentication results
		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @exception LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login failed, clean up state
		if (!loginSucceeded)
			return false;

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList();

		// Repository operations to get role principals
		Iterator roleListItr = roleList.iterator();
		while(roleListItr.hasNext())
		{
			String rolename = (String)roleListItr.next().toString().trim();
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				principalList.add(rolePrincipal);

				if (debug)
					logger.debug(this,
						LMN + " Added CSRolePrincipal " +
						rolename + " to principal ArrayList for user '" +
						username + "'");
			}
			else
			{
				if (debug)
					logger.debug(this,
						LMN + " CSRolePrincipal '" +
						rolename + "' already in Subject for '" +
						username + "', not added to principal List");
			}
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if (debug)
				logger.debug(this,
					LMN + " Added CSUserPrincipal '" + username +
					"' to principal List");
		}
		else
		{
			if (debug)
				logger.debug(this,
					LMN + " CSUserPrincipal '" + username +
					"' already in Subject, not added to principal List");
		}

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the AESO variety.

		if(debug)
		{
			logger.debug(this, LMN + " Adding " +
				AuthenticationMethod.CAMS_AESO_LOGIN.getName() +
				" AuthenticationMethod");
		}

		subject.getPublicCredentials().add(
			AuthenticationMethod.CAMS_AESO_LOGIN.getName());

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.agentSecret = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @exception LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.
			if (this.roleList != null)
			{
				this.roleList.clear();
				this.roleList = null;
			}

			this.loginSucceeded = false;
			this.username = null;
			this.agentSecret = null;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @exception LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if (this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if (this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.logger = null;
		this.callbackHandler = null;
		this.username = null;
		this.agentSecret = null;
		this.options = null;
		this.loginSucceeded = false;
		this.commitSucceeded = false;
		this.debug = false;

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username
	 * instance variable.
	 *
	 * @exception LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException, FailedLoginException
	{
		// This LoginModule requires a username and an agent_secret
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Populate callback values.
			callbackHandler.handle(callbacks);

			// Get the user name
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");
			
			// Get the agent secret
			this.agentSecret = ((MapCallback)callbacks[0]).getCallbackValue("agent_secret");
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available",
				e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (debug)
		{
			logger.debug(this, LMN + " User entered username = " + username);
			logger.debug(this, LMN + " AESO agent secret = " + agentSecret);
		}
	}

	/**
	 * Validate the LoginModules options Map object to ensure that all
	 * required properties exist and contain valid values.
	 *
	 * @exception LoginException if any of the following conditions occur:
	 * <UL>
	 * <LI>A CallbackHandler object is not available,
	 * <LI>the 'host' property is not available,
	 * <LI>the 'port' property is not available,
	 * <LI>the 'port' property is not a valid integer value,
	 * <LI>the port integer value is less than 1,
	 * <LI>the 'userBaseDN' property is not available,
	 * <LI>the 'roleBaseDN' property is not available,
	 * <LI>the 'roleAttrName' property is not available,
	 * <LI>the 'roleFilter' property is not available,
	 * <LI>the 'roleScope' property is not available,
	 * <LI>the 'roleScope' is not a valid role scope value.
	 * </UL>
	 */
	private void validateOptions() throws LoginException
	{
		if (callbackHandler == null)
			throw new LoginException(
				LMN + " Error: no CallbackHandler available");

		String host = (String)options.get("host");
		if ((host == null) || (host.trim().length() == 0))
			throw new LoginException(
				LMN + " Error: Required 'host' property missing");

		String port = (String)options.get("port");
		if ((port == null) || (port.trim().length() == 0))
			throw new LoginException(
				LMN + " Error: Required 'port' property missing");

		try
		{
			int portInt = Integer.parseInt(port);
			if (portInt < 1)
				throw new LoginException(LMN + " Error: 'port' " +
					"property must be an integer value greater than zero");
		}
		catch (NumberFormatException e)
		{
			throw new LoginException(LMN + " Error: " +
				"'port'property is not an integer value");
		}

		String userBaseDN = (String)options.get("userBaseDN");
		if ((userBaseDN == null) || "".equals(userBaseDN.trim()))
			throw new LoginException(
				LMN + " Error: Required 'userBaseDN' property missing");

		// Get and validate the roleBaseDN
		String roleBaseDN = (String)options.get("roleBaseDN");
		if ((roleBaseDN == null) || "".equals(roleBaseDN.trim()))
			throw new LoginException(LMN + " Error: Required " +
				"'roleBaseDN' property missing");

		// Get and validate the roleAttrName
		String roleAttrName = (String)options.get("roleAttrName");
		if ((roleAttrName == null) || "".equals(roleAttrName.trim()))
			throw new LoginException(LMN + " Error: Required " +
				"'roleAttrName' property missing");

		// Get and validate the roleFilter
		String roleFilter = (String)options.get("roleFilter");
		if ((roleFilter == null) || "".equals(roleFilter.trim()))
			throw new LoginException(LMN + " Error: Required " +
				"'roleFilter' property missing");

		// Get and validate the scope
		String roleScope = (String)options.get("roleScope");
		if ((!"BASE".equalsIgnoreCase(roleScope)) &&
			(!"ONE".equalsIgnoreCase(roleScope)) &&
			(!"SUB".equalsIgnoreCase(roleScope)))
			throw new LoginException(LMN + " Error: Required " +
				"Required option 'roleScope' must be 'BASE', 'ONE', or 'SUB'");
	}

	/**
	 * Create a LDAPConnection.
	 * @param port 
	 * @param host 
	 *
     * @param host The host name for the connection.
     * @param port The port for the connection.

	 * @return a LDAPConnection that is ready for use.
	 * @throws LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getConnection(String host, String port)
		throws LoginException
	{
		try
		{
			if (debug)
				logger.debug(this, LMN + " " +
					"Instantiating LDAPConnection: host=" + host +
					", port=" + port + ", useSSL=" + useSSL);

			// Get the connection timeout
			String connectTimeout = (String)options.get("connectTimeout");
			connectTimeout = (connectTimeout != null ? connectTimeout : "3000");

			// Create the connection
			LDAPConnection connection = null;
			if (useSSL)
			{
				// Warn if required SSL System property for trust store is not set
				String trustStoreProp =
					System.getProperty("javax.net.ssl.trustStore");
				if (trustStoreProp == null)
					logger.warning(this,
						"Missing required SSL system property: " +
						"javax.net.ssl.trustStore");

				// Use an SSL connection
				connection =
					new LDAPConnection(new LDAPJSSESecureSocketFactory());
			}
			else
			{
				// Use a non-SSL connection
				connection = new LDAPConnection();
			}

			// Create a new LDAPSocketFactory that will delegate to the
			// default factory, but also provides connection timeout
			LDAPSocketFactory socketFactory = new TimeoutLDAPSocketFactory(
				connection.getSocketFactory(), Long.parseLong(connectTimeout));

			connection = null;
			connection = new LDAPConnection(socketFactory);

			if (debug)
				logger.debug(this, LMN + " " +
					"Connecting to host=" + host +
					", port=" + port);

			// Debug to help customers determine LDAP response time 
			long timer = System.currentTimeMillis(); 

			// Connect to the directory
			connection.connect(host, Integer.parseInt(port));

			if (debug)
			{
				logger.debug(this, LMN + "Time to connect: '" +
					(System.currentTimeMillis() - timer) + "'");

				if (connection.isConnected() && connection.isConnectionAlive())
					logger.debug(this, LMN + " LDAP connection succeeded");
				else
					logger.debug(this, LMN + " LDAP connection failed");

				logger.debug(this, LMN + " Returning LDAPConnection");
			}

			return connection;
		}
		catch (LDAPException e)
		{
			logger.error(this, "LDAP error", e);

			// Dump the nested Throwables to get at the root cause
			int i = 1;
			Throwable cause = e.getCause();
			do
			{
				if (cause != null)
					logger.error(this, "LDAP error: CAUSE " + i + ": " +
						cause.getMessage());
			}
			while ((cause = cause.getCause()) != null);

			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
	}

	/*
	 * Translate a generic LDAP server LDAPException into a more specific
	 * LoginException and throw it.
	 * 
	 * @param e the LDAPException.
	 * @throws LoginException if not a more specific login failure.
	 * @throws FailedLoginException if invalid login credentials.
	 * @throws CredentialExpiredException if a login credential has expired.
	 * @throws AccountExpiredException if the user account has expired.
	 */
	private void genericLdapServerLDAPExceptionHandler(LDAPException e)
		throws LoginException, FailedLoginException,
			CredentialExpiredException, AccountExpiredException
	{
		// Check for credential (password) expired exception. Note the
		// message "password expired" is likely to be specific to the
		// SunOne LDAP server.
		if (e.getResultCode() == LDAPException.INVALID_CREDENTIALS)
		{
			if ((e.getLDAPErrorMessage() != null) &&
				(e.getLDAPErrorMessage().toLowerCase().startsWith("password expired")))
			{
				throw new CredentialExpiredException(
					LMN + " Error: " + e.getMessage() + "\n");
			}
			else
			{
				throw new FailedLoginException(
					LMN + " Invalid credentials: " +
					e.getLDAPErrorMessage() + "\n");
			}
		}
		else if (e.getResultCode() == LDAPException.UNWILLING_TO_PERFORM)
		{
			// Check for account expired exception. Red Hat Directory Server returns 
			// this message when "nsAccountLock" is set to "true".
			if ((e.getLDAPErrorMessage() != null) &&
				(e.getLDAPErrorMessage().toLowerCase().startsWith("account inactivated")))
			{
				throw new AccountExpiredException(
					LMN + " Error: " + e.getMessage() + "\n");
			}
			else
			{
				// Login failure: some other constraint violation
				throw new LoginException(
					LMN + " Error while trying to " +
						"bind user: " + e.getLDAPErrorMessage() + "\n");	
			}
		}
		else if (e.getResultCode() == LDAPException.CONSTRAINT_VIOLATION)
		{
			// Check for account expired exception, due to retry violation. The
			// message "Exceed password retry" is likely to be specific to the
			// SunOne LDAP server.
			if ((e.getLDAPErrorMessage() != null) &&
				(e.getLDAPErrorMessage().toLowerCase().startsWith("exceed password retry")))
			{
				throw new AccountExpiredException(
					LMN + " Error: " + e.getMessage() + "\n");
			}
			else
			{
				// Login failure: some other constraint violation
				throw new LoginException(
					LMN + " Error while trying to " +
						"bind user: " + e.getLDAPErrorMessage() + "\n");	
			}
		}
		else
		{
			// Login failure for other reason
			throw new LoginException(
				LMN + " Error while trying to " +
					"bind user: " + e.getLDAPErrorMessage() + "\n");	
		}
	}
	
	/**
	 * Translate a specific ActiveDirectory LDAPException into a specific
	 * LoginException and throw it.
	 * 
	 * @param e the LDAPException containing an ActiveDirectory specific message.
	 * @throws LoginException if not a more specific login failure.
	 * @throws FailedLoginException if invalid login credentials.
	 * @throws CredentialExpiredException if a login credential has expired.
	 * @throws AccountExpiredException if the user account has expired.
	 */
	private void activeDirectoryLDAPExceptionHandler(LDAPException e)
		throws LoginException, FailedLoginException,
			CredentialExpiredException, AccountExpiredException
	{
		// Normal login failure due to invalid credentials, check
		// Active Directory server message for subcode to determine the
		// most accurate exception to throw. Possible codes include:
		// 52b - invalid password (FailedLoginException)
		// 52e - invalid credentials (FailedLoginException)
		// 530 - invalid login time (FailedLoginException)
		// 532 - password expired (CredentialExpiredException)
		// 533 - account disabled (AccountExpiredException)
		// 701 - account expired (AccountExpiredException)
		// 773 - password reset required (CredentialExpiredException)
		if (e.getResultCode() == LDAPException.INVALID_CREDENTIALS)
		{
			if (e.getLDAPErrorMessage().indexOf("data 532") > 0)
			{
				throw new CredentialExpiredException(
					LMN + " Password expired: " +
					e.getLDAPErrorMessage() + "\n");
			}
			else if (e.getLDAPErrorMessage().indexOf("data 533") > 0)
			{
				throw new AccountExpiredException(
					LMN + " Account disabled: " +
					e.getLDAPErrorMessage() + "\n");
			}
			else if (e.getLDAPErrorMessage().indexOf("data 701") > 0)
			{
				throw new AccountExpiredException(
					LMN + " Account expired: " +
					e.getLDAPErrorMessage() + "\n");
			}
			else if (e.getLDAPErrorMessage().indexOf("data 773") > 0)
			{
				throw new CredentialExpiredException(
					LMN + " Password reset required: " +
					e.getLDAPErrorMessage() + "\n");
			}
			else
			{
				throw new FailedLoginException(
					LMN + " Invalid credentials: " +
					e.getLDAPErrorMessage() + "\n");
			}
		}
		else
		{
			// Login failure for reason other than invalid credentials
			throw new LoginException(
				LMN + " Error while trying to " +
					"bind user: " + e.getLDAPErrorMessage() + "\n");	
		}
	}

	/**
	 * Return the input pattern replaced with the match value.
	 *
	 * @param pattern The input pattern to use
	 * @param matchName The variable name to match
	 * @param matchValue The value to replace
	 * @return The input pattern with matched value
	 */
	private String replaceUsername(String pattern, String matchName,
		String matchValue)
	{
		String tmp = null;
		StringBuffer sb = new StringBuffer();
		StringTokenizer st = new StringTokenizer(pattern, "{}");
		while (st.hasMoreTokens())
		{
			tmp = st.nextToken();

			if (tmp.trim().equals(matchName))
				sb.append(matchValue);
			else
				sb.append(tmp);
		}

		return sb.toString();
	}

	/**
	 * Parse a string for values separated by a single input character. For
	 * example, "foo,goober,bar" could be parsed by specifying "," as the
	 * token. It would return a String[3] with foo, goober, and bar as
	 * the values.
	 *
	 * @param s the string to evaluate
	 * @param token the token that separates the String values
	 *
	 * @return a String array
	 */
	static String[] getStringArray(String s, String token)
	{
		// Create tokenizer and initialize array based on the number of tokens
		StringTokenizer st = new StringTokenizer(s.trim(), token);
		String[] sa = new String[st.countTokens()];

		// Parse the tokens and store them in the array
		int i = 0;
		while(st.hasMoreTokens())
		{
			sa[i] = st.nextToken();
			i++;
		}

		return sa;
	}
	
	/**
	 * A simple LDAPAuthHandler for following referrals
	 */
	private class AuthImpl implements LDAPAuthHandler
	{
		// The dn used to authenticate
		private String dn;

		// The password used to authenticate
		private byte[] passwd;

		/**
		 * Construct this instance using data
		 */
		private AuthImpl(String username, String password)
		{

			// Create an LDAPAuthProvider class to encapsulate the
			// authentication credentials.
			dn = username;
			try
			{
				passwd = password.getBytes("UTF8");
			}
			catch (UnsupportedEncodingException e)
			{
				// Should never get this
				logger.error(this,
					"Unable to encode password to UTF-8 for user: " + dn);
			}

			return;
		}

		/**
		 * Choose the login dn and password based on the host and port
		 * (this implementation simply reuses the original credentials)
		 */
		public LDAPAuthProvider getAuthProvider(String host, int port)
		{
			if (debug)
				logger.debug(this, "Following directory referral, rebinding " +
					"to: " + host + ":" + port);

			return new LDAPAuthProvider(dn, passwd);
		}
	}

} // End of class : AESOLdapLoginModule
