/**
 * Copyright (c) 1996-2011 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.cams.service.ActiveDirectoryGroupNameService;
import com.cafesoft.cams.service.UnboundIDLdapConnectionPoolService;

import com.cafesoft.core.lifecycle.LifecycleException;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.service.ServiceClient;
import com.cafesoft.core.service.ServiceException;
import com.cafesoft.core.service.ServiceFinder;

import com.unboundid.ldap.sdk.Attribute;
import com.unboundid.ldap.sdk.DN;
import com.unboundid.ldap.sdk.Entry;
import com.unboundid.ldap.sdk.Filter;
import com.unboundid.ldap.sdk.LDAPConnection;
import com.unboundid.ldap.sdk.LDAPException;
import com.unboundid.ldap.sdk.LDAPSearchException;
import com.unboundid.ldap.sdk.RDN;
import com.unboundid.ldap.sdk.SearchRequest;
import com.unboundid.ldap.sdk.SearchResult;
import com.unboundid.ldap.sdk.SearchScope;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;

import java.security.Principal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.StringTokenizer;

/**
 * <p><code>AESOUnboundIDActiveDirectoryLoginModule</code> is a JAAS LoginModule 
 * used with Cams "Automatic Enterprise Sign-On" (AESO) and Microsoft Active
 * Directory.</p>
 * <p>This LoginModule is designed to trust a user identity authenticated at the 
 * web layer, most likely via Kerberos, NTLM, HTTP Basic or HTTP Digest
 * authentication methods. The user is authenticated within the Cams security
 * context login succeeds within this LoginModule.<p>
 * <p>Responsibilities of this LoginModule include:
 * <ul>
 * <li>using a configured LDAP connection pool service to connect to Active 
 * Directory as an administrative user with permission to query user information.
 * <li>querying for existence of the user based on callback value "username".
 * <li>querying for group memberships and assigning role principals.
 * <li>populating a JAAS Subject with the user Principal and role Principals.
 * <li>adding the Authentication Method to the Subject public credential
 * cache based on the value of callback value "auth_type". Supported values
 * from "weakest" to "strongest" include: "PASSWORD", "NTLM", "Kerberos".
 * NOTE: the "auth_method" value is populated by proper configuration of 
 * the corresponding AESO callback value at the Cams AESO agent.
 * </ul>
 * Supported configuration options are listed below:
 * <ul>
 * <li><b>debug</b> - Enable/disable debugging [true|false].
 * 
 * <li><b>ldapPoolServiceId</b> - (required) The ID of a configured 
 * UnboundIDLdapConnectionPoolService in this security domain. When this 
 * option is configured, any LDAP connection parameters configured for this 
 * login module are ignored.
 * 
 * <li><b>defaultRoles</b> - (optional) A list of one or more default roles
 * separated by a comma to be added to the user session.</li>
 * 
 * <li><b>roleRegexPattern</b> - (optional) Limits roles saved in the user 
 * session to those that match a regular expression pattern. This option is 
 * useful when the role search returns many roles that are not used by the 
 * access control policy.
 * 
 * <li><b>domain</b> - (optional) The Active Directory domain to use. If this
 * value is present and the user name does not end with @domain, then append
 * "@" and the value of domain to the user name.</li>
 * 
 * <li><b>useRoleSearch</b> - (optional) A value of "true" (default) 
 * indicates that a role search will be attempted [true|false].</li>
 * 
 * <li><b>useDomainInRoleSearch</b> - (optional) A value of "true" (default) 
 * indicates that an optionally supplied "domain" used for authentication
 * will be included in the role search filter's {username} substitution value 
 * [true|false].</li>
 * 
 * <li><b>baseDN</b> - (conditional) The base distinguished name for the user 
 * entries. Not required if role search will not be performed.</li>
 * 
 * <li><b>scope</b> - (conditional) The scope for the role search. Valid values
 *   are "BASE", "ONE", and "SUB". Not required if role search will not be performed.</li>
 * 
 * <li><b>filter</b> - (conditional) The role search filter. Not required if role 
 * search will not be performed.</li>
 * 
 * <li><b>attribute</b> - (conditional) The attribute where the roles
 * are listed. Value should be either "memberOf" or "tokenGroups". If "tokenGroups" 
 * is used, then an Active Directory group name service must be configured
 * to lookup the group names by their object SIDs. Not required if role search will 
 * not be performed.</li>
 * 
 * <li><b>groupNameServiceId</b> - (conditional) The ID of a configured 
 * Active Directory group name service in this security domain. This option is
 * only valid if the "attribute" options is configured to "tokenGroups".</li>
 * 
 * <li><b>agentSecret</b> - The expected value for the "agent_secret" callback
 * value, which must be present in the authentication request. This value is 
 * used to authenticated Cams agents, which assert the identity of a user
 * already authenticated at the web layer. The agent secret is used to protect 
 * against impersonation attacks via carefully created HTTP requests 
 * attempting to subvert AESO component trust relationships.</li>
 * </ul>
 * 
 * WARNING: The values in this object can be accessed by other Cams components 
 * to enable SSO to other other systems, so administrator must take care that only 
 * trusted components have are used!</li> 
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.2 $ $Date: 2012/09/06 00:07:59 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 10/02/2008
 */
public class AESOUnboundIDActiveDirectoryLoginModule
	implements LoginModule, LoggerClient, ServiceClient
{
	//
	// Static Variables.
	//
	
	/**
	 * The LoginModule name (used in log messages throughout the code).
	 */
	private static final String LMN = "[AESOUnboundIDActiveDirectoryLoginModule]";
	
	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag indicating that a role search will be tried.
	 */
	private boolean useRoleSearch;

	/**
	 * Flag indicating that a domain value, if supplied, will be used in the 
	 * user name substitution value for a role search.
	 */
	private boolean useDomainInRoleSearch;
	
	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating user.
	 */
	private String username;

	/**
	 * The agent secret (used to protect against impersonation attacks).
	 */
	private String agentSecret;
	
	/**
	 * A default list of user roles to associate with an authenticated user.
	 */
	private List roleList;

	/**
	 * Used to lookup required Cams services.
	 */
	private ServiceFinder serviceFinder;

	/**
	 * Connection pool.
	 */
	private UnboundIDLdapConnectionPoolService lcps;

	//
	// Constructor
	//

	/**
	 * Default AESOUnboundIDActiveDirectoryLoginModule Constructor.
	 */
	public AESOUnboundIDActiveDirectoryLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.useRoleSearch = true;
		this.useDomainInRoleSearch = true;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.subject = null;
		this.username = null;
		this.agentSecret = null;
		this.roleList = new ArrayList();
		this.serviceFinder = null;
		this.lcps = null;
	}

	//
	// Implements the ServiceClient interface
	//

	/**
	 * Set the ServiceFinder.
	 *
	 * @param finder the class used by find Services by the ServiceClient.
	 */
	public void setServiceFinder(ServiceFinder finder)
	{
		this.serviceFinder = finder;
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code> after this 
	 * <code>LoginModule</code> has been instantiated. The purpose of this 
	 * method is to initialize this <code>LoginModule</code> with the relevant 
	 * information. If this <code>LoginModule</code> does not understand any 
	 * of the data stored in <code>sharedState</code> or <code>options</code> 
	 * parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> to prompt user
	 * for credentials such as user names, passwords, etc. 
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login <code>Configuration</code> 
	 * for this <code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.debug = "true".equalsIgnoreCase((String)options.get("debug"));
		this.useRoleSearch = 
			! "false".equalsIgnoreCase((String)options.get("useRoleSearch"));
		this.useDomainInRoleSearch = 
			! "false".equalsIgnoreCase((String)options.get("useDomainInRoleSearch"));
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates a <code>Subject</code>.  
	 * For example, it may prompt for <code>Subject</code> information such 
	 * as a user name and password and then attempt to verify the password. 
	 * This method saves the result of the authentication attempt as private 
	 * state within the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 * <code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's overall authentication 
	 * succeeded (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL 
	 * LoginModules succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt succeeded (checked 
	 * by retrieving the private state saved by the <code>login</code> method), 
	 * then this method associates relevant Principals and Credentials with 
	 * the <code>Subject</code> located in the <code>LoginModule</code>. If 
	 * this LoginModule's own authentication attempted failed, then this 
	 * method removes/destroys any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 * <code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt  succeeded (checked
	 * by retrieving the private state saved by the <code>login</code> method),
	 * then this method cleans up any state that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 * <code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 * <p>
	 * This method executes the following steps:
	 * <ol>
	 * <li>Checks preconditions
	 * <li>Handles callback objects
	 * <li>Get a connection to the Active Directory from a connection pool service
	 * <li>Queries user groups for role assignment
	 * <li>Release the connection
	 * <li>Reset state if authentication failed
	 * <li>Return results of authentication attempt
	 * </ol>
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @exception LoginException thrown if any error occurs while trying to
	 * authenticate the user.
	 * @exception FailedLoginException if authentication fails.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException,
		AccountExpiredException, CredentialExpiredException
	{
		// Check preconditions
		validateOptions();

		// Handle callback objects.
		handleCallbacks();

		// An agent secret must always be supplied as a callback value and
		// its value must match the expected value.
		String expectedAgentSecret = (String)options.get("agentSecret");
		if (agentSecret == null || agentSecret.length() == 0)
			throw new LoginException(LMN + " Error: Required callback " +
				"value 'agent_secret' does not exist or is empty");
		else if (agentSecret.equals(expectedAgentSecret) == false)
			throw new LoginException(LMN + " Error: Callback value 'agent_secret' " +
				"does not match expected value in option 'agentSecret'");
		
		// A user name must always be supplied
		if (this.username == null || this.username.length() == 0)
			throw new FailedLoginException(LMN + " Authentication failed");

		// User name supplied by AESO and authentication has succeeded
		this.loginSucceeded = true;

		/**
		 * If the username callback value has a backslash, then assume it
		 * is a UNC (Universal Naming Convention) name. In this case, chop off 
		 * the windows domain name because the UNC form does not work well 
		 * with Active Directory.
		 */
		int index = username.indexOf('\\');
		if (index > 0)
			this.username = this.username.substring(index + 1);
		
		// The fully-qualified username as entered by the user but 
		// may optionally have domain appended
		String fqUsername = this.username;

		// Optionally append the domain to the user name if supplied. Do
		// not modify username as it is required "as is" for the role filter.
		String domain = (String)options.get("domain");
		if ((domain != null) && ! ("".equals(domain.trim())))
			fqUsername = this.username + "@" + domain;

		if (debug)
			logger.debug(this, LMN + " Fully-qualified user name = '" + fqUsername + "'");

		if (this.debug)
			logger.debug(this, LMN + " Getting connection from pool");

		// Get a bound connection to Active Directory from the connectiong pool service
		LDAPConnection connection = null;
		connection = getPooledConnection();

		if (this.debug)
		{
			logger.debug(this, LMN + " Connection is connected: '" +
				connection.isConnected() + "'");
			logger.debug(this, LMN + " Connection statistics: " + 
				connection.getConnectionStatistics());
		}

		// Add default roles (if any) to list of roles
		String defaultRoles = (String)options.get("defaultRoles");
		if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
		{
			StringTokenizer st = new StringTokenizer(defaultRoles, ",");
			while(st.hasMoreTokens())
				roleList.add(st.nextElement().toString().trim());
		}

		// Associate roles from groups that the user is a member of

		// Check the required options when using a role search
		if (this.useRoleSearch)
		{
			String baseDN = (String)options.get("baseDN");
			String filter = (String)options.get("filter");
			String scope = (String)options.get("scope");

			// Set the search scope, the default of 0 is "BASE"
			int scopeInt = 0;
			if ("ONE".equalsIgnoreCase(scope))
				scopeInt = 1;
			else if ("SUB".equalsIgnoreCase(scope))
				scopeInt = 2;

			try
			{
				// Get and validate the attribute where roles are stored (the
				// "memberOf" groups attribute is used, but enable external 
				// configuration to supply an alternative attribute.

				// Specify the attributes that the search will return
				String[] attrs = {(String)options.get("attribute")};

				// Substitute the user name in the role search filter. Optionally,
				// don't use the supplied domain in the user name to enable searches
				// using attributes other than the userPrincipalName, such as 
				// sAMAccountName, uid, or cn, which don't use the fully-
				// qualified UPN.
				if (useDomainInRoleSearch)
					filter = replaceUsername(filter, "username", fqUsername);
				else
					filter = replaceUsername(filter, "username", this.username);

				// If supplied, only use roles that match roleRegexPattern
				String roleRegexPattern = (String)options.get("roleRegexPattern");
				
				// The "tokenGroups" attribute contains all groups of which a
				// a user is member (including primary and nested groups). However,
				// it is only returned if you search with the user's
				// distinguishedName. So, we first search to get the user's
				// distinguishedName, then we do a read to get the "tokenGroups".
				// Because the "tokenGroups" are returned as SIDs, we then
				// use a Cams service to efficiently obtain their names.
				if ("tokenGroups".equalsIgnoreCase(attrs[0]))
				{
					// Get the user's distinguishedName
					attrs[0] = "distinguishedName";

					if (this.debug)
					{
						logger.debug(this, LMN + 
							" Search for the user's 'distinguishedName' using:" +
							  "\n\tbaseDN: " + baseDN +
							  "\n\tfilter: " + filter +
							  "\n\t attrs: " + attrs[0]);
					}

					SearchRequest searchRequest = 
						new SearchRequest(baseDN, SearchScope.valueOf(scopeInt), 
						Filter.create(filter), attrs);

					// Search for the user entry attributes
					SearchResult searchResult = connection.search(searchRequest);

					// Process the search results to get the user dn
					if (searchResult != null)
					{
						List searchResultList = searchResult.getSearchEntries();
						Iterator searchResultIterator = searchResultList.iterator();
						if (searchResultIterator.hasNext())
						{
							// Get the next entries attibute
							Entry entry = (Entry)searchResultIterator.next();
							Collection attrCollection = entry.getAttributes();
							Iterator attrIterator = attrCollection.iterator();
							if (attrIterator.hasNext())
							{
								// Attribute may have more than one name/value pair
								Attribute attr = (Attribute) attrIterator.next();
								if (attr != null)
									baseDN = attr.getValue();
							}
						}
					}
					else
					{
						throw new LoginException ("Attribute 'distinguishedName' " +
							"not found for '" + filter + "'");
					}

					// Now that we have the user's dn, looking for tokenGroups
					attrs[0] = "tokenGroups";

					if (this.debug)
						logger.debug(this, LMN +
							" 'tokenGroups' attribute(s) for user '" + baseDN + "'");

					// Read the user's tokenGroups values
					Entry entry = connection.getEntry(baseDN, attrs);

					// Save the Active Directory groups as roles for commit method
					if (entry != null)
					{
						Attribute attribute = entry.getAttribute(attrs[0]);

						// Get an enumeration of the attribute values
						byte[][] allValues = attribute.getValueByteArrays();

						String role = null;
						String sid = null;

						if ((roleRegexPattern != null) && !("".equals(roleRegexPattern)))
						{
							Pattern pattern = Pattern.compile(roleRegexPattern);
							for (int i = 0; i < allValues.length; i++)
							{
								// Get the sid
								sid = Sid2String(allValues[i]);

								if (this.debug)
									logger.debug(this, LMN +
										" Getting cn for objectSid: '" + sid + "'");

								role = getGroupName(sid);

								// Role not found, use sid for role name and log error
								// GEG: Consider providing an externalized list mapping 
								// objectSid to cn if unresolved objectSid is common, which
								// is not expected to be the case.
								if (role == null)
								{
									role = sid;

									logger.error(this, LMN + 
										" Active Directory group name service returned 'null' cn for objectSid: '" + sid + "'");
								}
								else
								{
									if (debug)
										logger.debug(this, LMN + " SID role name: '" + role + "'");
								}

								if (pattern.matcher(role).matches())
								{
									roleList.add(role);

									if (this.debug)
										logger.debug(this, LMN +
											" Role: '" + role + 
											"' matched roleRegexPattern: '" + 
											roleRegexPattern + "'");
								}
								else
								{
									logger.debug(this, LMN + 
										" Role: '" + role + 
										"' did not match roleRegexPattern: '" + 
										roleRegexPattern + "'");
								}
							}
						}
						else
						{
							for (int i = 0; i < allValues.length; i++)
							{
								// Get the sid
								sid = Sid2String(allValues[i]);

								if (this.debug)
									logger.debug(this, LMN +
										" Getting cn for objectSid: '" + sid + "'");

								role = getGroupName(sid);

								// Role not found, use sid for role name and log error
								// GEG: Provide an externalized list of objectSid that
								// will not be used for role lookup if this is found to 
								// be a significant issue for customers.
								if (role == null)
								{
									role = sid;

									logger.error(this, LMN + 
										" Active Directory group name service returned 'null' cn for objectSid: '" + sid + "'");
								}
								else
								{
									if (debug)
										logger.debug(this, LMN + " SID role name: '" + role + "'");
								}

								this.roleList.add(role);
							}
						}
					}
				}

				// The role attribute is "memberOf", so we need to get the
				// get string values and explode them to get the group names
				else if ("memberOf".equalsIgnoreCase(attrs[0]))
				{
					if (this.debug)
					{
						logger.debug(this, LMN + 
							" Search for user's groups with:" +
							  "\n\tbaseDN: " + baseDN +
							  "\n\tfilter: " + filter +
							  "\n\t attrs: " + attrs[0]);
					}					

					// Search for the user entry attributes
					SearchResult searchResults = connection.search(
						baseDN, SearchScope.valueOf(scopeInt), filter, attrs);

					// Get the roles and save them for commits
					if (searchResults != null)
					{
						List searchResultsList = searchResults.getSearchEntries();
						Iterator searchResultsIterator = searchResultsList.iterator();
						while (searchResultsIterator.hasNext())
						{
							// Get the next entries attibute
							Entry entry = (Entry)searchResultsIterator.next();
							Collection attrCollection = entry.getAttributes();
							Iterator attrIterator = attrCollection.iterator();

							while (attrIterator.hasNext())
							{
								if ((roleRegexPattern != null) && !("".equals(roleRegexPattern)))
								{
									Pattern pattern = Pattern.compile(roleRegexPattern);

									// Attribute may have more than one name/value pair
									Attribute attr = (Attribute) attrIterator.next();
									String[] values = attr.getValues();
									String[] roles;
									for (int i = 0; i < values.length; i++)
									{
										// Parse the role name from the DN 
										RDN rdn = new RDN(DN.getRDNString(values[i]));
										roles = rdn.getAttributeValues();

										if (this.debug)
											logger.debug(this, LMN +
												" Parsed role '" + roles[0] + "' from dn '" +
												values[i] + "'");

										if (pattern.matcher(roles[0]).matches())
										{
											this.roleList.add(roles[0]);

											if (debug)
												logger.debug(this, LMN + 
													" Role: '" + roles[0] + 
													"' matched roleRegexPattern: '" + 
													roleRegexPattern + "'");
										}
										else
										{
											if (debug)
												logger.debug(this, LMN + 
													" Role: '" + roles[0] + 
													"' did not match roleRegexPattern: '" + 
													roleRegexPattern + "'");
										}
									}
								}
								else
								{
									// Attribute may have more than one name/value pair
									Attribute attr = (Attribute) attrIterator.next();
									String[] values = attr.getValues();
									String[] roles;
									for (int i = 0; i < values.length; i++)
									{
										// Parse the role name from the DN 
										RDN rdn = new RDN(DN.getRDNString(values[i]));
										roles = rdn.getAttributeValues();

										if (this.debug)
											logger.debug(this, LMN +
												" Parsed role: '" + roles[0] + "' from dn '" +
												values[i] + "'");

										this.roleList.add(roles[0]);
									}
								}
							}
						}
					}
					else
					{
						logger.debug(this, LMN +
						" No search results returned");
					}
				}
				else
				{
					logger.error(this, LMN +
						" Unsupported role search attribute, must be " +
						"'memberOf' or 'tokenGroups'");
				}
			}
			catch (LDAPSearchException e)
			{
				if (this.debug)
					logger.debug(this, LMN + " Error while " +
						"searching for groups for user: '" +
						this.username + "'", e);

				// Search filter failed
				throw new LoginException(LMN +
					" " + e.getExceptionMessage());
			}
			catch (LDAPException e)
			{
				if (this.debug)
					logger.debug(this, LMN + " Error while " +
						"searching for groups for user: '" +
						this.username + "'", e);

				// Search filter failed
				throw new LoginException(LMN +
					" " + e.getExceptionMessage());
			}
			finally
			{
				// Close the connection
				if (connection != null)
				{
					try
					{
						this.lcps.releaseConnection(connection);
					}
					catch(ServiceException e)
					{
						if (this.debug)
							logger.debug(this, LMN + 
								" Service error while returning connection to pool");

						throw new LoginException(e.getMessage());
					}
					catch(LifecycleException e)
					{
						if (this.debug)
							logger.debug(this, LMN + 
								" Lifecyle error while returning connection to pool");

						throw new LoginException(e.getMessage());
					}
				}
			}
		}

		if (debug)
			logger.debug(this, LMN + " Returning from authenticate(): loginSucceeded='" + loginSucceeded);
					
		// Return authentication results
		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @throws LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		if (debug)
			logger.debug(this, LMN + " Starting associateRules(): loginSucceeded=" + loginSucceeded);
		
		// The login failed, clean up state
		if (!loginSucceeded)
			return false;

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList();

		// Repository operations to get default role principals
		Iterator roleListItr = roleList.iterator();
		while(roleListItr.hasNext())
		{
			String rolename = (String)roleListItr.next().toString().trim();
			
			if (debug)
				logger.debug(this, LMN + " Creating CSRolePrincipal: rolename='" + rolename + "'");
		
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				if (debug)
					logger.debug(this,
						LMN + " Adding CSRolePrincipal " +
						rolename + " to principalList for: '" +
						this.username + "'");

				principalList.add(rolePrincipal);
			}
			else
			{
				if (debug)
					logger.debug(this,
						LMN + " CSRolePrincipal '" +
						rolename + "' already in principalList for '" +
						this.username + "', not added");
			}
		}

		if (debug)
			logger.debug(this, LMN + " Creating CSUserPrincipal: username='" + this.username + "'");
			
		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(this.username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			if (debug)
				logger.debug(this,
					LMN + " Adding CSUserPrincipal: '" + this.username +
					"' to principalList");

			principalList.add(userPrincipal);
		}
		else if (debug)
		{
			logger.debug(this,
				LMN + " CSUserPrincipal: '" + this.username +
				"' already in principalList, not added");
		}

		logger.debug(this, LMN + " Adding principalList to Subject");

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		if(debug)
		{
			logger.debug(this, LMN + " Adding " +
				AuthenticationMethod.CAMS_AESO_LOGIN.getName() +
				" authentication method");
		}

		subject.getPublicCredentials().add(AuthenticationMethod.CAMS_AESO_LOGIN);

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.agentSecret = null;

		if (debug)
			logger.debug(this, LMN + " Returning from associateRules(): commitSucceeded='" + commitSucceeded);
		
		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @throws LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.
			this.loginSucceeded = false;
			this.username = null;
			this.agentSecret = null;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @throws LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if(this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if (this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.logger = null;
		this.callbackHandler = null;
		this.username = null;
		this.agentSecret = null;
		this.options = null;
		this.loginSucceeded = false;
		this.commitSucceeded = false;
		this.debug = false;

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username and
	 * password instance variables with their appropriate value.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// This LoginModule requires a username and an agent_secret
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Get the callback values.
			callbackHandler.handle(callbacks);

			// Get the user name
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");
			
			// Get the agent secret
			this.agentSecret = ((MapCallback)callbacks[0]).getCallbackValue("agent_secret");
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available",
				e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (debug)
		{
			logger.debug(this, LMN + " User entered username = " + this.username);
			logger.debug(this, LMN + " AESO agent secret = " + agentSecret);
		}
	}

	/**
	 * Validate the LoginModules options Map object to ensure that all
	 * required properties exist and contain valid values.
	 *
	 * @exception LoginException if any of the following conditions occur:
	 * <ul>
	 * <li>A CallbackHandler object is not available</li>
	 * <li>the 'userDN' property is not available</li>
	 * <li>the 'baseDN' property is not available</li>
	 * <li>the 'attribute' property is not available</li>
	 * <li>the 'filter' property is not available</li>
	 * <li>the 'scope' property is not available or valid</li>
	 * </ul>
	 */
	private void validateOptions() throws LoginException
	{
		if (callbackHandler == null)
			throw new LoginException(LMN + " Error: no CallbackHandler available");

		String s = null;

		s = (String)options.get("ldapPoolServiceId");
		if ((s == null) || (s.trim().length() == 0))
			throw new LoginException(LMN +
				" Error: Required 'ldapPoolServiceId' property missing");

		if (this.useRoleSearch)
		{
			s = (String)options.get("baseDN");
			if ((s == null) || "".equals(s.trim()))
				throw new LoginException(LMN +
					" Error: Missing required option 'baseDN'");

			s = (String)options.get("filter");
			if ((s == null) || "".equals(s.trim()))
				throw new LoginException(LMN +
					" Error: Missing required option 'filter'");

			s = (String)options.get("scope");
			if ((!"BASE".equalsIgnoreCase(s)) &&
				(!"ONE".equalsIgnoreCase(s)) &&
				(!"SUB".equalsIgnoreCase(s)))
				throw new LoginException(LMN +
					" Error: Required option 'scope' must be 'BASE', 'ONE', or 'SUB'");

			s = (String)options.get("attribute");
			if ((s == null) || "".equals(s.trim()))
				throw new LoginException(LMN +
					" Error: Missing required option 'attribute'");
		}

		s = (String)options.get("agentSecret");
		if ((s == null) || (s.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'agentSecret' property missing");
	}

	/**
	 * Get an LDAP server connection from a connection pool.
	 *
	 * @return a LDAPConnection.
	 * @exception LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getPooledConnection()
		throws LoginException
	{
		// Get the connection values to use
		LDAPConnection connection = null;

		try
		{
			if (serviceFinder == null)
				throw new LoginException(LMN +
					" Error: ServiceFinder is null");

			String ldapPoolServiceId = (String)options.get("ldapPoolServiceId");
			if ((ldapPoolServiceId == null) || (ldapPoolServiceId.trim().length() == 0))
				throw new LoginException(LMN +
					" Error: Required property 'ldapPoolServiceId' missing");

			if (this.debug)
				logger.debug(this, LMN + " Using ldapPoolServiceId '" + 
					ldapPoolServiceId + "'");

			this.lcps = (UnboundIDLdapConnectionPoolService)
				serviceFinder.find(ldapPoolServiceId, 
				UnboundIDLdapConnectionPoolService.class);
			
			connection = lcps.getConnection();

			if (this.debug)
				logger.debug(this, LMN + " Got LDAP pool connection: " + 
				connection.getConnectionStatistics());
		}
		catch(ServiceException e)
		{
			logger.error(this, LMN + " Service error getting connection from pool", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		catch(LifecycleException e)
		{
			logger.error(this, LMN + " Lifecycle error getting connection from pool", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		catch(LDAPException e)
		{
			logger.error(this, LMN + " LDAP error getting connection from pool", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}

		return connection;
	}

	/**
	 * Return the input pattern replaced with the match value.
	 *
	 * @param pattern The input pattern to use
	 * @param matchName The variable name to match
	 * @param matchValue The value to replace
	 * @return The input pattern with matched value
	 */
	private String replaceUsername(String pattern, String matchName,
		String matchValue)
	{
		StringBuffer sb = new StringBuffer();
		StringTokenizer st = new StringTokenizer(pattern, "{}");
		while (st.hasMoreTokens())
		{
			String tmp = st.nextToken();
			if (tmp.trim().equals(matchName))
				sb.append(matchValue);
			else
				sb.append(tmp);
		}

		return sb.toString();
	}

	/**
	 * Gets the ActiveDirectoryGroupNameService for the LoginModule.
	 *
	 * @throws LoginException if an error occurs getting the
	 * ActiveDirectoryGroupNameService object.
	 */
	private String getGroupName(String sid)
		throws LoginException
	{
		String groupName = null;
		try
		{
			if (serviceFinder == null)
				throw new LoginException(
					LMN + " Error: ServiceFinder is null");

			String groupNameServiceId = (String)options.get("groupNameServiceId");
			if ((groupNameServiceId == null) || (groupNameServiceId.trim().length() == 0))
				throw new LoginException(LMN + " Error: " +
					"Required 'groupNameServiceId' property missing");

			if (debug)
			{
				logger.debug(this, "Using groupNameServiceId '" + groupNameServiceId + "'");
				logger.debug(this, "Getting group id '" + sid + "'");
			}

			ActiveDirectoryGroupNameService adgns = (ActiveDirectoryGroupNameService)
				serviceFinder.find(groupNameServiceId, ActiveDirectoryGroupNameService.class);
			groupName = adgns.getGroupNameBySid(sid);

			if (debug)
				logger.debug(this, "Returned group name '" + groupName + "'");
		}
		catch(ServiceException e)
		{
			logger.error(this, "Error finding ActiveDirectoryGroupNameService", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}

		return groupName;
	}

	/**
	 * Convert the native byte representation of a SID into string format
	 *
	 * @param sid a byte representation of a SID
	 * @return a string formatted representation of the SID
	 */
	private static String Sid2String(byte[] SID)
	{
		long version;
		long authority;
		long count;
		long rid;
		String strSID;
		strSID = "S";
		version = SID[0];
		strSID = strSID + "-" + Long.toString(version);
		authority = SID[4];
		for (int i = 0;i<4;i++)
		{
			authority <<= 8;
			authority += SID[4+i] & 0xFF;
		}
		strSID = strSID + "-" + Long.toString(authority);
		count = SID[2];
		count <<= 8;
		count += SID[1] & 0xFF;
		for (int j=0;j<count;j++)
		{
			rid = SID[11 + (j*4)] & 0xFF;
			for (int k=1;k<4;k++)
			{
				rid <<= 8;
				rid += SID[11-k + (j*4)] & 0xFF;
			}
			strSID = strSID + "-" + Long.toString(rid);
		}
		return strSID;
	}

} // End of class : AESOUnboundIDActiveDirectoryLoginModule

