/**
 * Copyright (c) 1996-2012 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-
 * INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR
 * ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.util.X509CertificateUtils;

import com.cafesoft.core.util.NamespaceMap;
import com.cafesoft.core.util.NamespaceHashMap;
import com.cafesoft.core.util.NamespaceMapUtils;
import com.cafesoft.core.util.StringUtils;
import com.cafesoft.core.util.TokenDelimiterSpec;

import com.novell.ldap.LDAPAttribute;
import com.novell.ldap.LDAPAuthHandler;
import com.novell.ldap.LDAPAuthProvider;
import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPConstraints;
import com.novell.ldap.LDAPReferralHandler;
import com.novell.ldap.LDAPReferralException;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;
import com.novell.ldap.LDAPSearchResults;
import com.novell.ldap.LDAPSocketFactory;
import com.novell.ldap.LDAPJSSESecureSocketFactory;

import java.security.Principal;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import java.security.cert.X509Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.io.ByteArrayInputStream;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * <p><code>AESOX509CertLdapLoginModule</code> is a JAAS LoginModule used
 * with Cams "Automatic Enterprise Sign-On" (AESO) and LDAP v3 compliant
 * directory servers.</p>
 * <p>This LoginModule is designed to trust a user identity authenticated
 * at the web layer via X.509 client certificate authentication and use
 * information available within the certificate to:
 * <ol>
 * <li>confirm the user exists by performing an LDAP query
 * <li>optionally assign more role principals used for Cams role based
 * access control (RBAC).
 * </ol>
 * </p>
 * <p>
 * This LoginModule attempts to bind to an administrator account DN, then
 * perform an LDAP query for an LDAP entry corresponding to user-specific
 * DN obtained from the X.509 certificate callback value.</p>
 * <p>
 * Role assignment is accomplished by use of an LDAP role filter, which
 * queries Groups to which the user is assigned and maps them directly to
 * roles.</p>
 * <ul>
 * <li><b>debug</b> - Enable/disable debugging [true|false].
 * <li><b>host</b> - The DNS name or IP address of the LDAP host.</li>
 * <li><b>port</b> - The host's LDAP service port.</li>
 * <li><b>useSSL</b> - Enables/disables use of SSL.</li>
 * <li><b>connectTimeout</b> - (Optional) The number of milliseconds to wait for
 * 	a connection to the LDAP server before timing out. For example: 3000</li>
 * <li><b>bindUsername</b> - The username or DN for an LDAP account with permissions
 * to query for user and group information.</li>
 * <li><b>bindPassword</b> - The password for the "bindUsername" account.</li>
 * <li><b>agentSecret</b> - The expected value for the "agent_secret" callback
 * value, which must be present in the authentication request. This value is 
 * used to authenticated Cams agents, which assert the identity of a user
 * already authenticated at the web layer. The agent secret is used to protect 
 * against impersonation attacks via carefully created HTTP requests 
 * attempting to subvert AESO component trust relationships.</li>
 * <li><b>cbValueNS</b> - The namespace used to reference callback values 
 * in other configuration options. For example, if cbValueNS is "cb", then
 * the value of a callback of name "email_addr" is referenced as: 
 * {cb:email_addr}
 * <li><b>cbValuesRequired</b> - A comma-separated list of callback values
 * that must be present in the authentication request before login will be
 * attempted. This list should name all callback values that are referenced
 * by other configuration options, plus the agent_secret and the x509_client_cert
 * callback values.
 * <li><b>userResultsNS</b> - The namespace used to reference attribute values 
 * returned from the user LDAP search in other configuration options. For example,
 * the userResultsNS is "user" and the "uid" LDAP attribute is returned with the 
 * user LDAP search results, then the "uid" value can be referenced in another
 * configuration option using: {user:uid}</li>
 * <li><b>userBaseDN</b> - The distinguished name at which to start a user 
 * search. This value should be the LDAP node closes to where user data is 
 * stored. For example: ou=people,dc=example,dc=com</li>
 * <li><b>userAttrNames</b> - The LDAP attribute names to be returned on user
 * searches. If the user LDAP entry is found, the values of these attributes
 * are made available for substitution in other configuration options, usually
 * the "userSubjectName" and "roleFilter".</li>
 * <li><b>userCertAttr</b> - The name of the LDAP attribute where the user's
 * X.509 certificate is store. The standard name is "userCertificate" and 
 * contains the DER-encoded (binary) certificate data.</b>
 * <li><b>userFilter</b> - The LDAP filter used to search for the user
 * LDAP entry and associated attributes. This filter will generally reference
 * a callback value that was created at the web layer by an AESO-enabled web
 * agent and obtained from the X.509 client certificate presented by the user.
 * For example, a user filter value of (mail={cb:email_addr}) indicates that
 * the user should be selected based on LDAP attribute "mail" with the value
 * matching the callback value named "email_addr".</li>
 * <li><b>userScope</b> - Specifies the LDAP search scope to use for user
 * searches. Valid values are BASE (search is done only at the base
 * DN specified by userBaseDN), ONE (search is done only at the first child 
 * tree node relative to userBaseDN), and SUB (search is done at userBaseDN
 * and all child nodes).</li>
 * <li><b>useRoleSearch</b> - If "true", then use the LDAP role search to
 * assign user roles (on successful login). If "false", then don't perform
 * the role search.
 * <li><b>roleBaseDN</b> - The base distinguished name to use for your
 * LDAP schema role searches. For example: ou=groups,dc=example,dc=com</li>
 * <li><b>roleAttrName</b> - The name of the attribute used to store the
 * role names (usually one name is used but can be a comma-delimited list).
 * For example: cn</li>
 * <li><b>roleFilter</b> - The filter to use for role searches. This value
 * generally needs to reference one or more callback values or values
 * returned from the user LDAP search. For example:
 * (uniqueMember=uid={user:uid},ou=people,dc=example,dc=com)</li>
 * <li><b>roleScope</b> - The LDAP search scope  to use for role searches. 
 * Valid values are BASE (search is done only at the base
 * DN specified by roleBaseDN), ONE (search is done only at the first child 
 * tree node relative to roleBaseDN), and SUB (search is done at roleBaseDN
 * and all child nodes).</li>
 * <li><b>defaultRoles</b> - (Optional) A list of one or more default roles
 * separated by a comma to be added to the user session.</li>
 * </ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.6 $ $Date: 2012/08/22 22:05:11 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 12/31/2008
 */
public class AESOX509CertLdapLoginModule
	implements LoginModule, LoggerClient
{
	//
	// static constants
	//

	/**
	 * The LoginModule name (used in log messages throughout the code).
	 */
	private static final String LMN = "[AESOX509CertLdapLoginModule]";
	
	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag indicating that SSL must be used.
	 */
	private boolean useSSL;

	/**
	 * Flag indicating that the LDAP directory is to be searched for roles.
	 */
	private boolean useRoleSearch;
	
	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The X509Certificate chain presented by the authenticated User.
	 */
	private X509Certificate[] certChain;
	
	// The authenticated user name.
	private String username;

	/**
	 * A list of user roles. This list is populated in the login method and
	 * used in the commit method.
	 */
	private List roleList;

	/**
	 * The agent secret (used to protect against impersonation attacks).
	 */
	private String agentSecret;

	private NamespaceMap namespaceMap;
	//
	// Constructor
	//

	/**
	 * Default AESOX509CertLdapLoginModule Constructor.
	 */
	public AESOX509CertLdapLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.useSSL = false;
		this.useRoleSearch = true;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.agentSecret = null;
		this.subject = null;
		this.certChain = null;
		this.roleList = new ArrayList();
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for accessing
	 * 	callback values (e.g. certChain).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.debug = "true".equals((String)options.get("debug"));
		this.useSSL = "true".equals((String)options.get("useSSL"));
		this.useRoleSearch = 
			! "false".equalsIgnoreCase((String)options.get("useRoleSearch"));

		this.namespaceMap = new NamespaceHashMap();
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @exception LoginException thrown if any error occurs while trying to
	 *	authenticate the user.
	 * @exception FailedLoginException if authentication fails.
	 * @exception AccountExpiredException if the user account is expired.
	 * @exception CredentialExpiredException if the user credential is expired.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException,
			AccountExpiredException, CredentialExpiredException
	{
		// Check preconditions
		if (callbackHandler == null)
			throw new LoginException(LMN + " Error: no CallbackHandler available");

		String host = (String)options.get("host");
		if ((host == null) || (host.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required 'host' property missing");

		String port = (String)options.get("port");
		if ((port == null) || (port.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required 'port' property missing");

		try
		{
			int portInt = Integer.parseInt(port);
			if (portInt < 1)
				throw new LoginException(LMN + " Error: " +
					"'port' property must be an integer value greater than zero");
		}
		catch (NumberFormatException e)
		{
			throw new LoginException(LMN + " Error: " +
				"'port' property is not an integer value");
		}

		String bindUsername = (String)options.get("bindUsername");
		if ((bindUsername == null) || (bindUsername.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'bindUsername' property missing");

		String bindPassword = (String)options.get("bindPassword");
		if ((bindPassword == null) || (bindPassword.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'bindPassword' property missing");

		String expectedAgentSecret = (String)options.get("agentSecret");
		if ((expectedAgentSecret == null) || (expectedAgentSecret.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'agentSecret' property missing");

		// Handle callback objects.
		handleCallbacks();

		// An agent secret must always be supplied as a callback value and
		// its value must match the expected value.
		if (agentSecret == null || agentSecret.length() == 0)
			throw new LoginException(LMN + " Error: Required callback value 'agent_secret' " +
				"does not exist or is empty");
		else if (agentSecret.equals(expectedAgentSecret) == false)
			throw new LoginException(LMN + " Error: Callback value 'agent_secret' " +
				"does not match expected value in option 'agentSecret'");

		// A certificate chain must always be supplied
		if (certChain == null || certChain.length == 0)
		{
			if (debug)
				logger.debug(this, "Authentication failed, missing client X509 certificate");

			throw new FailedLoginException(LMN + " Authentication failed");
		}

		// Get a connection to the LDAP repository
		LDAPConnection connection = getConnection(host, port);

		if (debug)
			logger.debug(this, LMN + " Successfully connected to LDAP Server");

		try
		{
			// Bind to LDAP Server using credentials with "admin" permissions.
			connection.bind(3, bindUsername, bindPassword.getBytes());
		}
		catch (LDAPException e)
		{
			if (debug)
				logger.debug(this, LMN + " An error occured while trying to bind user: '" + 
					bindUsername + "'" +
					"\n\tMessage: " + e.getMessage() +
					"\n\tLDAPErrorMessage: " + e.getLDAPErrorMessage() +
					"\n\tResultCode: " + e.getResultCode(), e);

			// Login failure during LDAP bind.
			throw new LoginException(LMN + " Error while trying to " +
					"bind user: " + e.getLDAPErrorMessage() + "\n");
		}

		if (debug)
			logger.debug(this, LMN + " Successful " +
				"bind to LDAP Server as user '" + bindUsername + "'");

		String subjectDN = null;
		TokenDelimiterSpec tds = new TokenDelimiterSpec("{", "}", ":");

		try
		{
			String s = (String)options.get("userBaseDN");
			if ((s == null) || "".equals(s.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'userBaseDN'");
			
			// Replace references to callback values in the userBaseDN.
			String userBaseDN = NamespaceMapUtils.replaceTokens(s, namespaceMap, tds);
			
			s = (String)options.get("userFilter");
			if ((s == null) || "".equals(s.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'userFilter'");

			// Replace references to callback values in the userFilter.
			String userFilter = NamespaceMapUtils.replaceTokens(s, namespaceMap, tds);

			// Get the user Subject DN from the certificate.
			Principal subjectPrincipal = certChain[0].getSubjectDN();
			subjectDN = subjectPrincipal.getName();

			if (debug)
				logger.debug(this, LMN + " Certificate contains subject DN='" + subjectDN + "'");

			String userScope = (String)options.get("userScope");
			if ((!"BASE".equalsIgnoreCase(userScope)) &&
				(!"ONE".equalsIgnoreCase(userScope)) &&
				(!"SUB".equalsIgnoreCase(userScope)))
				throw new LoginException(LMN + " " +
					"Error: Required option 'userScope' must be 'BASE', 'ONE', or 'SUB'");

			// Set the search scope, the default of 0 is "BASE"
			int scopeInt = 0;
			if ("ONE".equalsIgnoreCase(userScope))
				scopeInt = 1;
			else if ("SUB".equalsIgnoreCase(userScope))
				scopeInt = 2;
			
			// Set the search constraints to automatically follow referrals
			LDAPConstraints ldapConstraints = new LDAPConstraints();
			ldapConstraints.setReferralFollowing(true);

			// Create and set the referral handler
			LDAPReferralHandler referralHandler = new AuthImpl(bindUsername, bindPassword);
			ldapConstraints.setReferralHandler(referralHandler);

			// Set the search constraints
			connection.setConstraints(ldapConstraints);

			// Specify the attributes that the search will return
			String userAttrNames = (String)options.get("userAttrNames");
			if ((userAttrNames == null) || "".equals(userAttrNames.trim()))
				throw new LoginException(LMN + " " +
					"Error: Required option 'userAttrNames' is null or empty");

			String[] attrs = StringUtils.getStringArray(userAttrNames, ",");

			if (debug)
			{
				logger.debug(this, LMN + " Searching user LDAPEntry for:" +
					"\n\tuserBaseDN: " + userBaseDN +
					"\n\tuserFilter: " + userFilter +
					"\n\tuserScope: " + userScope +
					"\n\tuserAttrNames: " + userAttrNames);
			}

			// Search for the user entry attributes
			LDAPSearchResults searchResults = connection.search(userBaseDN,
				scopeInt, userFilter, attrs, false);

			// Process the search results to get the user dn
			if (searchResults.hasMore())
			{
				LDAPEntry entry = searchResults.next();
				if (entry != null)
				{
					if (debug)
						logger.debug(this, LMN + " login() succeeded");
					
					String userResultsNS = (String)options.get("userResultsNS");
					if ((userResultsNS == null) || "".equals(userResultsNS.trim()))
						userResultsNS = "user";
					
					String userCertAttrName = (String)options.get("userCertAttr");
					if ((userCertAttrName == null) || "".equals(userCertAttrName.trim()))
						userCertAttrName = "userCertificate";
					
					byte[] ldapUserCertBytes = null;
					
					// Put the user search attributes into the NamespaceMap
					// to they can be referenced by other config options.
					for (int i = 0; i < attrs.length; i++)
					{
						String attrName = attrs[i];
						LDAPAttribute attribute = entry.getAttribute(attrName);
						if (attribute != null)
						{
							if (debug)
								logger.debug(this, LMN +  " got attr named: " + attrName);
							
							if (attrName.equalsIgnoreCase(userCertAttrName))
							{
								if (debug)
									logger.debug(this, LMN + 
										" found user certificate attribute: " + 
										userCertAttrName + ". Saving attribute value as " +
										 "a byte array for certificate comparison.");

								ldapUserCertBytes = attribute.getByteValue();
							}
							else
							{
								String attrValue = attribute.getStringValue();
								
								if (debug)
									logger.debug(this, LMN + 
										" putting user attr: " + attrName + "='" +
										attrValue + "' into user result set");
								
								namespaceMap.put(userResultsNS, attrName, attrValue);
							}
						}
					}
					
					// Make sure user certificate data was queried from LDAP
					if (ldapUserCertBytes == null)
					{
						if (debug)
						{
							logger.debug(this, LMN + 
								" Did not find user certificate in LDAP attribute: " +
								userCertAttrName);
							logger.debug(this, LMN + 
								" Throwing FailedLoginException: can't compare certificates");
						}

						throw new FailedLoginException(
							LMN + " Authentication failed");
					}
					
					if (debug)
						logger.debug(this, LMN + " Attempting to create X509Certificate from byte array");

					// Create the certificate queried from the LDAP directory
					ByteArrayInputStream bais = new ByteArrayInputStream(ldapUserCertBytes);
					CertificateFactory cf = CertificateFactory.getInstance("X.509");
					X509Certificate ldapUserCert = (X509Certificate)cf.generateCertificate(bais);
					if (debug)
					{
						logger.debug(this, LMN + " Processing X509Certificate data queried via LDAP");
						logger.debug(this, LMN + " Successfully created X509Certificate from byte array");
						logger.debug(this, LMN + " ldapUserCert=\n" + ldapUserCert.toString());
						logger.debug(this, LMN + " Comparing Client-provided X509Certificate with the one obtained via LDAP");
					}
					
					// Compare the X.509 certificates using a full binary comparison.
					if (Arrays.equals(ldapUserCert.getEncoded(), certChain[0].getEncoded()))
					{
						if (debug)
						{
							logger.debug(this, LMN + " X509 certificates match");
							logger.debug(this, LMN + " Constructing authenticated username");
						}

						// The authenticated username is most likely an attribute that
						// was returned with the user search.
						String userSubjectName = (String)options.get("userSubjectName");
						if ((userSubjectName == null) || "".equals(userSubjectName.trim()))
							throw new LoginException("missing required option: userSubjectName");

						this.username = NamespaceMapUtils.replaceTokens(
							userSubjectName, namespaceMap, tds);

						if (debug)
							logger.debug(this, LMN + " login succeeded: username=" + username);
					
						this.loginSucceeded = true;
					}
					else
					{
						if (debug)
							logger.debug(this, "   X509 certificates don't match, login failed");
					}
				}
				else
				{
					if (debug)
						logger.debug(this, LMN + " login() failed");
								
					throw new FailedLoginException(
						LMN + " Authentication failed");
				}
			}
			else
			{
				if (debug)
				{
					logger.debug(this, LMN + " LDAPEntry not found for:" +
						"\n\tuserBaseDN: " + userBaseDN +
						"\n\tuserFilter: " + userFilter +
						"\n\tuserScope: " + userScope +
						"\n\tuserAttrNames: " + attrs);
				}
			
				throw new LoginException ("LDAPEntry not found for user");
			}
		}
		catch (LDAPException e)
		{
			if (debug)
			{
				logger.debug(this, LMN + 
					"Error while trying to search for subjectDN='" + subjectDN + "'" +
					"\n\tMessage: " + e.getMessage() +
					"\n\tLDAPErrorMessage: " + e.getLDAPErrorMessage() +
					"\n\tResultCode: " + e.getResultCode(), e);
			}

			// Login failure during LDAP bind.
			throw new LoginException(
				LMN + " Error while trying to " +
					"authenticate user: " + e.getLDAPErrorMessage() + "\n");
		}
		catch (CertificateException e)
		{
			if (debug)
				logger.debug(this, LMN + 
					"Error while trying to create user certificate queried from LDAP", e);

			// Login failure during LDAP bind.
			throw new LoginException(
				LMN + " Error while trying to " +
					"authenticate user: " + e.getMessage() + "\n");
		}

		// Add default roles (if any) to list of roles
		String defaultRoles = (String)options.get("defaultRoles");
		if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
		{
			StringTokenizer st = new StringTokenizer(defaultRoles, ",");
			while (st.hasMoreTokens())
				roleList.add(st.nextElement().toString().trim());
		}

		// If configured, search for roles stored in the LDAP directory.
		if (useRoleSearch)
		{
			String roleBaseDN = (String)options.get("roleBaseDN");
			if ((roleBaseDN == null) || "".equals(roleBaseDN.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'roleBaseDN'");

			String roleFilter = (String)options.get("roleFilter");
			if ((roleFilter == null) || "".equals(roleFilter.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'roleFilter'");

			String roleAttrName = (String)options.get("roleAttrName");
			if ((roleAttrName == null) || "".equals(roleAttrName.trim()))
				throw new LoginException(LMN + " Error: Missing required option 'roleAttrName'");

			String roleScope = (String)options.get("roleScope");
			if ((!"BASE".equalsIgnoreCase(roleScope)) &&
					(!"ONE".equalsIgnoreCase(roleScope)) &&
					(!"SUB".equalsIgnoreCase(roleScope)))
				throw new LoginException(LMN +
					" Error: Required option 'roleScope' must be 'BASE', 'ONE', or 'SUB'");

			// Set the search scope, the default of 0 is "BASE"
			int scopeInt = 0;
			if ("ONE".equalsIgnoreCase(roleScope))
				scopeInt = 1;
			else if ("SUB".equalsIgnoreCase(roleScope))
				scopeInt = 2;

			try
			{
				// Set the search constraints to automatically follow referrals
				LDAPConstraints ldapConstraints = new LDAPConstraints();
				ldapConstraints.setReferralFollowing(true);

				// Create and set the referral handler
				LDAPReferralHandler referralHandler = new AuthImpl(bindUsername, bindPassword);
				ldapConstraints.setReferralHandler(referralHandler);

				// Set the search constraints
				connection.setConstraints(ldapConstraints);

				// Substitute the user name in the role search filter. Optionally,
				// don't use the supplied domain in the user name to enable searches
				// using attributes other than the userPrincipalName, such as 
				// sAMAccountName, uid, or cn, which don't use the fully-
				// qualified UPN.
	
				String baseDN = NamespaceMapUtils.replaceTokens(roleBaseDN, namespaceMap, tds);
				String filter = NamespaceMapUtils.replaceTokens(roleFilter, namespaceMap, tds);

				String[] attrs = StringUtils.getStringArray(roleAttrName, ",");
				
				if (debug)
				{
					logger.debug(this, LMN + " Search " +
						"for the user's LDAP server groups using:" +
						"\n\tbaseDN: " + baseDN +
						"\n\tfilter: " + filter +
						"\n\t attrs: " + attrs[0]);
				}

				// Search for the user entry attributes
				LDAPSearchResults searchResults = connection.search(baseDN,
					scopeInt, filter, attrs, false);
					
				if (debug)
				{
					logger.debug(this, LMN + " After LDAP Search, searchResults=" + 
						searchResults);
					
					if (searchResults != null)
						logger.debug(this, LMN + " After LDAP Search, searchResults.hasMore()=" +
							searchResults.hasMore());
				}

				// Get the roles and save them for commits
				if (searchResults.hasMore())
				{
					LDAPEntry entry = searchResults.next();
					LDAPAttribute attribute = entry.getAttribute(attrs[0]);
					if (attribute != null)
					{
						// Get an enumeration of the attribute values
						Enumeration allValues = attribute.getStringValues();

						while (allValues.hasMoreElements())
						{
							// Get the role dn to add
							String role = (String)allValues.nextElement();

							if (debug)
								logger.debug(this, LMN + 
									" unexploded role '" + role +
									"' will be exploded");

							// Add the left most element of the role dn
							this.roleList.add(role);

							if (debug)
								logger.debug(this, LMN + 
									 " role '" + role + "' added to list");
						}
					}
				}
				else
				{
					logger.debug(this, LMN + " Search results were null");
				}
			}
			catch (LDAPReferralException e)
			{
				// Log any failed referrals.
				if (debug)
				{
					String[] referrals = e.getReferrals();
					for (int i = 0; i < referrals.length; i++)
					{
						logger.debug(this, LMN + " referral '" +
							referrals[i] + "' not found");
					}
				}
			}
			catch (LDAPException e)
			{
				if (debug)
					logger.debug(this, LMN + 
						" Error while searching for groups for subjectDN '" +
						subjectDN + "'", e);

				// Search filter failed
				throw new LoginException(
					LMN + " Error: " + e.getMessage() + "\n");
			}
			finally
			{
				// Close the connection
				if (connection != null)
				{
					try
					{
						connection.disconnect();
						connection = null;
					}
					catch(LDAPException e)
					{
						// Only log the exception
						logger.error(this, LMN + " Error while disconnecting", e);
					}
				}
			}
		}

		if (debug)
			logger.debug(this, LMN + 
				" Returning from authenticate(): loginSucceeded='" + loginSucceeded);
					
		// Return authentication results
		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @exception LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login failed, clean up state
		if (!loginSucceeded)
			return false;

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList();

		// Repository operations to get role principals
		Iterator roleListItr = roleList.iterator();
		while(roleListItr.hasNext())
		{
			String rolename = (String)roleListItr.next().toString().trim();
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				principalList.add(rolePrincipal);

				if (debug)
					logger.debug(this,
						LMN + " Added CSRolePrincipal " +
						rolename + " to principal ArrayList for user '" +
						username + "'");
			}
			else
			{
				if (debug)
					logger.debug(this,
						LMN + " CSRolePrincipal '" +
						rolename + "' already in Subject for '" +
						username + "', not added to principal List");
			}
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist

		Principal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if (debug)
				logger.debug(this,
					LMN + " Added CSUserPrincipal '" + username +
					"' to principal List");
		}
		else
		{
			if (debug)
				logger.debug(this,
					LMN + " CSUserPrincipal '" + username +
					"' already in Subject, not added to principal List");
		}

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the AESO variety.

		if(debug)
		{
			logger.debug(this, LMN + " Adding " +
				AuthenticationMethod.CAMS_AESO_LOGIN.getName() +
				" AuthenticationMethod");
		}

		subject.getPublicCredentials().add(
			AuthenticationMethod.CAMS_AESO_LOGIN.getName());

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.agentSecret = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @exception LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.
			if (this.roleList != null)
			{
				this.roleList.clear();
				this.roleList = null;
			}

			this.loginSucceeded = false;
			this.username = null;
			this.agentSecret = null;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @exception LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if (this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if (this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.logger = null;
		this.callbackHandler = null;
		this.username = null;
		this.agentSecret = null;
		this.options = null;
		this.loginSucceeded = false;
		this.commitSucceeded = false;
		this.debug = false;

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username
	 * instance variable.
	 *
	 * @exception LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException, FailedLoginException
	{
		// This LoginModule requires a username and an agent_secret
		MapCallback mapCallback = new MapCallback();
		Callback[] callbacks = new Callback[1];
		callbacks[0] = mapCallback;

		try
		{
			// Populate callback values.
			callbackHandler.handle(callbacks);

			// Put all callback values into a namespace map so they can
			// be referenced as substitution values in other options.
			String cbValueNS = (String)options.get("cbValueNS");
			if ((cbValueNS == null) || "".equals(cbValueNS))
				cbValueNS = "cb";

			Map cbValueMap = mapCallback.getCallbackValueMap();
			Iterator iter = cbValueMap.keySet().iterator();
			while (iter.hasNext())
			{
				String cbName = (String)iter.next();
				String cbValue = (String)cbValueMap.get(cbName);
				namespaceMap.put(cbValueNS, cbName, cbValue);
			}
			
			// Make sure all required callback values were provided.
			String s = (String)options.get("cbValuesRequired");
			if ((s == null) || "".equals(s.trim()))
				throw new LoginException("missing required configuration option: " + 
					"cbValuesRequired");
			
			String[] reqCBValues = StringUtils.getStringArray(s, ",");
			for (int i = 0; i < reqCBValues.length; i++)
			{
				String cbValue = (String)namespaceMap.get(cbValueNS, reqCBValues[i]);
				if ((cbValue == null) || "".equals(cbValue))
				{
					// If one of the required callback values is missing, then
					// insufficient credentials: login failed.
					if (debug)
						logger.debug(this, "missing required callback value: " + 
								reqCBValues[i]);
					
					throw new FailedLoginException(LMN + " Authentication failed");
				}
			}
	
			// Get the name of the callback value containing the client cert.
			// Set a default name if not provided.
			String cbValueX509ClientCert = (String)options.get("cbValue509ClientCert");
			if (cbValueX509ClientCert == null)
				cbValueX509ClientCert = "x509_client_cert";
			
			// Get the X.509 cert chain as a String.
			String certStr = mapCallback.getCallbackValue(cbValueX509ClientCert);
			if (certStr != null)
			{
				// Convert the cert callback value from a String to decoded form.
				this.certChain = X509CertificateUtils.decodeChainFromPEM(certStr);
			}
			
			// Get the agent secret
			this.agentSecret = mapCallback.getCallbackValue("agent_secret");
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available", e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() + " not available");
		}
		catch (CertificateException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available", e);

			throw new LoginException(LMN + " Error: " + e.getMessage());
		}

		if (debug)
		{
			if (this.certChain != null)
			{
				for (int i = 0; i < certChain.length; i++)
					logger.debug(this, "X509Certificate[" + i + "]=\n" +
					certChain[i].toString());
			}
			else
			{
				logger.debug(this, "No certificates supplied");
			}

			logger.debug(this, LMN + " AESO agent secret = " + agentSecret);
		}
	}

	/**
	 * Create a LDAPConnection.
	 * @param port 
	 * @param host 
	 *
     * @param host The host name for the connection.
     * @param port The port for the connection.

	 * @return a LDAPConnection that is ready for use.
	 * @throws LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getConnection(String host, String port)
		throws LoginException
	{
		try
		{
			if (debug)
				logger.debug(this, LMN + " " +
					"Instantiating LDAPConnection: host=" + host +
					", port=" + port + ", useSSL=" + useSSL);

			// Get the connection timeout
			String connectTimeout = (String)options.get("connectTimeout");
			connectTimeout = (connectTimeout != null ? connectTimeout : "3000");

			// Create the connection
			LDAPConnection connection = null;
			if (useSSL)
			{
				// Warn if required SSL System property for trust store is not set
				String trustStoreProp =
					System.getProperty("javax.net.ssl.trustStore");
				if (trustStoreProp == null)
					logger.warning(this,
						"Missing required SSL system property: " +
						"javax.net.ssl.trustStore");

				// Use an SSL connection
				connection =
					new LDAPConnection(new LDAPJSSESecureSocketFactory());
			}
			else
			{
				// Use a non-SSL connection
				connection = new LDAPConnection();
			}

			// Create a new LDAPSocketFactory that will delegate to the
			// default factory, but also provides connection timeout
			LDAPSocketFactory socketFactory = new TimeoutLDAPSocketFactory(
				connection.getSocketFactory(), Long.parseLong(connectTimeout));

			connection = new LDAPConnection(socketFactory);

			if (debug)
				logger.debug(this, LMN + " " +
					"Connecting to host=" + host +
					", port=" + port);

			// Connect to the directory
			connection.connect(host, Integer.parseInt(port));

			if (debug)
				logger.debug(this, LMN + " " +
					"Successfully connected to host=" + host +
					", port=" + port);

			if (debug)
				logger.debug(this, LMN + " Returning LDAPConnection");

			return connection;
		}
		catch (LDAPException e)
		{
			logger.error(this, "LDAP error", e);

			// Dump the nested Throwables to get at the root cause
			int i = 1;
			Throwable cause = e.getCause();
			do
			{
				if (cause != null)
					logger.error(this, "LDAP error: CAUSE " + i + ": " +
						cause.getMessage());
			}
			while ((cause = cause.getCause()) != null);

			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
	}

	/**
	 * A simple LDAPAuthHandler for following referrals
	 */
	private class AuthImpl implements LDAPAuthHandler
	{
		// The dn used to authenticate
		private String dn;

		// The password used to authenticate
		private byte[] passwd;

		/**
		 * Construct this instance using data
		 */
		private AuthImpl(String username, String password)
		{

			// Create an LDAPAuthProvider class to encapsulate the
			// authentication credentials.
			dn = username;
			try
			{
				passwd = password.getBytes("UTF8");
			}
			catch (UnsupportedEncodingException e)
			{
				// Should never get this
				logger.error(this,
					"Unable to encode password to UTF-8 for user: " + dn);
			}

			return;
		}

		/**
		 * Choose the login dn and password based on the host and port
		 * (this implementation simply reuses the original credentials)
		 */
		public LDAPAuthProvider getAuthProvider(String host, int port)
		{
			if (debug)
				logger.debug(this, "Following directory referral, rebinding " +
					"to: " + host + ":" + port);

			return new LDAPAuthProvider(dn, passwd);
		}
	}

} // End of class : AESOX509CertLdapLoginModule
