/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-
 * INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR
 * ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.cams.auth.login.userrepository.RepositoryUser;
import com.cafesoft.cams.auth.login.userrepository.UserRepository;
import com.cafesoft.cams.auth.login.userrepository.UserRepositoryException;

import com.cafesoft.cams.service.UserRepositoryService;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.service.ServiceClient;
import com.cafesoft.core.service.ServiceException;
import com.cafesoft.core.service.ServiceFinder;

import javax.security.auth.Subject;

import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;

import java.security.Principal;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * <code>AESOXmlLoginModule</code> implements the JAAS LoginModule interface 
 * and works in conjunction with Cams "Automatic Enterprise Sign-On" (AESO) to 
 * authenticate users and assign roles using information stored in a XML user
 * repository. This particular implementation uses a ServiceFinder object to 
 * access the XML user repository.
 * <P>
 * NOTE: This login module is appropriate for simple sites with a small number
 * of users or for Cams development and testing.
 * <P>
 * If support for one of these two authentication mechanisms is not configured,
 * the this login module will successfully authenticate the user named in the
 * "username" callback value and assign roles as configured in the XML user
 * repository.
 * <P>
 * All configuration parameters are listed below:
 * <Ul>
 * <LI><B>serviceId</B> - The unique name of the XML User Repository
 * Service that you are assigning to this login module.</LI>
 * <LI><B>debug</B> - Flag that enables/disables debugging.
 * <li><b>agentSecret</b> - The expected value for the "agent_secret" callback
 * value, which must be present in the authentication request. This value is 
 * used to authenticated Cams agents, which assert the identity of a user
 * already authenticated at the web layer. The agent secret is used to protect 
 * against impersonation attacks via carefully created HTTP requests 
 * attempting to subvert AESO component trust relationships.</li>
 * <li><B>defaultRoles</B> - (Optional) A list of one or more default roles
 * separated by a comma to be added to the user session on successful
 * authentication.</li>
 * </UL>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.3 $ $Date: 2009/08/14 23:49:51 $ $Author: norb $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 09/23/2008
 */
public class AESOXmlLoginModule
	implements LoginModule, ServiceClient, LoggerClient
{

	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Map that contains state shared with other configured LoginModules.
	 */
	private Map sharedState;

	/**
	 * Used by the LoginModule to lookup required Cams services.
	 */
	private ServiceFinder serviceFinder;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username supplied via Cams AESO.
	 */
	private String username;

	/**
	 * The agent secret value.
	 */
	private String agentSecret;

	/**
	 * The XMLUserRepository the LoginModule authenticates against.
	 */
	private UserRepository userRepository;

	//
	// Constructor
	//

	/**
	 * Default AESOXmlLoginModule Constructor.
	 */
	public AESOXmlLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.serviceFinder = null;
		this.sharedState = null;
		this.subject = null;
		this.username = null;
		this.agentSecret = null;
		this.userRepository = null;
	}

	//
	// Implements the ServiceClient interface
	//

	/**
	 * Set the ServiceFinder.
	 *
	 * @param finder the class used by find Services by the ServiceClient.
	 */
	public void setServiceFinder(ServiceFinder finder)
	{
		this.serviceFinder = finder;
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *			with the end user (prompting for usernames and
	 *			passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.sharedState = sharedState;
		this.debug = "true".equals((String)options.get("debug"));
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @throws LoginException thrown if any error occurs while trying to
	 * authenticate the user.
	 * @throws FailedLoginException if authentication fails.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException
	{
		String expectedAgentSecret = (String)options.get("agentSecret");
		if ((expectedAgentSecret == null) || (expectedAgentSecret.trim().length() == 0))
			throw new LoginException("[AESOXmlLoginModule] Error: Required " +
				"'agentSecret' property missing");

		handleCallbacks();

		// An agent secret must always be supplied as a callback value and
		// its value must match the expected value.
		if (agentSecret == null || agentSecret.length() == 0)
			throw new LoginException("[AESOXmlLoginModule] Error: Required callback value 'agent_secret' " +
				"does not exist or is empty");
		else if (agentSecret.equals(expectedAgentSecret) == false)
			throw new LoginException("[AESOXmlLoginModule] Error: Callback value 'agent_secret' " +
				"does not match expected value in option 'agentSecret'");

		// A username must always be supplied
		if (username == null || username.length() == 0)
			throw new FailedLoginException(
				"[AESOXmlLoginModule] Authentication failed");

		if (userRepository == null)
			getUserRepository();

		try
		{
			// Try to get the user
			RepositoryUser ru = userRepository.getUser(username);

			// If the user exists, get the user's password
			if (ru != null)
			{
				loginSucceeded = true;
			}
			else
			{
				if (debug)
				{
					// No user by that name
					logger.debug(this, "[AESOXmlLoginModule] User '" + username +
						"' does not exist");
				}
			}
		}
		catch(UserRepositoryException e)
		{
			logger.error(this, "Error accessing UserRepository", e);
			throw new LoginException("[AESOXmlLoginModule] Error: " +
				e.getMessage() + "\n");
		}

		// Authentication failed, clean up state and throw FailedLoginException.
		if(!loginSucceeded)
		{
			this.username = null;
			this.agentSecret = null;

			throw new FailedLoginException(
				"[AESOXmlLoginModule] Authentication failed");
		}

		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @throws LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login failed
		if(!loginSucceeded)
			return false;

		if(userRepository == null)
			getUserRepository();

		List principalList = new ArrayList(5);

		try
		{
			// Add default roles (if any) to list of roles
			String defaultRoles = (String)options.get("defaultRoles");
			if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
			{
				StringTokenizer st = new StringTokenizer(defaultRoles, ",");
				while(st.hasMoreTokens())
					principalList.add(
						new CSRolePrincipal(st.nextElement().toString().trim()));
			}

			// Get the user and the user roles
			RepositoryUser ru = userRepository.getUser(username);
			String[] r = ru.getRoles();

			// Add the user roles
			if((r != null) && (r.length > 0))
			{
				for(int i = 0; i < r.length; i++)
				{
					// Add the CSRolePrincipal to the Subject
					// only if it does not exist
					Principal rolePrincipal = new CSRolePrincipal(r[i].trim());
					if(!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
					{
						principalList.add(rolePrincipal);

						if(debug)
						{
							logger.debug(this,
								"[AESOXmlLoginModule] Added CSRolePrincipal " +
								r[i] + " to principal List for user '" +
								username + "'");
						}
					}
					else
					{
						if (debug)
						{
							logger.debug(this,
								"[AESOXmlLoginModule] CSRolePrincipal '" + r[i] +
								"' already in Subject for '" + username +
								"', not added to principal List");
						}
					}
				}
			}
		}
		catch (UserRepositoryException e)
		{
			logger.error(this, "Error in UserRepositoryService", e);
			throw new LoginException("[AESOXmlLoginModule] Error: " +
				e.getMessage() + "\n");
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if (debug)
			{
				logger.debug(this, "[AESOXmlLoginModule] Added CSUserPrincipal '" +
					username + "' to principal List");
			}
		}
		else
		{
			if (debug)
			{
				logger.debug(this, "[AESOXmlLoginModule] CSUserPrincipal '" +
					username + "' already in Subject, not added to principal List");
			}
		}

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		if(debug)
		{
			logger.debug(this, "[AESOXmlLoginModule] Adding " +
				AuthenticationMethod.PASSWORD.getName() +
				" AuthenticationMethod");
		}

		subject.getPublicCredentials().add(
			AuthenticationMethod.PASSWORD);

		commitSucceeded = true;

		// Clean up
		this.username = null;
		this.agentSecret = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @throws LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{

		// If authentication failed return false. If commit
		// succeeded log the user out return true. Otherwise,
		// Cleanup the login module and and return true.

		if(!loginSucceeded)
			return false;

		if(commitSucceeded)
			logout();
		else
			cleanup();

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @throws LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if(subject != null)
			subject.getPrincipals().clear();

		cleanup();

		return true;
	}

	/**
	 * Clean up the resources of the LoginModule
	 */
	private void cleanup()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.logger = null;
		this.loginSucceeded = false;
		this.options = null;
		this.serviceFinder = null;
		this.sharedState = null;
		this.subject = null;
		this.username = null;
		this.agentSecret = null;
		this.userRepository = null;
	}

	/**
	 * Populates Callback values.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// Check for a callbackHandler
		if (callbackHandler == null)
			throw new LoginException(
				"[AESOXmlLoginModule] Error: no CallbackHandler available");

		// This LoginModule requires a username
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Populate the callback values.
			callbackHandler.handle(callbacks);

			// The username is required.
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");

			// Get the agent secret
			this.agentSecret = ((MapCallback)callbacks[0]).getCallbackValue("agent_secret");
		}
		catch (IOException e)
		{
			logger.error(this, "[AESOXmlLoginModule] Login error occurred", e);
			throw new LoginException("[AESOXmlLoginModule] Error: " + e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				"[AESOXmlLoginModule] Login error occurred Callback not available", e);

			throw new LoginException("[AESOXmlLoginModule] Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (debug)
			logger.debug(this,
				"[AESOXmlLoginModule] User entered username = " + username);
	}

	/**
	 * Gets the XMLUserRepository for the LoginModule.
	 *
	 * @throws LoginException if an error occurs getting the
	 * UserRespository object.
	 */
	private void getUserRepository()
		throws LoginException
	{
		try
		{
			if(serviceFinder == null)
				throw new LoginException(
					"[AESOXmlLoginModule] Error: ServiceFinder is null");

			String serviceId = (String)options.get("serviceId");
			if((serviceId == null) || (serviceId.trim().length() == 0))
				throw new LoginException("[AESOXmlLoginModule] Error: Required " +
					"'serviceId' property missing");

			UserRepositoryService urs = (UserRepositoryService)
				serviceFinder.find(serviceId, UserRepositoryService.class);
			userRepository = urs.getUserRepository();
		}
		catch(ServiceException e)
		{
			logger.error(this, "Error finding UserRepositoryService", e);
			throw new LoginException("[AESOXmlLoginModule] Error: " +
				e.getMessage());
		}
	}

} // End of class : AESOXmlLoginModule

