/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose
 * such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;

import com.cafesoft.core.service.ServiceClient;
import com.cafesoft.core.service.ServiceFinder;

import com.cafesoft.core.util.Destroyable;

import java.util.Map;

import javax.security.auth.Subject;

import javax.security.auth.callback.CallbackHandler;

import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

/**
 * Abstract implementation of the LoginModule interface. This class provides
 * implementation for standard methods used to plug LoginModules into the Cams
 * policy server infrastructure. For instance, the ability to set the Logger
 * and ServiceFinder objects. In addition, this method has stubbed out the
 * methods that must be implemented by subclasses.
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.5 $ $Date: 2009/08/14 23:49:51 $ $Author $
 * @since 09/09/05
 * @see com.cafesoft.core.log.LoggerClient
 * @see com.cafesoft.core.service.ServiceClient
 * @see javax.security.auth.spi.LoginModule
 */
public abstract class AbstractLoginModule
	implements LoginModule, LoggerClient, ServiceClient, Destroyable
{
	//
	// Instance Variables
	//

	/**
	 * CallbackHandler that provides the means for accessing user supplied data.
	 * @since 1.0
	 */
	protected CallbackHandler callbackHandler;

	/**
	 * Enables/disables debugging.
	 * @since 1.0
	 */
	protected boolean debug;

	/**
	 * LoginModule's logger. This logger will write all debug to the security
	 * domain specific log file.
	 * @since 1.0
	 */
	protected Logger logger;

	/**
	 * The LoginModule's configuration options.
	 * @since 1.0
	 */
	protected Map options;

	/**
	 * ServiceFinder that makes Cams services available to the LoginModule.
	 * @since 1.0
	 */
	protected ServiceFinder serviceFinder;

	/**
	 * Map that contains state shared with other configured LoginModules.
	 * @since 1.0
	 */
	protected Map sharedState;

	/**
	 * The subject being authenticated.
	 */
	protected Subject subject;


	//
	// Constructor
	//

	/**
	 * Creates a new AbstractLoginModule.
	 *
	 * @since 1.0
	 */
	public AbstractLoginModule()
	{
		this.callbackHandler = null;
		this.debug = false;
		this.logger = null;
		this.options = null;
		this.serviceFinder = null;
		this.sharedState = null;
		this.subject = null;
	}

	//
	// Implements the LoggerClient interface.
	//

	/**
	 * Sets the logger.
	 *
	 * @param logger logs messages.
	 * @since 1.0
	 */
	public final void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the ServiceClient interface.
	//

	/**
	 * Set the ServiceFinder.
	 *
	 * @param finder the class used by find Services by the ServiceClient.
	 * @since 1.0
	 */
	public final void setServiceFinder(ServiceFinder finder)
	{
		this.serviceFinder = finder;
	}

	//
	// Implements the Destroyable interface.
	//

	/**
	 * Destroy the object and it's resources.
	 * <p>
	 * This method when called should free all resources used
	 * by this object. In addition, it should render the object
	 * unusable. Therefore, this method should only be called once
	 * per object.
	 *
	 * @since 1.0
	 */
	public void destroy()
	{
		this.callbackHandler = null;
		this.debug = false;
		this.logger = null;
		this.options = null;
		this.serviceFinder = null;
		this.sharedState = null;
		this.subject = null;
	}

	//
	// Implements the LoginModule interface.
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 * with the end user (prompting for usernames and passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login <code>Configuration</code>
	 * for this particular <code>LoginModule</code>.
	 * @since 1.0
	 */
	public void initialize(Subject subject,
						   CallbackHandler callbackHandler,
						   Map sharedState,
						   Map options)
	{
		this.callbackHandler = callbackHandler;
		this.debug = "true".equals((String)options.get("debug"));
		this.options = options;
		this.sharedState = sharedState;
		this.subject = subject;
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @throws LoginException if the authentication fails.
	 * @return true if the authentication succeeded, or false if this
	 * <code>LoginModule</code> should be ignored.
	 * @since 1.0
	 */
	public final boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @throws LoginException if the commit fails.
	 * @return true if this method succeeded, or false if this
	 * <code>LoginModule</code> should be ignored.
	 * @since 1.0
	 */
	public final boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @throws LoginException if the abort fails.
	 * @return true if this method succeeded, or false if this
	 * <code>LoginModule</code> should be ignored.
	 * @since 1.0
	 */
	public final boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @throws LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 * <code>LoginModule</code> should be ignored.
	 */
	public final boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Template Methods to be overridden
	//

	/**
	 * Authenticate the user.
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @throws LoginException if an error occurs authenticating.
	 * @since 1.0
	 */
	protected abstract boolean authenticate() throws LoginException;

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @throws LoginException if an error occurs associating user roles.
	 * @since 1.0
	 */
	protected abstract boolean associateRoles() throws LoginException;

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @throws LoginException if an error occurs aborting authentication.
	 * @since 1.0
	 */
	protected abstract boolean abortAuthentication() throws LoginException;

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @throws LoginException if an error occurs logging user out.
	 * @since 1.0
	 */
	protected abstract boolean logoutUser() throws LoginException;

} // End of class : AbstractLoginModule
