/**
 * Copyright (c) 1996-2012 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.service.ServiceClient;
import com.cafesoft.core.service.ServiceException;
import com.cafesoft.core.service.ServiceFinder;

import com.cafesoft.cams.service.ActiveDirectoryGroupNameService;

import com.novell.ldap.LDAPAuthHandler;
import com.novell.ldap.LDAPAuthProvider;
import com.novell.ldap.LDAPAttribute;
import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPConstraints;
import com.novell.ldap.LDAPDN;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;
import com.novell.ldap.LDAPReferralException;
import com.novell.ldap.LDAPReferralHandler;
import com.novell.ldap.LDAPSearchResults;
import com.novell.ldap.LDAPJSSESecureSocketFactory;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;
import java.io.UnsupportedEncodingException;

import java.security.Principal;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.StringTokenizer;

/**
 * Implements the LoginModule interface found in JAAS to authenticate users
 * against an Active Directory repository. This implementation attempts to
 * bind the connection with the user supplied credentials. If successful,
 * the user roles are retrieved and the user is authenticated.
 *
 * All configuration parameters are listed below:
 * <ul>
 * <li><b>host</b> - The DNS name or IP address of the Active Directory
 * host.</li>
 * <li><b>port</b> - The host's Active Directory port to use.</li>
 * <li><b>useSSL</b> - Enables/disables use of SSL.</li>
 * <li><b>connectTimeout</b> - (optional) The number of milliseconds to wait 
 * for a connection to the ActiveDirectory before timing out.</li>
 * <li><b>emptyPassword</b> - (optional) Enable empty passwords (true or false).
 * By default, the ActiveDirectoryLoginModule does not allow a password
 * to be empty. Set this value to true to allow use of empty passwords
 * when Active Directory is configured for and anonymous binds are
 * required.</li>
 * <li><b>defaultRoles</b> - (optional) A list of one or more default roles
 * separated by a comma to be added to the user session.</li>
 * <li><b>roleRegexPattern</b> - (optional) Limits roles saved in the user 
 * session to those that match a regular expression pattern. This option is 
 * useful when the role search returns many roles that are not used by the 
 * access control policy. 
 * <li><b>domain</b> - (optional) The Active Directory domain to use. If this
 * value is present and the user name does not end with @domain, then append
 * "@" and the value of domain to the user name before attempting to bind.</li>
 * <li><b>useRoleSearch</b> - (optional) A value of "true" (default) 
 * indicates that a role search will be attempted [true|false].</li>
 * <li><b>useDomainInRoleSearch</b> - (optional) A value of "true" (default) 
 * indicates that an optionally supplied "domain" used for authentication
 * will be included in the role search filter's {username} substitution value 
 * [true|false].</li>
 * <li><b>baseDN</b> - (conditional) The base distinguished name for the User entries.
 * <li><b>scope</b> - (conditional) The scope for the role search. Valid values
 *   are "BASE", "ONE", and "SUB".</li>
 * <li><b>filter</b> - (conditional) The role search filter.</li>
 * <li><b>roleAttr</b> - (conditional) The user entity attribute where the roles
 * are listed.</li>
 * <li><b>addPrivateCredentials</b> - (optional) A flag that indicates whether
 * or not a java.net.PasswordAuthentication object containing the user name and
 * password should be added to the Subject's private credential set on
 * successful authentication. 
 *
 * WARNING: The values in this object can be accessed by other Cams components 
 * to enable SSO to other other systems, so administrator must take care that only 
 * trusted components have are used!</li>
 * </ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.39 $ $Date: 2012/08/22 22:05:11 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 10/20/01
 */
public class ActiveDirectoryLoginModule
	implements LoginModule, LoggerClient, ServiceClient
{
	//
	// Static Variables.
	//

	/**
	 * The LoginModule name (used in log messages throughout the code).
	 */
	private static final String LMN = "[ActiveDirectoryLoginModule]";
	
	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag indicating that SSL must be used.
	 */
	private boolean useSSL;
	
	/**
	 * Flag indicating that a role search will be tried.
	 */
	private boolean useRoleSearch;

	/**
	 * Flag indicating that a domain value, if supplied, will be used in the 
	 * user name substitution value for a role search.
	 */
	private boolean useDomainInRoleSearch;
	
	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating user.
	 */
	private String username;

	/**
	 * The fully qualified bind username for the authenticating user .
	 */
	private String bindUsername;

	/**
	 * The password entered by the authenticating User.
	 */
	private String password;

	/**
	 * A default list of user roles to associate with an authenticated user.
	 */
	private List roleList;

	/**
	 * Used to lookup required Cams services.
	 */
	private ServiceFinder serviceFinder;

	/**
	 * A flag that enables/disables addition of credentials to the Subject
	 * private credential store.
	 */
	private boolean addPrivateCredentials;

	//
	// Constructor
	//

	/**
	 * Default ActiveDirectoryLoginModule Constructor.
	 */
	public ActiveDirectoryLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.useSSL = false;
		this.useRoleSearch = true;
		this.useDomainInRoleSearch = true;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.password = null;
		this.subject = null;
		this.username = null;
		this.roleList = new ArrayList();
		this.serviceFinder = null;
		this.addPrivateCredentials = false;
	}

	//
	// Implements the ServiceClient interface
	//

	/**
	 * Set the ServiceFinder.
	 *
	 * @param finder the class used by find Services by the ServiceClient.
	 */
	public void setServiceFinder(ServiceFinder finder)
	{
		this.serviceFinder = finder;
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *			with the end user (prompting for usernames and
	 *			passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.debug = "true".equalsIgnoreCase((String)options.get("debug"));
		this.useSSL = "true".equalsIgnoreCase((String)options.get("useSSL"));
		this.useRoleSearch = 
			! "false".equalsIgnoreCase((String)options.get("useRoleSearch"));
		this.useDomainInRoleSearch = 
			! "false".equalsIgnoreCase((String)options.get("useDomainInRoleSearch"));
		this.addPrivateCredentials = "true".equals(
			(String)options.get("addPrivateCredentials"));
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt  succeeded (checked
	 * by retrieving the private state saved by the <code>login</code> method),
	 * then this method cleans up any state that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 * <p>
	 * This method executes the following steps:
	 * <ol>
	 * <li>Checks preconditions
	 * <li>Handles callback objects
	 * <li>Get a connection to the Active Directory
	 * <li>Authenticates the users using bind (only method available
	 * with Active Directory)
	 * <li>Close the connection
	 * <li>Reset state if authentication failed
	 * <li>Return results of authentication attempt
	 * </ol>
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @exception LoginException thrown if any error occurs while trying to
	 * authenticate the user.
	 * @exception FailedLoginException if authentication fails.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException,
		AccountExpiredException, CredentialExpiredException
	{
		// Check preconditions
		validateOptions();

		// Handle callback objects.
		handleCallbacks();

		// A user name must always be supplied
		if (username == null || username.length() == 0)
			throw new FailedLoginException(LMN + " Authentication failed");

		// Optionally append the domain to the user name if supplied. Do
		// not modify username as it is required "as is" for the role filter.
		String domain = (String)options.get("domain");
		bindUsername = username;
		if ((domain != null) && ! ("".equals(domain.trim())))
			bindUsername = username + "@" + domain;

		if (debug)
			logger.debug(this,
				 LMN + " Will attempt to bind " +
				 "with user name = '" + bindUsername + "'");

		// A non-empty password must be supplied unless empty
		// passwords are permitted for this configuration.
		String emptyPassword = (String)options.get("emptyPassword");
		if ((emptyPassword == null) || ("false".equals(emptyPassword.trim())))
		{
			if ("".equals(this.password))
				throw new FailedLoginException(
					LMN + " Authentication failed");
		}
		else if (!("true".equals(emptyPassword.trim())))
		{
			throw new LoginException(LMN +
				" Error: Option 'emptyPassword' must be 'true' or 'false'");
		}

		// Get a connection to Active Directory
		LDAPConnection connection = getConnection();

		if (debug)
			logger.debug(this, LMN + " Connected to Active Directory: '" + 
				connection.isConnected() + "'");

		// Bind to Active Directory with the user supplied credentials
		try
		{
			// Authenticate the user by attempting to bind and get the user roles.
			// Bind the connection to Active Directory. If successful, the user
			// credentials are valid.
			connection.bind(3, bindUsername, password.getBytes());

			// Authentication has succeeded
			loginSucceeded = true;
		}
		catch (LDAPException e)
		{
			// Normal login failure due to invalid credentials, check
			// Active Directory server message for subcode to determine the
			// most accurate exception to throw. Possible codes include:
			// 52b - invalid password (FailedLoginException)
			// 52e - invalid credentials (FailedLoginException)
			// 530 - invalid login time (FailedLoginException)
			// 532 - password expired (CredentialExpiredException)
			// 533 - account disabled (AccountExpiredException)
			// 701 - account expired (AccountExpiredException)
			// 773 - password reset required (CredentialExpiredException)
			if (e.getResultCode() == e.INVALID_CREDENTIALS)
			{
				if (e.getLDAPErrorMessage().indexOf("data 532") > 0)
				{
					throw new CredentialExpiredException(LMN +
						" Password expired: " +
						e.getLDAPErrorMessage() + "\n");
				}
				else if (e.getLDAPErrorMessage().indexOf("data 533") > 0)
				{
					throw new AccountExpiredException(LMN +
						" Account disabled: " +
						e.getLDAPErrorMessage() + "\n");
				}
				else if (e.getLDAPErrorMessage().indexOf("data 701") > 0)
				{
					throw new AccountExpiredException(LMN +
						" Account expired: " +
						e.getLDAPErrorMessage() + "\n");
				}
				else if (e.getLDAPErrorMessage().indexOf("data 773") > 0)
				{
					throw new CredentialExpiredException(LMN +
						" Password reset required: " +
						e.getLDAPErrorMessage() + "\n");
				}
				else
				{
					throw new FailedLoginException(LMN +
						" Invalid credentials: " +
						e.getLDAPErrorMessage() + "\n");
				}
			}
			else
			{
				// Login failure for reason other than invalid credentials

				// The login failure is unexpected and possibly due to missconfiguration,
				// log the exception to facilitate customer resolution
				logger.error(this, LMN + " Error while trying to " +
					"bind user: '" + this.username + "'" +
					"\n\tMessage: " + e.getMessage() +
					"\n\tLDAPErrorMessage: " + e.getLDAPErrorMessage() +
					"\n\tResultCode: " + e.getResultCode(), e);

				throw new LoginException(LMN +
					" Error while trying to bind user: " +
					e.getLDAPErrorMessage() + "\n");
			}
		}

		if (debug)
			logger.debug(this, LMN + " Successful bind  as user: '" +
				bindUsername + "'");

		// Associate roles from groups that the user is a member of

		// Add default roles (if any) to list of roles
		String defaultRoles = (String)options.get("defaultRoles");
		if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
		{
			StringTokenizer st = new StringTokenizer(defaultRoles, ",");
			while(st.hasMoreTokens())
				roleList.add(st.nextElement().toString().trim());
		}

		// Check the required options when using a role search
		if (useRoleSearch)
		{
			String baseDN = (String)options.get("baseDN");
			if ((baseDN == null) || "".equals(baseDN.trim()))
				throw new LoginException(LMN +
					" Error: Missing required option 'baseDN'");

			String filter = (String)options.get("filter");
			if ((filter == null) || "".equals(filter.trim()))
				throw new LoginException(LMN +
					" Error: Missing required option 'filter'");

			String scope = (String)options.get("scope");
			if ((!"BASE".equalsIgnoreCase(scope)) &&
					(!"ONE".equalsIgnoreCase(scope)) &&
					(!"SUB".equalsIgnoreCase(scope)))
				throw new LoginException(LMN +
					" Error: Required option 'scope' must be 'BASE', 'ONE', or 'SUB'");

			// Set the search scope, the default of 0 is "BASE"
			int scopeInt = 0;
			if ("ONE".equalsIgnoreCase(scope))
				scopeInt = 1;
			else if ("SUB".equalsIgnoreCase(scope))
				scopeInt = 2;

			// Get and validate the attribute where roles are stored (the
			// "memberOf" groups attribute is often used but we'll enable
			// external configuration to supply an alternative attribute.
			String roleAttr = (String)options.get("roleAttr");
			if ((roleAttr == null) || "".equals(roleAttr.trim()))
				throw new LoginException(LMN +
					" Error: Missing required option 'roleAttr'");

			try
			{
				// Set the search constraints to automatically follow referrals
				LDAPConstraints ldapConstraints = new LDAPConstraints();
				ldapConstraints.setReferralFollowing(true);

				// Create and set the referral handler
				LDAPReferralHandler referralHandler = new AuthImpl();
				ldapConstraints.setReferralHandler(referralHandler);

				// Set the search constraints
				connection.setConstraints(ldapConstraints);

				// Specify the attributes that the search will return
				String[] attrs = {roleAttr};

				// Substitute the user name in the role search filter. Optionally,
				// don't use the supplied domain in the user name to enable searches
				// using attributes other than the userPrincipalName, such as 
				// sAMAccountName, uid, or cn, which don't use the fully-
				// qualified UPN.
				if (useDomainInRoleSearch)
					filter = replaceUsername(filter, "username", bindUsername);
				else
					filter = replaceUsername(filter, "username", username);

				// If supplied, only use roles that match roleRegexPattern
				String roleRegexPattern = (String)options.get("roleRegexPattern");
				
				// The "tokenGroups" attribute contains all groups of which a
				// a user is member (including primary and nested groups). However,
				// it is only returned if you search with the user's
				// distinguishedName. So, we first search to get the user's
				// distinguishedName, then we do a read to get the "tokenGroups".
				// Because the "tokenGroups" are returned as SIDs, we then
				// use a Cams service to efficiently obtain their names.
				if ("tokenGroups".equals(attrs[0]))
				{
					// Get the user's distinguishedName
					attrs[0] = "distinguishedName";

					if (debug)
					{
						logger.debug(this, LMN + 
							" Search for the user's 'distinguishedName' using:" +
							  "\n\tbaseDN: " + baseDN +
							  "\n\tfilter: " + filter +
							  "\n\t attrs: " + attrs[0]);
					}

					// Search for the user entry attributes
					LDAPSearchResults searchResults = connection.search(baseDN,
						scopeInt, filter, attrs, false);

					// Process the search results to get the user dn
					if (searchResults.hasMore())
					{
						LDAPEntry entry = searchResults.next();
						LDAPAttribute attribute = entry.getAttribute(attrs[0]);
						if (attribute != null)
							baseDN = attribute.getStringValue();
					}
					else
					{
						throw new LoginException ("Attribute 'distinguishedName' " +
							"not found for user '" + bindUsername + "'");
					}

					// Now that we have the user's dn, looking for tokenGroups
					attrs[0] = "tokenGroups";

					if (debug)
						logger.debug(this, LMN +
							"'tokenGroups' attribute(s) for user '" + baseDN + "'");

					// Read the user's tokenGroups values
					LDAPEntry entry = connection.read(baseDN, attrs);

					// Save the Active Directory groups as roles for commit method
					if (entry != null)
					{
						LDAPAttribute attribute = entry.getAttribute(attrs[0]);

						// Get an enumeration of the attribute values
						Enumeration allValues = attribute.getByteValues();
						String role = null;
						String sid = null;

						if ((roleRegexPattern != null) && !("".equals(roleRegexPattern)))
						{
							Pattern pattern = Pattern.compile(roleRegexPattern);
							while (allValues.hasMoreElements())
							{
								// Get the sid
								sid = Sid2String((byte[])allValues.nextElement());

								if (debug)
									logger.debug(this, LMN + " Search result SID: '" + sid + "'");
									
								role = getGroupName(sid);

								if (debug)
									logger.debug(this, LMN + " SID role name: '" + role + "'");

								if (pattern.matcher(role).matches())
								{
									this.roleList.add(role);
																			if (debug)
									logger.debug(this, LMN + 
										" Role: '" + role + 
										"' matched roleRegexPattern: '" + 
										roleRegexPattern + "'");
								}
								else
								{
									logger.debug(this, LMN + 
										" Role: '" + role + 
										"' did not match roleRegexPattern: '" + 
										roleRegexPattern + "'");
								}
							}
						}
						else
						{
							while (allValues.hasMoreElements())
							{
								// Get the sid
								sid = Sid2String((byte[]) allValues.nextElement());

								if (debug)
									logger.debug(this, LMN + " Search result SID: '" + sid + "'");

								role = getGroupName(sid);

								if (debug)
									logger.debug(this, LMN + " SID role name: '" + role + "'");

								this.roleList.add(role);
							}
						}
					}
				}

				// The role attribute is "memberOf", so we need to get the
				// get string values and explode them to get the group names
				else if ("memberOf".equals(attrs[0]))
				{
					if (debug)
					{
						logger.debug(this, LMN + 
							" Search for the user's Active Directory groups using:" +
							  "\n\tbaseDN: " + baseDN +
							  "\n\tfilter: " + filter +
							  "\n\t attrs: " + attrs[0]);
					}					

					// Search for the user entry attributes
					LDAPSearchResults searchResults = connection.search(baseDN,
						scopeInt, filter, attrs, false);

					// Get the roles and save them for commits
					if (searchResults.hasMore())
					{
						LDAPEntry entry = searchResults.next();
						LDAPAttribute attribute = entry.getAttribute(attrs[0]);
						if (attribute != null)
						{
							// Get an enumeration of the attribute values
							Enumeration allValues = attribute.getStringValues();
							String role = null;

							if ((roleRegexPattern != null) && !("".equals(roleRegexPattern)))
							{
								Pattern pattern = Pattern.compile(roleRegexPattern);
								while (allValues.hasMoreElements())
								{
									// Get the role dn to add
									role = (String)allValues.nextElement();

									if (debug)
										logger.debug(this, LMN + 
											" Exploding fully-qualifed role: '" + 
											role + "'");

									// Active Directory returns memberOf attributes as
									// fully qualified DNs, get the common name for each
									// by exploding the DN into parts
									String[] explodedDN = LDAPDN.explodeDN(role, true);

									role = explodedDN[0].trim();

									if (pattern.matcher(role).matches())
									{
										this.roleList.add(role);

										if (debug)
											logger.debug(this, LMN + 
												" Role: '" + role + 
												"' matched roleRegexPattern: '" + 
												roleRegexPattern + "'");
									}
									else
									{
										if (debug)
											logger.debug(this, LMN + 
												" Role: '" + role + 
												"' did not match roleRegexPattern: '" + 
												roleRegexPattern + "'");
									}
								}
							}
							else
							{
								while (allValues.hasMoreElements())
								{
									// Get the role dn to add
									role = (String)allValues.nextElement();

									if (debug)
										logger.debug(this, LMN + 
											" Exploding fully-qualifed role: '" + 
											role + "'");

									// Active Directory returns memberOf attributes as
									// fully qualified DNs, get the common name for each
									// by exploding the DN into parts
									String[] explodedDN = LDAPDN.explodeDN(role, true);

									role = explodedDN[0].trim();

									if (debug)
										logger.debug(this, LMN + 
											" Using role name: '" + role + "'");

									this.roleList.add(role);
								}
							}
						}
					}
					else
					{
						logger.debug(this, LMN +
						" No search results returned");
					}
				}
				else
				{
					logger.error(this, LMN +
						" Unsupported role search attribute, must be " +
						"'memberOf' or 'tokenGroups'");
				}
			}
			catch (LDAPReferralException e)
			{
				// Log any failed referrals.
				if (debug)
				{
					String[] referrals = e.getReferrals();
					for (int i = 0; i < referrals.length; i++)
					{
						logger.debug(this, LMN + " referral '" +
							referrals[i] + "' not found");
					}
				}
			}
			catch (LDAPException e)
			{
				if (debug)
					logger.debug(this, LMN + " An error " +
						"occured while searching for groups for user '" +
						this.username + "'", e);

				// Search filter failed
				throw new LoginException(LMN +
					" Error: " + e.getMessage() + "\n");
			}
			finally
			{
				// Close the connection
				if (connection != null)
				{
					try
					{
						connection.disconnect();
						connection = null;
					}
					catch(LDAPException e)
					{
						// Only log the exception
						logger.error(this, LMN +
							" Error while disconnecting",
							e);
					}
				}
			}
		}

		// Return authentication results
		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @throws LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login failed, clean up state
		if (!loginSucceeded)
			return false;

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList();

		// Repository operations to get default role principals
		Iterator roleListItr = roleList.iterator();
		while(roleListItr.hasNext())
		{
			String rolename = (String)roleListItr.next().toString().trim();
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				if (debug)
					logger.debug(this, LMN +
						" Adding CSRolePrincipal: '" +
						rolename + "' to principalList for: '" +
						username + "'");

				principalList.add(rolePrincipal);
			}
			else
			{
				if (debug)
					logger.debug(this, LMN +
						" CSRolePrincipal: '" +
						rolename + "' already in principalList for: '" +
						username + "', not added");
			}
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			if (debug)
				logger.debug(this, LMN +
					" Adding CSUserPrincipal: '" + username +
					"' to principalList");

			principalList.add(userPrincipal);
		}
		else if (debug)
		{
			logger.debug(this, LMN +
				" CSUserPrincipal: '" + username +
				"' already in principalList, not added");
		}

		logger.debug(this, LMN + " Adding principalList to Subject");

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		subject.getPublicCredentials().add(AuthenticationMethod.PASSWORD);

		if(debug)
		{
			logger.debug(this, LMN + " Added " +
				AuthenticationMethod.PASSWORD.getName() +
				" authentication method to Subject");
		}


		if (addPrivateCredentials)
		{
			if (debug)
				logger.debug(this, LMN + " Adding " +
					" 'java.net.PasswordAuthentication' instance" +
					" to private credentials");

			subject.getPrivateCredentials().add(
				new java.net.PasswordAuthentication(
					username, password.toCharArray()));
		}
		else
		{
			if (debug)
				logger.debug(this, LMN + " Did not add " +
					" 'java.net.PasswordAuthentication' instance" +
					" to private credentials");
		}

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.password = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @throws LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.
			this.loginSucceeded = false;
			this.username = null;
			this.password = null;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @throws LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if(this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if (this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.logger = null;
		this.callbackHandler = null;
		this.username = null;
		this.password = null;
		this.options = null;
		this.loginSucceeded = false;
		this.commitSucceeded = false;
		this.debug = false;

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username and
	 * password instance variables with their appropriate value.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// This LoginModule requires a username and a password
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Get the callback values and clear the password.
			callbackHandler.handle(callbacks);
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");
			this.password = ((MapCallback)callbacks[0]).getCallbackValue("password");
			((MapCallback)callbacks[0]).setCallbackValue("password", null);
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available",
				e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (debug)
		{
			logger.debug(this, LMN + " User entered username = " + username);
		}
	}

	/**
	 * Validate the LoginModules options Map object to ensure that all
	 * required properties exist and contain valid values.
	 *
	 * @exception LoginException if any of the following conditions occur:
	 * <UL>
	 * <LI>A CallbackHandler object is not available
	 * <LI>the 'host' property is not available
	 * <LI>the 'port' property is not available or a valid integer value
	 * <LI>the 'connectTimeout' property is not a valid integer value
	 * <LI>the 'userDN' property is not available
	 * <LI>the 'roleBaseDN' property is not available
	 * <LI>the 'roleAttrName' property is not available
	 * <LI>the 'roleFilter' property is not available
	 * <LI>the 'roleScope' property is not available
	 * <LI>the 'roleScope' is not a valid role scope value
	 * </UL>
	 */
	private void validateOptions() throws LoginException
	{
		if (callbackHandler == null)
			throw new LoginException(LMN +
				" Error: no CallbackHandler available");

		String host = (String)options.get("host");
		if ((host == null) || (host.trim().length() == 0))
			throw new LoginException(LMN +
				" Error: Required 'host' property missing");

		String s = (String)options.get("port");
		if ((s == null) || (s.trim().length() == 0))
			throw new LoginException(LMN + " Error: Required " +
				"'port' property missing");

		int port = 389;
		try
		{
			port = Integer.parseInt(s);
			if (port < 1)
				throw new LoginException(LMN + " Error: " +
					"'port' property must be greater than zero");
		}
		catch (NumberFormatException e)
		{
			throw new LoginException(LMN + " Error: " +
				"'port' property is not an integer value");
		}

		s = (String)options.get("connectTimeout");
		if ((s != null) && (s.trim().length() > 0))
		{
			try
			{
				int i = Integer.parseInt(s);
				if (i < 1)
					throw new LoginException(LMN + " Error: " +
					"'connectTimeout' property must be greater than zero");
			}
			catch (NumberFormatException e)
			{
				throw new LoginException(LMN + " Error: " +
					"'connectTimeout' property is not an integer");
			}
		}
	}

	/**
	 * Get an Active Directory connection.
	 *
	 * @return a LDAPConnection.
	 * @exception LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */

	private LDAPConnection getConnection()
		throws LoginException
	{
		// Get the connection values to use
		String host = (String)options.get("host");
		int port = Integer.parseInt((String)options.get("port"));
		int connectTimeout = 3000;

		// Get configuration connectTimeout, if available
		String s = (String)options.get("connectTimeout");
		if ((s != null) && (s.trim().length() > 0))
			connectTimeout = Integer.parseInt(s);

		try
		{
			if (debug)
				logger.debug(this, LMN + 
					" Instantiating LDAPConnection: host=" + host +
					", port=" + port + ", useSSL=" + useSSL +
					", connectTimeout=" + connectTimeout);

			// Initialize the connection
			LDAPConnection connection = null;

			if (useSSL)
			{
				// Warn if required SSL System property for trust store is not set
				String trustStoreProp = System.getProperty("javax.net.ssl.trustStore");
				if (trustStoreProp == null)
					logger.warning(this, "Missing required SSL system property: " +
						"javax.net.ssl.trustStore");

				connection = new LDAPConnection((new LDAPJSSESecureSocketFactory())); 
			}
			else
			{
				connection = new LDAPConnection(); 
			}

			// Set the connection timeout
			connection.setSocketTimeOut(connectTimeout);

			// Debug to help customers determine LDAP response time 
			long timer = System.currentTimeMillis(); 

			// Attempt connection
			connection.connect(host, port);

			if (debug)
			{
				logger.debug(this, LMN + " Time to connect: '" +
					(System.currentTimeMillis() - timer) + "'");

				if (connection.isConnected() && connection.isConnectionAlive())
					logger.debug(this, LMN + " LDAP connection succeeded");
				else
					logger.debug(this, LMN + " LDAP connection failed");

				logger.debug(this, LMN + " Returning LDAPConnection");
			}

			return connection;
		}
		catch (LDAPException e)
		{
			logger.error(this, "LDAP error", e);

			// Dump the nested Throwables to get at the root cause
			int i = 1;
			Throwable cause = e.getCause();
			do
			{
				if (cause != null)
					logger.error(this, "LDAP error: CAUSE " + i + ": " +
						cause.getMessage());
			}
			while ((cause = cause.getCause()) != null);

			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
	}

	/**
	 * Return the input pattern replaced with the match value.
	 *
	 * @param pattern The input pattern to use
	 * @param matchName The variable name to match
	 * @param matchValue The value to replace
	 * @return The input pattern with matched value
	 */
	private String replaceUsername(String pattern, String matchName,
		String matchValue)
	{
		StringBuffer sb = new StringBuffer();
		StringTokenizer st = new StringTokenizer(pattern, "{}");
		while (st.hasMoreTokens())
		{
			String tmp = st.nextToken();
			if (tmp.trim().equals(matchName))
				sb.append(matchValue);
			else
				sb.append(tmp);
		}

		return sb.toString();
	}

	/**
	 * Gets the ActiveDirectoryGroupNameService for the LoginModule.
	 *
	 * @throws LoginException if an error occurs getting the
	 * ActiveDirectoryGroupNameService object.
	 */
	private String getGroupName(String sid)
		throws LoginException
	{
		String groupName = null;
		try
		{
			if (serviceFinder == null)
				throw new LoginException(LMN +
					" Error: ServiceFinder is null");

			String serviceId = (String)options.get("serviceId");
			if ((serviceId == null) || (serviceId.trim().length() == 0))
				throw new LoginException(LMN +
					" Error: Required property 'serviceId' missing");

			if (debug)
			{
				logger.debug(this, "Using serviceId '" + serviceId + "'");
				logger.debug(this, "Getting group id '" + sid + "'");
			}

			ActiveDirectoryGroupNameService adgns = (ActiveDirectoryGroupNameService)
				serviceFinder.find(serviceId, ActiveDirectoryGroupNameService.class);
			groupName = adgns.getGroupNameBySid(sid);

			if (debug)
				logger.debug(this, "Returned group name '" + groupName + "'");
		}
		catch(ServiceException e)
		{
			logger.error(this, "Error finding ActiveDirectoryGroupNameService", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}

		return groupName;
	}

	/**
	 * Convert the native byte representation of a SID into string format
	 *
	 * @param sid a byte representation of a SID
	 * @return a string formatted representation of the SID
	 */
	private static String Sid2String(byte[] SID)
	{
		long version;
		long authority;
		long count;
		long rid;
		String strSID;
		strSID = "S";
		version = SID[0];
		strSID = strSID + "-" + Long.toString(version);
		authority = SID[4];
		for (int i = 0;i<4;i++)
		{
			authority <<= 8;
			authority += SID[4+i] & 0xFF;
		}
		strSID = strSID + "-" + Long.toString(authority);
		count = SID[2];
		count <<= 8;
		count += SID[1] & 0xFF;
		for (int j=0;j<count;j++)
		{
			rid = SID[11 + (j*4)] & 0xFF;
			for (int k=1;k<4;k++)
			{
				rid <<= 8;
				rid += SID[11-k + (j*4)] & 0xFF;
			}
			strSID = strSID + "-" + Long.toString(rid);
		}
		return strSID;
	}

	/**
	 * A simple LDAPAuthHandler for following referrals
	 */
	private class AuthImpl implements LDAPAuthHandler
	{
		// The dn used to authenticate
		private String dn;

		// The password used to authenticate
		private byte[] passwd;

		/**
		 * Construct this instance using data
		 */
		private AuthImpl()
		{

			// Create an LDAPAuthProvider class to encapsulate the
			// authentication credentials.
			dn = bindUsername;
			try
			{
				passwd = password.getBytes("UTF8");
			}
			catch (UnsupportedEncodingException e)
			{
				// Should never get this
				logger.error(this,
					"Unable to encode password to UTF-8 for user: " + dn);
			}

			return;
		}

		/**
		 * Choose the login dn and password based on the host and port
		 * (this implementation simply reuses the original credentials)
		 */
		public LDAPAuthProvider getAuthProvider(String host, int port)
		{
			if (debug)
				logger.debug(this, "Following directory referral, rebinding " +
					"to: " + host + ":" + port);

			return new LDAPAuthProvider(dn, passwd);
		}
	}
} // End of class : ActiveDirectoryLoginModule
