/**
 * Copyright (c) 1996-2012 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-
 * INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR
 * ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.cams.auth.login.userrepository.RepositoryUser;
import com.cafesoft.cams.auth.login.userrepository.UserRepository;
import com.cafesoft.cams.auth.login.userrepository.UserRepositoryException;

import com.cafesoft.cams.service.UserRepositoryService;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.service.ServiceClient;
import com.cafesoft.core.service.ServiceException;
import com.cafesoft.core.service.ServiceFinder;

import com.cafesoft.core.util.DigestString;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;

import java.io.BufferedReader;

import javax.security.auth.Subject;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.spi.LoginModule;

import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;

import java.lang.reflect.UndeclaredThrowableException;

import java.math.BigInteger;

import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;

import javax.net.ssl.HttpsURLConnection;

import java.security.GeneralSecurityException;
import java.security.MessageDigest;
import java.security.Principal;
import java.security.NoSuchAlgorithmException;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * Implements the LoginModule interface found in JAAS to authenticate client 
 * services against a XML user repository so an asserted user name can be used
 * to create a new Cams session. Authentication request must include:
 * <ul>
 * <li> The user's Cams session id (sid) at the Cams IdP.</li> 
 * <li> A One-Time Password (TOTP) generated by the Cams IdP CamsFederate servlet.</li>
 * <li> The asserted user name and roles.</li> 
 * <li> A service user name and password.</li> 
 * <li> Cams SP login config entry, security domain and original URL for 
 * redirect if authentication is successful.</li>
 * </ul>
 * The TOTP can be validated locally by Cams SP using this login module or remotely
 * by the Cams IdP CamsFederate servlet, which verifies the TOTP and the user's
 * the Cams session. Uses a ServiceFinder object to retrieve the XML user 
 * repository used for service client authentication. Roles associated with the 
 * user name in the XML user repository are assigned to the asserted user name 
 * as are asserted roles sent in the authentication request.  
 * <p>
 * Configuration parameters are:
 * <ul>
 * <li><b>debug</b> - Flag that enables/disables debugging.</li>
 * <li><b>serviceId</b> - (required) The unique name of the XML User Repository</li>
 * Service that you are assigning to this login module.
 * <li><b>sharedSecret</b> - (required) Concatenated with the Cams session id and 
 * hashed with "SHA-256") by both Cams IdP and SP to create the TOTP seed.</li>
 * <li><b>totpVerificationUrl</b> - (optional) The fully-qualified URL to the 
 * CamsFederate webapp that will be used to verify the TOTP, which is always in 
 * the cookie domain of the Cams IdP. The timeStepStart, timeStepSeconds, totpLength
 * and timeStepClockDriftOffset are conditionally used if totpVerificationUrl 
 * is <strong>NOT </strong> used.</li>
 * <li><b>timeStepStart</b> - (conditional) The TOTP time step start where "0" is 
 * UNIX epoch time. Defaults to "0" if not supplied.</li>
 * <li><b>timeStepSeconds</b> - (conditional) The time step in seconds. Defaults to 
 * recommend value of 30 if not supplied.</li> 
 * <li><b>totpLength</b> - (conditional) The TOTP length. Default to "8" if not
 * supplied.</li>
 * <li><b>timeStepClockDriftOffset</b> - (conditional) The number of time steps 
 * of clock drift in either forward or backward direction that will be acceptible 
 * to verify the TOTP. Defaults to "0" if not supplied.</li>
 * <li><b>defaultRoles</b> - (Optional) A list of one or more default roles
 * separated by a comma to be added to the user session.</li>
 * </ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.2 $ $Date: 2012/09/06 00:04:38 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 09/05/12
 */
public class CamsFederateXmlLoginModule
	implements LoginModule, ServiceClient, LoggerClient
{
	//
	// Static Constants
	//

	/**
	 * The login module name (used in log messages throughout the code).
	 */
	private static final String LMN = "[CamsFederateXmlLoginModule]";

	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Map that contains state shared with other configured LoginModules.
	 */
	private Map sharedState;

	/**
	 * Used by the LoginModule to lookup required Cams services.
	 */
	private ServiceFinder serviceFinder;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The authenticating service user name.
	 */
	private String username;

	/**
	 * The authenticating password.
	 */
	private String password;

	/**
	 * The asserted username for the Cams session.
	 */
	private String assertedUsername;

	/**
	 * The asserted roles for the Cams session.
	 */
	private String assertedRoles;

	/**
	 * The sid for the Cams IdP session.
	 */
	private String sid;

	/**
	 * The Cams IdP generated one-time password.
	 */
	private String totp;

	/**
	 * The TOTP time step start where 0 is UNIX epoch time.
	 */
	private static long timeStepStart = 0;

	/**
	 * The time step in seconds. 30 is recommended.
	 */
	private static long timeStepSeconds = 30;

	/**
	 * The TOTP length.
	 */
	private static String totpLength = "8";

	/**
	 * The number of time steps of clock drift in either forward or 
	 * backward direction that will be acceptible to verify the TOTP
	 */
	private static int timeStepClockDriftOffset = 0;

	/**
	 * The UserRepository to use.
	 */
	private UserRepository userRepository;

	//
	// Constructor
	//

	/**
	 * Default CamsFederateXmlLoginModule Constructor.
	 */
	public CamsFederateXmlLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.password = null;
		this.serviceFinder = null;
		this.sharedState = null;
		this.subject = null;
		this.username = null;
		this.sid = null;
		this.totp = null;
		this.assertedUsername = null;
		this.assertedRoles = null;
		this.userRepository = null;
	}

	//
	// Implements the ServiceClient interface
	//

	/**
	 * Set the ServiceFinder.
	 *
	 * @param finder the class used by find Services by the ServiceClient.
	 */
	public void setServiceFinder(ServiceFinder finder)
	{
		this.serviceFinder = finder;
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *		with the end user.
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *		<code>Configuration</code> for this particular
	 *		<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.sharedState = sharedState;
		this.debug = "true".equals((String)options.get("debug"));
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a user name and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @throws LoginException thrown if any error occurs while trying to
	 * authenticate the user.
	 * @throws FailedLoginException if authentication fails.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException
	{
		// Check preconditions
		validateOptions();

		// Handle callback objects.
		handleCallbacks();

		// A user name must be supplied
		if (this.username == null || this.username.length() == 0)
			throw new FailedLoginException(LMN + " Authentication failed");

		// An asserted user name must be supplied
		if (this.assertedUsername == null || this.assertedUsername.length() == 0)
			throw new FailedLoginException(LMN + " Authentication failed");

		// A Cams sid must be supplied
		if (this.sid == null || this.sid.length() == 0)
			throw new FailedLoginException(LMN + " Authentication failed");

		// A totp must be supplied
		if (this.totp == null || this.totp.length() == 0)
			throw new FailedLoginException(LMN + " Authentication failed");

		// Authenicate the service user
		if (this.userRepository == null)
			getUserRepository();

		try
		{
			// Try to get the service user name
			RepositoryUser ru = userRepository.getUser(this.username);

			// If the service user exists, get the user's password
			if (ru != null)
			{
				String rp = ru.getPassword();

				// The repository password may be "digested" and "salted" by use
				// of a "Message Digest" algorithm, if so, then we'll need to
				// digest and salt the user-entered password before comparison.
				String digestString = password;
				DigestString ds = new DigestString(rp);

				// Is the password clear text (or does it have unknown
				// algorithm)?
				if (ds.isDigestString())
				{
					// Get the digestString from the user-entered password using
					// the same ingredients found in the repository password;
					// the password, the digest algorithm, and possibly "salt"
					digestString = DigestString.createDigestString(password,
						ds.getAlgorithm(), ds.getSalt(), ds.getLabel());
				}

				// Compare the cleartext or digested passwords to authenticate
				// the service user
				if (rp.equals(digestString))
				{
					// Service user is authenticated, now attempt authentication
					// for the federated, asserted user

					// Verify the Cams sid and totp validity using a remote Cams IdP 
					// CamsFederate service. Also send the TOTP seed to use for the 
					// TOTP verification, which is a SHA-256 hash of the Cams sid and 
					// shared secret.
					String seed = getSHA256Hash(
						this.sid.substring(this.sid.lastIndexOf("-") + 1) + 
						(String)options.get("sharedSecret"));

					String queryParameters = 
						"sid=" + URLEncoder.encode(this.sid, "UTF-8") + 
						"&seed=" + URLEncoder.encode(seed, "UTF-8") + 
						"&totp=" + URLEncoder.encode(this.totp, "UTF-8");


					// Determine if verification will be performed remotely by 
					// TOTP originating Cams IdP or locally by Cams SP within
					// this login module
					String totpVerificationUrl = (String)options.get("totpVerificationUrl");
					if ((totpVerificationUrl != null) && (totpVerificationUrl.length() >0))
					{
						if (this.debug)
							logger.debug(this, LMN + " Cams IdP remotely validating TOTP at: " + totpVerificationUrl);

						URL url = new URL(totpVerificationUrl); 
						HttpURLConnection connection = (HttpURLConnection)url.openConnection();           
 
						// Use any HttpsURLConnection methods in this block 
						if (connection instanceof HttpsURLConnection) 
						{
							connection = (HttpsURLConnection) url.openConnection();           

							// Innerclass for development and testing with self-signed certs where 
							// where the subject may not the same as the hostname. Uncomment, add
							// the following imports and recompile to use.
							//
							//    import javax.net.ssl.HostnameVerifier;
							//    import javax.net.ssl.SSLSession;
							//
							//((HttpsURLConnection)connection).setHostnameVerifier(new HostnameVerifier()
							//{
							//	public boolean verify (String hostname, SSLSession session)
							//	{
							//		return true;
							//	}
							//});
						}

						// Set connection options
						connection.setDoOutput(true);
						connection.setDoInput(true);
						connection.setInstanceFollowRedirects(false); 
						connection.setRequestMethod("POST"); 
						connection.setRequestProperty("Content-Type", "application/x-www-form-urlencoded"); 
						connection.setRequestProperty("charset", "utf-8");
						connection.setRequestProperty("Content-Length", "" + Integer.toString(queryParameters.getBytes().length));
						connection.setUseCaches (false);

						// Write parameters
						OutputStreamWriter writer = new OutputStreamWriter(connection.getOutputStream());
						writer.write(queryParameters);
						writer.flush();

						// Get the response
						StringBuffer isValidTOTP = new StringBuffer();
						BufferedReader reader = new BufferedReader(new InputStreamReader(connection.getInputStream()));
						String line;
						while ((line = reader.readLine()) != null)
						{
							isValidTOTP.append(line);
						}

						writer.close();
						reader.close();
            
						// Log the response
						if (this.debug)
							logger.debug(this, LMN + " Cams IdP successfully validated TOTP: " + isValidTOTP.toString());

						// Check response for value "true" to determine if TOTP is valid
						if ("true".equalsIgnoreCase(isValidTOTP.toString()))
							loginSucceeded = true;
					}
					else
					{
						if (this.debug)
							logger.debug(this, LMN + " Cams SP locally validating TOTP: " + this.totp);

						// Verify supplied TOTP is valid and write response
						if (verifyTOTP(seed, this.totp))
						{
							loginSucceeded = true;

							// Log the response
							if (this.debug)
								logger.debug(this, LMN + " Cams SP successfully validated TOTP");
						}
					}
				}
			}
			else
			{
				if (this.debug)
				{
					// No user by that name
					logger.debug(this, LMN + " User '" + this.username +
						"' does not exist");
				}
			}
		}
		catch (UnsupportedEncodingException e)
		{
			logger.error(this, "Unsupported encoding", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage() + "\n");
		}
		catch (MalformedURLException e)
		{
			logger.error(this, "Malformed URL", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage() + "\n");
		}
		catch (IOException e)
		{
			logger.error(this, "IO error", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage() + "\n");
		}
		catch (NoSuchAlgorithmException e)
		{
			logger.error(this, "Unknown MessageDigestAlgorithm", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage() + "\n");
		}
		catch (UserRepositoryException e)
		{
			logger.error(this, "Error accessing UserRepository", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage() + "\n");
		}

		// Authentication failed, clean up state and throw FailedLoginException.
		if (!this.loginSucceeded)
		{
			cleanup();
			throw new FailedLoginException(LMN + " Authentication failed");
		}

		return this.loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @throws LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login failed
		if (!this.loginSucceeded)
			return false;

		if (this.userRepository == null)
			getUserRepository();

		List principalList = new ArrayList(5);

		try
		{
			// Add default roles (if any) to list of roles
			String defaultRoles = (String)options.get("defaultRoles");
			if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
			{
				StringTokenizer st = new StringTokenizer(defaultRoles, ",");
				while(st.hasMoreTokens())
					principalList.add(
						new CSRolePrincipal(st.nextElement().toString().trim()));
			}

			// Get the user and the user roles
			RepositoryUser ru = this.userRepository.getUser(this.username);
			String[] r = ru.getRoles();

			// Add the user roles from the service user
			if ((r != null) && (r.length > 0))
			{
				for (int i = 0; i < r.length; i++)
				{
					// Add the CSRolePrincipal to the Subject only if it does not exist
					Principal rolePrincipal = new CSRolePrincipal(r[i].trim());
					if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
					{
						principalList.add(rolePrincipal);

						if (this.debug)
						{
							logger.debug(this, LMN +
								" Added CSRolePrincipal " +
								r[i] + " to principal List for user '" +
								this.assertedUsername + "'");
						}
					}
					else
					{
						if (this.debug)
						{
							logger.debug(this, LMN +
								" CSRolePrincipal '" + r[i] +
								"' already in Subject for '" + this.assertedUsername +
								"', not added to principal List");
						}
					}
				}
			}

			// Add any asserted user roles
			if ((this.assertedRoles != null) && (this.assertedRoles.length() > 0))
			{
				String[] assertedUserRoles = this.assertedRoles.split(",");

				for (int i = 0; i < assertedUserRoles.length; i++)
				{
					// Add the CSRolePrincipal to the Subject only if it does not exist
					Principal rolePrincipal = new CSRolePrincipal(assertedUserRoles[i].trim());
					if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
					{
						principalList.add(rolePrincipal);

						if (this.debug)
						{
							logger.debug(this, LMN +
								" Added CSRolePrincipal " +
								assertedUserRoles[i] + " to principal List for user '" +
								this.assertedUsername + "'");
						}
					}
					else
					{
						if (this.debug)
						{
							logger.debug(this, LMN +
								" CSRolePrincipal '" + assertedUserRoles[i] +
								"' already in Subject for '" + this.assertedUsername +
								"', not added to principal List");
						}
					}
				}
			}
		}
		catch (UserRepositoryException e)
		{
			logger.error(this, "Error in UserRepositoryService", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage() + "\n");
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(this.assertedUsername);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if (this.debug)
			{
				logger.debug(this, LMN + " Added CSUserPrincipal '" +
					this.assertedUsername + "' to principal List");
			}
		}
		else
		{
			if (this.debug)
			{
				logger.debug(this, LMN + " CSUserPrincipal '" +
					this.assertedUsername + 
					"' already in Subject, not added to principal List");
			}
		}

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			this.subject.getPrincipals().add((Principal)principalList.get(i));

		// Add the authentication method to the Subject.
		this.subject.getPublicCredentials().add(AuthenticationMethod.XMLDIGSIG);

		if (this.debug)
		{
			logger.debug(this, LMN + " Adding " +
				AuthenticationMethod.XMLDIGSIG.getName() +
				" AuthenticationMethod");
		}

		return this.commitSucceeded = true;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @throws LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{

		// If authentication failed return false. If commit
		// succeeded log the user out return true. Otherwise,
		// Cleanup the login module and and return true.

		if (!this.loginSucceeded)
			return false;

		if (this.commitSucceeded)
		{
			logout();
		}
		else
		{
			cleanup();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @throws LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if (this.subject != null)
		{
			this.subject.getPrincipals().clear();
		}

		cleanup();

		return true;
	}

	/**
	 * Clean up the resources of the LoginModule
	 */
	private void cleanup()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.logger = null;
		this.loginSucceeded = false;
		this.options = null;
		this.serviceFinder = null;
		this.sharedState = null;
		this.subject = null;
		this.userRepository = null;
		this.totp = null;
		this.sid = null;
		this.username = null;
		this.password = null;
		this.assertedUsername = null;
		this.assertedRoles = null;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the user name and
	 * password instance variables with their appropriate value.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// This LoginModule requires a user name, password and an 
		// asserted user name
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Get the callback values and clear the password.
			callbackHandler.handle(callbacks);
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");
			this.password = ((MapCallback)callbacks[0]).getCallbackValue("password");
			this.totp = ((MapCallback)callbacks[0]).getCallbackValue("totp");
			this.sid = ((MapCallback)callbacks[0]).getCallbackValue("sid");
			this.assertedUsername = ((MapCallback)callbacks[0]).getCallbackValue("asserted_username");
			((MapCallback)callbacks[0]).setCallbackValue("password", null);
			this.assertedRoles = ((MapCallback)callbacks[0]).getCallbackValue("asserted_roles");
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available",
				e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (this.debug)
		{
			logger.debug(this, LMN + " Service user name = " + this.username);
			logger.debug(this, LMN + " Asserted user name = " + this.assertedUsername);
		}
	}

	/**
	 * Gets the UserRepository for the LoginModule.
	 *
	 * @throws LoginException if an error occurs getting the
	 * UserRespository object.
	 */
	private void getUserRepository()
		throws LoginException
	{
		try
		{
			if (this.serviceFinder == null)
				throw new LoginException(LMN +
					" Error: ServiceFinder is null");

			String serviceId = (String)this.options.get("serviceId");
			if ((serviceId == null) || (serviceId.trim().length() == 0))
				throw new LoginException(LMN + " Error: Required " +
					"'serviceId' property missing");

			UserRepositoryService urs = (UserRepositoryService)
				this.serviceFinder.find(serviceId, UserRepositoryService.class);
			this.userRepository = urs.getUserRepository();
		}
		catch(ServiceException e)
		{
			logger.error(this, "Error finding UserRepositoryService", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
	}

	/**
	 * Validate the LoginModules options Map object to ensure that all
	 * required properties exist and contain valid values.
	 *
	 * @exception LoginException if any of the following conditions occur:
	 */
	private void validateOptions() throws LoginException
	{
		if (callbackHandler == null)
			throw new LoginException(LMN + " Error: no CallbackHandler available");

		// Check for options required if "totpVerificationUrl" is not set,
		// meaning that this Cams SP login module will verify the TOTP.
		String s = (String)options.get("totpVerificationUrl");
		if ((s == null) || (s.trim().length() == 0))
		{
			s = (String)options.get("timeStepStart");
			if ((s == null) || (s.trim().length() == 0))
			{
				logger.warning(this, LMN +
					" Property 'timeStepStart' not found, using default: " + 
					timeStepStart);
			}
			else
			{
				try
				{
					timeStepStart = Long.parseLong(s);
				}
				catch (NumberFormatException e)
				{
					logger.warning(this, LMN +
						" Property 'timeStepStart' not a valid number, using default: " + 
						timeStepStart);
				}
			}

			s = (String)options.get("timeStepSeconds");
			if ((s == null) || (s.trim().length() == 0))
			{
				logger.warning(this, LMN +
					" Property 'timeStepStart' not found, using default: " + 
					timeStepSeconds);
			}
			else
			{
				try
				{
					timeStepSeconds = Long.parseLong(s);
				}
				catch (NumberFormatException e)
				{
					logger.warning(this, LMN +
						" Property 'timeStepSeconds' not a valid number, using default: " + 
						timeStepSeconds);
				}
			}

			s = (String)options.get("totpLength");
			if ((s == null) || (s.trim().length() == 0))
			{
				logger.warning(this, LMN +
					" Property 'totpLength' not found, using default: " + 
					totpLength);
			}
			else
			{
				try
				{
					// This variable is not used as an Integer but must be one
					Integer.parseInt(s);
					totpLength = s;
				}
				catch (NumberFormatException e)
				{
					logger.warning(this, LMN +
						" Property 'totpLength' not a valid number, using default: " + 
						totpLength);
				}
			}

			s = (String)options.get("timeStepClockDriftOffset");
			if ((s == null) || (s.trim().length() == 0))
			{
				logger.warning(this, LMN +
					" Property 'timeStepClockDriftOffset' not found, using default: " + 
					timeStepClockDriftOffset);
			}
			else
			{
				try
				{
					timeStepClockDriftOffset = Integer.parseInt(s);
				}
				catch (NumberFormatException e)
				{
					logger.warning(this, LMN +
						" Property 'timeStepClockDriftOffset' not a valid number, using default: " + 
						timeStepClockDriftOffset);
				}
			}
		}

		// A shared secret must be supplied
		s = (String)options.get("sharedSecret");
		if ((s == null) || (s.trim().length() == 0))
			throw new LoginException(LMN +
				" Error: Required 'sharedSecret' property missing");
	}

	/*
	 * Generate a SHA-256 hash from the supplied String value.
	 *
	 * @param s The String to hash 
	 *
	 * @return String The Hex value of the hash 
	 */
	private static String getSHA256Hash(String s)
		throws NoSuchAlgorithmException
	{
		StringBuffer sb = new StringBuffer();

		MessageDigest md = MessageDigest.getInstance("SHA-256");
		md.update(s.getBytes());
 
		byte byteData[] = md.digest();
 
		//convert the byte to hex format method 1
		for (int i = 0; i < byteData.length; i++) 
			sb.append(Integer.toString((byteData[i] & 0xff) + 0x100, 16).substring(1));
 
		return sb.toString();
	}

	/*
	 * Verify that a supplied TOTP is valid.
	 *
	 * @param seed The seed to use to validate the TOTP 
	 * @param sid The Cams session id (right hash substring only)
	 *
	 * @return String "true" if supplied TOTP is valid, else "false"
	 *
	 * @exception IOException if an input/output error occurs
	 */
	private boolean verifyTOTP(String seed, String totp)
		throws IOException
	{
		// Initialize TOTP variables
		boolean totpValid = false; 
		String time = "0";
		long T0 = timeStepStart;
		long X = timeStepSeconds;

		try
		{
			long T = (System.currentTimeMillis()/1000 - T0)/X;
			time = Long.toHexString(T).toUpperCase();
			while(time.length() < 16) time = "0" + time;

			String spLocalTOTP = generateTOTP512(seed, time, totpLength);

			if (this.debug)
				logger.debug(this, LMN + " Cams SP local TOTP: " + spLocalTOTP);

			if (totp.equals(spLocalTOTP))
			{
				// TOTP is valid, nothing else needs to be done
				totpValid = true;

				if (this.debug)
					logger.debug(this, LMN + " Cams IdP and SP TOTP match");
			}
			else if (timeStepClockDriftOffset > 0) // Use time step clock drift offset?
			{
				// Try to match using "clock drift" offset backward 
				for (int i = 0; i < timeStepClockDriftOffset; i++)
				{
					T = T + 1;
					time = Long.toHexString(T).toUpperCase();
					while(time.length() < 16) time = "0" + time;

					spLocalTOTP = generateTOTP512(seed, time, totpLength);

					if (this.debug)
						logger.debug(this, LMN + " Cams SP local TOTP: " + spLocalTOTP);

					if (totp.equals(spLocalTOTP))
					{
						totpValid = true;
						logger.info(this, LMN + " Detected " + (i + 1) + 
							" backward time steps of Cams SP clock drift");
						break;
					}
				}

				// Try to match using "clock drift" offset forwards 
				if (!totpValid)
				{
					// Reset to original time step
					T = T - timeStepClockDriftOffset;

					// Try to match using "clock drift" offset forward 
					for (int i = 0; i < timeStepClockDriftOffset; i++)
					{
						T = T - 1;
						time = Long.toHexString(T).toUpperCase();
						while(time.length() < 16) time = "0" + time;

						spLocalTOTP = generateTOTP512(seed, time, totpLength);

						if (this.debug)
							logger.debug(this, LMN + " Cams SP local TOTP: " + spLocalTOTP);

						if (totp.equals(spLocalTOTP))
						{
							totpValid = true;
							logger.info(this, LMN + " Detected " + (i + 1) + 
								" forward time steps of Cams SP clock drift");
							break;
						}
					}
				}
			}
		}
		catch (final Exception e)
		{
			logger.error(this, "ERROR: " + e);
		}

		return totpValid;
	}

	/**
	 * The following methods are an example implementation of the OATH TOTP algorithm.
	 * Visit www.openauthentication.org for more information. See also:
	 *
	 * http://tools.ietf.org/id/draft-mraihi-totp-timebased-06.html
	 *
	 * @author Johan Rydell, PortWise, Inc.
	 */

	/**
	 * This method uses the JCE to provide the crypto algorithm.
	 * HMAC computes a Hashed Message Authentication Code with the
	 * crypto hash algorithm as a parameter.
	 *
	 * @param crypto     the crypto algorithm (HmacSHA1, HmacSHA256, HmacSHA512)
	 * @param keyBytes   the bytes to use for the HMAC key
	 * @param text       the message or text to be authenticated.
	 */
	private static byte[] hmac_sha1(String crypto, byte[] keyBytes,
		byte[] text)
	{
		try 
		{
			Mac hmac;
			hmac = Mac.getInstance(crypto);
			SecretKeySpec macKey =
				new SecretKeySpec(keyBytes, "RAW");
			hmac.init(macKey);
			return hmac.doFinal(text);
		} 
		catch (GeneralSecurityException gse) 
		{
			throw new UndeclaredThrowableException(gse);
		}
	}

	/**
	 * This method converts HEX string to Byte[]
	 *
	 * @param hex   the HEX string
	 *
	 * @return      A byte array
	 */
	private static byte[] hexStr2Bytes(String hex)
	{
		// Adding one byte to get the right conversion
		// values starting with "0" can be converted
		byte[] bArray = new BigInteger("10" + hex,16).toByteArray();

		// Copy all the REAL bytes, not the "first"
		byte[] ret = new byte[bArray.length - 1];
		for (int i = 0; i < ret.length ; i++)
			ret[i] = bArray[i+1];
		return ret;
	}

	private static final int[] DIGITS_POWER
		// 0 1  2   3    4     5      6       7        8
		= {1,10,100,1000,10000,100000,1000000,10000000,100000000 };

	/**
	 * This method generates an TOTP value for the given set of parameters.
	 *
	 * @param key - the shared secret, HEX encoded
	 * @param time - a value that reflects a time
	 * @param returnDigits - number of digits to return
	 *
	 * @return - A numeric String in base 10 that includes {@link truncationDigits} digits
	 */
	public static String generateTOTP(String key, String time, String returnDigits)
	{
		return generateTOTP(key, time, returnDigits, "HmacSHA1");
	}

	/**
	 * This method generates an TOTP value for the given set of parameters.
	 *
	 * @param key - the shared secret, HEX encoded
	 * @param time - a value that reflects a time
	 * @param returnDigits - number of digits to return
	 *
	 * @return - A numeric String in base 10 that includes {@link truncationDigits} digits
	 */
	public static String generateTOTP256(String key, String time, String returnDigits)
	{
		return generateTOTP(key, time, returnDigits, "HmacSHA256");
	}

	/**
	 * This method generates an TOTP value for the given set of parameters.
	 *
	 * @param key - the shared secret, HEX encoded
	 * @param time - a value that reflects a time
	 * @param returnDigits - number of digits to return
	 *
	 * @return - A numeric String in base 10 that includes {@link truncationDigits} digits
	 */
	public static String generateTOTP512(String key, String time, String returnDigits)
	{
		return generateTOTP(key, time, returnDigits, "HmacSHA512");
	}

	/**
	 * This method generates an TOTP value for the given set of parameters.
	 *
	 * @param key - the shared secret, HEX encoded
	 * @param time - a value that reflects a time
	 * @param returnDigits - number of digits to return
	 * @param crypto - the crypto function to use
	 *
	 * @return - A numeric String in base 10 that includes {@link truncationDigits} digits
	 */
	private static String generateTOTP(String key, String time, String returnDigits, String crypto)
	{
		int codeDigits = Integer.decode(returnDigits).intValue();
		String result = null;
		byte[] hash;

		// Using the counter
		// First 8 bytes are for the movingFactor
		// Complaint with base RFC 4226 (HOTP)
		while(time.length() < 16 )
			time = "0" + time;

		// Get the HEX in a Byte[]
		byte[] msg = hexStr2Bytes(time);

		// Adding one byte to get the right conversion
		byte[] k = hexStr2Bytes(key);

		hash = hmac_sha1(crypto, k, msg);

		// Put selected bytes into result int
		int offset = hash[hash.length - 1] & 0xf;

		int binary =
			((hash[offset] & 0x7f) << 24) |
			((hash[offset + 1] & 0xff) << 16) |
			((hash[offset + 2] & 0xff) << 8) |
			(hash[offset + 3] & 0xff);

		int otp = binary % DIGITS_POWER[codeDigits];

		result = Integer.toString(otp);
		while (result.length() < codeDigits) 
		{
			result = "0" + result;
		}
		return result;
	}
} // End of class : CamsFederateXmlLoginModule

