/**
 * Copyright (c) 1996-2010 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 * OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
 * BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
 * OR ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.core.jdbc.JdbcUtils;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.service.ServiceClient;
import com.cafesoft.core.service.ServiceFinder;

import com.vasco.utils.AAL2Wrap;
import com.vasco.utils.KernelParms;

import java.security.Principal;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;
import java.io.UnsupportedEncodingException;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * Implements the LoginModule interface found in JAAS to authenticate users
 * with VASCO Digipass tokens against relational database using JDBC. The
 * following process flow is used:
 *
 * <ol>
 * <li>Always query the Digipass record values for the user, including:
 *  <ul>
 *   <li>Serial number
 *   <li>Authentication mode
 *   <li>Digipass data blob
 *  </ul>
 * <li>Determine if the authentication mode is response only (RO) or
 *  challenge/response (CR)
 * <li>If RO, use the password to try to authenticate the user
 * <li>If CR, and there is NO challenge, generate a challenge and
 *  return it to the user
 * <li>If CR, and there is a challenge, use it and the password
 *  to try to authenticate the user
 * <li>Always save the final Digipass data blob to the database to
 *  preserve state
 * </ol>
 *
 * See the VACMAN controller Java AAL2Wrap and KernelParms API javadoc,
 * included with the VACMAN controller distribution.
 *
 * NOTE: You must deploy the aal2wrap.jar and operating system specific
 * dlls/libraries into CAMS_HOME/lib. For Windows, use aal2sdk.dll and
 * aal2sdk.lib. For Linux/UNIX, use libaal2sdk-3.5.0.so and libaal2sdk.so.
 *
 * All configuration parameters are listed below:
 * <Ul>
 * <LI><B>jdbcDriver</B> - the JDBC Driver
 * <LI><B>jdbcUrl</B> - the database JDBC URL
 * <LI><B>jdbcUsername</B> - the database username
 * <LI><B>jdbcPassword</B> - the user's password on the database server
 * <LI><B>getDigipassPreparedStatement</B> - the prepared statement to
 * get a user's DIGIPASS
 * <LI><B>saveDigipassPreparedStatement</B> - the prepared statement to
 * update a user's DIGIPASS
 * <LI><B>serialNumberColumn</B> - the database column name for the DIGIPASS
 * serial number
 * <LI><B>authenticationModeColumn</B> - the database column name for the
 * authentication mode for the digipass:
 *    RO - Response Only
 *    CR - Challenge Response (stubbed out, not fully implemented)
 * <LI><B>digipassDataColumn</B> - the database column name for the DIGIPASS
 * data blob.
 * <LI><B>accountExpiredPreparedStatement</B> - (optional) the prepared
 * statement to determine if the account has expired. Prepared statement
 * must return a count of "1" if the account has expired, "0" if it has not.
 * This prepared statement is executed conditionally after the password has
 * been validated.
 * <LI><B>defaultRoles</B> - (Optional) A list of one or more default roles
 * separated by a comma to be added to the user session.
 * <LI><B>rolePreparedStatement</B> - the prepared statement to execute to
 * obtain the user's roles
 * <LI><B>roleColumn</B> - the database column name for the role field
 * <LI><B>kernelParameters</B> - the comma-delimited Vacman controller kernel
 * parameters to use as integers or hex values, for example:
 *
 *    "100,24,0,0,1,0,0,0,0,6,0,100,0,0,0x7FFFFF,0,0,0,0"
 *
 * <LI><B>responseLength</B> - the expected total length of the response after
 * a challenge (4 for pre-pended challenge plus the length dynamic response)
 *
 * </UL>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.9 $ $Date: 2010/12/17 21:15:41 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 01/23/06
 */
public class DigipassJdbcLoginModule
	implements LoginModule, ServiceClient, LoggerClient
{
	//
	// Static Constants
	//

	/**
	 * The login module name (used in log messages throughout the code).
	 */
	private static final String LMN = "[DigipassJdbcLoginModule]";

	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Map that contains state shared with other configured LoginModules.
	 */
	private Map sharedState;

	/**
	 * Used by the LoginModule to lookup required Cams services.
	 */
	private ServiceFinder serviceFinder;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating User.
	 */
	private String username;

	/**
	 * The password entered by the authenticating User.
	 */
	private String password;

	//
	// Constructor
	//

	/**
	 * Default LoginModule Constructor.
	 */
	public DigipassJdbcLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.password = null;
		this.serviceFinder = null;
		this.sharedState = null;
		this.subject = null;
		this.username = null;
	}

	//
	// Implements the ServiceClient interface
	//

	/**
	 * Set the ServiceFinder.
	 *
	 * @param finder the class used by find Services by the ServiceClient.
	 */
	public void setServiceFinder(ServiceFinder finder)
	{
		this.serviceFinder = finder;
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *			with the end user (prompting for usernames and
	 *			passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.sharedState = sharedState;
		this.debug = "true".equals((String)options.get("debug"));
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates a
	 * <code>Subject</code>. For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *  <code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @throws LoginException thrown if any error occurs while trying to
	 * authenticate the user.
	 * @throws FailedLoginException if authentication fails.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException,
		AccountExpiredException, CredentialExpiredException
	{
		handleCallbacks();

		// A user name must always be supplied
		if (username == null || username.length() == 0)
			throw new FailedLoginException( LMN + 
				" Authentication failed");

		// A non-empty password must be supplied
		if ("".equals(this.password))
			throw new FailedLoginException( LMN + 
				" Authentication failed");

		// Get JdbcDriver
		String jdbcDriver = (String)options.get("jdbcDriver");
		if ((jdbcDriver == null) || "".equals(jdbcDriver.trim()))
			throw new LoginException( LMN +
				" Error: Missing required option 'jdbcDriver'");

		// Get the JdbcUrl
		String jdbcUrl = (String)options.get("jdbcUrl");
		if ((jdbcUrl == null) || "".equals(jdbcUrl.trim()))
			throw new LoginException( LMN +
				" Error: Missing required option 'jdbcUrl'");

		// Get the JdbcUsername and JdbcPassword (empty may be valid)
		String jdbcUsername  = (String)options.get("jdbcUsername");
		String jdbcPassword  = (String)options.get("jdbcPassword");

		// Get prepared statment to query the Digipass record values
		String getDigipassPreparedStatement =
			(String)options.get("getDigipassPreparedStatement");
		if ((getDigipassPreparedStatement == null) ||
			"".equals(getDigipassPreparedStatement.trim()))
			throw new LoginException( LMN +
				" Error: Missing required option 'getDigipassPreparedStatement'");

		// Get prepared statment to update the Digipass data blob
		String saveDigipassPreparedStatement =
			(String)options.get("saveDigipassPreparedStatement");
		if ((saveDigipassPreparedStatement == null) ||
			"".equals(saveDigipassPreparedStatement.trim()))
			throw new LoginException( LMN +
				" Error: Missing required option 'saveDigipassPreparedStatement'");

		// Get database field serialNumberColumn name
		String serialNumberColumn = (String)options.get("serialNumberColumn");
		if ((serialNumberColumn == null) || "".equals(serialNumberColumn.trim()))
			throw new LoginException( LMN +
				" Error: Missing required option 'serialNumberColumn'");

		// Get database field authenticationModeColumn name
		String authenticationModeColumn =
			(String)options.get("authenticationModeColumn");

		if ((authenticationModeColumn == null) ||
			"".equals(authenticationModeColumn.trim()))
			throw new LoginException( LMN +
				" Error: Missing required option 'authenticationModeColumn'");

		// Get database field digipassDataColumn name
		String digipassDataColumn = (String)options.get("digipassDataColumn");
		if ((digipassDataColumn == null) || "".equals(digipassDataColumn.trim()))
			throw new LoginException( LMN +
				" Error: Missing required option 'digipassDataColumn'");

		// Get the expected response length (challenge plus dynamic response)
		// and convert to int
		String tmp = (String)options.get("responseLength");
		if ((tmp == null) || "".equals(tmp.trim()))
			throw new LoginException( LMN +
				" Error: Missing required option 'responseLength'");

		int responseLength = 11;
		try
		{
			responseLength = Integer.parseInt(tmp);
		}
		catch (NumberFormatException e)
		{
			throw new LoginException( LMN +
				" Error: Number format error while parsing " +
				"'responseLength' value");
		}

		// Get VACMAN controller kernel parameters
		tmp = (String)options.get("kernelParameters");
		if ((tmp == null) || "".equals(tmp.trim()))
			throw new LoginException( LMN +
				" Error: Missing required option 'kernelParameters'");

		// Kernel parameters are supplied as a comma-delimited string,
		// which must be parsed for int values
		StringTokenizer st = new StringTokenizer(tmp, ",");
		int[] kernelParameters = new int[st.countTokens()];
		int i = 0;
		try
		{
			String str = null;
			int radix = 10;
			while (st.hasMoreTokens())
			{
				str = st.nextToken();

				// Tokens that start with "0x" are followed by a hex values,
				// which must be converted to int values
				if (str.startsWith("0x"))
				{
					str = str.substring(2);
					radix = 16;
				}
				else
				{
					radix = 10;
				}

				kernelParameters[i] =
					Integer.parseInt(str, radix);

				i++;
			}
		}
		catch (NumberFormatException e)
		{
			throw new LoginException( LMN +
				" Error: Number format error while parsing " +
				"VACMAN controller kernel parameters");
		}

		Connection c = null;
		PreparedStatement s = null;
		ResultSet r = null;

		try
		{
			// Load the database driver
			JdbcUtils.loadDriver(jdbcDriver);

			// Get a JDBC Connection
			c = DriverManager.getConnection(
				jdbcUrl, jdbcUsername, jdbcPassword);

			// Query to get the user's Digipass database values
			s = c.prepareStatement(getDigipassPreparedStatement);
			s.setString(1, username);
			r = s.executeQuery();

			// Process query results
			if (r != null)
			{
				// Assuming a one-to-one mapping of user's to Digipass, there
				// should be one and only one row in the result set
				//
				// NOTE: It is valid but not recommended to map many to one and
				// one to many. Such configurations would require modifications
				// to support.
				if (r.next())
				{
					// The user's DIGIPASS serial number
					String serialNumber =
						r.getString(serialNumberColumn);

					// Get the authentication mode, valid options are response only
					// ("RO") and challenge response ("CR")
					String authenticationMode = r.getString(authenticationModeColumn);

					// Get the user's DIGIPASS data blob
					String digipassString =
						r.getString(digipassDataColumn);

					byte[] digipassBytes = digipassString.getBytes();

					if (debug)
					{
						logger.debug(this, LMN +
							" User's Digipass values:\n" +
							"\t      Serial number: '" + serialNumber + "'\n" +
							"\tAuthentication mode: '" + authenticationMode + "'\n" +
							"\t          Data blob: '" + digipassString + "'");
					}

					// Required Digipass record values cannot be null or empty
					if (((serialNumber != null) &&
						 (serialNumber.length() > 0)) &&
						((authenticationMode != null) &&
						 (authenticationMode.length() > 0)) &&
						((digipassBytes != null) &&
						 (digipassBytes.length > 0)))
					{
						// Initialize the VACMAN controller kernel
						KernelParms kernelParms = new KernelParms(kernelParameters);

						// Call the function of the dll (Windows) or library (Linux/UNIX)
						AAL2Wrap aal2Wrap = new AAL2Wrap();

						// Initialize challenge
						String challenge = null;

						// If the mode is CR and the request does not include a
						// challenge value, then a challenge is generated and returned
						// to the user to input in their Digipass token
						if ("CR".equals(authenticationMode))
						{
							// Parse the challenge and response from the user-supplied
							// password. The challenge is always the first 4 characters.
							// The response length is variable as configured in the
							// Digipass data blob.
							if (password.length() == responseLength)
							{
								challenge = password.substring(0,4);
								password = password.substring(4);

								// NOTE: Logging password, even in debug, could represent
								// a security hole if password is not dynamic
								if (debug)
								{
									logger.debug(this, LMN +
										" User's Digipass challenge/response:\n" +
										"\tChallenge: '" + challenge + "'\n" +
										"\t Response: '" + password + "'\n");
								}
							}

							// Generate challenge and return to user for display
							else
							{
								// Generate the challenge
								challenge = aal2Wrap.AAL2GenerateChallenge(
									digipassBytes, kernelParms);

								// NOTE: We currently print to standard error as there
								// does not yet exist a Cams mechanism to return the
								// value to the user. This will require implementation
								// of the Cams authentication exception feature.
								System.err.println("[INFO] Challenge is: '" +
									challenge + "'");

								if (debug)
									logger.debug(this, LMN +
										" Generated challenge: '" + challenge + "'\n");
							}
						}

						// If the mode is RO or CR with a non-null or non-empty
						// challenge, then attempt to authenticate the user
						if (("RO".equals(authenticationMode)) ||
							((("CR".equals(authenticationMode)) &&
							  (challenge != null) &&
							  (challenge.length() > 0))));
						{
							// Validate the user's password
							if (debug)
								logger.debug(this, LMN +
									" Attempting to validate user-supplied password");

							int validated = aal2Wrap.AAL2VerifyPassword(digipassBytes,
								kernelParms, password, challenge);

							if (validated != 0)
							{
								if (debug)
									logger.debug(this, LMN +
										" User-supplied password was NOT valid, VACMAN " +
										"controller return code: '" + validated + "'");
							}
							else
							{
								if (debug)
									logger.debug(this, LMN +
										" User-supplied password validated");

								// Set the login succeeded flag
								loginSucceeded = true;
							}

							// Always save the Digipass data to database to preserve
							// the state
							s = c.prepareStatement(saveDigipassPreparedStatement);

							// NOTE: Should encoding value be exernalized to facilitate
							// future Internalization?
							s.setString(1, new String(digipassBytes, "UTF-8"));
							s.setString(2, serialNumber);

							s.executeUpdate();
						}
					}

					// Null or empty authentication mode, FailedLoginException thrown later
					else
					{
						if (debug)
						{
							logger.debug(this, LMN +
								" Required authentication mode value missing");
						}
					}
				}

				// No user by that name, FailedLoginException thrown later
				else
				{
					if (debug)
					{
						logger.debug(this, LMN + " " +
							username + "' does not exist");
					}
				}
			}
			else
			{
				if (debug)
				{
					// Null result set, a FailedLoginException will be
					// thrown later
					logger.debug(this, LMN +
						" Result set null for query '" +
						s.toString() + "'");
				}
			}

			// Optionally check for account expired state only if login has
			// succeeded
			if (loginSucceeded)
			{
				// Throw an AccountExpiredException if account has expired
				String accountExpiredPreparedStatement =
					(String)options.get("accountExpiredPreparedStatement");

				if ((accountExpiredPreparedStatement != null) &&
					!("".equals(accountExpiredPreparedStatement.trim())))
				{
					s = c.prepareStatement(accountExpiredPreparedStatement);
					s.setString(1, username);
					r = s.executeQuery();

					if (r != null)
					{
						// Only one row should be returned
						if (r.next())
						{
							// Expecting "1" (expired) or "0" (not expired)
							if (r.getInt(1) == 1)
							{
								throw new AccountExpiredException(LMN + 
									" Error: Authentication " +
									"failed, account expired");
							}
						}
					}
				}
			}

		}
		catch (UnsupportedEncodingException e)
		{
			// Log the unsupported encoding exception for the Digipass
			// (UTF-8 is default)
			if (debug)
				logger.debug(this, "Unsupported encoding exception", e);

			throw new LoginException(LMN + 
				" Error: " + e.getMessage());
		}
		catch (SQLException e)
		{
			// Log the SQL error
			if (debug)
				JdbcUtils.err(e);

			throw new LoginException(LMN + e.getMessage());
		}
		finally
		{
			JdbcUtils.close(c,s,r);
		}

		// Authentication failed, clean up state and throw a FailedLoginException
		if (!loginSucceeded)
		{
			username = null;
			password = null;
			throw new FailedLoginException(LMN + 
				" Authentication failed");
		}

		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @throws LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login method failed, clean up
		if (!loginSucceeded)
			return false;

		// Add all principals to an ArrayList, then commit in single transaction
		List principalList = new ArrayList(5);

		// Add default roles (if any) to list of roles
		String defaultRoles = (String)options.get("defaultRoles");
		if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
		{
			StringTokenizer st = new StringTokenizer(defaultRoles, ",");
			while(st.hasMoreTokens())
				principalList.add(
					new CSRolePrincipal(st.nextElement().toString().trim()));
		}

		// Get the JDBC connection parameters (values validated in login method)
		String jdbcDriver = (String)options.get("jdbcDriver");
		String jdbcUrl = (String)options.get("jdbcUrl");
		String jdbcUsername  = (String)options.get("jdbcUsername");
		String jdbcPassword  = (String)options.get("jdbcPassword");

		// Get roles prepared statement used to query the user's role names
		String rolePreparedStatement =
			(String)options.get("rolePreparedStatement");

		if ((rolePreparedStatement == null) ||
			"".equals(rolePreparedStatement.trim()))
		{
			throw new LoginException(
				LMN + " Missing required option: rolePreparedStatement");
		}

		// Get role column (the column name in the result set for the password)
		String roleColumn = (String)options.get("roleColumn");
		if ((roleColumn == null) || "".equals(roleColumn.trim()))
			throw new LoginException(
				LMN + " Missing required option: roleColumn");

		// Get the roles for the authenticated user

		Connection c = null;
		PreparedStatement s = null;
		ResultSet r = null;

		try
		{
			// Load the database driver
			JdbcUtils.loadDriver(jdbcDriver);

			// Get the JDBC connection
			c = DriverManager.getConnection(
				jdbcUrl, jdbcUsername, jdbcPassword);

			// Query the user's roles
			s = c.prepareStatement(rolePreparedStatement);
			s.setString(1, username);
			r = s.executeQuery();

			// Get the roles from the result set
			if (r != null)
			{
				while (r.next())
				{
					String rolename = r.getString(roleColumn);
					if ((rolename != null) && ("".equals(rolename) == false))
						rolename = rolename.trim();
					else
					{
						if (debug)
							logger.debug(this,
								LMN + " Role query failed: " +
								"null returned from result set column '" +
								roleColumn + "'");

						throw new LoginException("Role assignment failed: " +
							"null or empty role name in column: " + roleColumn);
					}

					// Create the role Principal
					CSRolePrincipal rolePrincipal = new CSRolePrincipal(rolename);

					// Add the CSRolePrincipal if it does not already exist
					if (!subject.getPrincipals(rolePrincipal.getClass()).contains(rolePrincipal))
					{
						principalList.add(rolePrincipal);

						// Log that the CSRolePrincipal that was added
						if (debug)
						{
							logger.debug(this,
								LMN + " Added CSRolePrincipal " +
								rolename + " to principal ArrayList for user '" +
								username + "'");
						}
					}
					else
					{
						if(debug)
						{
							// Log the CSRolePrincipal was not added because it
							// already exists
							logger.debug(this,
								LMN + " CSRolePrincipal '" + username +
								"' already in Subject for '" + username +
								"', not added to " + "principal List");
						}
					}
				}
			}
		}
		catch (SQLException e)
		{
			// Log the SQL error
			if (debug)
				JdbcUtils.err(e);

			throw new LoginException(e.getMessage());
		}
		finally
		{
			JdbcUtils.close(c,s,r);
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		CSUserPrincipal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(userPrincipal.getClass()).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if(debug)
			{
				logger.debug(this, LMN + " Added CSUserPrincipal '" +
					username + "' to principal List");
			}
		}
		else
		{
			if(debug)
			{
				logger.debug(
					this, LMN + " CSUserPrincipal '" + username +
					"' already in Subject, not added to principal List");
			}
		}

		// Add all new principals to Subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		if(debug)
		{
			logger.debug(this, LMN + " Adding " +
				AuthenticationMethod.PASSWORD.getName() +
				" AuthenticationMethod");
		}

		subject.getPublicCredentials().add(
			AuthenticationMethod.PASSWORD);

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.password = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @throws LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// If authentication failed return false. If commit
		// succeeded log the user out return true. Otherwise,
		// Cleanup the login module and and return true.

		if(!loginSucceeded)
			return false;

		if(commitSucceeded)
			logout();
		else
			cleanup();

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @throws LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if(subject != null)
			subject.getPrincipals().clear();

		cleanup();

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username and
	 * password instance variables with their appropriate value.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// This LoginModule requires a username and a password
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Get the callback values and clear the password.
			callbackHandler.handle(callbacks);
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");
			this.password = ((MapCallback)callbacks[0]).getCallbackValue("password");
			((MapCallback)callbacks[0]).setCallbackValue("password", null);
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available",
				e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (debug)
		{
			logger.debug(this, LMN + " User entered username = " + username);
		}
	}
	
	/**
	 * Clean up the resources of the LoginModule
	 */
	private void cleanup()
	{
		callbackHandler = null;
		commitSucceeded = false;
		debug = false;
		logger = null;
		loginSucceeded = false;
		options = null;
		password = null;
		serviceFinder = null;
		sharedState = null;
		subject = null;
		username = null;
	}

} // End of class : DigipassJdbcLoginModule
