/**
 * Copyright (c) 1996-2012 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-
 * INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR
 * ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.novell.ldap.LDAPAttribute;
import com.novell.ldap.LDAPAttributeSet;
import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;
import com.novell.ldap.LDAPSearchResults;
import com.novell.ldap.LDAPJSSESecureSocketFactory;

import java.security.Principal;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.StringTokenizer;

/**
 * Implements the JAAS LoginModule interface to authenticate users
 * against a LDAP repository. Once a connection is obtained the LoginModule
 * attempts to authenticate the user by binding to the LDAP server using the
 * user's distinguished name and password. If the bind is successful, then
 * Cams authenticates the user and stores any roles found by the role filter
 * in the Cams user session.
 * <ul>
 * <li><b>debug</b> - Enable/disable debugging [true|false].
 * 
 * <li><b>host</b> - The hostname of the LDAP repository to connect to.
 * 
 * <li><b>port</b> - The port number of the LDAP repository.
 * 
 * <li><b>useSSL</b> - (Optional) Enable/disable SSL [true|false].
 * 
 * <li><b>ldapServerType</b> - A tag used to set the actual LDAP server
 * type, which is useful for identifying server-specific error codes and
 * translating them to specific login exceptions like:
 * AccountExpiredException, CredentialExpiredException, FailedLoginException,
 * and LoginException. Valid values include: "ACTIVE_DIRECTORY" and "GENERIC".
 * Additional LDAP server types may be added over time.
 * 
 * <li><b>connectTimeout</b> - (Optional) The number of milliseconds to wait
 * for a connection to the LDAP server before timing out (default = 3000).
 * 
 * <li><b>emptyPassword</b> - (Optional) Enable empty passwords (true or
 * false). By default, the LdapLoginModule does not allow a password to
 * be empty. Set this value to true to allow use of empty passwords
 * when anonymous binds are required.
 * 
 * <li><b>userDN</b> - The distinguished name pattern to match to find a
 * user. The username entered by the the user at login will be substituted
 * for the value "{username}". The LDAP login module will not successfully
 * find users in schemas that use a distinguished name that does not use
 * the username to distinguish users.
 * 
 * <li><b>defaultRoles</b> - (Optional) A list of one or more default roles
 * separated by a comma to be added to the user session.
 * 
 * <li><b>roleRegexPattern</b> - (optional) Limits roles saved in the user 
 * session to those that match a regular expression pattern. This option is 
 * useful when the role search returns many roles that are not used by the 
 * access control policy.  
 * 
 * <li><b>roleBaseDN</b> - The base distinguished name to use for your
 * LDAP schema role searches.
 * 
 * <li><b>roleScope</b> - Specifies the LDAP search scope to use for
 * role searches. Valid values are BASE (search is done only at the base
 * dn tree node), ONE (search is done only at the first child tree node
 * of the base dn), and SUB (search is done at the base dn node and all
 * child nodes).
 * 
 * <li><b>roleAttrName</b> - The name of the attribute used to store the
 * role names (usually one name is used but can be a comma-delimited list).
 * 
 * <li><b>roleFilter</b> - The filter to use to search for a users roles.
 * The username entered by the the user at login will be substituted for
 * the value "{username}".
 * </ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.45 $ $Date: 2012/08/22 22:05:12 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 10/20/01
 */
public class LdapLoginModule
	implements LoginModule, LoggerClient
{
	//
	// Static Constants
	//

	/**
	 * The login module name (used in log messages throughout the code).
	 */
	private static final String LMN = "[LdapLoginModule]";
	
	/**
	 * A tag used to identify a generic LDAP server type.
	 */
	private static final String GENERIC_LDAP_SERVER_TYPE = "GENERIC";
	
	/**
	 * A tag used to identify an Active Directory LDAP server type.
	 */
	private static final String ACTIVE_DIRECTORY_LDAP_SERVER_TYPE = "ACTIVE_DIRECTORY";
	
	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Map that contains state shared with other configured LoginModules.
	 */
	private Map sharedState;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag indicating that SSL must be used.
	 */
	private boolean useSSL;

	/**
	 * The LDAP server type, which is useful for customizing error handling.
	 */
	private String ldapServerType = GENERIC_LDAP_SERVER_TYPE;
	
	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating User.
	 */
	private String username;

	/**
	 * The password entered by the authenticating User.
	 */
	private String password;

	/**
	 * A list of user roles. This list is populated in the login method and
	 * used in the commit method.
	 */
	private List roleList;

	//
	// Constructor
	//

	/**
	 * Default LdapLoginModule Constructor.
	 */
	public LdapLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.useSSL = false;
		this.ldapServerType = GENERIC_LDAP_SERVER_TYPE;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.password = null;
		this.sharedState = null;
		this.subject = null;
		this.username = null;
		this.roleList = new ArrayList();
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *			with the end user (prompting for usernames and
	 *			passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.sharedState = sharedState;
		this.debug = "true".equals((String)options.get("debug"));
		this.useSSL = "true".equals((String)options.get("useSSL"));
		
		// Set the LDAP server type
		if (ACTIVE_DIRECTORY_LDAP_SERVER_TYPE.equals(options.get("ldapServerType")))
			this.ldapServerType = ACTIVE_DIRECTORY_LDAP_SERVER_TYPE;
		else
			this.ldapServerType = GENERIC_LDAP_SERVER_TYPE;
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 * <p>
	 * This method executes the following steps:
	 * <ol>
	 * <li>Checks preconditions
	 * <li>Handles callback objects
	 * <li>Get a connection to the LDAP repository
	 * <li>Authenticates the users by attepting and LDAP bind with the
	 * username and password
	 * <li>Return the connection to its ObjectPool
	 * <li>Reset state if authentication failed
	 * <li>Return results of authentication attempt
	 * </ol>
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @exception LoginException thrown if any error occurs while trying to
	 *	authenticate the user.
	 * @exception FailedLoginException if authentication fails.
	 * @exception AccountExpiredException if the user account is expired.
	 * @exception CredentialExpiredException if the user credential is expired.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException,
			AccountExpiredException, CredentialExpiredException
	{
		// Check preconditions
		validateOptions();

		// Handle callback objects.
		handleCallbacks();

		// A username must always be supplied
		if (username == null || username.length() == 0)
			throw new FailedLoginException(LMN + 
				" Authentication failed");

		// A non-empty password must be supplied unless empty
		// passwords are permitted for this configuration.
		String emptyPassword = (String)options.get("emptyPassword");
		if ((emptyPassword == null) || ("false".equals(emptyPassword.trim())))
		{
			if ("".equals(this.password))
				throw new FailedLoginException(LMN + 
					" Authentication failed");
		}
		else if (!("true".equals(emptyPassword.trim())))
		{
			throw new LoginException(LMN +
				" Error: Option 'emptyPassword' must be 'true' or 'false'");
		}

		// Get a connection to the LDAP repository
		LDAPConnection connection = getConnection();

		// Authenticate the users using an LDAP bind
		try
		{
			// Get the userDN
			String userDN = (String)options.get("userDN");

			// Get the user's fully qualifed dn
			String usernameDN = replaceUsername(userDN, "username", username);

			if (debug)
				logger.debug(this, LMN +
					" UsernameDN = " + usernameDN);

			loginSucceeded = bindAuthenticate(usernameDN, connection);

			if (debug)
				logger.debug(this, LMN +
					" login succeeded = " + loginSucceeded);

			if (loginSucceeded)
			{
				// Add default roles (if any) to list of roles
				String defaultRoles = (String)options.get("defaultRoles");
				if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
				{
					StringTokenizer st = new StringTokenizer(defaultRoles, ",");
					while(st.hasMoreTokens())
						roleList.add(st.nextElement().toString().trim());
				}

				// Get the roles for use later during the commit method.
				String roleBaseDN = (String)options.get("roleBaseDN");
				String roleAttrName = (String)options.get("roleAttrName");
				String roleFilter = (String)options.get("roleFilter");
				String roleScope = (String)options.get("roleScope");

				// Set the search scope, the default of 0 is "BASE"
				int scopeInt = 0;
				if ("ONE".equalsIgnoreCase(roleScope))
					scopeInt = 1;
				else if ("SUB".equalsIgnoreCase(roleScope))
					scopeInt = 2;

				// Set the search pattern to find users
				String filter =
					replaceUsername(roleFilter, "username", username);

				// Set the base dn where the search will be performed, and the
				// name of the role to use.
				String[] attrs = getStringArray(roleAttrName, ",");

				// Do the search.
				LDAPSearchResults searchResults = connection.search(
					roleBaseDN, scopeInt, filter, attrs, false);

				// Process the search results to get the role names
				//   -- The first while loop filters for the entries
				//   -- The second while loop filters for the rolenames
				//      and stores them to a principal and subject
				if (searchResults != null)
				{
					while (searchResults.hasMore())
					{
						// Get the next entries attibute
						LDAPEntry nextEntry = searchResults.next();
						LDAPAttributeSet attributeSet = nextEntry.
							getAttributeSet();
						Iterator ldapAttributes = attributeSet.iterator();

						// If supplied, only use roles that match roleRegexPattern regex
						String roleRegexPattern = (String)options.get("roleRegexPattern");

						while (ldapAttributes.hasNext())
						{
							// Attribute may have more than one name/value pair
							LDAPAttribute ldapAttribute = 
								(LDAPAttribute)ldapAttributes.next();
							Enumeration values = ldapAttribute.getStringValues();
							String role = null;

							if ((roleRegexPattern != null) && !("".equals(roleRegexPattern)))
							{
								Pattern pattern = Pattern.compile(roleRegexPattern);
								while(values.hasMoreElements())
								{
									role = (String)values.nextElement();

									if (pattern.matcher(role).matches())
									{
										roleList.add(role);

										if (this.debug)
											logger.debug(this, LMN +
												" Role: '" + role + 
												"' matched roleRegexPattern: '" + 
												roleRegexPattern + "'");
									}
									else
									{
										logger.debug(this, LMN + 
											" Role: '" + role + 
											"' did not match roleRegexPattern: '" + 
											roleRegexPattern + "'");
									}
								}
							}
							else
							{
								while(values.hasMoreElements())
								{
									role = (String)values.nextElement();

									roleList.add(role);

									if (debug)
										logger.debug(this, LMN +
										 " User has role: '" + role + "'");
								}
							}
						}
					}
				}
				else
				{
					if (debug)
						logger.debug(this, LMN + " No roles found for: '" + username + "'");
				}
			}
		}
		catch (LDAPException e)
		{
			// An error occurred looking up roles for the user therefore login
			// did not succeed. Although this deviates slightly from the JAAS
			// implementation of the commit method it basically performs this
			// error check prior to the commit method allowing it to fail fast.
			loginSucceeded = false;
			logger.error(this,  LMN + " Error occurred", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		finally
		{
			// Close the connection
			if (connection != null)
			{
				try
				{
					connection.disconnect();
					connection = null;
				}
				catch(LDAPException e)
				{
					// Catch the general exception here because we do not
					// need to distinguish between the exceptions that occurred.
					// We are just going to log the exception anyway.
					logger.error(this, LMN +
						" Error occurred while disconnecting", e);
				}
			}
		}

		// Reset state, authentication failed.
		if (!loginSucceeded)
		{
			this.username = null;
			this.password = null;

			throw new FailedLoginException(LMN +
				" Authentication failed");
		}

		// Return results of authentication attempt.
		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @exception LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login failed, clean up state
		if (!loginSucceeded)
			return false;

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList();

		// Repository operations to get role principals
		Iterator roleListItr = roleList.iterator();
		while(roleListItr.hasNext())
		{
			String rolename = (String)roleListItr.next().toString().trim();
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				principalList.add(rolePrincipal);

				if (debug)
					logger.debug(this,LMN +
						" Added CSRolePrincipal: '" +
						rolename + "' to principalList for: '" +
						username + "'");
			}
			else
			{
				if (debug)
					logger.debug(this,LMN +
						" CSRolePrincipal: '" +
						rolename + "' already in principalList for: '" +
						username + "', not added");
			}
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if (debug)
				logger.debug(this, LMN +
					" Added CSUserPrincipal: '" + username +
					"' to principalList");
		}
		else
		{
			if (debug)
				logger.debug(this, LMN +
					" CSUserPrincipal: '" + username +
					"' already in principalList, not added");
		}

		if (debug)
			logger.debug(this, LMN +
				" Adding principalList to Subject");

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		if(debug)
		{
			logger.debug(this, LMN + " Adding: '" +
				AuthenticationMethod.PASSWORD.getName() +
				"' AuthenticationMethod");
		}

		subject.getPublicCredentials().add(
			AuthenticationMethod.PASSWORD);

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.password = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @exception LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.
			if (this.roleList != null)
			{
				this.roleList.clear();
				this.roleList = null;
			}

			this.loginSucceeded = false;
			this.username = null;
			this.password = null;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @exception LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if (this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if (this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.logger = null;
		this.callbackHandler = null;
		this.username = null;
		this.password = null;
		this.options = null;
		this.loginSucceeded = false;
		this.commitSucceeded = false;
		this.debug = false;

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username and
	 * password instance variables with their appropriate value.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// This LoginModule requires a username and a password
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Get the callback values and clear the password.
			callbackHandler.handle(callbacks);
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");
			this.password = ((MapCallback)callbacks[0]).getCallbackValue("password");
			((MapCallback)callbacks[0]).setCallbackValue("password", null);
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available",
				e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (debug)
		{
			logger.debug(this, LMN + " User entered username = " + username);
		}
	}

	/**
	 * Validate the LoginModules options Map object to ensure that all
	 * required properties exist and contain valid values.
	 *
	 * @exception LoginException if any of the following conditions occur:
	 * <UL>
	 * <LI>A CallbackHandler object is not available
	 * <LI>the 'host' property is not available
	 * <LI>the 'port' property is not available or a valid integer value
	 * <LI>the 'connectTimeout' property is not a valid integer value
	 * <LI>the 'userDN' property is not available
	 * <LI>the 'roleBaseDN' property is not available
	 * <LI>the 'roleAttrName' property is not available
	 * <LI>the 'roleFilter' property is not available
	 * <LI>the 'roleScope' property is not available
	 * <LI>the 'roleScope' is not a valid role scope value
	 * </UL>
	 */
	private void validateOptions() throws LoginException
	{
		if (callbackHandler == null)
			throw new LoginException(LMN +
				" Error: no CallbackHandler available");

		String host = (String)options.get("host");
		if ((host == null) || (host.trim().length() == 0))
			throw new LoginException(LMN +
				" Error: Required 'host' property missing");

		String s = (String)options.get("port");
		if ((s == null) || (s.trim().length() == 0))
			throw new LoginException(LMN + 
				" Error: Required 'port' property missing");

		int port = 389;
		try
		{
			port = Integer.parseInt(s);
			if (port < 1)
				throw new LoginException("[ActiveDirectoryLoginModule] Error: " +
					"'port' property must be greater than zero");
		}
		catch (NumberFormatException e)
		{
			throw new LoginException("[ActiveDirectoryLoginModule] Error: " +
				"'port' property is not an integer value");
		}

		int timeout = 3000;
		s = null;
		s = (String)options.get("connectTimeout");
		if ((s != null) && (s.trim().length() > 0))
		{
			try
			{
				timeout = Integer.parseInt(s);
				if (timeout < 1)
					throw new LoginException("[ActiveDirectoryLoginModule] Error: " +
					"'connectTimeout' property must be greater than zero");
			}
			catch (NumberFormatException e)
			{
				throw new LoginException("[ActiveDirectoryLoginModule] Error: " +
					"'connectTimeout' property is not an integer");
			}
		}

		String userDN = (String)options.get("userDN");
		if ((userDN == null) || "".equals(userDN.trim()))
			throw new LoginException(LMN +
				" Error: Required 'userDN' property missing");

		String roleBaseDN = (String)options.get("roleBaseDN");
		if ((roleBaseDN == null) || "".equals(roleBaseDN.trim()))
			throw new LoginException(LMN + 
				" Error: Required 'roleBaseDN' property missing");

		String roleAttrName = (String)options.get("roleAttrName");
		if ((roleAttrName == null) || "".equals(roleAttrName.trim()))
			throw new LoginException(LMN + 
				" Error: Required 'roleAttrName' property missing");

		String roleFilter = (String)options.get("roleFilter");
		if ((roleFilter == null) || "".equals(roleFilter.trim()))
			throw new LoginException(LMN + 
				" Error: Required 'roleFilter' property missing");

		String roleScope = (String)options.get("roleScope");
		if ((!"BASE".equalsIgnoreCase(roleScope)) &&
			(!"ONE".equalsIgnoreCase(roleScope)) &&
			(!"SUB".equalsIgnoreCase(roleScope)))
			throw new LoginException(LMN + 
				"Error: Required option 'roleScope' must be 'BASE', 'ONE', or 'SUB'");
	}

	/**
	 * Get an LDAP server connection.
	 *
	 * @return a LDAPConnection.
	 * @exception LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getConnection()
		throws LoginException
	{
		// Get the connection values to use
		String host = (String)options.get("host");
		int port = Integer.parseInt((String)options.get("port"));
		int timeout = Integer.parseInt((String)options.get("connectTimeout"));

		try
		{
			if (debug)
				logger.debug(this, LMN + 
					" Instantiating LDAPConnection: host=" + host +
					", port=" + port + ", useSSL=" + useSSL +
					", connectTimeout=" + timeout);

			// Initialize the connection
			LDAPConnection connection = null; 

			if (useSSL)
			{
				// Warn if required SSL System property for trust store is not set
				String trustStoreProp = System.getProperty("javax.net.ssl.trustStore");
				if (trustStoreProp == null)
					logger.warning(this, "Missing required SSL system property: " +
						"javax.net.ssl.trustStore");

				connection = new LDAPConnection((new LDAPJSSESecureSocketFactory())); 
			}
			else
			{
				connection = new LDAPConnection(); 
			}

			// Set the connection timeout
			connection.setSocketTimeOut(timeout);

			// Debug to help customers determine LDAP response time 
			long timer = System.currentTimeMillis(); 

			// Attempt connection
			connection.connect(host, port);

			if (debug)
			{
				logger.debug(this, LMN + " Time to connect: '" +
					(System.currentTimeMillis() - timer) + "'");

				if (connection.isConnected() && connection.isConnectionAlive())
					logger.debug(this, LMN + " LDAP connection succeeded");
				else
					logger.debug(this, LMN + " LDAP connection failed");

				logger.debug(this, LMN + " Returning LDAPConnection");
			}

			return connection;
		}
		catch (LDAPException e)
		{
			logger.error(this, LMN + "LDAP error", e);

			// Dump the nested Throwables to get at the root cause
			int i = 1;
			Throwable cause = e.getCause();
			do
			{
				if (cause != null)
					logger.error(this, LMN + "LDAP error: CAUSE " + i + ": " +
						cause.getMessage());
			}
			while ((cause = cause.getCause()) != null);

			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
	}

	/**
	 * Authenticate using LDAP server bind.
	 *
	 * @param usernameDN the username's distinguished name.
	 * @param connection the LDAPConnection.
	 * @exception CredentialExpiredException if bind password is expired.
	 * @return true if authentication succeeded, false otherwise.
	 */
	private boolean bindAuthenticate(
		String usernameDN, LDAPConnection connection)
		throws LoginException, FailedLoginException,
			AccountExpiredException, CredentialExpiredException
	{
		try
		{
			connection.bind(3, usernameDN, password.getBytes());
			return true;
		}
		catch(LDAPException e)
		{
			if (ACTIVE_DIRECTORY_LDAP_SERVER_TYPE.equalsIgnoreCase(ldapServerType))
			{
				activeDirectoryLDAPExceptionHandler(e);
			}
			else
			{
				genericLdapServerLDAPExceptionHandler(e);
			}
			
			return false;
		}
	}

	/*
	 * Translate a generic LDAP server LDAPException into a more specific
	 * LoginException and throw it.
	 * 
	 * @param e the LDAPException.
	 * @throws LoginException if not a more specific login failure.
	 * @throws FailedLoginException if invalid login credentials.
	 * @throws CredentialExpiredException if a login credential has expired.
	 * @throws AccountExpiredException if the user account has expired.
	 */
	private void genericLdapServerLDAPExceptionHandler(LDAPException e)
		throws LoginException, FailedLoginException,
			CredentialExpiredException, AccountExpiredException
	{
		// Check for credential (password) expired exception. Note the
		// message "password expired" is likely to be specific to the
		// SunOne LDAP server.
		if (e.getResultCode() == LDAPException.INVALID_CREDENTIALS)
		{
			if ((e.getLDAPErrorMessage() != null) &&
				(e.getLDAPErrorMessage().toLowerCase().startsWith("password expired")))
			{
				throw new CredentialExpiredException(LMN +
					" Error: " + e.getMessage() + "\n");
			}
			else
			{
				throw new FailedLoginException(LMN +
					" Invalid credentials: " +
					e.getLDAPErrorMessage() + "\n");
			}
		}
		else if (e.getResultCode() == LDAPException.UNWILLING_TO_PERFORM)
		{
			// Check for account expired exception. Red Hat Directory Server returns 
			// this message when "nsAccountLock" is set to "true".
			if ((e.getLDAPErrorMessage() != null) &&
				(e.getLDAPErrorMessage().toLowerCase().startsWith("account inactivated")))
			{
				throw new AccountExpiredException(LMN +
					"Error: " + e.getMessage() + "\n");
			}
			else
			{
				// Login failure: some other constraint violation
				throw new LoginException(LMN +
					" Error while trying to " +
						"bind user: " + e.getLDAPErrorMessage() + "\n");	
			}
		}
		else if (e.getResultCode() == LDAPException.CONSTRAINT_VIOLATION)
		{
			// Check for account expired exception, due to retry violation. The
			// message "Exceed password retry" is likely to be specific to the
			// SunOne LDAP server.
			if ((e.getLDAPErrorMessage() != null) &&
				(e.getLDAPErrorMessage().toLowerCase().startsWith("exceed password retry")))
			{
				throw new AccountExpiredException(LMN +
					" Error: " + e.getMessage() + "\n");
			}
			else
			{
				// Login failure: some other constraint violation
				throw new LoginException(LMN +
					" Error while trying to bind user: " +
					e.getLDAPErrorMessage() + "\n");	
			}
		}
		else
		{
			// Login failure for other reason
			throw new LoginException(LMN +
				" Error while trying to bind user: " +
				e.getLDAPErrorMessage() + "\n");	
		}
	}
	
	/**
	 * Translate a specific ActiveDirectory LDAPException into a specific
	 * LoginException and throw it.
	 * 
	 * @param e the LDAPException containing an ActiveDirectory specific message.
	 * @throws LoginException if not a more specific login failure.
	 * @throws FailedLoginException if invalid login credentials.
	 * @throws CredentialExpiredException if a login credential has expired.
	 * @throws AccountExpiredException if the user account has expired.
	 */
	private void activeDirectoryLDAPExceptionHandler(LDAPException e)
		throws LoginException, FailedLoginException,
			CredentialExpiredException, AccountExpiredException
	{
		// Normal login failure due to invalid credentials, check
		// Active Directory server message for subcode to determine the
		// most accurate exception to throw. Possible codes include:
		// 52b - invalid password (FailedLoginException)
		// 52e - invalid credentials (FailedLoginException)
		// 530 - invalid login time (FailedLoginException)
		// 532 - password expired (CredentialExpiredException)
		// 533 - account disabled (AccountExpiredException)
		// 701 - account expired (AccountExpiredException)
		// 773 - password reset required (CredentialExpiredException)
		if (e.getResultCode() == LDAPException.INVALID_CREDENTIALS)
		{
			if (e.getLDAPErrorMessage().indexOf("data 532") > 0)
			{
				throw new CredentialExpiredException(LMN +
					" Password expired: " +
					e.getLDAPErrorMessage() + "\n");
			}
			else if (e.getLDAPErrorMessage().indexOf("data 533") > 0)
			{
				throw new AccountExpiredException(LMN +
					" Account disabled: " +
					e.getLDAPErrorMessage() + "\n");
			}
			else if (e.getLDAPErrorMessage().indexOf("data 701") > 0)
			{
				throw new AccountExpiredException(LMN +
					" Account expired: " +
					e.getLDAPErrorMessage() + "\n");
			}
			else if (e.getLDAPErrorMessage().indexOf("data 773") > 0)
			{
				throw new CredentialExpiredException(LMN +
					" Password reset required: " +
					e.getLDAPErrorMessage() + "\n");
			}
			else
			{
				throw new FailedLoginException(LMN +
					" Invalid credentials: " +
					e.getLDAPErrorMessage() + "\n");
			}
		}
		else
		{
			// Login failure for reason other than invalid credentials
			throw new LoginException(LMN +
				" Error while trying to " +
					"bind user: " + e.getLDAPErrorMessage() + "\n");	
		}
	}

	/**
	 * Return the input pattern replaced with the match value.
	 *
	 * @param pattern The input pattern to use
	 * @param matchName The variable name to match
	 * @param matchValue The value to replace
	 * @return The input pattern with matched value
	 */
	private String replaceUsername(String pattern, String matchName,
		String matchValue)
	{
		String tmp = null;
		StringBuffer sb = new StringBuffer();
		StringTokenizer st = new StringTokenizer(pattern, "{}");
		while (st.hasMoreTokens())
		{
			tmp = st.nextToken();

			if (tmp.trim().equals(matchName))
				sb.append(matchValue);
			else
				sb.append(tmp);
		}

		return sb.toString();
	}

	/**
	 * Parse a string for values separated by a single input character. For
	 * example, "foo,goober,bar" could be parsed by specifying "," as the
	 * token. It would return a String[3] with foo, goober, and bar as
	 * the values.
	 *
	 * @param s the string to evaluate
	 * @param token the token that separates the String values
	 *
	 * @return a String array
	 */
	static String[] getStringArray(String s, String token)
	{
		// Create tokenizer and initialize array based on the number of tokens
		StringTokenizer st = new StringTokenizer(s.trim(), token);
		String[] sa = new String[st.countTokens()];

		// Parse the tokens and store them in the array
		int i = 0;
		while(st.hasMoreTokens())
		{
			sa[i] = st.nextToken();
			i++;
		}

		return sa;
	}

} // End of class : LdapLoginModule
