/**
 * Copyright (c) 1996-2010 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-
 * INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR
 * ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.rsa.authagent.authapi.AuthSessionFactory;
import com.rsa.authagent.authapi.AuthSession;

import java.security.Principal;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;
import java.io.File;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * <code>RsaSecurIdLoginModule</code> implements a JAAS LoginModule that
 * authenticates RSA SecurId token users.
 *
 * <b>NOTE:</b>This module provides basic authentication support for users that
 * have an established PIN. This module treats RSA Authentication Manager
 * requests for "NEW_PIN_REQUIRED" and "NEXT_TOKEN_REQUIRED" as successful
 * authentication attempts. It provides no support for accepting a "New PIN"
 * or requesting a "Next Tokencode". Use an RSA-provided SecurId agent for
 * these use cases.
 * <ul>
 * <li><b>debug</b> - Enables/disables debugging for the LoginModule.</li>
 * <li><b>configPath</b> - The fully-qualified path to the
 * top-level RSA API configuration properties file. For example:
 * <br>(unix): /var/cams/conf/domains/system/rsa_api.properties
 * <br>(windows): c:/cams/conf/domains/system/rsa_api.properties</li>
 * <li><b>defaultRoles</b> - (Optional) A comma-separated list of one or
 * more default roles assigned to the user on successful authentication.
 * For example: SECURID</li>
 * <li><b>username_cb_name</b> - (Optional) By default, a "username" callback
 * is expected. If the login form uses a different name, then optionally
 * configure it, for example:
 *    <option name="username_cb_name" value="uid"/></li>
 * <li><b>password_cb_name</b> - (Optional) By default, a "password" callback
 * is expected. If the login form uses a different name, then optionally
 * configure it, for example:
 *    <option name="password_cb_name" value="code"/></li>
 * </ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.9 $ $Date: 2010/12/17 21:15:41 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 05/01/2006
 */
public class RsaSecurIdLoginModule
	implements LoginModule, LoggerClient
{

	//
	// Static Constants
	//

	/**
	 * The login module name (used in log messages throughout the code).
	 */
	private static final String LMN = "[RsaSecurIdLoginModule]";

	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating User.
	 */
	private String username;

	/**
	 * The password entered by the authenticating User.
	 */
	private String password;

	/**
	 * A list of user roles. This list is populated in the login method and
	 * used in the commit method.
	 */
	private List roleList;

	//
	// Constructor
	//

	/**
	 * Default RsaSecurIdLoginModule Constructor.
	 */
	public RsaSecurIdLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.username = null;
		this.password = null;
		this.subject = null;
		this.roleList = new ArrayList();
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *			with the end user (prompting for usernames and
	 *			passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.debug = "true".equals((String)options.get("debug"));
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 * <p>
	 * This method executes the following steps:
	 * <ol>
	 * <li>Checks preconditions
	 * <li>Handles callback objects
	 * <li>Get a connection to the LDAP repository
	 * <li>Authenticates the users by attepting and LDAP bind with the
	 * username and password
	 * <li>Return the connection to its ObjectPool
	 * <li>Reset state if authentication failed
	 * <li>Return results of authentication attempt
	 * </ol>
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @exception LoginException thrown if any error occurs while trying to
	 *	authenticate the user.
	 * @exception FailedLoginException if authentication fails.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException
	{
		// Check preconditions
		validateOptions();

		// Handle callback objects.
		handleCallbacks();

		// A username must always be supplied
		if (username == null || username.length() == 0)
		{
			if (debug)
				logger.debug(this, LMN +
					"Authentication failed: null or empty username");

			throw new FailedLoginException( LMN +
				" Authentication failed");
		}

		// A password in the form <pin><tokencode> must be supplied
		if (password == null || password.length() == 0)
		{
			if (debug)
				logger.debug(this, LMN + 
					" Authentication failed: null or empty password");

			throw new FailedLoginException( LMN +
				" Authentication failed");
		}

		// Authenticate the user
		try
		{
			String configPath = (String)options.get("configPath");

			if (debug)
				logger.debug(this, LMN +
					" Attempting to get the AuthSessionFactory. Using configPath=" +
					configPath);

			AuthSessionFactory asf = AuthSessionFactory.getInstance(configPath);

			if (debug)
				logger.debug(this, LMN +
					" Got the AuthSessionFactory, attempting to get an AuthSession");

			AuthSession session = asf.createUserSession();

			if (debug)
				logger.debug(this, LMN +
					" Got the AuthSession, checking username and passcode");

			int status = session.check(username, password);

			switch (status)
			{
		    	// Authentication was successful
				case AuthSession.ACCESS_OK:

					if (debug)
						logger.debug(this, LMN +
							" Got status: AuthSession.ACCESS_OK, login succeeded");

					loginSucceeded = true;
					break;

				// NEW_PIN_REQUIRED treated as a valid authentication
			   	case AuthSession.NEW_PIN_REQUIRED:

					if (debug)
						logger.debug(this, LMN +
							" Got status: AuthSession.NEW_PIN_REQUIRED, login succeeded");

					loginSucceeded = true;
					break;

				// NEXT_CODE_REQUIRED treated as a valid authentication
			   	case AuthSession.NEXT_CODE_REQUIRED:

					if (debug)
						logger.debug(this, LMN +
							" Got status: AuthSession.NEXT_CODE_REQUIRED, login succeeded");

					loginSucceeded = true;
			    	break;

		    	default:
					if (debug)
						logger.debug(this, LMN +
							" Got status: " + status + ", login failed");

					break;
			}

			// Done authenticating or attempting to authenticate, so close the session
			session.close();
		}
		catch (Exception e)
		{
			logger.error(this, LMN + " API error: " + e.getMessage());
			throw new LoginException( LMN + " API error");
		}

		// Reset state, authentication failed.
		if (!loginSucceeded)
		{
			this.username = null;
			this.password = null;

			throw new FailedLoginException( LMN +
				" Authentication failed");
		}

		// Return results of authentication attempt.
		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @exception LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login failed, clean up state
		if (!loginSucceeded)
			return false;

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList();

		// Add default roles (if any) to list of roles
		String defaultRoles = (String)options.get("defaultRoles");
		if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
		{
			StringTokenizer st = new StringTokenizer(defaultRoles, ",");
			while(st.hasMoreTokens())
				roleList.add(st.nextElement().toString().trim());
		}

		// Repository operations to get role principals
		Iterator roleListItr = roleList.iterator();
		while(roleListItr.hasNext())
		{
			String rolename = (String)roleListItr.next().toString().trim();
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				principalList.add(rolePrincipal);

				if (debug)
					logger.debug(this, LMN +
						" Added CSRolePrincipal " +
						rolename + " to principal ArrayList for user '" +
						username + "'");
			}
			else
			{
				if (debug)
					logger.debug(this, LMN +
						" CSRolePrincipal '" +
						rolename + "' already in Subject for '" +
						username + "', not added to principal List");
			}
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if (debug)
				logger.debug(this, LMN +
					" Added CSUserPrincipal '" + username +
					"' to principal List");
		}
		else
		{
			if (debug)
				logger.debug(this, LMN +
					" CSUserPrincipal '" + username +
					"' already in Subject, not added to principal List");
		}

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		if(debug)
		{
			logger.debug(this, LMN + " Adding " +
				AuthenticationMethod.HARDWARETOKEN.getName() +
				" AuthenticationMethod");
		}

		subject.getPublicCredentials().add(
			AuthenticationMethod.HARDWARETOKEN);

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.password = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @exception LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.
			if (this.roleList != null)
			{
				this.roleList.clear();
				this.roleList = null;
			}

			this.loginSucceeded = false;
			this.username = null;
			this.password = null;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @exception LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if (this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if (this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.logger = null;
		this.callbackHandler = null;
		this.username = null;
		this.password = null;
		this.options = null;
		this.loginSucceeded = false;
		this.commitSucceeded = false;
		this.debug = false;

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username and
	 * password instance variables with their appropriate value.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// This LoginModule requires a username and a password
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Get the callback values and clear the password.
			callbackHandler.handle(callbacks);

			// Get the username callback value
			String username_cb_name = (String)options.get("username_cb_name");
			if ((username_cb_name == null) || ("".equals(username_cb_name)))
				username_cb_name = "username";

			this.username = ((MapCallback)callbacks[0]).getCallbackValue(username_cb_name);

			// Get the username callback value
			String password_cb_name = (String)options.get("password_cb_name");
			if ((password_cb_name == null) || ("".equals(password_cb_name)))
				password_cb_name = "password";

			this.password = ((MapCallback)callbacks[0]).getCallbackValue(password_cb_name);
			((MapCallback)callbacks[0]).setCallbackValue(password_cb_name, null);
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available",
				e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (debug)
		{
			logger.debug(this, LMN + " User entered username = " + username);
		}
	}

	/**
	 * Validate the LoginModule's options Map object to ensure that all
	 * required properties exist and contain valid values.
	 *
	 * @exception LoginException if any of the following conditions occur:
	 * <ul>
	 * <li>A CallbackHandler object is not available,
	 * <li>the 'configPath' property is not available,
	 * </ul>
	 */
	private void validateOptions() throws LoginException
	{
		if (callbackHandler == null)
			throw new LoginException( LMN +
				" Error: no CallbackHandler available");

		String configPath = (String)options.get("configPath");
		if ((configPath == null) || "".equals(configPath.trim()))
			throw new LoginException( LMN +
				" Error: Required 'configPath' property missing");

		// Make sure the config file exists
		File f = new File(configPath);
		if (! f.exists())
		{
			logger.error(this, LMN + " Error: 'configPath' file: " +
				configPath + " does not exist");
			throw new LoginException( LMN +
				" Error: 'configPath' file: " +
				configPath + " does not exist");
		}
	}

} // End of class : RsaSecurIdLoginModule
