/**
 * Copyright (c) 1996-2010 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.util.DigestString;

import com.novell.ldap.LDAPAttribute;
import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPDN;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;

import java.security.Principal;
import java.security.NoSuchAlgorithmException;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * Implements the JAAS LoginModule interface to authenticate users
 * against a LDAP repository. Once a connection is obtained the LoginModule
 * attempts to authenticate the user.
 * The method of authentication is determined by the configuration
 * parameter "passwordValidationMethod". There are three methods supported:
 * <UL>
 * <LI><B>Cams</B> - Retreive the userPassword attribute value and let the
 * Cams LDAP login module do the password validation. The primary
 * advantage of this method is that you can take advantage of all
 * supported Cams password digests (including Unix Crypt, MD5, MD5
 * with salt, SHA, and SHA with salt).
 * <LI><B>Compare</B> - Performs an LDAP compare method call, which
 * delegates the comparison of the user's userPassword attribute to
 * the LDAP server. The primary advantage of this method is performance,
 * the check is performed on the server and a boolean value is returned
 * to the Cams LDAP login module indicating success or failure. Depending
 * upon the LDAP server, the compare method may do hashing to match
 * against a hashed value. This method may also provide greater security
 * as the LDAP server access control policy can be locked down to not
 * allow retreival of a user password (especially important if the
 * password is stored in cleartext).
 * <LI><B>Bind</B> - Attempts to bind to the LDAP server using the user's
 * distinguished name and password. If the bind is successful, then Cams
 * authenticates the user. The primary advantage of this method is that
 * it is the only one you can use if the LDAP server access control policy
 * prohibit retreival and comparison of the userPassword attribute. The
 * disadvantage is that binding has a greater performance hit than the
 * other two methods.
 * </UL>
 * All configuration parameters are listed below:
 * <Ul>
 * <LI><B>debug</B> - Enables/disables debugging for the LoginModule.
 * <LI><B>host</B> - The hostname of the LDAP repository to connect to.
 * <LI><B>port</B> - The port number of the LDAP repository.
 * <LI><B>passwordValidationMethod</B> - Specifies how the LDAP login
 * module will do password validation. Valid values are Cams (validation
 * is done by Cams LDAP login module using Password Digests), Compare
 * (validation is done by the LDAP server using the compare method call),
 * and Bind (validation is done the LDAP server using an LDAP connection
 * bind call).
 * <LI><B>adminUsername</B> - The username of the administrator to login with.
 * This value is only required if the passwordValidationMethod is Cams or
 * Compare.
 * <LI><B>adminPassword</B> - The password of the administrator to login with.
 * This value is only required if the passwordValidationMethod is Cams or
 * Compare.
 * <LI><B>userDN</B> - The distinguished name pattern to match to find a
 * user. The username entered by the the user at login will be substituted
 * for the value "{username}". The LDAP login module will not successfully
 * find users in schemas that use a distinguished name that does not use
 * the username to distinguish users.
 * <LI><B>roleAttrName</B> - The name of the attribute used to store the
 * role names.
 * </Ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.15 $ $Date: 2010/12/17 21:15:41 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 10/20/01
 */
public class SiteServerLoginModule
	implements LoginModule, LoggerClient
{
	//
	// Static Constants
	//

	/**
	 * The login module name (used in log messages throughout the code).
	 */
	private static final String LMN = "[SiteServerLoginModule]";

	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Map that contains state shared with other configured LoginModules.
	 */
	private Map sharedState;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating User.
	 */
	private String username;

	/**
	 * The password entered by the authenticating User.
	 */
	private String password;


	/**
	 * The password validation method (Cams/Compare/Bind).
	 */
	private String passwordValidationMethod;

	/**
	 * A list of user roles. This list is populated in the login method and
	 * used in the commit method.
	 */
	private List roleList;

	//
	// Constructor
	//

	/**
	 * Default SiteServerLoginModule Constructor.
	 */
	public SiteServerLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.password = null;
		this.sharedState = null;
		this.subject = null;
		this.sharedState = null;
		this.roleList = new ArrayList(10);
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *			with the end user (prompting for usernames and
	 *			passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.sharedState = sharedState;
		this.debug = "true".equals((String)options.get("debug"));
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 * <p>
	 * This method executes the following steps:
	 * <ol>
	 * <li>Checks preconditions
	 * <li>Handles callback objects
	 * <li>Get a connection to the LDAP repository
	 * <li>Authenticates the users using the appropriate authentication
	 * method defined by the passwordValidationMethod configuration
	 * parameter.
	 * <li>Return the connection to its ObjectPool
	 * <li>Reset state if authentication failed
	 * <li>Return results of authentication attempt
	 * </ol>
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @throws LoginException thrown if any error occurs while trying to
	 * 		authenticate the user.
	 * @throws FailedLoginException if authentication fails.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException
	{
		// Check preconditions
		validateOptions();

		// Handle callback objects.
		handleCallbacks();

		// Get a connection to the LDAP repository
		LDAPConnection connection = getConnection();

		// Authenticate the users using the appropriate authentication
		// method defined by the passwordValidationMethod configuration
		// parameter.

		try
		{

			// Get and userDN
			String userDN = (String)options.get("userDN");

			// Get the user's fully qualifed dn
			String usernameDN = replaceUsername (userDN, "username", username);

			if (debug)
				logger.debug(this, LMN + " UsernameDN = " + usernameDN);

			if("Cams".equalsIgnoreCase(passwordValidationMethod))
				loginSucceeded = camsAuthenticate(usernameDN, connection);
			else if("Compare".equalsIgnoreCase(passwordValidationMethod))
				loginSucceeded = compareAuthenticate(usernameDN, connection);
			else if("Bind".equalsIgnoreCase(passwordValidationMethod))
				loginSucceeded = bindAuthenticate(usernameDN, connection);
			else
			{
				throw new LoginException(LMN + " Unknown " +
					"passwordValidationMethod, method=" +
					passwordValidationMethod);
			}

			if(loginSucceeded)
			{
				// We are going to get the roles for use later with the commit
				// method.

				// Get the userDN (already validated in login)
				String roleAttrName = (String)options.get("roleAttrName");

				// Read the user's role attribute from the directory

				String[] roleAttr = {roleAttrName};
				LDAPEntry attrs = connection.read(usernameDN, roleAttr);
				LDAPAttribute groupsAttr = attrs.getAttribute(roleAttrName);

				// Get and associate with user the unique roles
				Enumeration allValues = groupsAttr.getStringValues();
				while (allValues.hasMoreElements())
				{
					String roleName = (String)allValues.nextElement();
					roleList.add(roleName);
				}

			}

		}
		catch (LDAPException e)
		{
			logger.error(this, LMN + " Error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		finally
		{
			// Close the connection
			if (connection != null)
			{
				try
				{
					connection.disconnect();
					connection = null;
				}
				catch(LDAPException e)
				{
					// Catch the general exception here because we do not
					// need to distinguish between the exceptions that occurred.
					// We are just going to log the exception anyway.
					logger.error(this, LMN +
						" Error occurred while disconnecting", e);
				}
			}
		}

		// Reset state, authentication failed.
		if (!loginSucceeded)
		{
			this.username = null;
			this.password = null;

			throw new FailedLoginException( LMN +
				" Authentication failed");
		}

		// Return results of authentication attempt.
		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned.
	 * @throws LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login failed, clean up state
		if (!loginSucceeded)
			return false;

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList(5);

		// Repository operations to get role principals
		Iterator roleListItr = roleList.iterator();
		while (roleListItr.hasNext())
		{
			// The role is returned by SiteServer as a DN, which
			// has the role common name as the first component:
			//
			//        "cn=myRole,ou=Groups,o=company"
			//
			// So we explode the DN and extact the role common name.
			String dn = (String)roleListItr.next();
			String[] sa = LDAPDN.explodeDN(dn, true);
			String rolename = sa[0].trim();
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				principalList.add(rolePrincipal);

				if (debug)
				{
					logger.debug(this, LMN +
						" Added CSRolePrincipal " +
						rolename + " to principal ArrayList for user '" +
						username + "'");
				}
			}
			else
			{
				if (debug)
				{
					logger.debug(this, LMN +
						" CSRolePrincipal '" +
						rolename + "' already in Subject for '" +
						username + "', not added to principal List");
				}
			}
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if (debug)
			{
				logger.debug(this, LMN +
					" Added CSUserPrincipal '" + username +
					"' to principal List");
			}
		}
		else
		{
			if (debug)
			{
				logger.debug(this, LMN +
					" CSUserPrincipal '" + username +
					"' already in Subject, not added to principal List");
			}
		}

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.password = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded false otherwise.
	 * @throws LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.

			if(this.roleList != null)
			{
				this.roleList.clear();
				this.roleList = null;
			}

			this.loginSucceeded = false;
			this.username = null;
			this.password = null;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded false otherwise.
	 * @throws LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if(this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if(this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.logger = null;
		this.callbackHandler = null;
		this.username = null;
		this.password = null;
		this.options = null;
		this.loginSucceeded = false;
		this.commitSucceeded = false;
		this.debug = false;

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username and
	 * password instance variables with their appropriate value.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// This LoginModule requires a username and a password
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Get the callback values and clear the password.
			callbackHandler.handle(callbacks);
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");
			this.password = ((MapCallback)callbacks[0]).getCallbackValue("password");
			((MapCallback)callbacks[0]).setCallbackValue("password", null);
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available",
				e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (debug)
		{
			logger.debug(this, LMN + " User entered username = " + username);
		}
	}

	/**
	 * Validate the LoginModules options Map object to ensure that all
	 * required properties exist and contain valid values.
	 *
	 * @throws LoginException if any of the following conditions occur:
	 * <UL>
	 * <LI>A CallbackHandler object is not available,
	 * <LI>the 'host' property is not available,
	 * <LI>the 'port' property is not available,
	 * <LI>the 'port' property is not a valid integer value,
	 * <LI>the port integer value is less than 1,
	 * <LI>the 'passwordValidationMethod' property is not available,
	 * <LI>the 'passwordValidationMethod' property is not a valid validation
	 * method,
	 * <LI>the 'adminUsername' property is not available (must be available
	 * for all validation methods exception Bind),
	 * <LI>the 'adminPassword' property is not available (must be available
	 * for all validation methods exception Bind),
	 * <LI>the 'userDN' property is not available,
	 * <LI>the 'roleAttrName' property is not available,
	 * <LI>the 'roleFilter' property is not available,
	 * </UL>
	 */
	private void validateOptions() throws LoginException
	{
		if (callbackHandler == null)
			throw new LoginException( LMN + " Error: " +
				"no CallbackHandler available");

		String host = (String)options.get("host");
		if ((host == null) || (host.trim().length() == 0))
			throw new LoginException( LMN + " Error: " +
				"Required 'host' property missing");

		String port = (String)options.get("port");
		if ((port == null) || (port.trim().length() == 0))
			throw new LoginException( LMN + " Error: " +
				"Required 'port' property missing");

		try
		{
			int portInt = Integer.parseInt(port);
			if (portInt < 1)
				throw new LoginException( LMN + " Error: " +
					"'port' property must be an integer value greater than zero");
		}
		catch (NumberFormatException e)
		{
			throw new LoginException( LMN + " Error: " +
				"'port'property is not an integer value");
		}

		passwordValidationMethod =
			(String)options.get("passwordValidationMethod");

		if ((!"Bind".equalsIgnoreCase(passwordValidationMethod)) &&
			(!"Compare".equalsIgnoreCase(passwordValidationMethod)) &&
			(!"Cams".equalsIgnoreCase(passwordValidationMethod)))
			throw new LoginException( LMN + " Error: Required" +
				" Required option 'passwordValidationMethod' must be " +
				"'Cams', 'Bind', or 'Compare'");

		if (!passwordValidationMethod.equalsIgnoreCase("Bind"))
		{
			String adminUsername = (String)options.get("adminUsername");
			if ((adminUsername == null) || (adminUsername.trim().length() == 0))
				throw new LoginException(LMN + " Error: " +
					"Required 'adminUsername' property missing");

			String adminPassword = (String)options.get("adminPassword");
			if ((adminPassword == null) || (adminPassword.trim().length() == 0))
				throw new LoginException(LMN + " Error: " +
					"Required 'adminPassword' property missing");
		}

		String userDN = (String)options.get("userDN");
		if ((userDN == null) || "".equals(userDN.trim()))
			throw new LoginException( LMN + " Error: Required" +
				" 'userDN' property missing");

		String roleAttrName = (String)options.get("roleAttrName");
		if ((roleAttrName == null) || "".equals(roleAttrName.trim()))
			throw new LoginException(LMN + " Error: Required" +
				" 'roleAttrName' property missing");

	}

	/**
	 * Create a LDAPConnection.
	 *
	 * @return a LDAPConnection that is ready for use.
	 * @throws LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getConnection()
		throws LoginException
	{
		try
		{
			// Get the host and port
			String host = (String)options.get("host");
			String port = (String)options.get("port");

			if (debug)
				logger.debug(this, LMN +
					" Instantiating LDAPConnection: host=" + host +
					", port=" + port);

			// Create the connection
			LDAPConnection connection = new LDAPConnection();

			if (debug)
				logger.debug(this, LMN +
					" Connecting to host=" + host +
					", port=" + port);

			// Connect to the directory
			connection.connect(host, Integer.parseInt(port));

			if (debug)
				logger.debug(this, LMN +
					" Successfully connected to host=" + host +
					", port=" + port);

			if(!passwordValidationMethod.equalsIgnoreCase("Bind"))
			{
				if (debug)
					logger.debug(this, LMN + 
						" PasswordValidationMethod=" + passwordValidationMethod +
						", binding as admin user");

				String adminUsername = (String)options.get("adminUsername");
				String adminPassword = (String)options.get("adminPassword");
				connection.bind(3, adminUsername, adminPassword.getBytes());
			}

			if (debug)
				logger.debug(this, LMN +
					" Returning LDAPConnection");

			return connection;
		}
		catch (LDAPException e)
		{
			logger.error(this, "LDAP error", e);
			throw new LoginException( LMN + " Error: " +
				e.getMessage());
		}
	}

	/**
	 * Authenticate using Cams password validation.
	 *
	 * @param usernameDN the username's distinguished name.
	 * @param connection the LDAPConnection.
	 * @return true if authentication succeeded, false otherwise.
	 * @throws LoginException if error occurs authenticating.
	 */
	private boolean camsAuthenticate(
		String usernameDN, LDAPConnection connection) throws LoginException
	{
		try
		{
			// Read the "userPassword" attribute for this user. If the user does
			// not exist, a "NO_SUCH_OBJECT" Exception is thrown.
			String[] sa = {"userPassword"};
			LDAPEntry attrs = connection.read(usernameDN, sa);
			LDAPAttribute passwordAttr = attrs.getAttribute("userPassword");

			// The user must have a userPassword attribute or a "LoginException"
			// is thrown as this is a miss configuration, alternatively, this
			// could be treated as a FailedLoginException pending customer
			// feedback
			if (passwordAttr == null)
				throw new LoginException(LMN + " User '" + username +
					"' does not have a userPassword attribute in the repository");

			// There can be one or more userPassword attributes, put them in an
			// array. However, we are only going to test the first password.
			String[] attr = passwordAttr.getStringValueArray();

			// Trim the supplied and repository passwords
			String ldapPassword = attr[0].trim();
			String digestString = password.trim();

			// The repository password may be "digested" and "salted" by use
			// of a "Message Digest" algorithm, if so, then we'll need to
			// digest and salt the user-entered password before comparison
			DigestString ds = new DigestString(ldapPassword);

			// Is the password clear text (or does it have an unknown
			// algorithm)?
			if (ds.isDigestString())
			{
				// Get the digestString from the user-entered password using
				// the same ingredients found in the repository password;
				// the password, the digest algorithm, and possibly "salt"
				digestString = DigestString.createDigestString(password,
					ds.getAlgorithm(), ds.getSalt(), ds.getLabel());
			}

			// Log the digest string and queried password for debug comparison
			if (debug)
			{
				logger.debug(this, LMN +
					" Repository queried password for user: '" +
					ldapPassword + "'");
				logger.debug(this, LMN +
					" User entered password as digestString: '" +
					digestString + "'");
			}

			// Compare the cleartext or digested passwords
			return ldapPassword.equals(digestString);

		}
		catch (LDAPException e)
		{
			if (debug)
				logger.debug(this, LMN + " username '" +
					this.username + "' not found", e);

			// If error is NO_SUCH_OBJECT, then username was not found and we'll
			// throw a FailedLoginException later; otherwise, throw a
			// LoginException
			if (e.getResultCode() != e.NO_SUCH_OBJECT)
				throw new LoginException( LMN +
					" Error: " + e.getMessage() + "\n");
		}
		catch (NoSuchAlgorithmException e)
		{
			logger.error(this, "Unrecogized MessageDigest algorithm", e);
			throw new LoginException( LMN +
				" Error: " + e.getMessage());
		}

		// We fell through therefore return false.
		return false;
	}

	/**
	 * Authenticate using the LDAP compare method.
	 *
	 * @param usernameDN the username's distinguished name.
	 * @param connection the LDAPConnection.
	 * @return true if authentication succeeded, false otherwise.
	 * @throws LoginException if error occurs authenticating.
	 */
	private boolean compareAuthenticate(
		String usernameDN, LDAPConnection connection) throws LoginException
	{
		try
		{
			LDAPAttribute passwordAttr = new LDAPAttribute(
				"userPassword", password);
			return connection.compare(usernameDN, passwordAttr);
		}
		catch(LDAPException e)
		{
			throw new LoginException( LMN +
				" Error: " + e.getMessage());
		}
	}

	/**
	 * Authenticate using LDAP server bind.
	 *
	 * @param usernameDN the username's distinguished name.
	 * @param connection the LDAPConnection.
	 * @return true if authentication succeeded, false otherwise.
	 */
	private boolean bindAuthenticate(
		String usernameDN, LDAPConnection connection)
	{
		try
		{
			connection.bind(3, usernameDN, password.getBytes());
			return true;
		}
		catch(LDAPException e)
		{
			// Should we check for a particular code here ???
			return false;
		}
	}

	/**
	 * Return the input pattern replaced with the match value.
	 *
	 * @param pattern The input pattern to use
	 * @param matchName The variable name to match
	 * @param matchValue The value to replace
	 * @return The input pattern with matched value
	 */
	private String replaceUsername(String pattern, String matchName,
		String matchValue)
	{
		String tmp = null;
		StringBuffer sb = new StringBuffer();
		StringTokenizer st = new StringTokenizer(pattern, "{}");
		while (st.hasMoreTokens())
		{
			tmp = st.nextToken();

			if (tmp.trim().equals(matchName))
				sb.append(matchValue);
			else
				sb.append(tmp);
		}

		return sb.toString();
	}
} // End of class : SiteServerLoginModule
