/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package com.cafesoft.cams.auth.login.module;

import com.novell.ldap.LDAPSocketFactory;

import java.io.IOException;

import java.net.Socket;
import java.net.UnknownHostException;

/**
 * <code>TimeoutLDAPSocketFactory</code> implements an LDAPSocketFactory that
 * provides connection timeout functionality. This is useful for Cams
 * LDAP-based LoginModules when the default LDAPSocketFactory takes too long
 * to fail when attempting to connect with an LDAP server, causing
 * Cams agents to timeout while waiting for an authentication request.
 * When multiple LoginModules are stacked, it may be desireable to enable
 * authentication success based on the results from other LoginModules.
 * This class essentially implements a way to interrupt an attempted
 * connection to an LDAP server, which can take nearly 1 minute to fail in
 * cases where the LDAP server host is unavailable due to a network outage.
 * <p>
 * This class will delegate to the specified LDAPSocketFactory if one is
 * provided or will create a Socket directly if necessary. Because a new
 * Thread is started to wait for the actual connection timeout, it may
 * continue running well after the enclosing LoginModule has exited. The
 * Thread will eventually stop and cleanup allocated resources.
 * <p>
 * The following example code shows how this class may be used:
 * <pre>
 * import com.cafesoft.cams.auth.login.module.TimeoutLDAPSocketFactory
 * ...
 *    // Create an LDAPConnection to see if default LDAPSocketFactory is set
 *    LDAPConnection connection = new LDAPConnection();
 *
 *    // Get the default LDAPSocketFactory (this may be null)
 *    LDAPSocketFactory defaultSocketFactory = connection.getSocketFactory();
 *
 *    // Create an LDAPSocketFactory that will timeout on connection
 *    // (The timeout is set to 500 milliseconds for this example)
 *    LDAPSocketFactory timeoutSocketFactory =
 *       new TimeoutLDAPSocketFactory(defaultSocketFactory, 500L);
 *
 *    // Now create a new LDAPConnection that uses the new factory
 *    connection = null;
 *    connection = new LDAPConnection(timeoutSocketFactory);
 *
 *    // Attempt to connect with the LDAP server
 *    connection.connect();
 * </pre>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.5 $ $Date: 2009/08/14 23:49:52 $ $Author: norb $
 * @see com.novell.ldap.LDAPSocketFactory
 * @since 02/05/04
 */
public class TimeoutLDAPSocketFactory
	implements LDAPSocketFactory, Runnable
{
	// The actual factory to which Socket creation will be delegated
	private LDAPSocketFactory factory = null;

	// The number of milliseconds to wait for connection before timing out
	private long timeout = 1000L;

	// The host to which this SocketFactory will attempt connection
	private String host = null;

	// The port to which this SocketFactory will attempt connection
	private int port = 383;

	// An IOException populated if unable to create the Socket
	private IOException socketException = null;

	// The Socket to be created
	private Socket socket = null;

	// A flag indicating whether connection has timed out
	private boolean hasTimedOut = false;

	/**
	 * Create a new TimeoutLDAPSocketFactory.
	 *
	 * @param factory the real LDAPSocketFactory to which socket creation
	 * 	will be delegated. If null, then a standard Socket will be created.
	 * @param timeout the number of milliseconds to wait for connection
	 * 	before timing out.
	 */
	public TimeoutLDAPSocketFactory(LDAPSocketFactory factory, long timeout)
	{
		this.factory = factory;
		this.timeout = timeout;
	}

	//
	// implementing LDAPSocketFactory
	//

	/**
	 * Create a new Socket to an LDAP server.
	 *
	 * @param host the host name or IP address
	 * @param port the TCP port
	 * @return the Socket
	 * @exception IOException if unable to connect to the server at the
	 * 	specified host/port within the configured timeout.
	 * @exception UnknownHostException if the host is unknown.
	 */
	public Socket createSocket(String host, int port)
		throws IOException, UnknownHostException
	{
		this.host = host;
		this.port = port;

		// Create and run a new Thread with this Object as the Runnable
		Thread t = new Thread(this);
		t.start();

		try
		{
			// Wait for the Thread to join this one for the specified timeout
			t.join(timeout);
		}
		catch (InterruptedException e)
		{
			// The Thread did not complete before the timeout.
			// Interrupt it.
			t.interrupt();
		}

		// If an error occurred creating the Socket, then throw the error
		if (socketException != null)
			throw socketException;

		// If the Socket is null, then the connect has not completed
		if (socket == null)
		{
			this.hasTimedOut = true;
			throw new IOException("Socket connection timed out: " +
				host + ":" + port + ", timeout=" + timeout + " (ms)");
		}

		// The Socket has been successfully created and is connected
		return socket;
	}

	//
	// implementing Runnable
	//

	/**
	 * Attempt to connect with the LDAP server
	 */
	public void run()
	{
		try
		{
			// If the LDAPSocketFactory is null, create a standard Socket
			if (factory == null)
			{
				// The LDAPSocketFactory is null, create a standard Socket
				this.socket = new Socket(host, port);
			}
			else
			{
				// Use the LDAPSocketFactory to create a Socket
				this.socket = factory.createSocket(host, port);
			}
		}
		catch (IOException e)
		{
			// Save the Exception in case it needs to be thrown by the
			// factory method
			this.socketException = e;
		}

		if (hasTimedOut)
		{
			// We've timed out waiting to connect

			try
			{
				// If the Socket was created, then close it.
				if (this.socket != null)
					this.socket.close();
			}
			catch (IOException e)
			{
				// Ignore the Exception: We just don't care
			}

			this.socket = null;
			this.socketException = null;
		}

		return;
	}

} // End of class : TimeoutLDAPSocketFactory
