/**
 * Copyright (c) 1996-2011 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.cams.service.ActiveDirectoryGroupNameService;
import com.cafesoft.cams.service.UnboundIDLdapConnectionPoolService;

import com.cafesoft.core.lifecycle.LifecycleException;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.service.ServiceClient;
import com.cafesoft.core.service.ServiceException;
import com.cafesoft.core.service.ServiceFinder;

import com.unboundid.ldap.sdk.Attribute;
import com.unboundid.ldap.sdk.DN;
import com.unboundid.ldap.sdk.Entry;
import com.unboundid.ldap.sdk.FailoverServerSet;
import com.unboundid.ldap.sdk.Filter;
import com.unboundid.ldap.sdk.LDAPConnection;
import com.unboundid.ldap.sdk.LDAPConnectionOptions;
import com.unboundid.ldap.sdk.LDAPException;
import com.unboundid.ldap.sdk.LDAPSearchException;
import com.unboundid.ldap.sdk.RDN;
import com.unboundid.ldap.sdk.ResultCode;
import com.unboundid.ldap.sdk.RoundRobinServerSet;
import com.unboundid.ldap.sdk.SearchRequest;
import com.unboundid.ldap.sdk.SearchResult;
import com.unboundid.ldap.sdk.SearchScope;
import com.unboundid.ldap.sdk.SingleServerSet;
import com.unboundid.util.ssl.SSLUtil;
import com.unboundid.util.ssl.TrustAllTrustManager;

import javax.net.ssl.SSLSocketFactory;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;

import java.security.Principal;
import java.security.GeneralSecurityException;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.StringTokenizer;

/**
 * Implements the LoginModule interface found in JAAS to authenticate users
 * against an Active Directory repository. This implementation attempts to
 * bind the connection with the user supplied credentials. If successful,
 * the user roles are retrieved and the user is authenticated.
 *
 * All configuration parameters are listed below:
 * <ul>
 * <li><b>ldapPoolServiceId</b> - (optional) The ID of a configured 
 * UnboundIDLdapConnectionPoolService in this security domain. When this 
 * option is configured, any LDAP connection parameters configured for this 
 * login module are ignored.
 * 
 * <li><b>connectionHost</b> - A comma-delimited list of DNS names or IP 
 * addresses for connection LDAP servers. See <strong>serverSetType</strong> for 
 * configuration options to use more than one host.</li>
 * 
 * <li><b>connectionPort</b> - A comma-delimited list of one or more ports for 
 * LDAP server connections. Specify a port of the same type (LDAP or LDAPS) for
 * each host in the <strong>connectionHost</strong> list. Default values are 389 
 * for LDAP and 636 for LDAPS.</li>
 * 
 * <li><b>useSSL</b> - (optional) Enable/disable SSL [true|false]. 
 * 
 * <li><b>serverSetType</b> - (optional) Can be: 
 * <b>SINGLE</b> - (default) Connects to a single server. 
 * <b>FAILOVER</b> - Attempts to establish connections to servers in the order 
 * they are provided. If the first server is unavailable, then it will attempt 
 * to connect to the second, then to the third, etc. 
 * <b>ROUND_ROBIN</b> - Uses a round-robin algorithm to select the server to 
 * which the connection should be established. Any number of servers may be 
 * included in this server set, and each request will attempt to retrieve a 
 * connection to the next server in the list, circling back to the beginning of 
 * the list as necessary. If a server is unavailable when an attempt is made to 
 * establish a connection to it, then the connection will be established to the 
 * next available server in the set. 
 * 
 * WARNING: Except for the connectionHost and connectionPort parameters, all
 * other parameters for this login module are shared and must be the same for
 * each server when more than one are configured in a ServerSet.</li>
 * 
 * <li><b>connectionTimeout</b> - (optional) The number of milliseconds to wait 
 * for a connection to the LDAP server before failing (default=3000)</li>
 * 
 * <li><b>responseTimeout</b> - (optional) The number of milliseconds to wait 
 * for a response from the server before failing (default=3000)</li> 
 * 
 * <li><b>emptyPassword</b> - (optional) Enable empty passwords (true or false).
 * By default, the UnboundIDActiveDirectoryLoginModule does not allow a password
 * to be empty. Set this value to true to allow use of empty passwords
 * when Active Directory is configured for and anonymous binds are
 * required.</li>
 * 
 * <li><b>defaultRoles</b> - (optional) A list of one or more default roles
 * separated by a comma to be added to the user session.</li>
 * 
 * <li><b>roleRegexPattern</b> - (optional) Limits roles saved in the user 
 * session to those that match a regular expression pattern. This option is 
 * useful when the role search returns many roles that are not used by the 
 * access control policy.
 * 
 * <li><b>domain</b> - (optional) The Active Directory domain to use. If this
 * value is present and the user name does not end with @domain, then append
 * "@" and the value of domain to the user name before attempting to bind.</li>
 * 
 * <li><b>useRoleSearch</b> - (optional) A value of "true" (default) 
 * indicates that a role search will be attempted [true|false].</li>
 * 
 * <li><b>useDomainInRoleSearch</b> - (optional) A value of "true" (default) 
 * indicates that an optionally supplied "domain" used for authentication
 * will be included in the role search filter's {username} substitution value 
 * [true|false].</li>
 * 
 * <li><b>baseDN</b> - (conditional) The base distinguished name for the user entries.
 * 
 * <li><b>scope</b> - (conditional) The scope for the role search. Valid values
 *   are "BASE", "ONE", and "SUB".</li>
 * 
 * <li><b>filter</b> - (conditional) The role search filter.</li>
 * 
 * <li><b>attribute</b> - (conditional) The attribute where the roles
 * are listed. Value should be either "memberOf" or "tokenGroups". If "tokenGroups" 
 * is used, then an Active Directory group name service must be configured
 * to lookup the group names by their object SIDs.</li>
 * 
 * <li><b>groupNameServiceId</b> - (conditional) The ID of a configured 
 * Active Directory group name service in this security domain. This option is
 * only valid if the "attribute" options is configured to "tokenGroups". 
 * 
 * <li><b>addPrivateCredentials</b> - (optional) A flag that indicates whether
 * or not a java.net.PasswordAuthentication object containing the user name and
 * password should be added to the Subject's private credential set on
 * successful authentication. 
 *
 * WARNING: The values in this object can be accessed by other Cams components 
 * to enable SSO to other other systems, so administrator must take care that only 
 * trusted components have are used!</li>
 * </ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.9 $ $Date: 2012/03/09 00:34:16 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 10/20/01
 */
public class UnboundIDActiveDirectoryLoginModule
	implements LoginModule, LoggerClient, ServiceClient
{
	//
	// Class variables.
	//

	/**
	 * The LoginModule name (used in log messages throughout the code).
	 */
	private static final String LMN = "[UnboundIDActiveDirectoryLoginModule]";
	
	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag indicating that a role search will be tried.
	 */
	private boolean useRoleSearch;

	/**
	 * Flag indicating that a domain value, if supplied, will be used in the 
	 * user name substitution value for a role search.
	 */
	private boolean useDomainInRoleSearch;
	
	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating user.
	 */
	private String username;

	/**
	 * The fully qualified bind username for the authenticating user .
	 */
	private String bindUsername;

	/**
	 * The password entered by the authenticating User.
	 */
	private String password;

	/**
	 * A default list of user roles to associate with an authenticated user.
	 */
	private List roleList;

	/**
	 * Used to lookup required Cams services.
	 */
	private ServiceFinder serviceFinder;

	/**
	 * Connection pool.
	 */
	private UnboundIDLdapConnectionPoolService lcps = null; 

	/**
	 * A flag that enables/disables addition of credentials to the Subject
	 * private credential store.
	 */
	private boolean addPrivateCredentials;

	//
	// Constructor
	//

	/**
	 * Default UnboundIDActiveDirectoryLoginModule Constructor.
	 */
	public UnboundIDActiveDirectoryLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.useRoleSearch = true;
		this.useDomainInRoleSearch = true;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.password = null;
		this.subject = null;
		this.username = null;
		this.roleList = new ArrayList();
		this.addPrivateCredentials = false;
	}

	//
	// Implements the ServiceClient interface
	//

	/**
	 * Set the ServiceFinder.
	 *
	 * @param finder the class used by find Services by the ServiceClient.
	 */
	public void setServiceFinder(ServiceFinder finder)
	{
		this.serviceFinder = finder;
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *			with the end user (prompting for usernames and
	 *			passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.debug = "true".equalsIgnoreCase((String)options.get("debug"));
		this.useRoleSearch = 
			! "false".equalsIgnoreCase((String)options.get("useRoleSearch"));
		this.useDomainInRoleSearch = 
			! "false".equalsIgnoreCase((String)options.get("useDomainInRoleSearch"));
		this.addPrivateCredentials = "true".equals(
			(String)options.get("addPrivateCredentials"));
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt  succeeded (checked
	 * by retrieving the private state saved by the <code>login</code> method),
	 * then this method cleans up any state that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 * <p>
	 * This method executes the following steps:
	 * <ol>
	 * <li>Checks preconditions
	 * <li>Handles callback objects
	 * <li>Get a connection to the Active Directory
	 * <li>Authenticates the users using bind (only method available
	 * with Active Directory)
	 * <li>Close the connection
	 * <li>Reset state if authentication failed
	 * <li>Return results of authentication attempt
	 * </ol>
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @exception LoginException thrown if any error occurs while trying to
	 * authenticate the user.
	 * @exception FailedLoginException if authentication fails.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException,
		AccountExpiredException, CredentialExpiredException
	{
		// Check preconditions
		validateOptions();

		// Handle callback objects.
		handleCallbacks();

		// A user name must always be supplied
		if (username == null || username.length() == 0)
			throw new FailedLoginException(LMN + " Authentication failed");

		// Optionally append the domain to the user name if supplied. Do
		// not modify username as it is required "as is" for the role filter.
		String domain = (String)options.get("domain");
		bindUsername = username;
		if ((domain != null) && ! ("".equals(domain.trim())))
			bindUsername = username + "@" + domain;

		if (this.debug)
			logger.debug(this, LMN + 
				" User name to bind: '" + bindUsername + "'");

		// A non-empty password must be supplied unless empty
		// passwords are permitted for this configuration.
		String emptyPassword = (String)options.get("emptyPassword");
		if ((emptyPassword == null) || ("false".equals(emptyPassword.trim())))
		{
			if ("".equals(this.password))
				throw new FailedLoginException(
					LMN + " Authentication failed");
		}
		else if (!("true".equals(emptyPassword.trim())))
		{
			throw new LoginException(LMN +
				" Error: Option 'emptyPassword' must be 'true' or 'false'");
		}

		// Get a connection to Active Directory
		LDAPConnection connection = null;

		if (((options.containsKey("ldapPoolServiceId"))) && 
		    ((options.get("ldapPoolServiceId") != null)))
		{
			if (this.debug)
				logger.debug(this, LMN + " Getting connection from pool");

			connection = getPooledConnection();
		}
		else
		{
			if (this.debug)
				logger.debug(this, LMN + " Creating new connection");

			connection = getConnection();
		}

		if (this.debug)
		{
			logger.debug(this, LMN + " Connection is connected: '" +
				connection.isConnected() + "'");
			logger.debug(this, LMN + " Connection statistics: " + 
				connection.getConnectionStatistics());
		}

		// Bind to Active Directory with the user supplied credentials
		try
		{
			if (this.debug)
				logger.debug(this, LMN + " Attempting bind for user: '" + 
					bindUsername + "'");

			// Authenticate the user by attempting to bind and get the user roles.
			// Bind the connection to Active Directory. If successful, the user
			// credentials are valid.
			connection.bind(bindUsername, password);

			if (this.debug)
				logger.debug(this, LMN + " Bind successful");

			// Authentication has succeeded
			this.loginSucceeded = true;
		}
		catch (LDAPException e)
		{
			activeDirectoryLDAPExceptionHandler(e);
		}
		finally
		{
			// Close the connection if there was an exception
			if (!loginSucceeded)
			{
				if (connection != null)
				{
					if (((options.containsKey("ldapPoolServiceId"))) && 
					((options.get("ldapPoolServiceId") != null)))
					{
						try
						{
							lcps.releaseConnection(connection);
						}
						catch(ServiceException e)
						{
							if (this.debug)
								logger.debug(this, LMN + " Service error while returning connection to pool");

							throw new LoginException(e.getMessage());
						}
						catch(LifecycleException e)
						{
							if (this.debug)
								logger.debug(this, LMN + " Lifecyle error while returning connection to pool");

							throw new LoginException(e.getMessage());
						}
					}
					else
					{
						connection.close();
						connection = null;
					}
				}

			}
		}

		if (this.debug)
			logger.debug(this, LMN + " Successfully bound user: '" +
				bindUsername + "'");

		// Add default roles (if any) to list of roles
		String defaultRoles = (String)options.get("defaultRoles");
		if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
		{
			StringTokenizer st = new StringTokenizer(defaultRoles, ",");
			while(st.hasMoreTokens())
				roleList.add(st.nextElement().toString().trim());
		}

		// Associate roles from groups that the user is a member of

		// Check the required options when using a role search
		if (this.useRoleSearch)
		{
			String baseDN = (String)options.get("baseDN");
			String filter = (String)options.get("filter");
			String scope = (String)options.get("scope");

			// Set the search scope, the default of 0 is "BASE"
			int scopeInt = 0;
			if ("ONE".equalsIgnoreCase(scope))
				scopeInt = 1;
			else if ("SUB".equalsIgnoreCase(scope))
				scopeInt = 2;

			try
			{
				// Get and validate the attribute where roles are stored (the
				// "memberOf" groups attribute is used, but enable external 
				// configuration to supply an alternative attribute.

				// Specify the attributes that the search will return
				String[] attrs = {(String)options.get("attribute")};

				// Substitute the user name in the role search filter. Optionally,
				// don't use the supplied domain in the user name to enable searches
				// using attributes other than the userPrincipalName, such as 
				// sAMAccountName, uid, or cn, which don't use the fully-
				// qualified UPN.
				if (useDomainInRoleSearch)
					filter = replaceUsername(filter, "username", bindUsername);
				else
					filter = replaceUsername(filter, "username", username);

				// If supplied, only use roles that match roleRegexPattern
				String roleRegexPattern = (String)options.get("roleRegexPattern");
				
				// The "tokenGroups" attribute contains all groups of which a
				// a user is member (including primary and nested groups). However,
				// it is only returned if you search with the user's
				// distinguishedName. So, we first search to get the user's
				// distinguishedName, then we do a read to get the "tokenGroups".
				// Because the "tokenGroups" are returned as SIDs, we then
				// use a Cams service to efficiently obtain their names.
				if ("tokenGroups".equalsIgnoreCase(attrs[0]))
				{
					// Get the user's distinguishedName
					attrs[0] = "distinguishedName";

					if (this.debug)
					{
						logger.debug(this, LMN + 
							" Search for the user's 'distinguishedName' using:" +
							  "\n\tbaseDN: " + baseDN +
							  "\n\tfilter: " + filter +
							  "\n\t attrs: " + attrs[0]);
					}

					SearchRequest searchRequest = 
						new SearchRequest(baseDN, SearchScope.valueOf(scopeInt), 
						Filter.create(filter), attrs);

					// Search for the user entry attributes
					SearchResult searchResult = connection.search(searchRequest);

					// Process the search results to get the user dn
					if (searchResult != null)
					{
						List searchResultList = searchResult.getSearchEntries();
						Iterator searchResultIterator = searchResultList.iterator();
						if (searchResultIterator.hasNext())
						{
							// Get the next entries attibute
							Entry entry = (Entry)searchResultIterator.next();
							Collection attrCollection = entry.getAttributes();
							Iterator attrIterator = attrCollection.iterator();
							if (attrIterator.hasNext())
							{
								// Attribute may have more than one name/value pair
								Attribute attr = (Attribute) attrIterator.next();
								if (attr != null)
									baseDN = attr.getValue();
							}
						}
					}
					else
					{
						throw new LoginException ("Attribute 'distinguishedName' " +
							"not found for user '" + bindUsername + "'");
					}

					// Now that we have the user's dn, looking for tokenGroups
					attrs[0] = "tokenGroups";

					if (this.debug)
						logger.debug(this, LMN +
							" 'tokenGroups' attribute(s) for user '" + baseDN + "'");

					// Read the user's tokenGroups values
					Entry entry = connection.getEntry(baseDN, attrs);

					// Save the Active Directory groups as roles for commit method
					if (entry != null)
					{
						Attribute attribute = entry.getAttribute(attrs[0]);

						// Get an enumeration of the attribute values
						byte[][] allValues = attribute.getValueByteArrays();

						String role = null;
						String sid = null;

						if ((roleRegexPattern != null) && !("".equals(roleRegexPattern)))
						{
							Pattern pattern = Pattern.compile(roleRegexPattern);
							for (int i = 0; i < allValues.length; i++)
							{
								// Get the sid
								sid = Sid2String(allValues[i]);

								if (this.debug)
									logger.debug(this, LMN +
										" Getting cn for objectSid: '" + sid + "'");

								role = getGroupName(sid);

								// Role not found, use sid for role name and log error
								// GEG: Consider providing an externalized list mapping 
								// objectSid to cn if unresolved objectSid is common, which
								// is not expected to be the case.
								if (role == null)
								{
									role = sid;

									logger.error(this, LMN + 
										" Active Directory group name service returned 'null' cn for objectSid: '" + sid + "'");
								}
								else
								{
									if (debug)
										logger.debug(this, LMN + " SID role name: '" + role + "'");
								}

								if (pattern.matcher(role).matches())
								{
									roleList.add(role);

									if (this.debug)
										logger.debug(this, LMN +
											" Role: '" + role + 
											"' matched roleRegexPattern: '" + 
											roleRegexPattern + "'");
								}
								else
								{
									logger.debug(this, LMN + 
										" Role: '" + role + 
										"' did not match roleRegexPattern: '" + 
										roleRegexPattern + "'");
								}
							}
						}
						else
						{
							for (int i = 0; i < allValues.length; i++)
							{
								// Get the sid
								sid = Sid2String(allValues[i]);

								if (this.debug)
									logger.debug(this, LMN +
										" Getting cn for objectSid: '" + sid + "'");

								role = getGroupName(sid);

								// Role not found, use sid for role name and log error
								// GEG: Provide an externalized list of objectSid that
								// will not be used for role lookup if this is found to 
								// be a significant issue for customers.
								if (role == null)
								{
									role = sid;

									logger.error(this, LMN + 
										" Active Directory group name service returned 'null' cn for objectSid: '" + sid + "'");
								}
								else
								{
									if (debug)
										logger.debug(this, LMN + " SID role name: '" + role + "'");
								}

								this.roleList.add(role);
							}
						}
					}
				}

				// The role attribute is "memberOf", so we need to get the
				// get string values and explode them to get the group names
				else if ("memberOf".equalsIgnoreCase(attrs[0]))
				{
					if (this.debug)
					{
						logger.debug(this, LMN + 
							" Search for user's groups with:" +
							  "\n\tbaseDN: " + baseDN +
							  "\n\tfilter: " + filter +
							  "\n\t attrs: " + attrs[0]);
					}					

					// Search for the user entry attributes
					SearchResult searchResults = connection.search(
						baseDN, SearchScope.valueOf(scopeInt), filter, attrs);

					// Get the roles and save them for commits
					if (searchResults != null)
					{
						List searchResultsList = searchResults.getSearchEntries();
						Iterator searchResultsIterator = searchResultsList.iterator();
						while (searchResultsIterator.hasNext())
						{
							// Get the next entries attibute
							Entry entry = (Entry)searchResultsIterator.next();
							Collection attrCollection = entry.getAttributes();
							Iterator attrIterator = attrCollection.iterator();

							while (attrIterator.hasNext())
							{
								if ((roleRegexPattern != null) && !("".equals(roleRegexPattern)))
								{
									Pattern pattern = Pattern.compile(roleRegexPattern);

									// Attribute may have more than one name/value pair
									Attribute attr = (Attribute) attrIterator.next();
									String[] values = attr.getValues();
									String[] roles;
									for (int i = 0; i < values.length; i++)
									{
										// Parse the role name from the DN 
										RDN rdn = new RDN(DN.getRDNString(values[i]));
										roles = rdn.getAttributeValues();

										if (this.debug)
											logger.debug(this, LMN +
												" Parsed role '" + roles[0] + "' from dn '" +
												values[i] + "'");

										if (pattern.matcher(roles[0]).matches())
										{
											this.roleList.add(roles[0]);

											if (debug)
												logger.debug(this, LMN + 
													" Role: '" + roles[0] + 
													"' matched roleRegexPattern: '" + 
													roleRegexPattern + "'");
										}
										else
										{
											if (debug)
												logger.debug(this, LMN + 
													" Role: '" + roles[0] + 
													"' did not match roleRegexPattern: '" + 
													roleRegexPattern + "'");
										}
									}
								}
								else
								{
									// Attribute may have more than one name/value pair
									Attribute attr = (Attribute) attrIterator.next();
									String[] values = attr.getValues();
									String[] roles;
									for (int i = 0; i < values.length; i++)
									{
										// Parse the role name from the DN 
										RDN rdn = new RDN(DN.getRDNString(values[i]));
										roles = rdn.getAttributeValues();

										if (this.debug)
											logger.debug(this, LMN +
												" Parsed role: '" + roles[0] + "' from dn '" +
												values[i] + "'");

										this.roleList.add(roles[0]);
									}
								}
							}
						}
					}
					else
					{
						logger.debug(this, LMN +
						" No search results returned");
					}
				}
				else
				{
					logger.error(this, LMN +
						" Unsupported role search attribute, must be " +
						"'memberOf' or 'tokenGroups'");
				}
			}
			catch (LDAPSearchException e)
			{
				if (this.debug)
					logger.debug(this, LMN + " Error while " +
						"searching for groups for user: '" +
						this.username + "'", e);

				// Search filter failed
				throw new LoginException(LMN +
					" " + e.getExceptionMessage());
			}
			catch (LDAPException e)
			{
				if (this.debug)
					logger.debug(this, LMN + " Error while " +
						"searching for groups for user: '" +
						this.username + "'", e);

				// Search filter failed
				throw new LoginException(LMN +
					" " + e.getExceptionMessage());
			}
			finally
			{
				// Close the connection
				if (connection != null)
				{
					if (((options.containsKey("ldapPoolServiceId"))) && 
					    ((options.get("ldapPoolServiceId") != null)))
					{
						try
						{
							this.lcps.releaseConnection(connection);
						}
						catch(ServiceException e)
						{
							if (this.debug)
								logger.debug(this, LMN + " Service error while returning connection to pool");

							throw new LoginException(e.getMessage());
						}
						catch(LifecycleException e)
						{
							if (this.debug)
								logger.debug(this, LMN + " Lifecyle error while returning connection to pool");

							throw new LoginException(e.getMessage());
						}
					}
					else
					{
						connection.close();
						connection = null;
					}
				}
			}
		}

		// Return authentication results
		return this.loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @throws LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login failed, clean up state
		if (!this.loginSucceeded)
			return false;

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList();

		// Repository operations to get default role principals
		Iterator roleListItr = roleList.iterator();
		while(roleListItr.hasNext())
		{
			String rolename = (String)roleListItr.next().toString().trim();
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				principalList.add(rolePrincipal);

				if (this.debug)
					logger.debug(this, LMN +
						" Added CSRolePrincipal " +
						rolename + " to principal ArrayList for user '" +
						username + "'");
			}
			else
			{
				if (this.debug)
					logger.debug(this, LMN +
						" CSRolePrincipal '" +
						rolename + "' already in Subject for '" +
						username + "', not added to principal List");
			}
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if (this.debug)
				logger.debug(this, LMN +
					" Added CSUserPrincipal '" + username +
					"' to principal List");
		}
		else if (this.debug)
		{
			logger.debug(this, LMN +
				" CSUserPrincipal '" + username +
				"' already in Subject, not added to principal List");
		}

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		if (this.debug)
		{
			logger.debug(this, LMN + " Adding " +
				AuthenticationMethod.PASSWORD.getName() +
				" authentication method");
		}

		subject.getPublicCredentials().add(AuthenticationMethod.PASSWORD);

		if (addPrivateCredentials)
		{
			if (this.debug)
				logger.debug(this, LMN + " Adding 'java.net.PasswordAuthentication' " +
					"instance to private credentials");

			subject.getPrivateCredentials().add(
				new java.net.PasswordAuthentication(
					username, password.toCharArray()));
		}
		else
		{
			if (this.debug)
				logger.debug(this, LMN + " Did not add " +
					" 'java.net.PasswordAuthentication' instance" +
					" to private credentials");
		}

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.password = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @throws LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!this.loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.
			this.loginSucceeded = false;
			this.username = null;
			this.password = null;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @throws LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if(this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if (this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.addPrivateCredentials = false;
		this.bindUsername = null;
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.logger = null;
		this.loginSucceeded = false;
		this.options = null;
		this.password = null;
		this.roleList = null;
		this.serviceFinder = null;
		this.useDomainInRoleSearch = false;
		this.useRoleSearch = false;
		this.username = null;

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username and
	 * password instance variables with their appropriate value.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// This LoginModule requires a username and a password
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Get the callback values and clear the password.
			callbackHandler.handle(callbacks);
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");
			this.password = ((MapCallback)callbacks[0]).getCallbackValue("password");
			((MapCallback)callbacks[0]).setCallbackValue("password", null);
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available",
				e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (this.debug)
		{
			logger.debug(this, LMN + " User entered username = " + username);
		}
	}

	/**
	 * Validate the LoginModules options Map object to ensure that all
	 * required properties exist and contain valid values.
	 *
	 * @exception LoginException if any of the following conditions occur:
	 * <ul>
	 * <li>A CallbackHandler object is not available</li>
	 * <li>the 'connectionHost' property is not available (if connection pool is not used)</li>
	 * <li>the 'connectionPort' property is not available or a valid integer (if connection pool is not used)</li>
	 * <li>the 'connectionTimeout' property is not a valid integer (if connection pool is not used)</li>
	 * <li>the 'responseTimeout' property is not a valid integer (if connection pool is not used)</li>
	 * <li>the 'userDN' property is not available</li>
	 * <li>the 'baseDN' property is not available</li>
	 * <li>the 'attribute' property is not available</li>
	 * <li>the 'filter' property is not available</li>
	 * <li>the 'scope' property is not available or valid</li>
	 * </ul>
	 */
	private void validateOptions() throws LoginException
	{
		if (callbackHandler == null)
			throw new LoginException(LMN + " Error: no CallbackHandler available");

		String s = null;

		// Only validate connection options if not using a connection
		// pool service
		if (((!options.containsKey("ldapPoolServiceId"))) || 
		    ((options.get("ldapPoolServiceId") == null)))
		{
			s = (String)options.get("connectionHost");
			if ((s == null) || (s.trim().length() == 0))
				throw new LoginException(LMN +
					" Error: Required 'connectionHost' property missing");

			s = (String)options.get("connectionPort");
			if ((s == null) || (s.trim().length() == 0))
				throw new LoginException(LMN + 
					" Error: Required 'connectionPort' property missing");

			s = (String)options.get("connectionTimeout");
			if ((s != null) && (s.trim().length() > 0))
			{
				try
				{
					if (Integer.parseInt(s) < 1)
						throw new LoginException(LMN + " Error: " +
						"'connectionTimeout' property must be greater than zero");
				}
				catch (NumberFormatException e)
				{
					throw new LoginException(LMN + " Error: " +
						"'connectionTimeout' property is not an integer");
				}
			}

			s = (String)options.get("responseTimeout");
			if ((s != null) && (s.trim().length() > 0))
			{
				try
				{
					if (Integer.parseInt(s) < 1)
						throw new LoginException(LMN + " Error: " +
						"'responseTimeout' property must be greater than zero");
				}
				catch (NumberFormatException e)
				{
					throw new LoginException(LMN + " Error: " +
						"'responseTimeout' property is not an integer");
				}
			}
		}

		if (this.useRoleSearch)
		{
			s = (String)options.get("baseDN");
			if ((s == null) || "".equals(s.trim()))
				throw new LoginException(LMN +
					" Error: Missing required option 'baseDN'");

			s = (String)options.get("filter");
			if ((s == null) || "".equals(s.trim()))
				throw new LoginException(LMN +
					" Error: Missing required option 'filter'");

			s = (String)options.get("scope");
			if ((!"BASE".equalsIgnoreCase(s)) &&
				(!"ONE".equalsIgnoreCase(s)) &&
				(!"SUB".equalsIgnoreCase(s)))
				throw new LoginException(LMN +
					" Error: Required option 'scope' must be 'BASE', 'ONE', or 'SUB'");

			s = (String)options.get("attribute");
			if ((s == null) || "".equals(s.trim()))
				throw new LoginException(LMN +
					" Error: Missing required option 'attribute'");
		}

	}

	/**
	 * Get an LDAP server connection.
	 *
	 * @return a LDAPConnection.
	 * @exception LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getConnection()
		throws LoginException
	{
		LDAPConnection connection = null;

		// Get the connection values to use
		String[] connectionHost = getStringArray((String)options.get("connectionHost"), ",");
		int[] connectionPort = getIntArray((String)options.get("connectionPort"), ",");
		boolean useSSL = "true".equals((String)options.get("useSSL"));
		String serverSetType = (String)options.get("serverSetType");
		int connectionTimeout = Integer.parseInt((String)options.get("connectionTimeout"));
		long responseTimeout = (long)Integer.parseInt((String)options.get("responseTimeout"));

		try
		{
			if (this.debug)
				logger.debug(this, LMN + " Instantiating LDAP connection");

			// Initialize the connection
			SSLSocketFactory sslSocketFactory = null;

			// Set the connection timeout
			LDAPConnectionOptions connectionOptions = new LDAPConnectionOptions();
			connectionOptions.setConnectTimeoutMillis(connectionTimeout);
			connectionOptions.setResponseTimeoutMillis(responseTimeout);

			if (this.debug)
				logger.debug(this, LMN + " LDAP connection options: " +
					connectionOptions.toString());

			if (useSSL)
			{
				SSLUtil sslUtil = new SSLUtil(new TrustAllTrustManager());
				sslSocketFactory = sslUtil.createSSLSocketFactory();
			}

			long timer = System.currentTimeMillis(); 

			// Use "SINGLE" ServerSet if configuration parameter is not 
			// supplied to support legacy integrations
			if (serverSetType == null)
				serverSetType = "SINGLE";

			if (serverSetType.equalsIgnoreCase("FAILOVER"))
			{
				if (this.debug)
					logger.debug(this, LMN + " Using 'FAILOVER' ServerSet for LDAP connection"); 

				FailoverServerSet failoverServerSet = 
					new FailoverServerSet(connectionHost, 
					connectionPort, sslSocketFactory, 
					connectionOptions);

				connection = failoverServerSet.getConnection();
			}
			else if (serverSetType.equalsIgnoreCase("ROUND_ROBIN"))
			{
				if (this.debug)
					logger.debug(this, LMN + " Using 'ROUND_ROBIN' ServerSet for LDAP connection"); 

				RoundRobinServerSet roundRobinServerSet = 
					new RoundRobinServerSet(connectionHost, 
					connectionPort, sslSocketFactory, 
					connectionOptions);

				connection = roundRobinServerSet.getConnection();
			}
			else // Fall through is "SINGLE"
			{
				if (this.debug)
					logger.debug(this, LMN + " Using 'SINGLE' ServerSet for LDAP connection"); 

				SingleServerSet singleServerSet = 
					new SingleServerSet(connectionHost[0], 
					connectionPort[0], sslSocketFactory,
					connectionOptions);

				connection = singleServerSet.getConnection();
			}

			if (this.debug)
			{
				logger.debug(this, LMN + " Time to connect: '" +
					(System.currentTimeMillis() - timer) + "'");

				if (connection.isConnected())
				{
					logger.debug(this, LMN + " LDAP connection succeeded to: '" + 
						connection.getConnectedAddress() + ":" +
						connection.getConnectedPort() + "'");
				}
				else
				{
					logger.debug(this, LMN + " LDAP connection failed");
				}

				logger.debug(this, LMN + " Returning LDAP connection");
			}

			return connection;
		}
		catch (GeneralSecurityException e)
		{
			logger.error(this, LMN + " General security error", e);

			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		catch (LDAPException e)
		{
			logger.error(this, LMN + "LDAP error", e);

			// Dump the nested Throwables to get at the root cause
			int i = 1;
			Throwable cause = e.getCause();
			do
			{
				if (cause != null)
					logger.error(this, LMN + "LDAP error: CAUSE " + i + ": " +
						cause.getMessage());
			}
			while ((cause = cause.getCause()) != null);

			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
	}

	/**
	 * Get an LDAP server connection from a connection pool.
	 *
	 * @return a LDAPConnection.
	 * @exception LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getPooledConnection()
		throws LoginException
	{
		// Get the connection values to use
		LDAPConnection connection = null;

		try
		{
			if (serviceFinder == null)
				throw new LoginException(LMN +
					" Error: ServiceFinder is null");

			String ldapPoolServiceId = (String)options.get("ldapPoolServiceId");
			if ((ldapPoolServiceId == null) || (ldapPoolServiceId.trim().length() == 0))
				throw new LoginException(LMN +
					" Error: Required property 'ldapPoolServiceId' missing");

			if (this.debug)
				logger.debug(this, LMN + " Using ldapPoolServiceId '" + 
					ldapPoolServiceId + "'");

			this.lcps = (UnboundIDLdapConnectionPoolService)
				serviceFinder.find(ldapPoolServiceId, 
				UnboundIDLdapConnectionPoolService.class);
			
			connection = lcps.getConnection();

			if (this.debug)
				logger.debug(this, LMN + " Got LDAP pool connection: " + 
				connection.getConnectionStatistics());
		}
		catch(ServiceException e)
		{
			logger.error(this, LMN + " Service error getting connection from pool", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		catch(LifecycleException e)
		{
			logger.error(this, LMN + " Lifecycle error getting connection from pool", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		catch(LDAPException e)
		{
			logger.error(this, LMN + " LDAP error getting connection from pool", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}

		return connection;
	}

	/**
	 * Translate a specific ActiveDirectory LDAPException into a specific
	 * LoginException and throw it.
	 * 
	 * @param e the LDAPException containing an ActiveDirectory specific message.
	 * @throws LoginException if not a more specific login failure.
	 * @throws FailedLoginException if invalid login credentials.
	 * @throws CredentialExpiredException if a login credential has expired.
	 * @throws AccountExpiredException if the user account has expired.
	 */
	private void activeDirectoryLDAPExceptionHandler(LDAPException e)
		throws LoginException, FailedLoginException,
			CredentialExpiredException, AccountExpiredException
	{
		if (this.debug)
			logger.debug(this, LMN + " Error while binding connection" +
				"\n\tMessage: " + e.getExceptionMessage());

		// Normal login failure due to invalid credentials, check
		// Active Directory server message for subcode to determine the
		// most accurate exception to throw. Possible codes include:
		// 52e - invalid credentials (FailedLoginException)
		// 525 - user not found (FailedLoginException)
		// 530 - invalid login time (FailedLoginException)
		// 532 - password expired (CredentialExpiredException)
		// 533 - account disabled (AccountExpiredException)
		// 701 - account expired (AccountExpiredException)
		// 773 - password reset required (CredentialExpiredException)
		// 775 - user account locked (AccountExpiredException)
		// 2030 - no such object (FailedLoginException)
		if (e.getResultCode().equals(ResultCode.INVALID_CREDENTIALS))
		{
			if (e.getExceptionMessage().indexOf("data 52e") > 0)
			{
				if(this.debug)
					logger.debug(this, LMN + " Failed credentials (data 52e)");

				throw new FailedLoginException("Failed credentials (data 52e");
			}
			if (e.getExceptionMessage().indexOf("data 532") > 0)
			{
				if(this.debug)
					logger.debug(this, LMN + " Password expired (data 532)");

				throw new CredentialExpiredException("Password expired (data 532)");
			}
			else if (e.getExceptionMessage().indexOf("data 533") > 0)
			{
				if(this.debug)
					logger.debug(this, LMN + " Account disabled (data 533)");

				throw new AccountExpiredException("Account disabled (data 533)");
			}
			else if (e.getExceptionMessage().indexOf("data 701") > 0)
			{
				if(this.debug)
					logger.debug(this, LMN + " Account expired (data 701)");

				throw new AccountExpiredException("Account expired (data 701)");
			}
			else if (e.getExceptionMessage().indexOf("data 773") > 0)
			{
				if(this.debug)
					logger.debug(this, LMN + " Password reset required (data 773)");

				throw new CredentialExpiredException("Password reset required (data 773)");
			}
			else
			{
				if(this.debug)
					logger.debug(this, LMN + " " + e);

				throw new FailedLoginException(e.getExceptionMessage());
			}
		}
		else
		{
			// Other reason
			logger.error(this, LMN + " Error while binding connection: " + e);

			throw new LoginException(e.getExceptionMessage());
		}
	}

	/**
	 * Return the input pattern replaced with the match value.
	 *
	 * @param pattern The input pattern to use
	 * @param matchName The variable name to match
	 * @param matchValue The value to replace
	 * @return The input pattern with matched value
	 */
	private String replaceUsername(String pattern, String matchName,
		String matchValue)
	{
		StringBuffer sb = new StringBuffer();
		StringTokenizer st = new StringTokenizer(pattern, "{}");
		while (st.hasMoreTokens())
		{
			String tmp = st.nextToken();
			if (tmp.trim().equals(matchName))
				sb.append(matchValue);
			else
				sb.append(tmp);
		}

		return sb.toString();
	}

	/**
	 * Gets the ActiveDirectoryGroupNameService for the LoginModule.
	 *
	 * @throws LoginException if an error occurs getting the
	 * ActiveDirectoryGroupNameService object.
	 */
	private String getGroupName(String sid)
		throws LoginException
	{
		String groupName = null;
		try
		{
			if (serviceFinder == null)
				throw new LoginException(LMN +
					" Error: ServiceFinder is null");

			String groupNameServiceId = (String)options.get("groupNameServiceId");
			if ((groupNameServiceId == null) || (groupNameServiceId.trim().length() == 0))
				throw new LoginException(LMN +
					" Error: Required property 'groupNameServiceId' missing");

			if (this.debug)
			{
				logger.debug(this, LMN + " Using groupNameServiceId '" + groupNameServiceId + "'");
				logger.debug(this, LMN + " Getting group id '" + sid + "'");
			}

			ActiveDirectoryGroupNameService adgns = (ActiveDirectoryGroupNameService)
				serviceFinder.find(groupNameServiceId, ActiveDirectoryGroupNameService.class);

			groupName = adgns.getGroupNameBySid(sid);

			if (this.debug)
				logger.debug(this, LMN + " Returned group name '" + groupName + "'");
		}
		catch(ServiceException e)
		{
			logger.error(this, "Error finding ActiveDirectoryGroupNameService", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}

		return groupName;
	}

	/**
	 * Convert the native byte representation of a SID into string format
	 *
	 * @param sid a byte representation of a SID
	 * @return a string formatted representation of the SID
	 */
	private static String Sid2String(byte[] SID)
	{
		long version;
		long authority;
		long count;
		long rid;
		String strSID;
		strSID = "S";
		version = SID[0];
		strSID = strSID + "-" + Long.toString(version);
		authority = SID[4];
		for (int i = 0;i<4;i++)
		{
			authority <<= 8;
			authority += SID[4+i] & 0xFF;
		}
		strSID = strSID + "-" + Long.toString(authority);
		count = SID[2];
		count <<= 8;
		count += SID[1] & 0xFF;
		for (int j=0;j<count;j++)
		{
			rid = SID[11 + (j*4)] & 0xFF;
			for (int k=1;k<4;k++)
			{
				rid <<= 8;
				rid += SID[11-k + (j*4)] & 0xFF;
			}
			strSID = strSID + "-" + Long.toString(rid);
		}
		return strSID;
	}

	/**
	 * Parse a string for values separated by a single input character. For
	 * example, "foo,goober,bar" could be parsed by specifying "," as the
	 * token. It would return a String[3] with foo, goober, and bar as
	 * the values.
	 *
	 * @param s the string to evaluate
	 * @param token the token that separates the String values
	 *
	 * @return a String array
	 */
	static String[] getStringArray(String s, String token)
	{
		// Create tokenizer and initialize array based on the number of tokens
		StringTokenizer st = new StringTokenizer(s.trim(), token);
		String[] sa = new String[st.countTokens()];

		// Parse the tokens and store them in the array
		int i = 0;
		while(st.hasMoreTokens())
		{
			sa[i] = st.nextToken();
			i++;
		}

		return sa;
	}

	/**
	 * Parse a string for int values separated by a single input character. 
	 * For example, "389,5389,389" could be parsed by specifying "," as the
	 * token. It would return a int[3] with 389, 389 and 389 as the values.
	 *
	 * @param s the string to evaluate
	 * @param token the token that separates the String values
	 * 
	 * @throws LoginException if values in option are not integers
	 *
	 * @return a int array
	 */
	static int[] getIntArray(String s, String token)
		throws LoginException
	{
		// Create tokenizer and initialize array based on the number of tokens
		StringTokenizer st = new StringTokenizer(s.trim(), token);
		int[] ia = new int[st.countTokens()];

		// Parse the tokens and store them in the array
		int i = 0;
		while(st.hasMoreTokens())
		{
			try
			{
				ia[i] = Integer.parseInt(st.nextToken());
			}
			catch (NumberFormatException e)
			{
				throw new LoginException("A 'connectionPort' value is invalid, must be an integer");
			}

			i++;
		}

		return ia;
	}
} // End of class : UnboundIDActiveDirectoryLoginModule
