/**
 * Copyright (c) 1996-2011 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-
 * INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR
 * ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.cams.service.UnboundIDLdapConnectionPoolService;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.service.ServiceClient;
import com.cafesoft.core.service.ServiceException;
import com.cafesoft.core.service.ServiceFinder;

import com.cafesoft.core.lifecycle.LifecycleException;

import com.unboundid.ldap.sdk.Attribute;
import com.unboundid.ldap.sdk.Entry;
import com.unboundid.ldap.sdk.FailoverServerSet;
import com.unboundid.ldap.sdk.LDAPConnection;
import com.unboundid.ldap.sdk.LDAPConnectionOptions;
import com.unboundid.ldap.sdk.LDAPException;
import com.unboundid.ldap.sdk.LDAPSearchException;
import com.unboundid.ldap.sdk.ResultCode;
import com.unboundid.ldap.sdk.RoundRobinServerSet;
import com.unboundid.ldap.sdk.SearchResult;
import com.unboundid.ldap.sdk.SearchScope;
import com.unboundid.ldap.sdk.SingleServerSet;
import com.unboundid.util.ssl.SSLUtil;
import com.unboundid.util.ssl.TrustAllTrustManager;

import java.security.Principal;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;
import javax.net.ssl.SSLSocketFactory;

import java.io.IOException;

import java.security.GeneralSecurityException;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.StringTokenizer;

/**
 * Implements the JAAS LoginModule interface to authenticate users
 * against a LDAP repository. Once a connection is obtained the LoginModule
 * attempts to authenticate the user by binding to the LDAP server using the
 * user's distinguished name and password. If the bind is successful, then
 * Cams authenticates the user and stores any roles found by the role filter
 * in the Cams user session.
 * <ul>
 * <li><b>debug</b> - (optional) Enable/disable debugging [true|false].
 * <li><b>ldapPoolServiceId</b> - (optional) The service ID configured in 
 * security-domain.xml for an UnboundID LDAP connection pooling service. When 
 * configured, the connection pooling service uses its LDAP connection parameters 
 * and any configured for this login module are ignored.
 * <li><b>connectionHost</b> - (conditional) A comma-delimited list of DNS names 
 * or IP addresses for connection LDAP servers. See <strong>serverSetType</strong> 
 * for configuration options to use more than one host.</li>
 * <li><b>connectionPort</b> - (conditional) A comma-delimited list of one or 
 * more ports for LDAP server connections. Specify a port of the same type (LDAP 
 * or LDAPS) for each host in the <strong>connectionHost</strong> list. Default 
 * values are 389 for LDAP and 636 for LDAPS.</li>
 * <li><b>serverSetType</b> - (optional) Can be: 
 * <b>SINGLE</b> - (default) Connects to a single server. 
 * <b>FAILOVER</b> - Attempts to establish connections to servers in the order 
 * they are provided. If the first server is unavailable, then it will attempt 
 * to connect to the second, then to the third, etc. 
 * <b>ROUND_ROBIN</b> - Uses a round-robin algorithm to select the server to 
 * which the connection should be established. Any number of servers may be 
 * included in this server set, and each request will attempt to retrieve a 
 * connection to the next server in the list, circling back to the beginning of 
 * the list as necessary. If a server is unavailable when an attempt is made to 
 * establish a connection to it, then the connection will be established to the 
 * next available server in the set. 
 * 
 * WARNING: Except for the connectionHost and connectionPort parameters, all
 * other parameters for this login module are shared and must be the same for
 * each server when more than one are configured in a ServerSet. 
 * </li>
 * <li><b>connectionTimeout</b> - (optional) The number of milliseconds to wait 
 * for a connection to the LDAP server before failing (default=3000)</li>
 * 
 * <li><b>responseTimeout</b> - (optional) The number of milliseconds to wait 
 * for a response from the server before failing (default=3000)</li>
 * 
 * <li><b>useSSL</b> - (optional) Enable/disable SSL [true|false].
 * 
 * <li><b>ldapServerType</b> - (optional) A tag used to set the actual LDAP server
 * type, which is useful for identifying server-specific error codes and
 * translating them to specific login exceptions like: AccountExpiredException, 
 * CredentialExpiredException, FailedLoginException, and LoginException. Valid 
 * values include: 
 * <b>ACTIVE_DIRECTORY</b> - Active Directory and ADAM (AD LDS)
 * <b>GENERIC</b> - (default) All LDAP servers that are not ACTIVE_DIRECTORY
 * 
 * <li><b>emptyPassword</b> - (optional) Enable empty passwords (true or
 * false). By default, passwords cannot be empty. Set this value to true to 
 * allow use of empty passwords when anonymous binds are required.
 * 
 * <li><b>defaultRoles</b> - (optional) A list of one or more default roles
 * separated by a comma to be added to the user session.
 * 
 * <li><b>roleRegexPattern</b> - (optional) Limits roles saved in the user 
 * session to those that match a regular expression pattern. This option is 
 * useful when the role search returns many roles that are not used by the 
 * access control policy.
 * 
 * <li><b>userDN</b> - (required) The distinguished name pattern to match to find 
 * a user. The username entered by the the user at login will be substituted
 * for the value "{username}". The LDAP login module will not successfully
 * find users in schemas that use a distinguished name that does not use
 * the username to distinguish users.
 * 
 * <li><b>baseDN</b> - (required) The base distinguished name to use for your
 * LDAP schema role searches.
 * 
 * <li><b>scope</b> - (required) Specifies the LDAP search scope to use for
 * role searches. Valid values are BASE (search is done only at the base
 * dn tree node), ONE (search is done only at the first child tree node
 * of the base dn), and SUB (search is done at the base dn node and all
 * child nodes).
 * 
 * <li><b>attribute</b> - (required) The name of the attribute used to store 
 * the role names.
 * 
 * <li><b>filter</b> - (required) The filter to use to search for a users 
 * roles. The username entered by the the user at login will be substituted for
 * the value "{username}".
 * </ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.12 $ $Date: 2011/05/04 17:17:33 $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 01/18/11
 */
public class UnboundIDLdapLoginModule
	implements LoginModule, LoggerClient, ServiceClient
{
	//
	// Class variables
	//

	/**
	 * The login module name (used in log messages throughout the code).
	 */
	private static final String LMN = "[UnboundIDLdapLoginModule]";
	
	/**
	 * A tag used to identify a generic LDAP server type.
	 */
	private static final String GENERIC_LDAP_SERVER_TYPE = "GENERIC";
	
	/**
	 * A tag used to identify an Active Directory LDAP server type.
	 */
	private static final String ACTIVE_DIRECTORY_LDAP_SERVER_TYPE = "ACTIVE_DIRECTORY";
	
	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Map that contains state shared with other configured LoginModules.
	 */
	private Map sharedState;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * The LDAP server type, which is useful for customizing error handling.
	 */
	private String ldapServerType = GENERIC_LDAP_SERVER_TYPE;
	
	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating User.
	 */
	private String username;

	/**
	 * The password entered by the authenticating User.
	 */
	private String password;

	/**
	 * A list of user roles. This list is populated in the login method and
	 * used in the commit method.
	 */
	private List roleList;

	/**
	 * Used to lookup required Cams services.
	 */
	private ServiceFinder serviceFinder;

	/**
	 * Connection pool.
	 */
	private UnboundIDLdapConnectionPoolService lcps;

	//
	// Constructor
	//

	/**
	 * Default constructor.
	 */
	public UnboundIDLdapLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.ldapServerType = GENERIC_LDAP_SERVER_TYPE;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.password = null;
		this.sharedState = null;
		this.subject = null;
		this.username = null;
		this.roleList = new ArrayList();
		this.serviceFinder = null;
		this.lcps = null;
	}

	//
	// Implements the ServiceClient interface
	//

	/**
	 * Set the ServiceFinder.
	 *
	 * @param finder the class used by find Services by the ServiceClient.
	 */
	public void setServiceFinder(ServiceFinder finder)
	{
		this.serviceFinder = finder;
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *			with the end user (prompting for usernames and
	 *			passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.sharedState = sharedState;
		this.debug = "true".equals((String)options.get("debug"));
		
		// Set the LDAP server type
		if (ACTIVE_DIRECTORY_LDAP_SERVER_TYPE.equals(options.get("ldapServerType")))
			this.ldapServerType = ACTIVE_DIRECTORY_LDAP_SERVER_TYPE;
		else
			this.ldapServerType = GENERIC_LDAP_SERVER_TYPE;
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 * <p>
	 * This method executes the following steps:
	 * <ol>
	 * <li>Checks preconditions
	 * <li>Handles callback objects
	 * <li>Get a connection to the LDAP repository
	 * <li>Authenticates the users by attepting and LDAP bind with the
	 * username and password
	 * <li>Return the connection to its ObjectPool
	 * <li>Reset state if authentication failed
	 * <li>Return results of authentication attempt
	 * </ol>
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @exception LoginException thrown if any error occurs while trying to
	 *	authenticate the user.
	 * @exception FailedLoginException if authentication fails.
	 * @exception AccountExpiredException if the user account is expired.
	 * @exception CredentialExpiredException if the user credential is expired.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException,
			AccountExpiredException, CredentialExpiredException
	{
		// Check preconditions
		validateOptions();

		// Handle callback objects.
		handleCallbacks();

		// A username must always be supplied
		if (username == null || username.length() == 0)
			throw new FailedLoginException(LMN + 
				" Authentication failed");

		// A non-empty password must be supplied unless empty
		// passwords are permitted for this configuration.
		String emptyPassword = (String)options.get("emptyPassword");
		if ((emptyPassword == null) || ("false".equals(emptyPassword.trim())))
		{
			if ("".equals(this.password))
				throw new FailedLoginException(LMN + 
					" Authentication failed");
		}
		else if (!("true".equals(emptyPassword.trim())))
		{
			throw new LoginException(LMN +
				" Error: Option 'emptyPassword' must be 'true' or 'false'");
		}

		// Get a connection to LDAP 
		LDAPConnection connection = null;
	
		// Authenticate the users using an LDAP bind
		try
		{
			if (((options.containsKey("ldapPoolServiceId"))) && 
			((options.get("ldapPoolServiceId") != null)))
			{
				if (this.debug)
					logger.debug(this, LMN + " Getting connection from pool");

				connection = getPooledConnection();
			}
			else
			{
				if (this.debug)
					logger.debug(this, LMN + " Creating new connection");

				connection = getConnection();
			}

			if (this.debug)
			{
				logger.debug(this, LMN + " Connection is connected: '" +
					connection.isConnected() + "'");
				logger.debug(this, LMN + " Connection statistics: " + 
					connection.getConnectionStatistics());
			}

			// Get the user's fully qualifed dn
			String userDN = replaceUsername(
				(String)options.get("userDN"), "username", username);

			if (this.debug)
				logger.debug(this, LMN + " Attempting bind with userDN: '" + 
					userDN + "'");

			connection.bind(userDN, password);

			this.loginSucceeded = true; 

			if (this.debug)
				logger.debug(this, LMN + " Successfully bound user: '" + 
					userDN + "'");
		}
		catch (LDAPException e)
		{
			// An error occurred binding connection therefore login
			// did not succeed. 
			loginSucceeded = false;

			if (ACTIVE_DIRECTORY_LDAP_SERVER_TYPE.equalsIgnoreCase(ldapServerType))
				activeDirectoryLDAPExceptionHandler(e);
			else
				genericLdapServerLDAPExceptionHandler(e);
		}
		finally
		{
			// Close the connection if there was an exception
			if (!loginSucceeded)
			{
				if (connection != null)
				{
					if (((options.containsKey("ldapPoolServiceId"))) && 
					((options.get("ldapPoolServiceId") != null)))
					{
						try
						{
							lcps.releaseConnection(connection);
						}
						catch(ServiceException e)
						{
							if (this.debug)
								logger.debug(this, LMN + " Service error while returning connection to pool");

							throw new LoginException(e.getMessage());
						}
						catch(LifecycleException e)
						{
							if (this.debug)
								logger.debug(this, LMN + " Lifecyle error while returning connection to pool");

							throw new LoginException(e.getMessage());
						}
					}
					else
					{
						connection.close();
						connection = null;
					}
				}

			}
		}

		// Get the user's roles 
		try
		{
			// Add default roles (if any) to list of roles
			String defaultRoles = (String)options.get("defaultRoles");
			if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
			{
				StringTokenizer st = new StringTokenizer(defaultRoles, ",");
				while(st.hasMoreTokens())
					roleList.add(st.nextElement().toString().trim());
			}

			// Get the roles for use later during the commit method.
			String baseDN = (String)options.get("baseDN");
			String[] attrs = getStringArray((String)options.get("attribute"), ",");
			String filter = 
				replaceUsername((String)options.get("filter"), "username", username);
			String scope = (String)options.get("scope");

			// Set the search scope, the default of 0 is "BASE"
			int scopeInt = SearchScope.BASE_INT_VALUE;
			if ("ONE".equalsIgnoreCase(scope))
				scopeInt = SearchScope.ONE_INT_VALUE;
			else if ("SUB".equalsIgnoreCase(scope))
				scopeInt = SearchScope.SUB_INT_VALUE;

			// Set the base dn where the search will be performed, and the
			// name of the role to use.

			// Do the search.
			if (this.debug)
			{
				logger.debug(this, LMN + 
					" Search for user's groups with:" +
					  "\n\tbaseDN: " + baseDN +
					  "\n\tfilter: " + filter +
					  "\n\t attrs: " + attrs[0]);
			}

			SearchResult searchResults = connection.search(
				baseDN, SearchScope.valueOf(scopeInt), filter, attrs);

			if (this.debug)
				logger.debug(this, LMN + " Done searching, result code is: '" + 
					searchResults.getResultCode() + "'");

			// Process the search results to get the role names
			//   -- The first while loop filters for the entries
			//   -- The second while loop filters for the rolenames //      and stores them to a principal and subject
			if (searchResults != null)
			{
				List searchResultsList = searchResults.getSearchEntries();
				Iterator searchResultsIterator = searchResultsList.iterator();
				while (searchResultsIterator.hasNext())
				{
					// Get the next entries attibute
					Entry entry = (Entry)searchResultsIterator.next();
					Collection attrCollection = entry.getAttributes();
					Iterator attrIterator = attrCollection.iterator();

					// If supplied, only use roles that match roleRegexPattern 
					String roleRegexPattern = (String)options.get("roleRegexPattern");

					while (attrIterator.hasNext())
					{
						// Attribute may have more than one name/value pair
						Attribute attr = (Attribute) attrIterator.next();
						String[] values = attr.getValues();
						if ((roleRegexPattern != null) && !("".equals(roleRegexPattern)))
						{
							Pattern pattern = Pattern.compile(roleRegexPattern);
							for (int i = 0; i < values.length; i++)
							{
								if (pattern.matcher(values[i]).matches())
								{
									roleList.add(values[i]);

									if (this.debug)
										logger.debug(this, LMN +
											" Role: '" + values[i] + 
											"' matched roleRegexPattern: '" + 
											roleRegexPattern + "'");

									break;
								}
								else
								{
									logger.debug(this, LMN + 
										" Role: '" + values[i] + 
										"' did not match roleRegexPattern: '" + 
										roleRegexPattern + "'");
								}
							}
						}
						else
						{
							for (int i = 0; i < values.length; i++)
							{
								roleList.add(values[i]);

								if (this.debug)
									logger.debug(this, LMN +
									 " User has role: '" + values[i] + "'");
							}
						}
					}
				}
			}
			else
			{
				if (this.debug)
					logger.debug(this, LMN +
						" No roles found for user: '" + username + "'");
			}
		}
		catch (LDAPSearchException e)
		{
			// An error occurred looking up roles for therefore login
			// did not succeed. Although this deviates slightly from the JAAS
			// implementation of the commit method it basically performs this
			// error check prior to the commit method allowing it to fail fast.
			if (this.debug)
				logger.debug(this, LMN + " Error while " +
					"searching for groups for user: '" +
					this.username + "'", e);

			// Search filter failed
			throw new LoginException(LMN + " " + e.getMessage());
		}
		catch (LDAPException e)
		{
			if (this.debug)
				logger.debug(this, LMN + " Error while " +
					"searching for groups for user: '" +
					this.username + "'", e);

			// Search filter failed
			throw new LoginException(LMN + " " + e.getMessage());
		}
		finally
		{
			// Close the connection
			if (connection != null)
			{
				if (((options.containsKey("ldapPoolServiceId"))) && 
				    ((options.get("ldapPoolServiceId") != null)))
				{
					try
					{
						lcps.releaseConnection(connection);
					}
					catch(ServiceException e)
					{
						if (this.debug)
							logger.debug(this, LMN + " Service error while returning connection to pool");

						throw new LoginException(e.getMessage());
					}
					catch(LifecycleException e)
					{
						if (this.debug)
							logger.debug(this, LMN + " Lifecyle error while returning connection to pool");

						throw new LoginException(e.getMessage());
					}
				}
				else
				{
					connection.close();
					connection = null;
				}
			}
		}

		// Reset state, authentication failed.
		if (!loginSucceeded)
		{
			this.username = null;
			this.password = null;

			throw new FailedLoginException(LMN + " Authentication failed");
		}

		// Return results of authentication attempt.
		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @exception LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login failed, clean up state
		if (!loginSucceeded)
			return false;

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList();

		// Repository operations to get role principals
		Iterator roleListItr = roleList.iterator();
		while(roleListItr.hasNext())
		{
			String rolename = (String)roleListItr.next().toString().trim();
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				principalList.add(rolePrincipal);

				if (this.debug)
					logger.debug(this, LMN +
						" Added CSRolePrincipal: '" + rolename + 
						"' to principalList for: '" + username + "'");
			}
			else
			{
				if (this.debug)
					logger.debug(this, LMN +
						" CSRolePrincipal: '" + rolename + 
						"' already in principalList for: '" + 
						username + "', did not add");
			}
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if (this.debug)
				logger.debug(this, LMN +
					" Added CSUserPrincipal: '" + username +
					"' to principalList");
		}
		else
		{
			if (this.debug)
				logger.debug(this, LMN +
					" CSUserPrincipal: '" + username +
					"' already in principalList did not add");
		}

		if (this.debug)
			logger.debug(this, LMN + " Adding principalList to Subject");

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		if (this.debug)
		{
			logger.debug(this, LMN + " Adding " +
				AuthenticationMethod.PASSWORD.getName() +
				" AuthenticationMethod");
		}

		subject.getPublicCredentials().add(
			AuthenticationMethod.PASSWORD);

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.password = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @exception LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.
			if (this.roleList != null)
			{
				this.roleList.clear();
				this.roleList = null;
			}

			this.loginSucceeded = false;
			this.username = null;
			this.password = null;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @exception LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if (this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if (this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.logger = null;
		this.callbackHandler = null;
		this.username = null;
		this.password = null;
		this.options = null;
		this.loginSucceeded = false;
		this.commitSucceeded = false;
		this.debug = false;

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username and
	 * password instance variables with their appropriate value.
	 *
	 * @throws LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException
	{
		// This LoginModule requires a username and a password
		Callback[] callbacks = new Callback[1];
		callbacks[0] = new MapCallback();

		try
		{
			// Get the callback values and clear the password.
			callbackHandler.handle(callbacks);
			this.username = ((MapCallback)callbacks[0]).getCallbackValue("username");
			this.password = ((MapCallback)callbacks[0]).getCallbackValue("password");
			((MapCallback)callbacks[0]).setCallbackValue("password", null);
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " +
				e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this, LMN + 
				" Callback not available", e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() +
				" not available");
		}

		if (this.debug)
		{
			logger.debug(this, LMN + " User entered username = " + username);
		}
	}

	/**
	 * Validate the LoginModules options Map object to ensure that all
	 * required properties exist and contain valid values.
	 *
	 * @exception LoginException if any of the following validations fails:
	 * <ul>
	 * <li>A CallbackHandler object is not available</li>
	 * <li>the 'connectionHost' is not available</li>
	 * <li>the 'connectionPort' is not available</li>
	 * <li>the 'connectionTimeout' is not a valid integer</li>
	 * <li>the 'responseTimeout' is not a valid integer</li>
	 * <li>the 'userDN' is not available</li>
	 * <li>the 'baseDN' is not available</li>
	 * <li>the 'attribute' is not available</li>
	 * <li>the 'filter' is not available</li>
	 * <li>the 'scope' is not available</li>
	 * <li>the 'scope' is not a valid role scope value</li>
	 * </uL>
	 */
	private void validateOptions() throws LoginException
	{
		if (callbackHandler == null)
			throw new LoginException(LMN +
				" Error: no CallbackHandler available");

		String s = null;

		// Only validate connection options if not using a connection
		// pool service
		if (((!options.containsKey("ldapPoolServiceId"))) || 
		    ((options.get("ldapPoolServiceId") == null)))
		{
			s = (String)options.get("connectionHost");
			if ((s == null) || (s.trim().length() == 0))
				throw new LoginException(LMN +
					" Error: Required 'connectionHost' property missing");

			s = (String)options.get("connectionPort");
			if ((s == null) || (s.trim().length() == 0))
				throw new LoginException(LMN + 
					" Error: Required 'connectionPort' property missing");

			s = (String)options.get("connectionTimeout");
			if ((s != null) && (s.trim().length() > 0))
			{
				try
				{
					if (Integer.parseInt(s) < 1)
						throw new LoginException(LMN + " Error: " +
						"'connectionTimeout' property must be greater than zero");
				}
				catch (NumberFormatException e)
				{
					throw new LoginException(LMN + " Error: " +
						"'connectionTimeout' property is not an integer");
				}
			}

			s = (String)options.get("responseTimeout");
			if ((s != null) && (s.trim().length() > 0))
			{
				try
				{
					if (Integer.parseInt(s) < 1)
						throw new LoginException(LMN + " Error: " +
						"'responseTimeout' property must be greater than zero");
				}
				catch (NumberFormatException e)
				{
					throw new LoginException(LMN + " Error: " +
						"'responseTimeout' property is not an integer");
				}
			}
		}

		String userDN = (String)options.get("userDN");
		if ((userDN == null) || "".equals(userDN.trim()))
			throw new LoginException(LMN +
				" Error: Required 'userDN' property missing");

		String baseDN = (String)options.get("baseDN");
		if ((baseDN == null) || "".equals(baseDN.trim()))
			throw new LoginException(LMN + 
				" Error: Required 'baseDN' property missing");

		String attr = (String)options.get("attribute");
		if ((attr == null) || "".equals(attr.trim()))
			throw new LoginException(LMN + 
				" Error: Required 'attribute' property missing");

		String filter = (String)options.get("filter");
		if ((filter == null) || "".equals(filter.trim()))
			throw new LoginException(LMN + 
				" Error: Required 'filter' property missing");

		String scope = (String)options.get("scope");
		if ((!"BASE".equalsIgnoreCase(scope)) &&
			(!"ONE".equalsIgnoreCase(scope) &&
			(!"SUB".equalsIgnoreCase(scope))))
			throw new LoginException(LMN + 
				"Error: Required option 'scope' must be 'BASE', 'ONE', or 'SUB'");
	}

	/**
	 * Get an LDAP server connection.
	 *
	 * @return a LDAPConnection.
	 * @exception LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getConnection()
		throws LoginException
	{
		// Get the connection values to use
		String[] connectionHost = getStringArray((String)options.get("connectionHost"), ",");
		int[] connectionPort = getIntArray((String)options.get("connectionPort"), ",");
		boolean useSSL = "true".equals((String)options.get("useSSL"));
		String serverSetType = (String)options.get("serverSetType");
		int connectionTimeout = Integer.parseInt((String)options.get("connectionTimeout"));
		long responseTimeout = (long)Integer.parseInt((String)options.get("responseTimeout"));

		try
		{
			if (this.debug)
				logger.debug(this, LMN + " Instantiating LDAP connection");

			// Initialize the connection
			LDAPConnection connection = null;
			SSLSocketFactory sslSocketFactory = null;

			// Set the connection timeout
			LDAPConnectionOptions connectionOptions = new LDAPConnectionOptions();
			connectionOptions.setConnectTimeoutMillis(connectionTimeout);
			connectionOptions.setResponseTimeoutMillis(responseTimeout);

			if (useSSL)
			{
				SSLUtil sslUtil = new SSLUtil(new TrustAllTrustManager());
				sslSocketFactory = sslUtil.createSSLSocketFactory();
			}

			long timer = System.currentTimeMillis(); 

			// Use "SINGLE" ServerSet if configuration parameter is not 
			// supplied to support legacy integrations
			if (serverSetType == null)
				serverSetType = "SINGLE";

			if (serverSetType.equalsIgnoreCase("FAILOVER"))
			{
				if (this.debug)
					logger.debug(this, LMN + " Using 'FAILOVER' ServerSet for LDAP connection"); 

				FailoverServerSet failoverServerSet = 
					new FailoverServerSet(connectionHost, 
					connectionPort, sslSocketFactory, 
					connectionOptions);

				connection = failoverServerSet.getConnection();
			}
			else if (serverSetType.equalsIgnoreCase("ROUND_ROBIN"))
			{
				if (this.debug)
					logger.debug(this, LMN + " Using 'ROUND_ROBIN' ServerSet for LDAP connection"); 

				RoundRobinServerSet roundRobinServerSet = 
					new RoundRobinServerSet(connectionHost, 
					connectionPort, sslSocketFactory, 
					connectionOptions);

				connection = roundRobinServerSet.getConnection();
			}
			else // Fall through is "SINGLE"
			{
				if (this.debug)
					logger.debug(this, LMN + " Using 'SINGLE' ServerSet for LDAP connection"); 

				SingleServerSet singleServerSet = 
					new SingleServerSet(connectionHost[0], 
					connectionPort[0], sslSocketFactory,
					connectionOptions);

				connection = singleServerSet.getConnection();
			}

			if (this.debug)
			{
				logger.debug(this, LMN + " Time to connect: '" +
					(System.currentTimeMillis() - timer) + "'");

				if (connection.isConnected())
				{
					logger.debug(this, LMN + " LDAP connection succeeded to: '" + 
						connection.getConnectedAddress() + ":" +
						connection.getConnectedPort() + "'");
				}
				else
				{
					logger.debug(this, LMN + " LDAP connection failed");
				}

				logger.debug(this, LMN + " Returning LDAP connection");
			}

			return connection;
		}
		catch (GeneralSecurityException e)
		{
			logger.error(this, LMN + " General security error", e);

			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		catch (LDAPException e)
		{
			logger.error(this, LMN + " LDAP error", e);

			// Dump the nested Throwables to get at the root cause
			int i = 1;
			Throwable cause = e.getCause();
			do
			{
				if (cause != null)
					logger.error(this, LMN + " LDAP error: CAUSE " + i + ": " +
						cause.getMessage());
			}
			while ((cause = cause.getCause()) != null);

			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
	}

	/**
	 * Get an LDAP server connection from a connection pool.
	 *
	 * @return a LDAPConnection.
	 * @exception LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getPooledConnection()
		throws LoginException
	{
		// Get the connection values to use
		LDAPConnection connection = null;

		try
		{
			if (serviceFinder == null)
				throw new LoginException(LMN +
					" Error: ServiceFinder is null");

			String ldapPoolServiceId = (String)options.get("ldapPoolServiceId");

			if (this.debug)
				logger.debug(this, LMN + " Using ldapPoolServiceId '" + ldapPoolServiceId + "'");

			this.lcps = (UnboundIDLdapConnectionPoolService)
				serviceFinder.find(ldapPoolServiceId, UnboundIDLdapConnectionPoolService.class);

			connection = lcps.getConnection();

			if (this.debug)
				logger.debug(this, LMN + " Got LDAP connection: " + 
				connection.getConnectionStatistics());
		}
		catch(ServiceException e)
		{
			logger.error(this, LMN + " Service error while getting connection from pool", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		catch(LifecycleException e)
		{
			logger.error(this, LMN + " Lifecycle error while getting connection from pool", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		catch(LDAPException e)
		{
			logger.error(this, LMN + " LDAP error while getting connection from pool", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}

		return connection;
	}

	/*
	 * Translate a generic LDAP server LDAPException into a more specific
	 * LoginException and throw it.
	 * 
	 * @param e the LDAPException.
	 * @throws LoginException if not a more specific login failure.
	 * @throws FailedLoginException if invalid login credentials.
	 * @throws CredentialExpiredException if a login credential has expired.
	 * @throws AccountExpiredException if the user account has expired.
	 */
	private void genericLdapServerLDAPExceptionHandler(LDAPException e)
		throws LoginException, FailedLoginException,
			CredentialExpiredException, AccountExpiredException
	{
		// GEG todo: Test against SunONE LDAP to determine if more
		// specific information is available for failed conditions 
		if (this.debug)
			logger.debug(this, LMN + " LDAPException while binding and searching for roles" +
				"\n\tMessage: " + e.getExceptionMessage());

		if (e.getResultCode().equals(ResultCode.INVALID_CREDENTIALS))
		{
			throw new FailedLoginException();
		}
		else
		{
			throw new LoginException();	
		}
	}
	
	/**
	 * Translate a specific ActiveDirectory LDAPException into a specific
	 * LoginException and throw it.
	 * 
	 * @param e the LDAPException containing an ActiveDirectory specific message.
	 * @throws LoginException if not a more specific login failure.
	 * @throws FailedLoginException if invalid login credentials.
	 * @throws CredentialExpiredException if a login credential has expired.
	 * @throws AccountExpiredException if the user account has expired.
	 */
	private void activeDirectoryLDAPExceptionHandler(LDAPException e)
		throws LoginException, FailedLoginException,
			CredentialExpiredException, AccountExpiredException
	{
		if (this.debug)
			logger.debug(this, LMN + " LDAPException while binding " +
				"\n\tMessage: " + e.getExceptionMessage());

		// Normal login failure due to invalid credentials, check
		// Active Directory server message for subcode to determine the
		// most accurate exception to throw. Possible codes include:
		// 52e - invalid credentials (FailedLoginException)
		// 525 - user not found (FailedLoginException)
		// 530 - invalid login time (FailedLoginException)
		// 532 - password expired (CredentialExpiredException)
		// 533 - account disabled (AccountExpiredException)
		// 701 - account expired (AccountExpiredException)
		// 773 - password reset required (CredentialExpiredException)
		// 775 - user account locked (AccountExpiredException)
		// 2030 - no such object (FailedLoginException)
		if (e.getResultCode().equals(ResultCode.INVALID_CREDENTIALS))
		{
			if (e.getExceptionMessage().indexOf("data 52e") > 0)
			{
				if(this.debug)
					logger.debug(this, LMN + " Failed credentials (data 52e)");

				throw new FailedLoginException("Failed credentials (data 52e)");
			}
			if (e.getExceptionMessage().indexOf("data 532") > 0)
			{
				if(this.debug)
					logger.debug(this, LMN + " Password expired (data 532)");

				throw new CredentialExpiredException("Password expired (data 532)");
			}
			else if (e.getExceptionMessage().indexOf("data 533") > 0)
			{
				if(this.debug)
					logger.debug(this, LMN + " Account disabled (data 533)");

				throw new AccountExpiredException("Account disabled (data 533)");
			}
			else if (e.getExceptionMessage().indexOf("data 701") > 0)
			{
				if(this.debug)
					logger.debug(this, LMN + " Account expired (data 701)");

				throw new AccountExpiredException("Account expired (data 701)");
			}
			else if (e.getExceptionMessage().indexOf("data 773") > 0)
			{
				if(this.debug)
					logger.debug(this, LMN + " Password reset required (data 773)");

				throw new CredentialExpiredException("Password reset required (data 773)");
			}
			else
			{
				if(this.debug)
					logger.debug(this, LMN + " " + e);

				throw new FailedLoginException(e.getExceptionMessage());
			}
		}
		else
		{
			logger.error(this, LMN + " Error while binding connection: " + e);

			throw new LoginException(e.getExceptionMessage());
		}
	}

	/**
	 * Return the input pattern replaced with the match value.
	 *
	 * @param pattern The input pattern to use
	 * @param matchName The variable name to match
	 * @param matchValue The value to replace
	 * @return The input pattern with matched value
	 */
	private String replaceUsername(String pattern, String matchName,
		String matchValue)
	{
		String tmp = null;
		StringBuffer sb = new StringBuffer();
		StringTokenizer st = new StringTokenizer(pattern, "{}");
		while (st.hasMoreTokens())
		{
			tmp = st.nextToken();

			if (tmp.trim().equals(matchName))
				sb.append(matchValue);
			else
				sb.append(tmp);
		}

		return sb.toString();
	}

	/**
	 * Parse a string for values separated by a single input character. For
	 * example, "foo,goober,bar" could be parsed by specifying "," as the
	 * token. It would return a String[3] with foo, goober, and bar as
	 * the values.
	 *
	 * @param s the string to evaluate
	 * @param token the token that separates the String values
	 *
	 * @return a String array
	 */
	static String[] getStringArray(String s, String token)
	{
		// Create tokenizer and initialize array based on the number of tokens
		StringTokenizer st = new StringTokenizer(s.trim(), token);
		String[] sa = new String[st.countTokens()];

		// Parse the tokens and store them in the array
		int i = 0;
		while(st.hasMoreTokens())
		{
			sa[i] = st.nextToken();
			i++;
		}

		return sa;
	}

	/**
	 * Parse a string for int values separated by a single input character. 
	 * For example, "389,5389,389" could be parsed by specifying "," as the
	 * token. It would return a int[3] with 389, 389 and 389 as the values.
	 *
	 * @param s the string to evaluate
	 * @param token the token that separates the String values
	 * 
	 * @throws LoginException if values in option are not integers
	 *
	 * @return a int array
	 */
	static int[] getIntArray(String s, String token)
		throws LoginException
	{
		// Create tokenizer and initialize array based on the number of tokens
		StringTokenizer st = new StringTokenizer(s.trim(), token);
		int[] ia = new int[st.countTokens()];

		// Parse the tokens and store them in the array
		int i = 0;
		while(st.hasMoreTokens())
		{
			try
			{
				ia[i] = Integer.parseInt(st.nextToken());
			}
			catch (NumberFormatException e)
			{
				throw new LoginException(LMN + " A 'connectionPort' value is invalid, must be an integer");
			}

			i++;
		}

		return ia;
	}
} // End of class : UnboundIDLdapLoginModule
