/**
 * Copyright (c) 1996-2010 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-
 * INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR
 * ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.util.X509CertificateUtils;

import java.security.Principal;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidAlgorithmParameterException;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import javax.security.auth.spi.LoginModule;

import javax.security.auth.x500.X500Principal;

import java.security.cert.X509Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.CertPathValidatorException;
import java.security.cert.PKIXCertPathValidatorResult;

import java.io.File;
import java.io.IOException;

import java.math.BigInteger;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * <code>X509CertificateFileLoginModule</code> implements a JAAS
 * <code>LoginModule</code> that authenticates users presenting an
 * <code>X509Certificate</code> chain. The X509 certificate repository
 * uses a simple file-based scheme to lookup certificates by serial number,
 * which is required to be unique by Certificate Authority.
 * <ul>
 * <li><b>debug</b> - Enables/disables debugging for the LoginModule.</li>
 * <li><b>caCertFile</b> - The path to the directory containing
 * Certificate Authority (CA) certificate. The IssuerDN of a certificate
 * in the user-supplied certificate chain must match the IssuerDN of the
 * CA certificate.</li>
 * <li><b>userCertDir</b> - The path to the directory containing
 * issued certificates that are used for comparision against those
 * supplied by users attempting to authenticate. The certificate files
 * must be have names in the form: serial.pem where "serial" is the unique
 * serial number for CA issued certificate. For example: 05.pem.</li>
 * <li><b>defaultRoles</b> - (Optional) A comma-delimited list of
 * default roles to be assigned to a successfully authenticated user.</li>
 * <li><b>X509_cb_name</b> - (Optional) By default, a "X509" callback
 * is expected. If the login request uses a different name, then optionally
 * configure it, for example:
 *    <option name="X509_cb_name" value="X509_cert"/></li>
 * </ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.10 $ $Date: 2010/12/17 21:15:42 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 08/26/2005
 */
public class X509CertificateFileLoginModule
	implements LoginModule, LoggerClient
{
	//
	// static constants
	//

	/**
	 * The LoginModule name (used in log messages throughout the code).
	 */
	private static final String LMN = "[X509CertificateFileLoginModule]";

	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Map that contains state shared with other configured LoginModules.
	 */
	private Map sharedState;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The X509Certificate chain supplied by the authenticating user.
	 */
	private X509Certificate[] certChain;

	/**
	 * A list of user roles. This list is populated in the login method and
	 * used in the commit method.
	 */
	private List roleList;

	//
	// Constructor
	//

	/**
	 * Create a new X509CertificateFileLoginModule instance.
	 */
	public X509CertificateFileLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.certChain = null;
		this.sharedState = null;
		this.subject = null;
		this.roleList = new ArrayList();
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for
	 * 	getting access to authentication credentials.
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.sharedState = sharedState;
		this.debug = "true".equals((String)options.get("debug"));
	}

	/**
	 * Athenticate a <code>Subject</code> (phase 1).
	 *
	 * <p>The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		if (debug)
			logger.debug(this, "validating LoginModule configuration options");

		// Check preconditions
  		try
		{
			validateOptions();
		}
		catch (LoginException e)
	  	{
			logger.error(this, "Error validating options", e);
			throw e;
		}

		if (debug)
			logger.debug(this, "handling LoginModule Callbacks");

		// Handle callback objects.
		handleCallbacks();

		if (debug)
			logger.debug(this, "checking X509Certificate chain validity period");

		// A certificate chain must always be specified
		if (certChain == null || certChain.length == 0)
		{
			if (debug)
				logger.debug(this, "authentication failed: " +
					"null or empty X09Certificate chain");

			throw new FailedLoginException(LMN + " Authentication failed");
		}

		// NKK: Verify that this step is necessary here.
		// X509CertficateUtils.validate(...) may already do this.
		//
		// Check the X509Certificate validity period
		for (int i = 0; i < certChain.length; i++)
		{
			X509Certificate cert = certChain[i];

			if (debug)
				logger.debug(this, "X509Certificate[" + i + "]=\n" +
					cert.toString());

			try
			{
				cert.checkValidity();
			}
			catch (CertificateExpiredException e)
			{
				throw new CredentialExpiredException(LMN +
					" X509 certificate expired at: " +
					cert.getNotAfter());
			}
			catch (CertificateNotYetValidException e)
			{
				throw new FailedLoginException(LMN +
					" X509 certificate not valid until: " +
					cert.getNotBefore());
			}

			if (debug)
				logger.debug(this, "X509Certificate[" + i + "] is valid");
		}

		if (debug)
			logger.debug(this, "loading the certificate authority X509Certificate");

		X509Certificate caCert = null;

		try
		{
			// Load the Certificate Authority certificate
			caCert = X509CertificateUtils.decodeFromPEM(
				new File((String)options.get("caCertFile")));
		}
		catch (CertificateException e)
		{
			logger.error(this, LMN + " error loading CA certificate");

			throw new LoginException(LMN + " error loading CA certificate");
		}

		if (debug)
			logger.debug(this, LMN + " performing PKIX certificate path validation");

		try
		{
			// Validate the certificate chain
			PKIXCertPathValidatorResult r =
				X509CertificateUtils.validate(certChain, caCert);
		}
		catch (CertPathValidatorException e)
	  	{
			throw new FailedLoginException(LMN +
				" invalid certificate chain: " +
				e.getMessage());
		}
		catch (CertificateException e)
	  	{
			logger.error(this, LMN +
				"error when validating certificate", e);

			throw new LoginException(LMN +
				"error when validating certificate");
		}
		catch (NoSuchAlgorithmException e)
	  	{
			logger.error(this, LMN +
				"error when validating certificate", e);

			throw new LoginException(LMN +
				"error when validating certificate");
		}
		catch (InvalidAlgorithmParameterException e)
	  	{
			logger.error(this, LMN +
				"error when validating certificate", e);

			throw new LoginException(LMN +
				"error when validating certificate");
		}

		// Client certificate is the first cert of the chain.
		X509Certificate transmittedClientCert = certChain[0];

		// Get the serial number of the certificate in order to lookup cert.
		BigInteger serialNumber = transmittedClientCert.getSerialNumber();

		// Create the name of the file
		String clientCertFilename =
			(String)options.get("userCertDir") + "/" + serialNumber.toString() + ".pem";

		File clientCertFile = new File(clientCertFilename);

		X509Certificate storedClientCert = null;

		try
		{
			storedClientCert =
				X509CertificateUtils.decodeFromPEM(clientCertFile);
		}
		catch(CertificateException e)
		{
			logger.error(this, LMN +
				" failed to load " + clientCertFilename, e);

			throw new LoginException(LMN +
				" failed to load " + clientCertFilename);
		}

		Principal transPrincipal = transmittedClientCert.getSubjectDN();
		Principal storedPrincipal = storedClientCert.getSubjectDN();

		if(transPrincipal.equals(storedPrincipal))
		{
			loginSucceeded = true;
		}

		if (loginSucceeded)
		{
			if (debug)
				logger.debug(this, "login succeeded");

			// Add default roles (if any) to list of roles
			String defaultRoles = (String)options.get("defaultRoles");
			if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
			{
				StringTokenizer st = new StringTokenizer(defaultRoles, ",");
				while(st.hasMoreTokens())
				{
					String roleName = st.nextElement().toString().trim();
					if (debug)
						logger.debug(this, "adding role name: " + roleName);

					roleList.add(roleName);
				}
			}
		}

		// Reset state, authentication failed.
		if (!loginSucceeded)
		{
			this.certChain = null;

			throw new FailedLoginException(LMN +
				" Authentication failed");
		}

		// Return results of authentication attempt.
		return loginSucceeded;
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		// The login failed, clean up state
		if (!loginSucceeded)
			return false;

		// Get the Subject distinguished name
  		X500Principal subjectPrincipal = certChain[0].getSubjectX500Principal();
		String subjectName = subjectPrincipal.getName();

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList();

		// Repository operations to get role principals
		Iterator roleListItr = roleList.iterator();
		while(roleListItr.hasNext())
		{
			String rolename = (String)roleListItr.next().toString().trim();
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				principalList.add(rolePrincipal);

				if (debug)
					logger.debug(this, LMN +
						" Added CSRolePrincipal " +
						rolename + " to principal ArrayList for user '" +
						subjectName + "'");
			}
			else
			{
				if (debug)
					logger.debug(this, LMN +
						" CSRolePrincipal '" +
						rolename + "' already in Subject for '" +
						subjectName + "', not added to principal List");
			}
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(subjectName);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if (debug)
				logger.debug(this, LMN +
					" Added CSUserPrincipal '" + subjectName +
					"' to principal List");
		}
		else
		{
			if (debug)
				logger.debug(this, LMN +
					" CSUserPrincipal '" + subjectName +
					"' already in Subject, not added to principal List");
		}

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		if (debug)
		{
			logger.debug(this, LMN + " Adding " +
				AuthenticationMethod.X509.getName() + 
				" AuthenticationMethod");
		}

		subject.getPublicCredentials().add(
			AuthenticationMethod.X509);

		this.commitSucceeded = true;

		// Clean up
		return commitSucceeded;
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.
			if (this.roleList != null)
			{
				this.roleList.clear();
				this.roleList = null;
			}

			this.loginSucceeded = false;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		// Cleanup Object references
		if (this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if (this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.logger = null;
		this.callbackHandler = null;
		this.options = null;
		this.loginSucceeded = false;
		this.commitSucceeded = false;
		this.debug = false;

		return true;
	}

	//
	// Private Methods
	//

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. In turn if successful this method will assign the username
	 * instance variable.
	 *
	 * @exception LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException, FailedLoginException
	{
		// This LoginModule requires a X509 digital cert 
		MapCallback mapCallback = new MapCallback();
		Callback[] callbacks = new Callback[1];
		callbacks[0] = mapCallback;

		try
		{
			// Populate callback values.
			callbackHandler.handle(callbacks);

			// Get the name of the callback value containing the client cert.
			// Set a default name if not provided.
			String X509_cb_name = (String)options.get("X509_cb_name");
			if ((X509_cb_name == null) || ("".equals(X509_cb_name)))
				X509_cb_name = "x509_client_cert";
			
			// Get the X.509 cert chain as a String.
			String certStr = mapCallback.getCallbackValue(X509_cb_name);
			if (certStr != null)
			{
				// Convert the cert callback value from a String to decoded form.
				this.certChain = X509CertificateUtils.decodeChainFromPEM(certStr);
			}
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available", e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() + " not available");
		}
		catch (CertificateException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available", e);

			throw new LoginException(LMN + " Error: " + e.getMessage());
		}

		if (debug)
		{
			if (this.certChain != null)
			{
				for (int i = 0; i < certChain.length; i++)
					logger.debug(this, "X509Certificate[" + i + "]=\n" +
					certChain[i].toString());
			}
			else
			{
				logger.debug(this, "No certificates supplied");
			}
		}
	}

	/**
	 * Validate the LoginModules options Map object to ensure that all
	 * required properties exist and contain valid values.
	 *
	 * @exception LoginException if any of the following conditions occur:
	 * <UL>
	 * <LI>A CallbackHandler object is not available,
	 * <LI>the 'caCertFile' property is not available,
	 * <LI>the 'caCertFile' references a file that does not exist
	 * <LI>the 'caCertFile' references a file that cannot be read
	 * <LI>the 'userCertDir' property is not available,
	 * <LI>the 'userCertDir' references a directory that does not exist
	 * </UL>
	 */
	private void validateOptions() throws LoginException
	{
		if (callbackHandler == null)
			throw new LoginException(LMN +
				" Error: no CallbackHandler available");

		String s = (String)options.get("caCertFile");
		if ((s == null) || "".equals(s.trim()))
			throw new LoginException(LMN +
				" Error: Required 'caCertFile'" +
				"	property missing");

		File f = new File(s);
		if (f.exists() == false)
			throw new LoginException(LMN +
				" Error: caCertFile: " +
					f.getAbsolutePath() + " does not exist");

		if (f.canRead() == false)
			throw new LoginException(LMN +
				" Error: caCertFile: " +
					f.getAbsolutePath() + " cannot be read, no permission");

		s = (String)options.get("userCertDir");
		if ((s == null) || "".equals(s.trim()))
			throw new LoginException(LMN +
				" Error: Required 'userCertDir'" +
				" property missing");

		f = new File(s);
		if (f.exists() == false)
			throw new LoginException(LMN +
				" Error: userCertDir: " +
					f.getAbsolutePath() + " directory does not exist");

		if (f.isDirectory() == false)
			throw new LoginException(LMN +
				" Error: userCertDir: " +
					f.getAbsolutePath() + " is not a directory");
	}
} // End of class : X509CertificateFileLoginModule
