/**
 * Copyright (c) 1996-2012 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-
 * INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR
 * ITS DERIVATIVES.
 */

package com.cafesoft.cams.auth.login.module;

import com.cafesoft.cams.auth.AuthenticationMethod;
import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.CSUserPrincipal;

import com.cafesoft.cams.auth.callback.MapCallback;

import com.cafesoft.core.log.Logger;
import com.cafesoft.core.log.LoggerClient;
import com.cafesoft.core.log.StderrLogger;

import com.cafesoft.core.util.X509CertificateUtils;

import com.novell.ldap.LDAPAttribute;
import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;
import com.novell.ldap.LDAPSearchResults;
import com.novell.ldap.LDAPSocketFactory;
import com.novell.ldap.LDAPJSSESecureSocketFactory;

import java.security.Principal;

import javax.security.auth.Subject;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;

import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import java.security.cert.X509Certificate;
import java.security.cert.CertificateException;

import javax.security.auth.spi.LoginModule;

import java.io.IOException;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * <p><code>X509CertificateLdapLoginModule</code> implements a JAAS LoginModule
 * that uses LDAP to authenticate a user presenting a username, password, and an 
 * X509 client certificate.</p>
 * <p>
 * This LoginModule attempts to bind to a user-specific LDAP DN which
 * may be constructed by substituting the value of configuration option
 * "userDN" with a username parameter. The LDAP bind is attempted using
 * the user-specified password, but only if all three authentication callback
 * parameters are present.</p>
 * <p>
 * If the LDAP bind succeeds, the LDAP userCertificate attribute is
 * compared with the X509 certificate presented by the client. If the
 * certificates match (exactly), then the first phase of the login
 * transaction continues to assignment of roles, else login fails due
 * an invalid certificate.</p>
 * <p>
 * Role assignment is accomplished by use of an LDAP role filter, which
 * queries Groups to which the user assigned and maps them directly to
 * roles.</p>
 * <p>
 * The following configuration options are available:</p>
 * <ul>
 * <li><b>debug</b> - Enables/disables debugging for the LoginModule.</li>
 * <li><b>host</b> - The hostname of the LDAP server.</li>
 * <li><b>port</b> - The port number of the LDAP server.</li>
 * <li><b>useSSL</b> - Enables/disables use of SSL to the LDAP server.</li>
 * <li><b>connectTimeout</b> - (Optional) The number of milliseconds to wait
 * for a connection to the LDAP server before timing out.</li>
 * <li><b>emptyPassword</b> - (Optional) Enable empty passwords (true or
 * false). By default, the X509CertificateLdapLoginModule does not allow a password to
 * be empty. Set this value to true to allow use of empty passwords
 * when anonymous binds are required.</li>
 * <li><b>userDN</b> - The distinguished name pattern to match to find a
 * user. The username callback values entered by the the user as login credentials
 * will be substituted for the values "{username}". For example:<br>
 * uid={username},ou=accounting,o=mycompany,c=US</li>
 * <li><b>defaultRoles</b> - (Optional) A comma-delimited list of one or
 * more roles that will be assigned to successfully authenticated users.
 * If not specified (or empty), no default roles will be assigned.</li>
 * <li><b>roleBaseDN</b> - The base distinguished name to use for your
 * LDAP schema role searches.</li>
 * <li><b>roleScope</b> - Specifies the LDAP search scope to use for
 * role searches. Valid values are BASE (search is done only at the base
 * dn tree node), ONE (search is done only at the first child tree node
 * of the base dn), and SUB (search is done at the base dn node and all
 * child nodes).</li>
 * <li><b>roleAttrName</b> - The name of the attribute used to store the
 * role names.</li>
 * <li><b>roleFilter</b> - The filter to use to search for a users roles.
 * The username entered by the the user at login will be substituted for
 * the value "{username}".</li>
 * <li><b>username_cb_name</b> - (Optional) By default, a "username" callback
 * is expected. If the login form uses a different name, then optionally
 * configure it, for example:
 *    <option name="username_cb_name" value="uid"/></li> 
 * <li><b>password</b> - (Optional) By default, a "password" callback
 * is expected. If the login form uses a different name, then optionally
 * configure it, for example:
 *    <option name="password_cb_name" value="pin"/></li>
 * <li><b>X509_cb_name</b> - (Optional) By default, a "X509" callback
 * is expected. If the login request uses a different name, then optionally
 * configure it, for example:
 *    <option name="X509_cb_name" value="X509_cert"/></li>
 * </ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.12 $ $Date: 2012/08/22 22:05:12 $ $Author: gary $
 * @see javax.security.auth.Subject
 * @see javax.security.auth.callback.CallbackHandler
 * @see javax.security.auth.spi.LoginModule
 * @since 07/10/2006
 */
public class X509CertificateLdapLoginModule
	implements LoginModule, LoggerClient
{
	//
	// static constants
	//

	/**
	 * The LoginModule name (used in log messages throughout the code).
	 */
	private static final String LMN = "[X509CertificateLdapLoginModule]";

	//
	// Instance Variables
	//

	/**
	 * The subject being authenticated.
	 */
	private Subject subject;

	/**
	 * Provides the means for accessing user supplied data.
	 */
	private CallbackHandler callbackHandler;

	/**
	 * Configuration options for this LoginModule.
	 */
	private Map options;

	/**
	 * Map that contains state shared with other configured LoginModules.
	 */
	private Map sharedState;

	/**
	 * Used to the LoginModule to log security domain-specific messages.
	 */
	private Logger logger;

	/**
	 * Flag that indicates if debug is enabled/disabled.
	 */
	private boolean debug;

	/**
	 * Flag indicating that SSL must be used.
	 */
	private boolean useSSL;

	/**
	 * Flag that indicates if user authentication succeeded.
	 */
	private boolean loginSucceeded;

	/**
	 * A flag used to indicate login commit() transaction succeeded.
	 */
	private boolean commitSucceeded;

	/**
	 * The username entered by the authenticating User.
	 */
	private String username;

	/**
	 * The password entered by the authenticating User.
	 */
	private String password;

	/**
	 * The X509Certificate chain presented by the authenticated User.
	 */
	private X509Certificate[] certChain;

	/**
	 * A list of user roles. This list is populated in the login method and
	 * used in the commit method.
	 */
	private List roleList;

	//
	// Constructor
	//

	/**
	 * Default X509CertificateLdapLoginModule Constructor.
	 */
	public X509CertificateLdapLoginModule()
	{
		this.callbackHandler = null;
		this.commitSucceeded = false;
		this.debug = false;
		this.useSSL = false;
		this.logger = StderrLogger.DEFAULT_STDERR_LOGGER;
		this.loginSucceeded = false;
		this.options = null;
		this.password = null;
		this.sharedState = null;
		this.subject = null;
		this.username = null;
		this.certChain = null;
		this.roleList = new ArrayList();
	}

	//
	// Implements the LoggerClient interface
	//

	/**
	 * Sets the logger.
	 *
	 * @param  logger logs messages
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	//
	// Implements the LoginModule interface
	//

	/**
	 * Initialize this LoginModule.
	 *
	 * <p> This method is called by the <code>LoginContext</code>
	 * after this <code>LoginModule</code> has been instantiated.
	 * The purpose of this method is to initialize this
	 * <code>LoginModule</code> with the relevant information.
	 * If this <code>LoginModule</code> does not understand
	 * any of the data stored in <code>sharedState</code> or
	 * <code>options</code> parameters, they can be ignored.
	 *
	 * @param subject the <code>Subject</code> to be authenticated.
	 * @param callbackHandler a <code>CallbackHandler</code> for communicating
	 *			with the end user (prompting for usernames and
	 *			passwords, for example).
	 * @param sharedState state shared with other configured LoginModules.
	 * @param options options specified in the login
	 *			<code>Configuration</code> for this particular
	 *			<code>LoginModule</code>.
	 */
	public void initialize(Subject subject, CallbackHandler callbackHandler,
		Map sharedState, Map options)
	{
		this.subject = subject;
		this.callbackHandler = callbackHandler;
		this.options = options;
		this.sharedState = sharedState;
		this.debug = "true".equals((String)options.get("debug"));
		this.useSSL = "true".equals((String)options.get("useSSL"));
	}

	/**
	 * Method to authenticate a <code>Subject</code> (phase 1).
	 *
	 * <p> The implementation of this method authenticates
	 * a <code>Subject</code>.  For example, it may prompt for
	 * <code>Subject</code> information such as a username and password
	 * and then attempt to verify the password. This method saves the
	 * result of the authentication attempt as private state within
	 * the LoginModule.
	 *
	 * @exception LoginException if the authentication fails
	 * @return true if the authentication succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean login() throws LoginException
	{
		return authenticate();
	}

	/**
	 * Method to commit the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication succeeded
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * succeeded).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method associates relevant
	 * Principals and Credentials with the <code>Subject</code> located in the
	 * <code>LoginModule</code>.  If this LoginModule's own
	 * authentication attempted failed, then this method removes/destroys
	 * any state that was originally saved.
	 *
	 * @exception LoginException if the commit fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean commit() throws LoginException
	{
		return associateRoles();
	}

	/**
	 * Method to abort the authentication process (phase 2).
	 *
	 * <p> This method is called if the LoginContext's
	 * overall authentication failed.
	 * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
	 * did not succeed).
	 *
	 * <p> If this LoginModule's own authentication attempt
	 * succeeded (checked by retrieving the private state saved by the
	 * <code>login</code> method), then this method cleans up any state
	 * that was originally saved.
	 *
	 * @exception LoginException if the abort fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean abort() throws LoginException
	{
		return abortAuthentication();
	}

	/**
	 * Method which logs out a <code>Subject</code>.
	 *
	 * <p>An implementation of this method might remove/destroy a Subject's
	 * Principals and Credentials.
	 *
	 * @exception LoginException if the logout fails
	 * @return true if this method succeeded, or false if this
	 *			<code>LoginModule</code> should be ignored.
	 */
	public boolean logout() throws LoginException
	{
		return logoutUser();
	}

	//
	// Private Methods
	//

	/**
	 * Authenticate the user.
	 * <p>
	 * This method executes the following steps:
	 * <ol>
	 * <li>Checks preconditions
	 * <li>Handles callback objects
	 * <li>Get a connection to the LDAP repository
	 * <li>Authenticates the users by attepting and LDAP bind with the
	 * username and password
	 * <li>Return the connection to its ObjectPool
	 * <li>Reset state if authentication failed
	 * <li>Return results of authentication attempt
	 * </ol>
	 *
	 * @return true if the authentication succeeds, false otherwise.
	 * @exception LoginException thrown if any error occurs while trying to
	 *	authenticate the user.
	 * @exception FailedLoginException if authentication fails.
	 */
	private boolean authenticate()
		throws LoginException, FailedLoginException
	{
		this.loginSucceeded = false;

		// Check preconditions
		validateOptions();

		// Handle callback objects.
		handleCallbacks();

		// A username must always be supplied
		if (username == null || username.length() == 0)
		{
			if (debug)
				logger.debug(this, "Authentication failed, missing username");

			throw new FailedLoginException(LMN +
				" Authentication failed");
		}

		// A non-empty password must be supplied unless empty
		// passwords are permitted for this configuration.
		String emptyPassword = (String)options.get("emptyPassword");
		if ((emptyPassword == null) || ("false".equals(emptyPassword.trim())))
		{
			if ("".equals(this.password))
			{
				if (debug)
					logger.debug(this, "Authentication failed, missing password " +
					"when emptyPassword=false");

				throw new FailedLoginException(LMN +
					" Authentication failed");
			}
		}
		else if (!("true".equals(emptyPassword.trim())))
		{
			throw new LoginException(LMN +
				"Option 'emptyPassword' must be 'true' or 'false'");
		}

		if (certChain == null || certChain.length == 0)
		  {
			if (debug)
				logger.debug(this, "Authentication failed, missing client X509 certificate");

			throw new FailedLoginException(LMN +
				" Authentication failed");
		  }

		// Get a connection to the LDAP repository
		LDAPConnection connection = getConnection();

		// Authenticate the users using an LDAP bind

		try
		{
			// Get the userDN
			String userDN = (String)options.get("userDN");

			// Get the user's fully qualifed dn
			String expandedDN = replaceToken(userDN, "username", username);

			if (debug)
				logger.debug(this, "Expanded userDN = " + expandedDN);

			// Bind to the user DN using the specified password
			if (bindAuthenticate(expandedDN, connection))
			   {
				if (debug)
					logger.debug(this, "Successful LDAP bind");

				// Compare the user specified X509 cert with the one in LDAP
				byte[] encodedClientCert = certChain[0].getEncoded();
				LDAPAttribute attr =
					new LDAPAttribute("usercertificate", encodedClientCert);

				if (connection.compare(expandedDN, attr))
				{
					if (debug)
						logger.debug(this, "User-provided certificate matches " +
							"the one in LDAP, assigning roles");

					// Populate a List of roles that will be assigned on commit
					populateRoleList(connection);

					if (debug)
						logger.debug(this, "login succeeded");

					 this.loginSucceeded = true;
				}
				else if (debug)
				{
					logger.debug(this, "User-provided certificate does not " +
						"match the one in LDAP");
				}
			}
			else if (debug)
			{
				logger.debug(this, "Unsuccessful LDAP bind");
			}
		}
		catch (LDAPException e)
		{
			if (e.getResultCode() == e.NO_SUCH_ATTRIBUTE)
			{
				// A usercertificate attribute was not found for this user,
				// fall through for normal failed authentication.
				logger.error(this, LMN +
					 " No usercertificate " +
					 "attribute for user '" + username + "'", e);
			}
			else
			{
				// An error occurred looking up roles for the user therefore login
				// did not succeed. Although this deviates slightly from the JAAS
				// implementation of the commit method it basically performs this
				// error check prior to the commit method allowing it to fail fast.
				logger.error(this, LMN, e);
				throw new LoginException(LMN + e.getMessage());
			}
		}
		catch (CertificateException e)
		{
			// An error occurred getting the encoded client certificate
			logger.error(this, LMN, e);
			throw new LoginException(LMN + e.getMessage());
		}
		finally
		{
			// Close the connection
			if (connection != null)
			{
				try
				{
					connection.disconnect();
					connection = null;
				}
				catch(LDAPException e)
				{
					// Catch the general exception here because we do not
					// need to distinguish between the exceptions that occurred.
					// We are just going to log the exception anyway.
					logger.error(this, LMN +
						" Error occurred while disconnecting", e);
				}
			}
		}

		// Reset state, authentication failed.
		if (!loginSucceeded)
		{
			this.username = null;
			this.password = null;
			this.certChain = null;

			throw new FailedLoginException(LMN + " Authentication failed");
		}

		// Return results of authentication attempt.
		return loginSucceeded;
	}

	/**
	 * Associate the roles with the user.
	 *
	 * @return true if roles successfully are assigned, false otherwise.
	 * @exception LoginException if an error occurs associating roles.
	 */
	private boolean associateRoles()
		throws LoginException
	{
		// The login failed, clean up state
		if (!loginSucceeded)
			return false;

		// Initialize ldap connection object
		ArrayList principalList = new ArrayList();

		// Repository operations to get role principals
		Iterator roleListItr = roleList.iterator();
		while(roleListItr.hasNext())
		{
			String rolename = (String)roleListItr.next().toString().trim();
			Principal rolePrincipal = new CSRolePrincipal(rolename);

			// Add the CSRolePrincipal to the Subject
			// only if it does not exist
			if (!subject.getPrincipals(CSRolePrincipal.class).contains(rolePrincipal))
			{
				principalList.add(rolePrincipal);

				if (debug)
					logger.debug(this,
						"Added CSRolePrincipal " + rolename +
						" to principal ArrayList for user '" + username + "'");
			}
			else
			{
				if (debug)
					logger.debug(this,
						"CSRolePrincipal '" + rolename +
						"' already in Subject for '" + username +
						"', not added to principal List");
			}
		}

		// If commit gets this far, there is always a relevant CSUserPrincipal
		// for this user, but only add it to the Subject if it does not exist
		Principal userPrincipal = new CSUserPrincipal(username);
		if (!subject.getPrincipals(CSUserPrincipal.class).contains(userPrincipal))
		{
			principalList.add(userPrincipal);

			if (debug)
				logger.debug(this, "Added CSUserPrincipal '" + username +
					"' to principal List");
		}
		else
		{
			if (debug)
				logger.debug(this, "CSUserPrincipal '" + username +
					"' already in Subject, not added to principal List");
		}

		// Add all principals to subject
		for (int i = 0; i < principalList.size(); i++)
			subject.getPrincipals().add((Principal)principalList.get(i));

		// Before we leave we must add to the Subject that the authentication
		// method was of the password variety.

		if(debug)
		{
			logger.debug(this, "Adding " +
				AuthenticationMethod.PASSWORD.getName() + " AuthenticationMethod");
			logger.debug(this, "Adding " +
				AuthenticationMethod.X509.getName() + " AuthenticationMethod");
		}

		subject.getPublicCredentials().add(AuthenticationMethod.PASSWORD);
		subject.getPublicCredentials().add(AuthenticationMethod.X509);

		this.commitSucceeded = true;

		// Clean up
		this.username = null;
		this.password = null;
		this.certChain = null;

		return commitSucceeded;
	}

	/**
	 * Abort authentication attempt.
	 *
	 * @return true if abort succeeded, false otherwise.
	 * @exception LoginException if any error occurs logging out.
	 */
	private boolean abortAuthentication()
		throws LoginException
	{
		// Authentication for this LoginModule did not succeed
		if (!loginSucceeded)
		{
			return false;
		}
		else if (!commitSucceeded)
		{
			// Authentication for this LoginModule succeeded, but commit failed.
			if (this.roleList != null)
			{
				this.roleList.clear();
				this.roleList = null;
			}

			this.loginSucceeded = false;
			this.username = null;
			this.password = null;
			this.certChain = null;
			this.logger = null;
		}
		else
		{
			// Authentication and commit for this LoginModule's succeeded, but
			// another relevant LoginModule failed
			logout();
		}

		return true;
	}

	/**
	 * Log an authenticated user out.
	 *
	 * @return true if logout succeeded, false otherwise.
	 * @exception LoginException if an error occurs logging out.
	 */
	private boolean logoutUser()
		throws LoginException
	{
		// Cleanup Object references
		if (this.subject != null)
		{
			this.subject.getPrincipals().clear();
			this.subject = null;
		}

		if (this.roleList != null)
		{
			this.roleList.clear();
			this.roleList = null;
		}

		this.logger = null;
		this.callbackHandler = null;
		this.username = null;
		this.password = null;
		this.certChain = null;
		this.options = null;
		this.loginSucceeded = false;
		this.commitSucceeded = false;
		this.debug = false;

		return true;
	}

	/**
	 * Creates Callbacks and executes the defined CallbackHandlers handle
	 * method. This CallbackHandler requires a username, password and X509
	 * digital certificate. 
	 *
	 * @exception LoginException if any error occurs executing the callbacks
	 */
	private void handleCallbacks()
		throws LoginException, FailedLoginException
	{
		// This LoginModule requires a username, password and X509 digital cert
		MapCallback mapCallback = new MapCallback();
		Callback[] callbacks = new Callback[1];
		callbacks[0] = mapCallback;

		try
		{
			// Populate callback values.
			callbackHandler.handle(callbacks);

			// Get the username callback value
			String username_cb_name = (String)options.get("username_cb_name");
			if ((username_cb_name == null) || ("".equals(username_cb_name)))
				username_cb_name = "username";

			this.username = ((MapCallback)callbacks[0]).getCallbackValue(username_cb_name);

			// Get the username callback value
			String password_cb_name = (String)options.get("password_cb_name");
			if ((password_cb_name == null) || ("".equals(password_cb_name)))
				password_cb_name = "password";

			this.password = ((MapCallback)callbacks[0]).getCallbackValue(password_cb_name);
			((MapCallback)callbacks[0]).setCallbackValue(password_cb_name, null);

			// Get the name of the callback value containing the client cert.
			// Set a default name if not provided.
			String X509_cb_name = (String)options.get("X509_cb_name");
			if ((X509_cb_name == null) || ("".equals(X509_cb_name)))
				X509_cb_name = "x509_client_cert";
			
			// Get the X.509 cert chain as a String.
			// Get the X.509 cert chain as a String.
			String certStr = mapCallback.getCallbackValue(X509_cb_name);
			if (certStr != null)
			{
				// Convert the cert callback value from a String to decoded form.
				this.certChain = X509CertificateUtils.decodeChainFromPEM(certStr);
			}
		}
		catch (IOException e)
		{
			logger.error(this, LMN + " Login error occurred", e);
			throw new LoginException(LMN + " Error: " + e.getMessage());
		}
		catch (UnsupportedCallbackException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available", e);

			throw new LoginException(LMN + " Error: " +
				e.getMessage() + ", " + e.getCallback().toString() + " not available");
		}
		catch (CertificateException e)
		{
			logger.error(this,
				LMN + " Login error occurred Callback not available", e);

			throw new LoginException(LMN + " Error: " + e.getMessage());
		}

		if (debug)
		{
			if (this.certChain != null)
			{
				for (int i = 0; i < certChain.length; i++)
					logger.debug(this, "X509Certificate[" + i + "]=\n" +
					certChain[i].toString());
			}
			else
			{
				logger.debug(this, "No certificates supplied");
			}
		}
	}


	/**
	 * Validate the LoginModules options Map object to ensure that all
	 * required properties exist and contain valid values.
	 *
	 * @exception LoginException if any of the following conditions occur:
	 * <UL>
	 * <LI>A CallbackHandler object is not available,
	 * <LI>the 'host' property is not available,
	 * <LI>the 'port' property is not available,
	 * <LI>the 'port' property is not a valid integer value,
	 * <LI>the port integer value is less than 1,
	 * <LI>the 'userDN' property is not available,
	 * <LI>the 'roleBaseDN' property is not available,
	 * <LI>the 'roleAttrName' property is not available,
	 * <LI>the 'roleFilter' property is not available,
	 * <LI>the 'roleScope' property is not available,
	 * <LI>the 'roleScope' is not a valid role scope value.
	 * </UL>
	 */
	private void validateOptions() throws LoginException
	{
		if (callbackHandler == null)
			throw new LoginException(LMN +
				"No CallbackHandler available");

		String host = (String)options.get("host");
		if ((host == null) || (host.trim().length() == 0))
			throw new LoginException(LMN +
				"Required 'host' option missing");

		String port = (String)options.get("port");
		if ((port == null) || (port.trim().length() == 0))
			throw new LoginException(LMN +
				"Required 'port' option missing");

		try
		{
			int portInt = Integer.parseInt(port);
			if (portInt < 1)
				throw new LoginException(LMN +
					"Required 'port' option is not an integer value");
		}
		catch (NumberFormatException e)
		{
			throw new LoginException(LMN +
				"Required 'port' option is not an integer value");
		}

		String userDN = (String)options.get("userDN");
		if ((userDN == null) || "".equals(userDN.trim()))
			throw new LoginException(LMN +
				"Required 'userDN' option missing");

		// Get and validate the roleBaseDN
		String roleBaseDN = (String)options.get("roleBaseDN");
		if ((roleBaseDN == null) || "".equals(roleBaseDN.trim()))
			throw new LoginException(LMN +
				"Required 'roleBaseDN' option missing");

		// Get and validate the roleAttrName
		String roleAttrName = (String)options.get("roleAttrName");
		if ((roleAttrName == null) || "".equals(roleAttrName.trim()))
			throw new LoginException(LMN +
				"Required 'roleAttrName' option missing");

		// Get and validate the roleFilter
		String roleFilter = (String)options.get("roleFilter");
		if ((roleFilter == null) || "".equals(roleFilter.trim()))
			throw new LoginException(LMN +
				"Required 'roleFilter' option missing");

		// Get and validate the scope
		String roleScope = (String)options.get("roleScope");
		if ((!"BASE".equalsIgnoreCase(roleScope)) &&
			(!"ONE".equalsIgnoreCase(roleScope)) &&
			(!"SUB".equalsIgnoreCase(roleScope)))
			throw new LoginException(LMN +
				"Required 'roleScope' option must be 'BASE', 'ONE', or 'SUB'");
	}

	/**
	 * Create a LDAPConnection.
	 *
	 * @return a LDAPConnection that is ready for use.
	 * @throws LoginException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getConnection()
		throws LoginException
	{
		try
		{
			// Get the host and port
			String host = (String)options.get("host");
			String port = (String)options.get("port");

			if (debug)
				logger.debug(this, "Instantiating LDAPConnection: host = " +
					host + ", port = " + port + ", useSSL = " + useSSL);

			// Get the connection timeout
			String connectTimeout = (String)options.get("connectTimeout");
			connectTimeout = (connectTimeout != null ? connectTimeout : "1000");

			// Create the connection
			LDAPConnection connection = null;
			if (useSSL)
			{
				// Warn if required SSL System property for trust store is not set
				String trustStoreProp =
					System.getProperty("javax.net.ssl.trustStore");
				if (trustStoreProp == null)
					logger.warning(this,
						"Missing required SSL system property: " +
						"javax.net.ssl.trustStore");

				// Use an SSL connection
				connection =
					new LDAPConnection(new LDAPJSSESecureSocketFactory());
			}
			else
			{
				// Use a non-SSL connection
				connection = new LDAPConnection();
			}

			// Create a new LDAPSocketFactory that will delegate to the
			// default factory, but also provides connection timeout
			LDAPSocketFactory socketFactory = new TimeoutLDAPSocketFactory(
				connection.getSocketFactory(), Long.parseLong(connectTimeout));

			connection = null;
			connection = new LDAPConnection(socketFactory);

			if (debug)
				logger.debug(this, "Connecting to host=" + host + ", port=" + port);

			// Connect to the directory
			connection.connect(host, Integer.parseInt(port));

			if (debug)
				logger.debug(this, "Successfully connected to host=" + host +
					", port=" + port);

			if (debug)
				logger.debug(this, "Returning LDAPConnection");

			return connection;
		}
		catch (LDAPException e)
		{
			logger.error(this, "LDAP error", e);

			// Dump the nested Throwables to get at the root cause
			int i = 1;
			Throwable cause = e.getCause();
			do
			{
				if (cause != null)
					logger.error(this, "LDAP error CAUSE: " + i + ": " +
						cause.getMessage());
			}
			while ((cause = cause.getCause()) != null);

			throw new LoginException(LMN + e.getMessage());
		}
	}

	/**
	 * Authenticate using LDAP server bind.
	 *
	 * @param usernameDN the username's distinguished name.
	 * @param connection the LDAPConnection.
	 * @exception CredentialExpiredException if bind password is expired.
	 * @return true if authentication succeeded, false otherwise.
	 */
	private boolean bindAuthenticate(
		String usernameDN, LDAPConnection connection)
		throws AccountExpiredException, CredentialExpiredException
	{
		try
		{
			connection.bind(3, usernameDN, password.getBytes());
			return true;
		}
		catch(LDAPException e)
		{
			// Check for credential (password) expired exception. Note the
			// message "password expired!" is likely to be specific to the
			// SunOne LDAP server.
			if (e.getResultCode() == e.INVALID_CREDENTIALS)
				if ((e.getLDAPErrorMessage() != null) &&
					(e.getLDAPErrorMessage().equalsIgnoreCase("password expired!")))
					throw new CredentialExpiredException(
						LMN + e.getMessage() + "\n");

			// Check for account expired exception, due to retry violation. The
			// message "Exceed password retry" is likely to be specific to the
			// SunOne LDAP server.
			if (e.getResultCode() == e.CONSTRAINT_VIOLATION)
				if ((e.getLDAPErrorMessage() != null) &&
					(e.getLDAPErrorMessage().startsWith("Exceed password retry")))
					throw new AccountExpiredException(
						LMN + e.getMessage() + "\n");

			return false;
		}
	}

	/**
	 * Populate a List of roles that will be assigned to the user on commit.
	 *
	 * @param connection the LDAPConnection, which must have successfully
	 * 	executed a bind() request for the authenticating user.
	 * @exception LDAPException if an error searching the LDAP directory for
	 *	the assigned user roles
	 */
	private void populateRoleList(LDAPConnection connection)
		throws LDAPException
	{
		// Add default roles (if any) to list of roles
		String defaultRoles = (String)options.get("defaultRoles");
		if ((defaultRoles != null) && !("".equals(defaultRoles.trim())))
		{
			if (debug)
				logger.debug(this, "Adding defaultRoles: " + defaultRoles);

			StringTokenizer st = new StringTokenizer(defaultRoles, ",");
			while(st.hasMoreTokens())
			   {
				String roleName = st.nextElement().toString().trim();
				roleList.add(roleName);
			   }
		}
		else if (debug)
		{
			   logger.debug(this, "No defaultRoles to be added");
		}

		// Get the roles for use later during the commit method.
		String roleBaseDN = (String)options.get("roleBaseDN");
		String roleAttrName = (String)options.get("roleAttrName");
		String roleFilter = (String)options.get("roleFilter");
		String roleScope = (String)options.get("roleScope");

		// Set the search scope, the default of 0 is "BASE"
		int scopeInt = LDAPConnection.SCOPE_BASE;
		if ("ONE".equalsIgnoreCase(roleScope))
			scopeInt = LDAPConnection.SCOPE_ONE;
		else if ("SUB".equalsIgnoreCase(roleScope))
			scopeInt = LDAPConnection.SCOPE_SUB;

		// Set the search pattern to find roles
		String filter = replaceToken(roleFilter, "username", username);

		// Set the base dn where the search will be performed, and the
		// name of the role to use.
		String[] attr = {roleAttrName};

		if (debug)
			logger.debug(this, "Searching LDAP for assigned roles using:\n" +
				"\t roleBaseDN=" + roleBaseDN + "\n" +
				"\t scopeInt=" + scopeInt + "\n" +
				"\t filter=" + filter + "\n" +
				"\t attr=" + attr[0]);

		// Do the search.
		LDAPSearchResults searchResults = connection.search(
			roleBaseDN, scopeInt, filter, attr, false);

		if (debug)
			logger.debug(this, "LDAP assigned roles found: " +
				searchResults.hasMore());

		// Process the search results to get the role names
		//   -- The first while loop filters for the entries
		//   -- The second while loop filters for the rolenames
		//      and stores them to a principal and subject
		while (searchResults.hasMore())
		{
			LDAPEntry nextEntry = searchResults.next();
			LDAPAttribute attribute = nextEntry.getAttribute(roleAttrName);

			Enumeration allValues = attribute.getStringValues();
			while (allValues.hasMoreElements())
			{
				String role = (String)allValues.nextElement();

				if (debug)
					logger.debug(this, "Adding LDAP assigned role: " + role);

				roleList.add(role);
			}
		}
	}

	/**
	 * Return the input pattern replaced with the match value.
	 *
	 * @param pattern The input pattern to use
	 * @param matchName The variable name to match
	 * @param matchValue The value to replace
	 * @return The input pattern with matched value
	 */
	private String replaceToken(String pattern, String matchName,
		String matchValue)
	{
		String tmp = null;
		StringBuffer sb = new StringBuffer();
		StringTokenizer st = new StringTokenizer(pattern, "{}");
		while (st.hasMoreTokens())
		{
			tmp = st.nextToken();

			if (tmp.trim().equals(matchName))
				sb.append(matchValue);
			else
				sb.append(tmp);
		}

		return sb.toString();
	}
} // End of class : X509CertificateLdapLoginModule

