/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package examples.session;

import com.cafesoft.cams.Config;
import com.cafesoft.cams.ConfigException;

import com.cafesoft.cams.auth.SubjectUtils;

import com.cafesoft.cams.session.ManagedSession;
import com.cafesoft.cams.session.ManagedSessionEvent;
import com.cafesoft.cams.session.ManagedSessionEventType;
import com.cafesoft.cams.session.ManagedSessionEventHandler;

import com.cafesoft.core.log.Logger;

import com.cafesoft.core.util.Base64;

import javax.security.auth.Subject;

import java.net.PasswordAuthentication;

import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * <code>HttpBasicAuthManagedSessionEventHandler</code> demonstrates
 * how a <code>java.net.PasswordAuthentication</code> token added to an
 * an authenticated <code>Subject</code>s private credential store can be
 * converted to an HTTP Basic Authorization header value and set as a session
 * attribute.  This value may then be used by Cams web agents supporting
 * HTTP Basic Authorization header insertion for user impersonation.
 *
 * @version $Revision: 1.5 $ $Date: 2009/08/14 23:52:25 $
 * @author Norbert K. Kuhnert
 */
public class HttpBasicAuthManagedSessionEventHandler
	implements ManagedSessionEventHandler
{
	/** The Config object provided at initialization */
	private Config config;

	/** The flag that enables/disables DEBUG-level messages. */
	private boolean debug = false;

	/**
	 * The name of the session attribute to be populated.
	 */
	private String sessionAttrName = "basic_authorization";

	/**
	 * The namespace of the session attribute to be populated.
	 */
	private String sessionAttrNamespace = "private";

	/**
	 * A HashSet used to exclude specific usernames from having the session
	 * attribute value set.
	 */
	private Set excludedUsersSet = new HashSet(1);

	/**
	 * Initialize the ManagedSessionEventHandler.
	 *
	 * @param config a Config object that provides access to configuration
	 * 	parameters and a runtime Context.
	 */
	public void initialize(Config config)
		throws ConfigException
	{
		this.config = config;

		// Just dump the Config/Context to the supplied Logger
		Logger logger = config.getContext().getLogger();
		logger.info(this, "... initializing: class=" +
			this.getClass().getName());

		Iterator i = config.getInitParameterNames();
		if (i.hasNext())
		{
			while (i.hasNext())
			{
				String name = (String)i.next();

				if (debug)
					logger.debug(this, "   ... param: " +
						name + "=" +
						config.getInitParameter(name));

				if ("session-attr.namespace".equals(name))
				{
					this.sessionAttrNamespace = config.getInitParameter(name);
				}
				else if ("session-attr.name".equals(name))
				{
					this.sessionAttrName = config.getInitParameter(name);
				}
				else if ("excluded-users".equals(name))
				{
					String excludedUsers = config.getInitParameter(name);
					if ((excludedUsers != null) && (! "".equals(excludedUsers.trim())))
		 			{
						StringTokenizer st = new StringTokenizer(excludedUsers, ",");
						while (st.hasMoreTokens())
		  				{
							String username = st.nextToken();

							if (debug)
								logger.debug(this,
									"adding user to excluded-users set: " +
				  					username);

							excludedUsersSet.add(username);
		  				}
					}
				}
				else
				{
					if (debug)
						logger.debug(this,
							"ignoring unrecognized initialiation parameter: " +
		  					name);
				}
			}
		}
		else
		{
			if (debug)
				logger.debug(this, "   ... no initialization params");
		}

		logger.info(this, "... using parameters: " +
			"session-attr.namespace=" + this.sessionAttrNamespace +
			", session-attr.name=" + this.sessionAttrName);

		logger.info(this, "... initialized: class=" +
			this.getClass().getName());
	}

	/**
	 * Enable or disable debugging.
	 *
	 * @param enable if true, enable debugging, else disable debugging.
	 * @return the previous value of the debug flag, which is useful for
	 *	restoring the debug state if temporarily changed.
	 */
	public boolean setDebug(boolean enable)
	{
		boolean prevDebug = this.debug;
		this.debug = enable;
		return prevDebug;
	}

	/**
	 * Handle a ManagedSessionEvent.
	 *
	 * @param event the ManagedSessionEvent to be handled.
	 */
	public void handleManagedSessionEvent(ManagedSessionEvent event)
	{
		Logger logger = config.getContext().getLogger();

		if (debug)
			logger.debug(this, "... handling ManagedSessionEvent: class=" +
				this.getClass().getName());

		ManagedSession session = event.getManagedSession();

		// If the session was just created, then look for the Subject's
		// PasswordAuthentication token.
		if (event.getEventType() == ManagedSessionEventType.SESSION_CREATED)
		{
			Subject subject = session.getSubject();

			if (debug)
				logger.debug(this, "... checking list of excluded users");

			String username = SubjectUtils.getUserName(subject);
			if (username != null)
	   		{
				if (excludedUsersSet.contains(username))
				{
					if (debug)
						logger.debug(this, "... found excluded user: " +
							username + ", no session attribute will be set");
					return;
				}
			}

			if (debug)
				logger.debug(this, "... looking for PasswordAuthentication in " +
					"private credentials set");

			Set privateSet = subject.getPrivateCredentials(
				PasswordAuthentication.class);
			PasswordAuthentication[] pwArray =
				(PasswordAuthentication[])privateSet.toArray(
					new PasswordAuthentication[0]);
			if (pwArray.length > 0)
			{
				if (debug)
					logger.debug(this, "... found PasswordAuthentication token");

				PasswordAuthentication pwAuth = pwArray[0];

				// Format the the Basic Authorization header value
				StringBuffer buff1 = new StringBuffer(pwAuth.getUserName());
				buff1.append(":");
				buff1.append(pwAuth.getPassword());
				String base64Encoded = new String(
					Base64.encode(buff1.toString().getBytes()));

				StringBuffer basicHeaderValue = new StringBuffer("Basic ");
				basicHeaderValue.append(base64Encoded);

				if (debug)
					logger.debug(this, "... adding session attribute " +
						"namespace=" + this.sessionAttrNamespace +
						", name=" + this.sessionAttrName +
						basicHeaderValue.toString());

				session.putAttribute("private", "basic_authorization",
					basicHeaderValue.toString());
			}
			else
			{
				if (debug)
					logger.debug(this, "... did not find PasswordAuthentication token");
			}
		}

		if (debug)
			logger.debug(this, "... handled ManagedSessionEvent: class=" +
				this.getClass().getName());
	}

} // End of class: HttpBasicAuthManagedSessionEventHandler


