/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package examples.session;

import com.cafesoft.cams.Config;
import com.cafesoft.cams.ConfigException;

import com.cafesoft.cams.auth.SubjectUtils;

import com.cafesoft.cams.session.ManagedSession;
import com.cafesoft.cams.session.ManagedSessionEvent;
import com.cafesoft.cams.session.ManagedSessionEventType;
import com.cafesoft.cams.session.ManagedSessionEventHandler;

import com.cafesoft.core.log.Logger;

import com.cafesoft.core.util.Base64;

import javax.security.auth.Subject;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;

import java.net.PasswordAuthentication;

import java.util.Iterator;
import java.util.HashSet;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * <code>JdbcHttpBasicAuthManagedSessionEventHandler</code> uses the 
 * authenticated Subject name and an SQL query to lookup Cams JITSO
 * credentials for HTTP Basic authentication. The username and password
 * queried via JDBC are use to format an HTTP Basic Authorization header 
 * value and set as a session attribute.  This value may then be used by 
 * Cams web agents supporting HTTP Basic Authorization header insertion 
 * for user impersonation.
 *
 * Configuration parameters are:
 *
 * <ul>
 * <li>session-attr.namespace - (optional) The namespace into which the HTTP Basic
 *      authorization header value is to be added as a Cams session attribute.
 *      The default value is "private", which ensures that Cams agents won't make
 *      the attribute visible as a Cams HTTP request header</li>
 * <li>session-attr.name - (optional) The name of the Cams session attribute
 *      corresponding to the HTTP Basic header value. The default value is:
 *      "basic_authorization"</li>
 * <li>excluded-users - (optional) A comma-delimited list of one or more authenticated
 * 		users to exclude. You should include the username used for authenticating
 * 		Cams webagents and special users (e.g. "admin") for whom resources requiring
 * 		HTTP Basic authentication are not available.</li>
 * <li>jdbcDriver - The JDBC driver class</li>
 * <li>jdbcUrl - The database connection url</li>
 * <li>jdbcUser - The database connection username</li>
 * <li>jdbcPassword - The database connection password</li>
 * <li>sql - The SQL query string to retrieve the HTTP Basic username and password
 *      values. /value pairs</li>
 * <li>usernameColumn - The name of the result set column containing the HTTP Basic
 *      username for the authenticated Cams subject</li>
 * <li>PasswordColumn - The name of the result set column containing the HTTP Basic
 *      password for the authenticated Cams subject</li>
 *
 * @version $Revision: 1.2 $ $Date: 2009/07/24 16:02:34 $
 * @author Norbert K. Kuhnert
 */
public class JdbcHttpBasicAuthManagedSessionEventHandler
	implements ManagedSessionEventHandler
{
	/** The Config object provided at initialization */
	private Config config;

	/** The flag that enables/disables DEBUG-level messages. */
	private boolean debug = false;

	/**
	 * The name of the session attribute to be populated.
	 */
	private String sessionAttrName = "basic_authorization";

	/**
	 * The namespace of the session attribute to be populated.
	 */
	private String sessionAttrNamespace = "private";

	/**
	 * A HashSet used to exclude specific usernames from having the session
	 * attribute value set.
	 */
	private Set excludedUsersSet = new HashSet(1);

	/** The loaded JDBC Driver. */
	private Driver driver;

	/** The loaded JDBC username/password Connection Properties. */
	private Properties jdbcProp;

	/**
	 * Initialize the ManagedSessionEventHandler.
	 *
	 * @param config a Config object that provides access to configuration
	 * 	parameters and a runtime Context.
	 */
	public void initialize(Config config)
		throws ConfigException
	{
		this.config = config;

		// Just dump the Config/Context to the supplied Logger
		Logger logger = config.getContext().getLogger();
		logger.info(this, "... initializing: class=" +
			this.getClass().getName());

		// Check for required configuration parameters
		String jdbcDriver = config.getInitParameter("jdbcDriver");
		if ((jdbcDriver == null) || "".equals(jdbcDriver.trim()))
			throw new ConfigException("Missing required parameter 'jdbcDriver'");

		String jdbcUrl = config.getInitParameter("jdbcUrl");
		if ((jdbcUrl == null) || "".equals(jdbcUrl.trim()))
			throw new ConfigException("Missing required parameter 'jdbcUrl'");

		String sql = config.getInitParameter("sql");
		if ((sql == null) || "".equals(sql.trim()))
			throw new ConfigException("Missing required parameter 'sql'");

		String usernameColumn = config.getInitParameter("usernameColumn");
		if ((usernameColumn == null) || "".equals(usernameColumn.trim()))
			throw new ConfigException("Missing required parameter 'usernameColumn'");

		String passwordColumn = config.getInitParameter("passwordColumn");
		if ((passwordColumn == null) || "".equals(passwordColumn.trim()))
			throw new ConfigException("Missing required parameter 'passwordColumn'");

		// Make sure the JDBC Driver loads
		try
		{
			this.driver = (Driver)Class.forName(jdbcDriver).newInstance();
		}
		catch (Exception e)
		{
			throw new ConfigException(
				"Error loading jdbcDriver: " + jdbcDriver, e);
		}

		// Set additional, sometimes optional jdbc parameters
		this.jdbcProp = new Properties();
		jdbcProp.put("user", config.getInitParameter("jdbcUser", ""));
		jdbcProp.put("password", config.getInitParameter("jdbcPassword", ""));

		// Set optional parameters.
		this.sessionAttrNamespace = config.getInitParameter(
			"session-attr.namespace", "private");
		this.sessionAttrName = config.getInitParameter(
			"session-attr.name", "basic_authorization");

		String excludedUsers = config.getInitParameter("excluded-users");
		if ((excludedUsers != null) && (! "".equals(excludedUsers.trim())))
		{
			StringTokenizer st = new StringTokenizer(excludedUsers, ",");
			while (st.hasMoreTokens())
			{
				String username = st.nextToken();

				if (debug)
					logger.debug(this,
						"adding user to excluded-users set: " + username);

				excludedUsersSet.add(username);
			}
		}

		logger.info(this, "... initialized: class=" +
			this.getClass().getName());
	}

	/**
	 * Enable or disable debugging.
	 *
	 * @param enable if true, enable debugging, else disable debugging.
	 * @return the previous value of the debug flag, which is useful for
	 *	restoring the debug state if temporarily changed.
	 */
	public boolean setDebug(boolean enable)
	{
		boolean prevDebug = this.debug;
		this.debug = enable;
		return prevDebug;
	}

	/**
	 * Handle a ManagedSessionEvent.
	 *
	 * @param event the ManagedSessionEvent to be handled.
	 */
	public void handleManagedSessionEvent(ManagedSessionEvent event)
	{
		Logger logger = config.getContext().getLogger();

		if (debug)
			logger.debug(this, "... handling ManagedSessionEvent: class=" +
				this.getClass().getName());

		ManagedSession session = event.getManagedSession();

		// The session was just created.
		if (event.getEventType() == ManagedSessionEventType.SESSION_CREATED)
		{
			Subject subject = session.getSubject();

			// Ignore excluded users

			if (debug)
				logger.debug(this, "... checking list of excluded users");

			String subjectName = SubjectUtils.getUserName(subject);
			if (subjectName != null)
	   		{
				if (excludedUsersSet.contains(subjectName))
				{
					if (debug)
						logger.debug(this, "... found excluded user: " +
							subjectName + ", no session attribute will be set");
					return;
				}
			}

			if (debug)
				logger.debug(this, "... querying HTTP Basic username and password " +
					"for authenticated subject: " + subjectName);

			// Get the PasswordAuthentication credentials for user.
			PasswordAuthentication pwAuth = queryPasswordAuthentication(subjectName);
			if (pwAuth != null)
			{
				if (debug)
					logger.debug(this, "... found PasswordAuthentication token");
			}
			else
			{
				if (debug)
					logger.debug(this, "... did not find PasswordAuthentication token " +
						"for authenticated subject: " + subjectName + 
						", no session attribute will be set");
				return;
			}

			// Format the the Basic Authorization header value
			StringBuffer buff1 = new StringBuffer(pwAuth.getUserName());
			buff1.append(":");
			buff1.append(pwAuth.getPassword());
			String base64Encoded = new String(
				Base64.encode(buff1.toString().getBytes()));

			StringBuffer basicHeaderValue = new StringBuffer("Basic ");
			basicHeaderValue.append(base64Encoded);

			if (debug)
				logger.debug(this, "... adding session attribute " +
					"namespace=" + this.sessionAttrNamespace +
					", name=" + this.sessionAttrName +
					basicHeaderValue.toString());

			session.putAttribute("private", "basic_authorization",
				basicHeaderValue.toString());
		}

		if (debug)
			logger.debug(this, "... handled ManagedSessionEvent: class=" +
				this.getClass().getName());
	}

	/**
	 * Query the HTTP Basic username and password for a given authenticated subject
	 * username and return the credentials as a PasswordAuthentication object.
	 *
	 * @param subjectName the username associated with the Cams/JAAS authenticated Subject.
	 * @return a PasswordAuthentication object or null if no credentials are available for
	 * 	the authenticated subject or if an error occurs during the query.
	 */
	private PasswordAuthentication queryPasswordAuthentication(String subjectName)
	{
		Logger logger = config.getContext().getLogger();

		PasswordAuthentication pwAuth = null;
		Connection c = null;
		PreparedStatement s = null;
		ResultSet r = null;

		try
		{
			// Connect to the database
			c = driver.connect(config.getInitParameter("jdbcUrl"), jdbcProp);

			// Execute and process the SQL query
			s = c.prepareStatement(config.getInitParameter("sql"));
			s.setString(1, subjectName);
			r = s.executeQuery();

			if (r.next())
			{
				String basicAuthUsername = r.getString(
					config.getInitParameter("usernameColumn"));
				String basicAuthPassword = r.getString(
					config.getInitParameter("passwordColumn"));

				if ((basicAuthUsername != null) && 
					(basicAuthPassword != null))
				{
					if (debug)
						logger.debug(this, "HTTP Basic username and password " +
							"obtained for authenticated subject: " + subjectName);
		
					// Create the PasswordAuthentication instance
					pwAuth = new PasswordAuthentication(
						basicAuthUsername, basicAuthPassword.toCharArray());
				}
				else
				{
					if (debug)
						logger.debug(this, "HTTP Basic username and/or password " +
							"are null for authenticated subject: " + subjectName);
				}
			}
			else
			{
				logger.info(this, "HTTP Basic username and password " +
					"not found for subject: " + subjectName);
			}
		}
		catch (SQLException e)
		{
			logger.error(this, "error querying HTTP Basic username and password " +
				"for subject: " + subjectName + ", Error=" + e);
		}
		finally
		{
			try
			{
				if (r != null)
					r.close();

				if (s != null)
					s.close();

				if (c != null)
					c.close();
			}
			catch (SQLException e)
			{
				logger.error(this, "error closing database connection " + e);
			}
		}

		return pwAuth;
	}

} // End of class: HttpBasicAuthManagedSessionEventHandler


