/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose
 * such Confidential Information and shall use it only in accordance
 * with the terms of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 * OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
 * BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
 * OR ITS DERIVATIVES.
 */
package examples.session;

import com.cafesoft.cams.Config;
import com.cafesoft.cams.ConfigException;

import com.cafesoft.cams.session.ManagedSession;
import com.cafesoft.cams.session.ManagedSessionEvent;
import com.cafesoft.cams.session.ManagedSessionEventType;
import com.cafesoft.cams.session.ManagedSessionEventHandler;

import com.cafesoft.cams.auth.SubjectUtils;

import com.cafesoft.core.log.Logger;

import java.util.Properties;
import java.util.StringTokenizer;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;

import java.sql.Timestamp;

/**
 * A handler for <code>ManagedSessionEvents</code> that queries the last
 * login date/time for a user and updates the last login date/time with the
 * current session's creation date/time.
 *
 * <ul>
 * <li>jdbcDriver - The JDBC driver class</li>
 * <li>url - The database connection url</li>
 * <li>user - (optional) The database connection username</li>
 * <li>password - (optional) The database connection password</li>
 * <li>queryPreparedStatement - The SQL prepared statement to fetch the last
 * login date. For example:
 * 	"SELECT LAST_LOGIN FROM USER_ACCOUNTS WHERE USER_NAME = ?".
 * The username will be substituted for the "?" in the prepared statement.</li>
 * <li>updatePreparedStatement - The SQL prepared statement to update the last
 * login date. For example:
 * 	"UPDATE USER_ACCOUNTS SET LAST_LOGIN = ? WHERE USER_NAME = ?".
 * The current session date/time will be substituted for the first "?" in
 * the prepared statement and the userName will be substituted for second.</li>
 * for the second in the prepared statement.</li>
 * <li>nameSpace - The session attribute's namespace</li>
 * <li>excludeUsers - (optional) A comma-delimited list of one or more user's
 * to exclude</li>
 * </ul>
 *
 * @version $Revision: 1.4 $ $Date: 2009/08/14 23:52:25 $
 */
public class JdbcLastLoginManagedSessionEventHandler
	implements ManagedSessionEventHandler
{
	/** The Config object provided at initialization */
	private Config config;

	/** The flag that enables/disables DEBUG-level messages.  */
	private boolean debug = false;

	/** Logs messages.  */
	private Logger logger;

	/** The loaded JDBC Driver. */
	private Driver driver;

	/** The loaded JDBC username/password Connection Properties. */
	private Properties jdbcProp;

	/**
	 * Initialize the ManagedSessionEventHandler.
	 *
	 * @param config a Config object that provides access to configuration
	 * 	parameters and a runtime Context.
	 */
	public void initialize(Config config) throws ConfigException
	{
		this.config = config;
		this.logger = config.getContext().getLogger();

		// Check for required configuration parameters
		String jdbcDriver = config.getInitParameter("jdbcDriver");
		if ((jdbcDriver == null) || "".equals(jdbcDriver.trim()))
			throw new ConfigException("Missing required parameter 'jdbcDriver'");

		String url = config.getInitParameter("url");
		if ((url == null) || "".equals(url.trim()))
			throw new ConfigException("Missing required parameter 'url'");

		String queryPreparedStatement =
			config.getInitParameter("queryPreparedStatement");
		if ((queryPreparedStatement == null) ||
			("".equals(queryPreparedStatement.trim())))
			throw new ConfigException(
			"Missing required parameter 'queryPreparedStatement'");

		String updatePreparedStatement =
			config.getInitParameter("updatePreparedStatement");
		if ((updatePreparedStatement == null) ||
			("".equals(updatePreparedStatement.trim())))
			throw new ConfigException(
				"Missing required parameter 'updatePreparedStatement'");

		String nameSpace = config.getInitParameter("nameSpace");
		if ((nameSpace == null) || "".equals(nameSpace.trim()))
			throw new ConfigException("Missing required parameter 'nameSpace'");

		// Make sure the JDBC Driver loads
		try
		{
			this.driver = (Driver)Class.forName(jdbcDriver).newInstance();
		}
		catch (Exception e)
		{
			throw new ConfigException(
				"Error loading jdbcDriver: " + jdbcDriver, e);
		}

		// Create the JDBC Connection Properties (user/password)
		// so we don't have to do it for every Session created
		this.jdbcProp = new Properties();
		jdbcProp.put("user", config.getInitParameter("user", ""));
		jdbcProp.put("password", config.getInitParameter("password", ""));
	}

	/**
	 * Enable or disable debugging.
	 *
	 * @param enable if true, enable debugging, else disable debugging.
	 * @return the previous value of the debug flag, which is useful for
	 *	restoring the debug state if temporarily changed.
	 */
	public boolean setDebug(boolean enable)
	{
		boolean prevDebug = this.debug;
		this.debug = enable;
		return prevDebug;
	}

	/**
	 * Handle a ManagedSessionEvent.
	 *
	 * @param event the ManagedSessionEvent to be handled.
	 */
	public void handleManagedSessionEvent(ManagedSessionEvent event)
	{
		if(debug)
		{
			logger.debug(this, "Handling ManagedSessionEvent: class=" +
						getClass().getName());
		}

		// If the session was just created, then add an attribute.
		if (event.getEventType() == ManagedSessionEventType.SESSION_CREATED)
		{
			// Get the user's name
			ManagedSession session = event.getManagedSession();
			String userName = SubjectUtils.getUserName(session.getSubject());

			// If a user is excluded from this event, ignore it!
			String excludeUsers = config.getInitParameter("excludeUsers");
			if ((excludeUsers != null) && (excludeUsers.length() > 0))
			{
				StringTokenizer st = new StringTokenizer(excludeUsers, ",");

				while (st.hasMoreTokens())
				{
					if (st.nextToken().equals(userName))
					{
						if (debug)
							logger.debug(this,
										 "Ignoring user '" + userName + "'");

						return;
					}
				}
			}

			Connection c = null;
			PreparedStatement s = null;
			ResultSet r = null;

			try
			{
				// Connect to the database
				c = driver.connect(config.getInitParameter("url"), jdbcProp);

				// Execute the last login query prepared statement
				s = c.prepareStatement(
					config.getInitParameter("queryPreparedStatement"));
				s.setString(1, userName);
				r = s.executeQuery();
				if (r.next())
				{
					ResultSetMetaData rsmd = r.getMetaData();

					// Get the account balance for the user
					String value = r.getString(1);

					// Persist the value into the session object
					session.putAttribute(config.getInitParameter("nameSpace"),
						rsmd.getColumnLabel(1), value);
				}
				else
				{
					logger.info(this, "Last login value for '" + userName +
						"' does not exist");
				}

				// Execute the last login update prepared statement
				s = c.prepareStatement(
					config.getInitParameter("updatePreparedStatement"));
				s.setTimestamp(1, new Timestamp(session.getCreationTime()));
				s.setString(2, userName);
				s.executeUpdate();
			}
			catch (SQLException e)
			{
				logger.error(this,
					"error querying or updating last login value for " +
					userName + e);
			}
			finally
			{
				try
				{
					if (r != null)
						r.close();

					if (s != null)
						s.close();

					if (c != null)
						c.close();
				}
				catch (SQLException e)
				{
					logger.error(this, "error closing database connection " + e);
				}
			}
		}

		if(debug)
		{
			logger.debug(this, "Handled ManagedSessionEvent: class=" +
				this.getClass().getName());
		}
	}

} // End of class: JdbcLastLoginManagedSessionEventHandler
