/**
 * Copyright (c) 1996-2011 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose
 * such Confidential Information and shall use it only in accordance
 * with the terms of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 * OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
 * BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
 * OR ITS DERIVATIVES.
 */
package examples.session;

import com.cafesoft.cams.Config;
import com.cafesoft.cams.ConfigException;

import com.cafesoft.cams.session.ManagedSession;
import com.cafesoft.cams.session.ManagedSessionEvent;
import com.cafesoft.cams.session.ManagedSessionEventType;
import com.cafesoft.cams.session.ManagedSessionEventHandler;

import com.cafesoft.cams.auth.SubjectUtils;

import com.cafesoft.cams.session.SessionId;
import com.cafesoft.core.log.Logger;

import java.lang.NumberFormatException;

import java.util.Properties;
import java.util.StringTokenizer;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.PreparedStatement;
import java.sql.SQLException;

/**
 * A JDBC handler for <code>ManagedSessionEvents</code> that writes and removes
 * sessions from a relational database based on create, expire and close managed
 * session events. The database schema example for MySQL is:
 * 
 * CREATE TABLE `session_tracker_log` (
 * `session_id` varchar(128) NOT NULL default '',
 * `security_domain` varchar(20) NOT NULL default '',
 * `user_name` varchar(45) NOT NULL default '',
 * `created` datetime NOT NULL default '0000-00-00 00:00:00',
 * PRIMARY KEY  (`session_id`),
 * UNIQUE KEY `session_id` (`session_id`)
 *) ENGINE=MyISAM DEFAULT CHARSET=latin1
 *
 * Configuration parameters are:
 *
 * <ul>
 * <li>driver - The JDBC driver class</li>
 * <li>url - The database JDBC connection url</li>
 * <li>user - The database connection username</li>
 * <li>password - The database connection password</li>
 * <li>purgeSessionsSQL - The SQL to purge sessions at Cams policy server 
 * startup. In the event of an abrupt shutdown, sessions may exist in the database
 * that are no longer valid. This prepared statement is used to do a one time
 * purge of such sessions. Care should be taken to utilize a query that only
 * purges sessions created by the given policy server instance and security 
 * domain within the instance. For example: 
 * "DELETE FROM session_tracker_log WHERE session_id * like 'MyCamsCluster-MyCamsServer-system%'"
 * </li>
 * <li>createSessionSQL - The SQL to write a create session event to the 
 * database. Parameters are required to write the session id, security domain,
 * user name and creation time. For example:
 * "INSERT INTO session_tracker_log (session_id, security_domain, user_name, created) VALUES (?,?,?,?)"
 * </li>
 * <li>deleteSessionSQL - The SQL to delete an existing session from the database
 * based on a session CLOSED or EXPIRED event. Matching is done on the session id,
 * which must by definition be unique. For example:
 * "DELETE FROM session_tracker_log WHERE session_id = ?" 
 * </li>
 * <li>utcOffset - (optional) A value in the format "+hh:mm" that expresses the
 * UTC offset for the Cams user session value. hh must be between 0 and 13 and mm
 * must be "00" or "30". If supplied, the session created time is written to the 
 * database normalized to UTC.</li>
 * <li>excludeUsers - (optional) A comma-delimited list of one or more user's to 
 * exclude. This option is particularly useful when the session event handler
 * is used in the system security domain to exclude Cams web agent logins.</li>
 * </ul>
 * 
 * <p>A sample configuration looks like:
 * 
 * <pre>
 * <session-event-handler
 *  className="examples.session.JdbcSessionTrackerManagedSessionEventHandler"
 *  debug="false">
 *  <param-list>
 *    <param name="driver"           value="org.apache.commons.dbcp.PoolingDriver"/>
 *    <param name="url"              value="jdbc:apache:commons:dbcp:camsusers-pool"/>
 *    <param name="user"             value=""/>
 *    <param name="password"         value=""/>
 *    <param name="purgeSessionsSQL" value="DELETE FROM session_tracker_log WHERE session_id like 'MyCamsCluster-MyCamsServer-system%'"/>
 *    <param name="createSessionSQL" value="INSERT INTO session_tracker_log (session_id, security_domain, user_name, created) VALUES (?,?,?,?)"/>
 *    <param name="deleteSessionSQL" value="DELETE FROM session_tracker_log WHERE session_id = ?"/>
 *    <param name="excludeUsers"     value="cams-web-agent"/>
 *  </param-list>
 * </session-event-handler>
 * </pre>
 *
 * @version $Revision: 1.3 $ $Date: 2011/08/12 21:15:04 $
 */
public class JdbcSessionTrackerManagedSessionEventHandler
	implements ManagedSessionEventHandler
{
	/** 
	 * The Config object provided at initialization 
	 */
	private Config config;

	/** 
	 * The flag that enables/disables DEBUG-level messages.  
	 */
	private boolean debug = false;

	/** 
	 * Logs messages.  
	 */
	private Logger logger;

	/** 
	 * The loaded JDBC Driver. 
	 */
	private Driver driver;

	/** 
	 * The JDBC URL. 
	 */
	private String url;

	/** 
	 * The loaded JDBC username/password Connection Properties. 
	 */
	private Properties jdbcProp;

	/** 
	 * The loaded JDBC username/password Connection Properties. 
	 */
	private boolean purgeSessions = true;

	/** 
	 * The utcOffset expressed in milliseconds
	 */
	private int utcOffset = 0;

	/**
	 * Initialize the ManagedSessionEventHandler.
	 *
	 * @param config a Config object that provides access to configuration
	 * 	parameters and a runtime Context.
	 */
	public void initialize(Config config) throws ConfigException
	{
		this.config = config;
		this.logger = config.getContext().getLogger();

		// Check for required configuration parameters
		String s = config.getInitParameter("driver");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'driver'");

		this.url = config.getInitParameter("url");
		if ((this.url == null) || "".equals(this.url.trim()))
			throw new ConfigException("Missing required parameter 'url'");

		s = config.getInitParameter("purgeSessionsSQL");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'purgeSessionsSQL'");

		s = config.getInitParameter("createSessionSQL");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'createSessionSQL'");

		s = config.getInitParameter("deleteSessionSQL");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'deleteSessionSQL'");

		// Make sure the JDBC Driver loads
		try
		{
			this.driver = 
				(Driver)Class.forName(config.getInitParameter("driver")).newInstance();
		}
		catch (Exception e)
		{
			throw new ConfigException(
				"Error loading JDBC driver: " + driver, e);
		}

		// Create the JDBC Connection Properties (user/password)
		// so we don't have to do it for every Session created
		this.jdbcProp = new Properties();
		jdbcProp.put("user", config.getInitParameter("user"));
		jdbcProp.put("password", config.getInitParameter("password"));

		// If supplied, validate the format of the GMT offset
		s = config.getInitParameter("utcOffset");
		if ((s != null) && s.length() > 0)
		{
			if (!((s.startsWith("+")) || (s.startsWith("-"))))
			{
				throw new ConfigException("Optional parameter 'utcOffset' must start with '+' or '-'");
			}

			if (s.length() != 6)
			{
				throw new ConfigException("Optional parameter 'utcOffset' must have 5 characters in format '+hhmm'");
			}

			int hours = 0;
			try
			{
				hours = Integer.parseInt(s.substring(2,3));

				if (hours > 13)
					throw new ConfigException("Optional parameter 'utcOffset' hour value must be in range '00' to '13'");
			}
			catch (NumberFormatException e)
			{
				throw new ConfigException("Optional parameter 'utcOffset' hour value must be '00' to '13'", e);
			}

			int minutes = 0;
			try
			{
				minutes = Integer.parseInt(s.substring(5,6));

				if (!((minutes == 00) || (minutes == 30)))
					throw new ConfigException("Optional parameter 'utcOffset' minute value must be either '00' or '30'");
			}
			catch (NumberFormatException e)
			{
				throw new ConfigException("Optional parameter 'utcOffset' minute value must be either '00' or '30'", e);
			}

			if (debug)
			{
				logger.debug(this, "utcOffset config value: " + s);
				logger.debug(this, "utcOffset hours: " + hours);
				logger.debug(this, "utcOffset minutes: " + minutes);
			}

			// Convert GMT offset to milliseconds
			utcOffset = ((hours * 60) + minutes) * 60 * 1000;
			if (s.startsWith("+"))
				utcOffset = -utcOffset;

			if (debug)
				logger.debug(this, "utcOffset in milliseconds: -" + utcOffset);
		}
	}

	/**
	 * Enable or disable debugging.
	 *
	 * @param enable if true, enable debugging, else disable debugging.
	 * @return the previous value of the debug flag, which is useful for
	 *	restoring the debug state if temporarily changed.
	 */
	public boolean setDebug(boolean enable)
	{
		boolean prevDebug = this.debug;
		this.debug = enable;
		return prevDebug;
	}

	/**
	 * Handle a ManagedSessionEvent.
	 *
	 * @param event the ManagedSessionEvent to be handled.
	 */
	public void handleManagedSessionEvent(ManagedSessionEvent event)
	{
		if(debug)
		{
			logger.debug(this, "Handling ManagedSessionEvent: class=" +
						getClass().getName());
		}

		// Only purge once per policy server invocation on the first
		// pass through this code.
		if(purgeSessions)
		{
			if(debug)
				logger.debug(this, "Purging all sessions from database for this Cams policy server");

			// Create a new thread to perform database purge.
			JdbcSessionTrackerRunnable runnable = 
				new JdbcSessionTrackerRunnable(driver, url, jdbcProp, 
				"PURGE", null, null, null, logger);
			Thread t = new Thread(runnable);
			t.start();

			if (debug)
				logger.debug(this, "Invoked thread to perform database purge asynchronously");

			// Only purge on initial request after policy server startup
			purgeSessions = false;
		}

		// Get session information 
		ManagedSession session = event.getManagedSession();
		String userName = SubjectUtils.getUserName(session.getSubject());
		SessionId sid = session.getId();

		// If a user is excluded, ignore!
		String excludeUsers = config.getInitParameter("excludeUsers");
		if ((excludeUsers != null) && (excludeUsers.length() > 0))
		{
			StringTokenizer st = new StringTokenizer(excludeUsers, ",");

			while (st.hasMoreTokens())
			{
				if (st.nextToken().equals(userName))
				{
					if (debug)
						logger.debug(this, "Ignoring user '" + userName + "'");

					return;
				}
			}
		}

		// Handle session CREATED, EXPIRED and CLOSED events. EXPIRED and
		// CLOSED events are handled the same from a database deletion
		// perspective, but let's make it clear in debug which we're 
		// responding to by using seperate "else if".

		if (event.getEventType() == ManagedSessionEventType.SESSION_CREATED)
		{
			if(debug)
				logger.debug(this, "Handling SESSION_CREATED event for sessionId: " + 
					sid.toString());

			// Create a new thread to perform database operation.
			JdbcSessionTrackerRunnable runnable = 
				new JdbcSessionTrackerRunnable(driver, url, jdbcProp, 
				"SESSION_CREATED", sid, session, userName, logger);
			Thread t = new Thread(runnable);
			t.start();

			if (debug)
				logger.debug(this, "Invoked thread to perform database operation asynchronously");
		}
		else if (event.getEventType() == ManagedSessionEventType.SESSION_CLOSED)
		{
			if(debug)
				logger.debug(this, "Handling SESSION_CLOSED event for sessionId: " + 
					sid.toString());

			// Create a new thread to perform database operation.
			JdbcSessionTrackerRunnable runnable = 
				new JdbcSessionTrackerRunnable(driver, url, jdbcProp, 
				"SESSION_CLOSED", sid, session, userName, logger);
			Thread t = new Thread(runnable);
			t.start();

			if (debug)
				logger.debug(this, "Invoked thread to perform database operation asynchronously");
		}
		else if (event.getEventType() == ManagedSessionEventType.SESSION_EXPIRED)
		{
			if(debug)
				logger.debug(this, "Handling SESSION_EXPIRED event for sessionId: " + 
					sid.toString());

			// Create a new thread to perform database operation.
			JdbcSessionTrackerRunnable runnable = 
				new JdbcSessionTrackerRunnable(driver, url, jdbcProp, 
				"SESSION_EXPIRED", sid, session, userName, logger);
			Thread t = new Thread(runnable);
			t.start();

			if (debug)
				logger.debug(this, "Invoked thread to perform database operation asynchronously");
		}

		if(debug)
		{
			logger.debug(this, "Handled ManagedSessionEvent: class=" +
				this.getClass().getName());
		}
	}

	/**
	 * <code>JdbcLoggerRunnable</code> implements a Runnable object that
	 * logs a message to a database.
	 */
	private final class JdbcSessionTrackerRunnable implements Runnable
	{
		private Driver driver;
		private String url;
		private Properties jdbcProp;
		private String action;
		private SessionId sid;
		private ManagedSession session;
		private String userName;
		private Logger logger;

		/**
		 * Create a new JdbcSessionTrackerRunnable.
		 *
		 * @param driver the JDBC driver to use.
		 * @param url the JDBC URL to use.
		 * @param jdbcProp the JDBC user name and password to use.
		 * @param action the SQL action to perform. 
		 * @param sid the user's Cams session ID. 
		 * @param session the user's Cams session. 
		 * @param logger the Logger for logging error messages.
		 */
		public JdbcSessionTrackerRunnable(Driver driver, String url, Properties jdbcProp,
			String action, SessionId sid, ManagedSession session, String userName,
			Logger logger)
		{
			this.driver = driver;
			this.url = url;
			this.jdbcProp = jdbcProp;
			this.action = action;
			this.sid = sid;
			this.session = session;
			this.userName = userName;
			this.logger = logger;
		}

		//
		// implementing Runnable
		//

		/**
		 * Log the authentication event.
		 */
		public void run()
		{
			Connection c = null;
			PreparedStatement s = null;

			try
			{
				// Connect to the database
				c = this.driver.connect(config.getInitParameter("url"), this.jdbcProp);

				if(debug)
				{
					logger.debug(this, 
					"Creating prepared statement and executing update: " + action);
				}

				// If the session was just created, then add an attribute.
				if (action.equals("SESSION_CREATED"))
				{
					s = c.prepareStatement(config.getInitParameter("createSessionSQL"));
					s.setString(1, sid.toString());
					s.setString(2, sid.getSecurityDomainName());
					s.setString(3, userName);
					s.setTimestamp(4, new java.sql.Timestamp((session.getCreationTime() + utcOffset)));
				}
				else if (action.equals("SESSION_CLOSED"))
				{
					s = c.prepareStatement(config.getInitParameter("deleteSessionSQL"));
					s.setString(1, sid.toString());
				}
				else if (action.equals("SESSION_EXPIRED"))
				{
					s = c.prepareStatement(config.getInitParameter("deleteSessionSQL"));
					s.setString(1, sid.toString());
				}
				else if (action.equals("PURGE"))
				{
					s = c.prepareStatement(config.getInitParameter("purgeSessionsSQL"));
				}

				s.execute();

				if(debug)
				{
					logger.debug(this, 
					"Done executing update: " + action);
				}
			}
			catch (SQLException e)
			{
				logger.error(this, "Error performing database operation: " + e);
			}
			finally
			{
				try
				{
					if (s != null)
						s.close();

					if (c != null)
						c.close();
				}
				catch (SQLException e)
				{
					logger.error(this, "Error closing database connection " + e);
				}
			}
		}
	}

} // End of class: JdbcSessionTrackerManagedSessionEventHandler
