/**
 * Copyright (c) 1996-2011 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose
 * such Confidential Information and shall use it only in accordance
 * with the terms of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 * OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
 * BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
 * OR ITS DERIVATIVES.
 */
package examples.session;

import com.cafesoft.cams.Config;
import com.cafesoft.cams.ConfigException;

import com.cafesoft.cams.session.ManagedSession;
import com.cafesoft.cams.session.ManagedSessionEvent;
import com.cafesoft.cams.session.ManagedSessionEventType;
import com.cafesoft.cams.session.ManagedSessionEventHandler;

import com.cafesoft.cams.auth.SubjectUtils;

import com.cafesoft.core.log.Logger;

import com.cafesoft.core.ldap.LdapUtils;

import com.novell.ldap.LDAPAttribute;
import com.novell.ldap.LDAPAttributeSet;
import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;
import com.novell.ldap.LDAPSearchResults;
import com.novell.ldap.LDAPJSSESecureSocketFactory;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.StringTokenizer;

/**
 * A general purpose handler for <code>ManagedSessionEvents</code>
 * that queries a configurable, user-specific attribute from a LDAP repository
 * and inserts the corresponding attribute into a Cams ManagedSession.
 * Configuration parameters are:
 *
 * <ul>
 * <li><b>connectionHost</b> - The DNS name or IP address of the LDAP server</li>
 * <li><b>connectionPort</b> - The port to use (default = 389)</li>
 * <li><b>ldapVersion</b> - (optional, default is 3) The LDAP server protocol 
 * version</li>
 * <li><b>connectionTimeout</b> - (optional default is 3000) The number of 
 * milliseconds to wait for a connectionto the LDAP server before timing out</li>
 * <li><b>useSSL</b> - (optional, default is false) Enables/disables use of 
 * SSL [true|false]</li>
 * <li><b>loginDN</b> - The LDAP repository login dn </li>
 * <li><b>loginPassword</b> - The LDAP repository login password</li>
 * <li><b>searchScope</b> - The LDAP search scope (default = 1)</li>
 * <li><b>searchBase</b> - The repository tree search base
 * (e.g. "ou=people,dc=mycompany,dc=com"</li>
 * <li><b>searchPattern</b> - The LDAP search pattern (e.g. "(uid={username})")</li>
 * <li><b>searchAttributes</b> - A comma-delimited list of one or more LDAP attributes
 * to be returned by the search</li>
 * <li><b>nameSpace</b> - The session attribute namespace for storing search results</li>
 * <li><b>excludeUsers</b> - (optional) A comma-delimited list of one or more user's
 * to exclude</li>
 * </ul>
 *
 * @author Cafesoft LLC
 * @version $Revision: 1.19 $ $Date: 2011/01/25 20:04:45 $
 * @author Gary Gwin
 */
public class LdapUserAttributeManagedSessionEventHandler
	implements ManagedSessionEventHandler
{
	/**
	 * The Config object provided at initialization
	 */
	private Config config;

	/**
	 * The flag that enables/disables DEBUG-level messages.
	 */
	private boolean debug = false;

	/**
	 * The LDAP server host.
	 */
	private String connectionHost;

	/**
	 * The LDAP server port.
	 */
	private int connectionPort = 389;

	/**
	 * The LDAP server version.
	 */
	private int ldapVersion = 3;

	/**
	 * The LDAP server connection timeout.
	 */
	private int connectionTimeout = 3000;

	/**
	 * The flag to specify use of SSL LDAP server connetions. 
	 */
	private boolean useSSL = false;

	/**
	 * The LDAP server login DN.
	 */
	private String loginDN;

	/**
	 * The LDAP server login password.
	 */
	private String loginPassword;

	/**
	 * The search base DN.
	 */
	private String searchBase;

	/**
	 * The search filter pattern.
	 */
	private String searchPattern;

	/**
	 * The search scope.
	 */
	private int searchScope = 1;

	/**
	 * The search attributes to return.
	 */
	private String searchAttributes;

	/**
	 * The Cams session namespace to use with the returned values.
	 */
	private String nameSpace;

	/**
	 * Initialize the ManagedSessionEventHandler.
	 *
	 * @param config a Config object that provides access to configuration
	 * 	parameters and a runtime Context.
	 */
	public void initialize(Config config)
		throws ConfigException
	{
		this.config = config;

		Logger logger = config.getContext().getLogger();
		logger.info(this, "Initializing class: '" + this.getClass().getName() + "'");

		// Check for required configuration parameters
		String s = config.getInitParameter("connectionHost");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'connectionHost'");
		else
			this.connectionHost = s;

		s = config.getInitParameter("connectionPort");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'connectionPort'");
		else
		{
			try
			{
				this.connectionPort = Integer.parseInt(s);
			}
		  catch (NumberFormatException e)
		  {
				throw new ConfigException("Parameter 'connectionPort' not an integer");
		  }
		}

		// Optional ldapVersion
		s = config.getInitParameter("ldapVersion");
		if ((s != null) && (s.trim().length() > 0))
		{
		  try
		  {
				this.ldapVersion = Integer.parseInt(s);
		  }
		  catch (NumberFormatException e)
		  {
				throw new ConfigException("Parameter 'ldapVersion' not an integer");
		  }
		}

		// Optional connectionTimeout
		s = config.getInitParameter("connectionTimeout");
		if ((s != null) && (s.trim().length() > 0))
		{
			try
			{
				this.connectionTimeout = Integer.parseInt(s);
			}
			catch (NumberFormatException e)
			{
				throw new ConfigException("Parameter 'connectionTimeout' not an integer");
			}
		}

		// Optional useSSL
		s = config.getInitParameter("useSSL");
		if ((s != null) && (s.trim().length() > 0))
			this.useSSL = "true".equals(s);

		s = config.getInitParameter("loginDN");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'loginDN'");
		else
			this.loginDN = s;

		s = config.getInitParameter("loginPassword");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'loginPassword'");
		else
			this.loginPassword = s;

		s = config.getInitParameter("searchScope");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'searchScope'");
		else
		{
			// Set the search scope, the default of 0 is "BASE"
			if ("BASE".equalsIgnoreCase(s))
				this.searchScope = 0;
			else if ("ONE".equalsIgnoreCase(s))
				this.searchScope = 1;
			else if ("SUB".equalsIgnoreCase(s))
				this.searchScope = 2;
			else
				throw new ConfigException("Required parameter 'searchScope' " +
				"must be 'BASE', 'ONE' or 'SUB'");
		}

		s = config.getInitParameter("searchBase");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'searchBase'");
		else
			this.searchBase = s;

		s = config.getInitParameter("searchPattern");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'searchPattern'");
		else
			this.searchPattern = s;

		s = config.getInitParameter("searchAttributes");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'searchAttributes'");
		else
			this.searchAttributes = s;

		s = config.getInitParameter("nameSpace");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'nameSpace'");
		else
			this.nameSpace = s;
	}

	/**
	 * Enable or disable debugging.
	 *
	 * @param enable if true, enable debugging, else disable debugging.
	 * @return the previous value of the debug flag, which is useful for
	 *	restoring the debug state if temporarily changed.
	 */
	public boolean setDebug(boolean enable)
	{
		boolean prevDebug = this.debug;
		this.debug = enable;
		return prevDebug;
	}

	/**
	 * Handle a ManagedSessionEvent.
	 *
	 * @param event the ManagedSessionEvent to be handled.
	 */
	public void handleManagedSessionEvent(ManagedSessionEvent event)
	{
		Logger logger = config.getContext().getLogger();

		if(debug)
		{
			logger.debug(this, "Handling ManagedSessionEvent: class=" +
						getClass().getName());
		}

		// If the session was just created, then add an attribute.
		if (event.getEventType() == ManagedSessionEventType.SESSION_CREATED)
		{
			// Flag for existance of the search attribute
			boolean noAttr = false;

			// Get the userName
			ManagedSession session = event.getManagedSession();
			String userName = SubjectUtils.getUserName(session.getSubject());

			// If a user is excluded from this event, ignore it!
			String excludeUsers = config.getInitParameter("excludeUsers");
			if ((excludeUsers != null) && (excludeUsers.length() > 0))
			{
				StringTokenizer st = new StringTokenizer(excludeUsers, ",");

				while (st.hasMoreTokens())
				{
					if (st.nextToken().equals(userName))
					{
						if (debug)
							logger.debug(this, "Ignoring user '" + userName + "'");

						return;
					}
				}
			}

			// Initialize the connection
			LDAPConnection connection = null;

			try
			{
				// Connect and bind to the LDAP server
				connection = getConnection(logger);

				// Verify the connection 
				if (connection == null || !connection.isConnected())
					throw new LDAPException("Connection is null or not connected", 91, null);

				connection.bind(ldapVersion, config.getInitParameter("loginDN"),
					config.getInitParameter("loginPassword").getBytes());

				// Set the search pattern to find attribute for the user
				String searchFilter = LdapUtils.replaceUsername
					(config.getInitParameter("searchPattern"),
					"username", userName);

				// Get the list of attributes to return from the search
				StringTokenizer st = new StringTokenizer(
					config.getInitParameter("searchAttributes"),",");

				String[] searchAttributes = new String[st.countTokens()];
				int i = 0;
				while (st.hasMoreTokens())
				{
					searchAttributes[i] = st.nextToken();

					if (debug)
						logger.debug(this, "Return attribute '" +
							searchAttributes[i] + "' with search results");

					i++;
				}

				// Perform the search
				LDAPSearchResults searchResults =
					connection.search(config.getInitParameter("searchBase"),
					searchScope, searchFilter, searchAttributes, false);

				// Get all the attributes for the first entity
				if (searchResults.hasMore())
				{				
					// Get all the attributes for the first entity
					LDAPEntry nextEntry = searchResults.next();
					LDAPAttributeSet attributeSet = nextEntry.getAttributeSet();

					Iterator ldapAttributes = attributeSet.iterator();

					if (debug)
					{
						logger.debug(this, "Entity count from search is '" +
							searchResults.getCount() + "'");

						logger.debug(this, "Attribute set size is '" +
							attributeSet.size() + "'");
					}

					// Loop over all attributes
					while (ldapAttributes.hasNext())
					{
						LDAPAttribute attribute =
							(LDAPAttribute)ldapAttributes.next();

						String attributeName = attribute.getName();

						// Field to store attribute values
						String value = null;

						// An attribute may have multiple values, get them all
						Enumeration allValues = attribute.getStringValues();

						// If a value exists, get the first one only
						if (allValues != null)
							value = (String)allValues.nextElement();
						else
							noAttr = true;

						if (debug)
						{
							logger.debug(this,
								"Search attribute name/value pair is '" +
								attributeName + "'/'" + value + "'");
						}

						// Persist the number into the session object
						session.putAttribute(config.getInitParameter("nameSpace"),
							attributeName, value);

						// An value does not exist for this userName, log an error
						if (noAttr)
						{
							logger.info(this, "Value for search attribute '" +
								attributeName + "' for user'" +
								userName + "' does not exist");
						}
					}
				}
				else if (debug)
				{
					logger.debug(this, "No search results for user '" +
						userName + "'");
				}
			}
			catch (LDAPException e)
			{
				logger.error(this, "An LDAP error occured while searching for " +
					"attribute(s) '" + config.getInitParameter("searchAttributes") +
					"' for user '" + userName + "' " + e);
			}
			finally
			{
				// Disconnect from the LDAP Server
				if (connection != null)
				{
					try
					{
						connection.disconnect();
					}
					catch (LDAPException e)
					{
						logger.error(this, "error closing LDAP connection " + e);
					}
				}
			}
		}

		if(debug)
		{
			logger.debug(this, "Handled ManagedSessionEvent: class=" +
				this.getClass().getName());
		}
	}

	//
	// Private methods
	//

	/**
	 * Create a LDAPConnection.
	 *
	 * @param logger A logger to use
	 * @return a LDAPConnection that is ready for use.
	 * @throws LDAPException is a connection problem occurs
	 */
	private LDAPConnection getConnection(Logger logger)
		throws LDAPException
	{
		if (debug)
			logger.debug(this, "Instantiating LDAPConnection: host=" + 
				this.connectionHost + ", port=" + this.connectionPort + 
				", useSSL=" + this.useSSL + ", connectionTimeout=" +
				this.connectionTimeout);

		// Intialize connection 
		LDAPConnection connection = null;

		try
		{
			if (useSSL)
			{
				// Issue a warning if certain System properties are not set
				String sslProtocolProp =
					System.getProperty("java.protocol.handler.pkgs");
				if (sslProtocolProp == null)
					logger.warning(this,
						"Missing required SSL system property: " +
						"java.protocol.handler.pkgs");

				String trustStoreProp =
					System.getProperty("javax.net.ssl.trustStore");
				if (trustStoreProp == null)
					logger.warning(this,
						"Missing required SSL system property: " +
						"javax.net.ssl.trustStore");

				connection = new LDAPConnection((new LDAPJSSESecureSocketFactory())); 
			}
			else
			{
				connection = new LDAPConnection(); 
			}

			// Set the connection timeout
			connection.setSocketTimeOut(this.connectionTimeout);

			// Debug to help customers determine LDAP response time 
			long timer = System.currentTimeMillis(); 
			
			// Connect to the directory
			connection.connect(this.connectionHost, this.connectionPort);

			if (debug)
			{
				logger.debug(this, "Time to connect: '" +
					(System.currentTimeMillis() - timer) + "'");

				if (connection.isConnected() && connection.isConnectionAlive())
					logger.debug(this, "LDAP Connection succeeded");
				else
					logger.debug(this, "LDAP Connection failed");

				logger.debug(this, "Returning LDAPConnection");
			}
		}
		catch (LDAPException e)
		{
			logger.error(this, "LDAP error", e);

			// Dump the nested Throwables to get at the root cause
			int i = 1;
			Throwable cause = e.getCause();
			do
			{
				if (cause != null)
					logger.error(this, "LDAP error: CAUSE " + i + ": " +
						cause.getMessage());
			}
			while ((cause = cause.getCause()) != null);
		}

		return connection;
	}
} // End of class: LdapUserAttributeManagedSessionEventHandler

