/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package examples.session;

import com.cafesoft.cams.Config;
import com.cafesoft.cams.ConfigException;

import com.cafesoft.cams.session.ManagedSession;
import com.cafesoft.cams.session.ManagedSessionEvent;
import com.cafesoft.cams.session.ManagedSessionEventType;
import com.cafesoft.cams.session.ManagedSessionEventHandler;

import com.cafesoft.core.log.Logger;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;

/**
 * <code>Sample2ManagedSessionEventHandler</code> implements a
 * <code>ManagedSessionEventHandler</code> that is useful for testing
 * the addition of multiple session attribute namespaces and values for
 * every authenticated user in the enclosing security domain. The following
 * security domain component registration shows how multiple attributes
 * may be added to each user's session at session creation time.
 * <pre>
 * <session-event-handler
 *    className="examples.session.Sample2ManagedSessionEventHandler"
 *    debug="true">
 *    <param-list>
 *       <param name="attr-prefix" value="xyz"/>
 *       <param name="xyz:customer:first_name" value="Howard"/>
 *       <param name="xyz:customer:last_name" value="Hughes"/>
 *       <param name="xyz:customer:middle_name" value="Marvin"/>
 *       <param name="xyz:address:city" value="San Diego"/>
 *       <param name="xyz:address:state" value="CA"/>
 *       <param name="xyz:address:country" value="US"/>
 *       <param name="xyz:account:type" value="Checking"/>
 *    </param-list>
 * </session-event-handler>
 * </pre>
 * The attr-prefix value is used by this component to distinguish configuration
 * parameters used to populated a users session from non-session attributes,
 * including those inherited from the enclosing security domain. The example
 * above shows how to configure hard-coded session data for three different
 * attribute namespaces: customer, address, and account.
 *
 * @version $Revision: 1.5 $ $Date: 2009/08/14 23:52:25 $
 * @author Norbert K. Kuhnert
 */
public class Sample2ManagedSessionEventHandler
	implements ManagedSessionEventHandler
{
	/** The Config object provided at initialization */
	private Config config;

	/** The flag that enables/disables DEBUG-level messages. */
	private boolean debug = false;

	/** A parsed array of AttributeTriplet instances. */
	private AttributeTriplet[] attrTripletArray = null;

	/**
	 * Initialize the ManagedSessionEventHandler.
	 *
	 * @param config a Config object that provides access to configuration
	 * 	parameters and a runtime Context.
	 */
	public void initialize(Config config)
		throws ConfigException
	{
		this.config = config;

		// Just dump the Config/Context to the supplied Logger
		Logger logger = config.getContext().getLogger();
		logger.info(this, "... initializing: class=" +
			this.getClass().getName());

		// Get the attribute parameter prefix used to distinguish
		// session attributes from inherited configuration parameters.
		String attrPrefix = config.getInitParameter("attr-prefix");
		if (attrPrefix == null)
	  		throw new ConfigException("Missing required parameter: attr-prefix");

		attrPrefix = attrPrefix + ":";
		int nsStartIndex = attrPrefix.length();
		List attrTripletList = new ArrayList();
		Iterator i = config.getInitParameterNames();
		if (i.hasNext())
		{
			while (i.hasNext())
			{
				String name = (String)i.next();
				if (! name.startsWith(attrPrefix))
					continue;

				String attrValue = config.getInitParameter(name);

				logger.info(this, "   ... param: " +
					name + "=" + attrValue);

				// If a colon exists in the name, then split the fully-
				// qualified attribute name from the prefix.
				String fqAttrName = name.substring(nsStartIndex);
				String attrNamespace = null;
				String attrName = null;
				int index = fqAttrName.indexOf(':');
				if (index > 0)
				{
					attrNamespace = fqAttrName.substring(0, index);
					attrName = fqAttrName.substring(index+1);
				}
				else
				{
					attrNamespace = null; // no colon implies default namespace
					attrName = fqAttrName;
				}

				attrTripletList.add(new AttributeTriplet(
					attrNamespace, attrName, attrValue));
			}
		}
		else
		{
			logger.info(this, "   ... no initialization params");
		}

		this.attrTripletArray = (AttributeTriplet[])attrTripletList.toArray(
			new AttributeTriplet[0]);

		logger.info(this, "... initialized: class=" +
			this.getClass().getName());
	}

	/**
	 * Enable or disable debugging.
	 *
	 * @param enable if true, enable debugging, else disable debugging.
	 * @return the previous value of the debug flag, which is useful for
	 *	restoring the debug state if temporarily changed.
	 */
	public boolean setDebug(boolean enable)
	{
		boolean prevDebug = this.debug;
		this.debug = enable;
		return prevDebug;
	}

	/**
	 * Handle a ManagedSessionEvent.
	 *
	 * @param event the ManagedSessionEvent to be handled.
	 */
	public void handleManagedSessionEvent(ManagedSessionEvent event)
	{
		Logger logger = config.getContext().getLogger();

		if(debug)
		{
			logger.debug(this, "... handling ManagedSessionEvent: class=" +
				this.getClass().getName());
		}

		ManagedSession session = event.getManagedSession();
		logger.info(this, "   ... sessionId=" + session.getId());
		logger.info(this, "   ... event type=" + event.getEventType().toString());

		// If the session was just created, then add all attributes.
		if (event.getEventType() == ManagedSessionEventType.SESSION_CREATED)
		{
			for (int i = 0; i < attrTripletArray.length; i++)
			{
				AttributeTriplet a = attrTripletArray[i];
				session.putAttribute(a.getName(), a.getNamespace(), a.getValue());

				if (debug)
				{
					if (a.getNamespace() != null)
		 			{
						logger.debug(this, "... added attribute: " +
							a.getNamespace() + ":" + a.getName() + "=" +
							a.getValue());
		 			}
					else
		 			{
						logger.debug(this, "... added attribute: " +
							a.getName() + "=" + a.getValue());
		 			}
				}
			}
		}

		if(debug)
		{
			logger.debug(this, "... handled ManagedSessionEvent: class=" +
				this.getClass().getName());
		}
	}

	//
	// inner classes
	//

	/**
	 * A simple data object for a session attribute including
	 * the namespace, attribute name, and attribute value.
	 */
	public class AttributeTriplet
	{
		private String namespace;
		private String name;
		private String value;

		public AttributeTriplet(String namespace, String name, String value)
		{
			this.namespace = namespace;
			this.name = name;
			this.value = value;
		}

		public String getNamespace()
		{
			return namespace;
		}

		public String getName()
		{
			return name;
		}

		public String getValue()
		{
			return value;
		}
	}

} // End of class: Sample2ManagedSessionEventHandler


