/**
 * Copyright (c) 1996-2009 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package examples.session;

import com.cafesoft.cams.Config;
import com.cafesoft.cams.ConfigException;

import com.cafesoft.cams.session.ManagedSession;
import com.cafesoft.cams.session.ManagedSessionEvent;
import com.cafesoft.cams.session.ManagedSessionEventType;
import com.cafesoft.cams.session.ManagedSessionEventHandler;

import com.cafesoft.core.log.Logger;

import java.util.Iterator;

/**
 * <code>SampleManagedSessionEventHandler</code> demonstrates a basic handler
 * for <code>ManagedSessionEvent</code>s.
 *
 * @version $Revision: 1.11 $ $Date: 2009/08/14 23:52:25 $
 * @author Norbert K. Kuhnert
 */
public class SampleManagedSessionEventHandler
	implements ManagedSessionEventHandler
{
	/** The Config object provided at initialization */
	private Config config;

	/** The flag that enables/disables DEBUG-level messages. */
	private boolean debug = false;

	/**
	 * Initialize the ManagedSessionEventHandler.
	 *
	 * @param config a Config object that provides access to configuration
	 * 	parameters and a runtime Context.
	 */
	public void initialize(Config config)
		throws ConfigException
	{
		this.config = config;

		// Just dump the Config/Context to the supplied Logger
		Logger logger = config.getContext().getLogger();
		logger.info(this, "... initializing: class=" +
			this.getClass().getName());

		Iterator i = config.getInitParameterNames();
		if (i.hasNext())
		{
			while (i.hasNext())
			{
				String name = (String)i.next();
				logger.info(this, "   ... param: " +
					name + "=" +
					config.getInitParameter(name));
			}
		}
		else
		{
			logger.info(this, "   ... no initialization params");
		}

		logger.info(this, "... initialized: class=" +
			this.getClass().getName());
	}

	/**
	 * Enable or disable debugging.
	 *
	 * @param enable if true, enable debugging, else disable debugging.
	 * @return the previous value of the debug flag, which is useful for
	 *	restoring the debug state if temporarily changed.
	 */
	public boolean setDebug(boolean enable)
	{
		boolean prevDebug = this.debug;
		this.debug = enable;
		return prevDebug;
	}

	/**
	 * Handle a ManagedSessionEvent.
	 *
	 * @param event the ManagedSessionEvent to be handled.
	 */
	public void handleManagedSessionEvent(ManagedSessionEvent event)
	{
		Logger logger = config.getContext().getLogger();
		
		if(debug)
		{
			logger.debug(this, "... handling ManagedSessionEvent: class=" +
				this.getClass().getName());
		}

		ManagedSession session = event.getManagedSession();
		logger.info(this, "   ... sessionId=" + session.getId());
		logger.info(this, "   ... event type=" + event.getEventType().toString());

		// If the session was just created, then add an attribute.
		String namespace = "example-namespace";
		if (event.getEventType() == ManagedSessionEventType.SESSION_CREATED)
		{
			session.putAttribute("name1", namespace, "value1");
			session.putAttribute("name2", namespace, "value2");
		}

		// Dump the name/value pairs in the "example-namespace"
		String[] attrName = session.getAttributeNames(namespace);
		for (int i = 0; i < attrName.length; i++)
		{
			logger.info(this, "   ... Attribute: " +
				namespace + ": " +
				attrName[i] + "=" +
				session.getAttribute(attrName[i], namespace));
		}

		if(debug)
		{
			logger.debug(this, "... handled ManagedSessionEvent: class=" +
				this.getClass().getName());
		}
	}

} // End of class: SampleManagedSessionEventHandler


