/**
 * Copyright (c) 1996-2012 Cafesoft, LLC. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cafesoft, LLC. ("Confidential Information"). You shall not disclose
 * such Confidential Information and shall use it only in accordance
 * with the terms of the license agreement you entered into with Cafesoft.
 *
 * CAFESOFT MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 * OR NON-INFRINGEMENT. CAFESOFT SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED
 * BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
 * OR ITS DERIVATIVES.
 */
package examples.session;

import com.cafesoft.cams.auth.CSRolePrincipal;
import com.cafesoft.cams.auth.SubjectUtils;

import com.cafesoft.cams.Config;
import com.cafesoft.cams.ConfigException;

import com.cafesoft.cams.session.ManagedSession;
import com.cafesoft.cams.session.ManagedSessionEvent;
import com.cafesoft.cams.session.ManagedSessionEventType;
import com.cafesoft.cams.session.ManagedSessionEventHandler;

import com.cafesoft.security.engine.service.StandardUnboundIDLdapConnectionPoolService;

import com.cafesoft.core.log.Logger;

import com.cafesoft.core.ldap.LdapUtils;

import com.cafesoft.core.lifecycle.LifecycleException;

import com.cafesoft.core.service.ServiceException;

import com.unboundid.ldap.sdk.Attribute;
import com.unboundid.ldap.sdk.Entry;
import com.unboundid.ldap.sdk.FailoverServerSet;
import com.unboundid.ldap.sdk.RoundRobinServerSet;
import com.unboundid.ldap.sdk.SingleServerSet;
import com.unboundid.ldap.sdk.LDAPConnection;
import com.unboundid.ldap.sdk.LDAPConnectionOptions;
import com.unboundid.ldap.sdk.LDAPException;
import com.unboundid.ldap.sdk.LDAPSearchException;
import com.unboundid.ldap.sdk.ResultCode;
import com.unboundid.ldap.sdk.SearchResult;
import com.unboundid.ldap.sdk.SearchScope;
import com.unboundid.util.ssl.SSLUtil;
import com.unboundid.util.ssl.TrustAllTrustManager;

import javax.net.ssl.SSLSocketFactory;

import java.security.GeneralSecurityException;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

/**
 * A general purpose handler for <code>ManagedSessionEvents</code>
 * that queries a configurable, user-specific attribute from a LDAP repository
 * and inserts the corresponding attribute into a Cams ManagedSession using the
 * UnboundID LDAP SDK:
 * 
 *       http://www.unboundid.com/products/ldapsdk/
 * 
 * Configuration parameters are:
 *
 * <ul>
* <li><b>ldapPoolServiceId</b> - (optional) The service ID configured in 
 * security-domain.xml for an UnboundID LDAP connection pooling service. When 
 * configured, the connection pooling service uses its LDAP connection parameters 
 * and any configured for this login module are ignored.
 * <li><b>connectionHost</b> - (conditional) A comma-delimited list of DNS names 
 * or IP addresses for connection LDAP servers. See <strong>serverSetType</strong> 
 * for configuration options to use more than one host.</li>
 * <li><b>connectionPort</b> - (conditional) A comma-delimited list of one or 
 * more ports for LDAP server connections. Specify a port of the same type (LDAP 
 * or LDAPS) for each host in the <strong>connectionHost</strong> list. Default 
 * values are 389 for LDAP and 636 for LDAPS.</li> 
 * <li><b>serverSetType</b> - (optional with default of <strong>SINGLE</strong>) 
 * Can be: 
 * <b>SINGLE</b> - Connects to a single server. It may be used in cases where a 
 * ServerSet is required but only a single server is needed. 
 * <b>FAILOVER</b> - Attempts to establish connections to servers in the order 
 * they are provided. If the first server is unavailable, then it will attempt 
 * to connect to the second, then to the third, etc. 
 * <b>ROUND_ROBIN</b> - Uses a round-robin algorithm to select the server to 
 * which the connection should be established. Any number of servers may be 
 * included in this server set, and each request will attempt to retrieve a 
 * connection to the next server in the list, circling back to the beginning of 
 * the list as necessary. If a server is unavailable when an attempt is made to 
 * establish a connection to it, then the connection will be established to the 
 * next available server in the set. 
 * 
 * WARNING: Except for the connectionHost and connectionPort parameters, all
 * other parameters for this login module are shared and must be the same for
 * each server when more than one are configured in a ServerSet. 
 * </li>
 * <li><b>connectionTimeout</b> - (optional with default of 3000) The number of 
 * milliseconds to wait for a connectionto the LDAP server before failing</li>
 * <li><b>responseTimeout</b> - (optional with default of 3000) The number of 
 * milliseconds to wait for a response from the server before failing
 * <li><b>loginDN</b> - (optional) The LDAP login dn. Only supply when the client
 * component requires bound connections.</li>
 * <li><b>loginPassword</b> - (optional) The LDAP login password.</li>
 * <li><b>useSSL</b> - (optional, default is false) Enables/disables use of 
 * SSL [true|false]</li>
 * <li><b>baseDN</b> - (required) The base DN to search in the repository 
 * (e.g. "ou=people,dc=mycompany,dc=com"</li>
 * <li><b>filter</b> - (required) The LDAP search pattern (e.g. "(uid={username})")</li>
 * <li><b>scope</b> - (optional) The LDAP search scope (default=ONE)</li>
 * <li><b>attributes</b> - A comma-delimited list of one or more LDAP attributes
 * to be returned by the search</li>
 * <li><b>nameSpace</b> - (required) The session attribute namespace for storing search results</li>
 * <li><b>excludeUsers</b> - (optional) A comma-delimited list of one or more user's
 * to exclude</li>
 * <li><b>includeUsersInRoles</b> - (optional) A comma-delimited list of one or more
 * user roles that, if supplied, a user must have at least one for this session 
 * event handler to attempt the search for attributes.</li>
 * </ul>
 *
 * @version $Revision: 1.6 $ $Date: 2012/05/15 19:07:07 $
 * @author Cafesoft LLC
 */
public class UnboundIDLdapUserAttributeManagedSessionEventHandler
	implements ManagedSessionEventHandler
{
	//
	// Class variables
	//

	/**
	 * The service name
	 */
	private static final String SEHN = "[UnboundIDLdapUserAttributeSessionEventHandler]";

	//
	// Instance variables
	//

	/**
	 * The Config object provided at initialization
	 */
	private Config config;

	/**
	 * The flag that enables/disables DEBUG-level messages.
	 */
	private boolean debug = false;

	/**
	 * The LDAP server host.
	 */
	private String[] connectionHost;

	/**
	 * The LDAP server port.
	 */
	private int[] connectionPort;

	/**
	 * The LDAP server connection timeout.
	 */
	private int connectionTimeout = 3000;

	/**
	 * The LDAP server response timeout.
	 */
	private long responseTimeout = 3000;

	/**
	 * The flag to specify use of SSL LDAP server connetions. 
	 */
	private boolean useSSL = false;

	/**
	 * The LDAP server login DN.
	 */
	private String loginDN;

	/**
	 * The LDAP server login password.
	 */
	private String loginPassword;

	/**
	 * The search base DN.
	 */
	private String baseDN;

	/**
	 * The search filter pattern.
	 */
	private String filter;

	/**
	 * The search scope.
	 */
	private SearchScope scope = SearchScope.ONE;

	/**
	 * The search attributes to return.
	 */
	private String[] attributes;

	/**
	 * The Cams session namespace to use with the returned values.
	 */
	private String nameSpace;

	/**
	 * Connection pool.
	 */
	private StandardUnboundIDLdapConnectionPoolService ulcps;

	/**
	 * Initialize the ManagedSessionEventHandler.
	 *
	 * @param config a Config object that provides access to configuration
	 * 	parameters and a runtime Context.
	 */
	public void initialize(Config config)
		throws ConfigException
	{
		this.config = config;

		Logger logger = config.getContext().getLogger();
		logger.info(this, "Initializing class: '" + this.getClass().getName() + "'");

		// Warn if any configuration parameters are not in known universe 
		// for this component
		String[] validParamNames = {
			"ldapPoolServiceId",
			"connectionHost",
			"connectionPort",
			"serverSetType",
			"connectionTimeout",
			"responseTimeout",
			"loginDN",
			"loginPassword",
			"useSSL",
			"baseDN", 
			"filter",
			"scope",
			"attributes",
			"nameSpace",
			"excludeUsers",
			"includeUsersInRoles"};
		Iterator paramNames = config.getInitParameterNames();
		String s = null;
		int count = 0;
		while(paramNames.hasNext())
		{
			s = (String)paramNames.next();

			// Skip check for context supplied parameters 
			if ((!s.startsWith("cams.")) && (!"logPrefix".equals(s)))
			{
				count = 0;
				while (count < validParamNames.length)
				{
					// Parameter is in known universe
					if (s.equals(validParamNames[count]))
						break;

					count++;

					// Parameter is not in known universe
					if (count == validParamNames.length)
						logger.warning(this, SEHN + 
							" Parameter name is unknown: " + s);
				}
			}
		}

		// Validate options
		s = config.getInitParameter("ldapPoolServiceId");
		if ((s != null) && (s.length() > 0))
		{
			// Get the LDAP connection pool to use for LDAP operations
			try
			{
				this.ulcps = (StandardUnboundIDLdapConnectionPoolService)
					config.getContext().getServiceFinder().find(s, 
					StandardUnboundIDLdapConnectionPoolService.class);

				if (this.debug)
					logger.debug(this, SEHN + " Connection pooling service initialized");
			}
			catch (ServiceException e)
			{
				throw new ConfigException(SEHN + " Unable to load connection pooling service" + e);
			}
		}
		else
		{
			// Validate and load connection params to create a new 
			// connection for each  LDAP operation
			s = config.getInitParameter("connectionHost");
			if ((s == null) || "".equals(s.trim()))
				throw new ConfigException("Missing required parameter 'connectionHost'");
			else
				this.connectionHost = getStringArray(s, ",");

			s = config.getInitParameter("connectionPort");
			if ((s == null) || "".equals(s.trim()))
				throw new ConfigException("Missing required parameter 'connectionPort'");
			else
				this.connectionPort = getIntArray(s, ",");

			// Optional connectionTimeout
			s = config.getInitParameter("connectionTimeout");
			if ((s != null) && (s.trim().length() > 0))
			{
				try
				{
					this.connectionTimeout = Integer.parseInt(s);
				}
				catch (NumberFormatException e)
				{
					throw new ConfigException("Parameter 'connectionTimeout' not an integer");
				}
			}

			// Optional responseTimeout
			s = config.getInitParameter("responseTimeout");
			if ((s != null) && (s.trim().length() > 0))
			{
				try
				{
					this.responseTimeout = (long)Integer.parseInt(s);
				}
				catch (NumberFormatException e)
				{
					throw new ConfigException("Parameter 'responseTimeout' not an integer");
				}
			}

			s = config.getInitParameter("loginDN");
			if ((s == null) || "".equals(s.trim()))
				throw new ConfigException("Missing required parameter 'loginDN'");
			else
				this.loginDN = s;

			s = config.getInitParameter("loginPassword");
			if ((s == null) || "".equals(s.trim()))
				throw new ConfigException("Missing required parameter 'loginPassword'");
			else
				this.loginPassword = s;

			// Optional useSSL flag
			s = config.getInitParameter("useSSL");
			if ((s != null) && (s.trim().length() > 0))
				this.useSSL = "true".equals(s);
		}
	
		// Load and validate search parameters
		s = config.getInitParameter("scope");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'scope'");
		else
		{
			// Set the search scope, the default of 0 is "BASE"
			if ("BASE".equalsIgnoreCase(s))
				this.scope = SearchScope.BASE;
			else if ("ONE".equalsIgnoreCase(s))
				this.scope = SearchScope.ONE;
			else if ("SUB".equalsIgnoreCase(s))
				this.scope = SearchScope.SUB;
			else
				throw new ConfigException("Required parameter 'scope' " +
				"must be 'BASE', 'ONE' or 'SUB'");
		}

		s = config.getInitParameter("baseDN");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'baseDN'");
		else
			this.baseDN = s;

		s = config.getInitParameter("filter");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'filter'");
		else
			this.filter = s;

		s = config.getInitParameter("attributes");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'attributes'");
		else
			this.attributes = getStringArray(s, ",");

		s = config.getInitParameter("nameSpace");
		if ((s == null) || "".equals(s.trim()))
			throw new ConfigException("Missing required parameter 'nameSpace'");
		else
			this.nameSpace = s;
	}

	/**
	 * Enable or disable debugging.
	 *
	 * @param enable if true, enable debugging, else disable debugging.
	 * @return the previous value of the debug flag, which is useful for
	 *	restoring the debug state if temporarily changed.
	 */
	public boolean setDebug(boolean enable)
	{
		boolean prevDebug = this.debug;
		this.debug = enable;
		return prevDebug;
	}

	/**
	 * Handle a ManagedSessionEvent.
	 *
	 * @param event the ManagedSessionEvent to be handled.
	 */
	public void handleManagedSessionEvent(ManagedSessionEvent event)
	{
		Logger logger = config.getContext().getLogger();

		if(this.debug)
			logger.debug(this, SEHN + 
				" Handling ManagedSessionEvent: class=" +
				getClass().getName());

		// If the session was just created, then add an attribute.
		if (event.getEventType() == ManagedSessionEventType.SESSION_CREATED)
		{
			// Get the userName
			ManagedSession session = event.getManagedSession();
			String userName = SubjectUtils.getUserName(session.getSubject());

			// If a user is excluded from this event, ignore it!
			String excludeUsers = config.getInitParameter("excludeUsers");
			if ((excludeUsers != null) && (excludeUsers.length() > 0))
			{
				StringTokenizer st = new StringTokenizer(excludeUsers, ",");

				while (st.hasMoreTokens())
				{
					if (st.nextToken().equals(userName))
					{
						if (this.debug)
							logger.debug(this, SEHN + " Ignoring user '" + userName + "'");

						return;
					}
				}
			}

			// If a user does not have a required role, ignore it!
			String includeUsersInRoles = config.getInitParameter("includeUsersInRoles");
			if ((includeUsersInRoles != null) && (includeUsersInRoles.length() > 0))
			{
				StringTokenizer st = new StringTokenizer(includeUsersInRoles, ",");

				CSRolePrincipal role = null;
				boolean ignoreUser = true;
				while (st.hasMoreTokens())
				{
					role = new CSRolePrincipal(st.nextToken());
					if (session.getSubject().getPrincipals().contains(role))
					{
						if (debug)
							logger.debug(this, SEHN + 
							" User in includeUsersInRole role '" + role.toString() + "'");
						ignoreUser = false;
						break;
					}
				}

				if (ignoreUser)
				{
					if (debug)
						logger.debug(this, SEHN + 
						" User not in includeUsersInRole '" + userName + "'");
	
					return;
				}
			}

			// Initialize the connection
			LDAPConnection connection = null;

			try
			{
				// If connection pool is initialized, try to use it
				if ((this.ulcps != null) && (this.ulcps.isStarted()))
				{
					if (this.debug)
						logger.debug(this, SEHN + " Getting connection from pool");

					connection = getPooledConnection();
				}
				else
				{
					if (this.debug)
						logger.debug(this, SEHN + " Creating connection from parameters");

					connection = getConnection();

					if (this.debug)
						logger.debug(this, SEHN + " Binding connection");

					connection.bind(config.getInitParameter("loginDN"),
						config.getInitParameter("loginPassword"));
				}

				// Verify the connection 
				if (connection == null || !connection.isConnected())
					throw new LDAPException(ResultCode.CONNECT_ERROR, 
						"LDAP connection is null or not connected");

				// Set the search pattern to find attribute for the user
				String searchFilter = LdapUtils.replaceUsername
					(this.filter, "username", userName);

				// Perform the search
				SearchResult searchResult = connection.search(
					this.baseDN, scope, searchFilter, this.attributes);

				if ((searchResult != null) && (searchResult.getEntryCount() > 0))
				{
					// Get all the attributes for the first entity
					List searchResultsList = searchResult.getSearchEntries();
					Iterator searchResultsIterator = searchResultsList.iterator();
					while (searchResultsIterator.hasNext())
					{				
						// Get the next entries attibute
						Entry entry = (Entry)searchResultsIterator.next();
						Collection attrCollection = entry.getAttributes();
						Iterator attrIterator = attrCollection.iterator();

						if (this.debug)
						{
							logger.debug(this, SEHN + " Entity count from search is '" +
								searchResult.getEntryCount() + "'");

							logger.debug(this, SEHN + " Attribute set size is '" +
								attrCollection.size() + "'");
						}
						
						// GEG: Unlike other LDAP servers, the UnboundId Directory 
						// Server does not store or return dn as an attribute. If 
						// specified as an attribute to be returned, handle it.
						// GEG: TODO: Enhance code with parameter to specific directory
						// server
						for (int k = 0; k < this.attributes.length; k++)
						{
							if (this.attributes[k].equals("dn"))
							{
								// Persist the number into the session object
								session.putAttribute(config.getInitParameter("nameSpace"),
									"dn", entry.getDN());
							}
						}

						while (attrIterator.hasNext())
						{
							// Attribute may have more than one name/value pair
							Attribute attr = (Attribute) attrIterator.next();
							String[] values = attr.getValues();
							for (int j = 0; j < values.length; j++)
							{
								String attributeName = attr.getName();

								if (this.debug)
								{
									logger.debug(this, SEHN +
										" Search attribute name/value pair is '" +
										attributeName + "'/'" + values[j] + "'");
								}

								// Persist the number into the session object
								session.putAttribute(config.getInitParameter("nameSpace"),
									attributeName, values[j]);
							}
						}
					}
				}
				else if (this.debug)
				{
					logger.debug(this, SEHN + " No search results for '" +
						userName + "'");
				}
			}
			catch (ServiceException e)
			{
				logger.error(this, SEHN + " A ServiceException occured while searching for " +
					"attribute(s) '" + this.attributes +
					"' for user '" + userName + "' " + e);
			}
			catch (LDAPException e)
			{
				logger.error(this, SEHN + " A LDAP error occured while searching for " +
					"attribute(s) '" + this.attributes +
					"' for user '" + userName + "' " + 
					new LDAPSearchException(e).toString());

			}
			finally
			{
				// Close the connection
				if (connection != null)
				{
					if ((this.ulcps != null) && (this.ulcps.isStarted()))
					{
						// Try to return connection to pool
						try
						{
							ulcps.releaseConnection(connection);
						}
						catch(ServiceException e)
						{
							if (this.debug)
								logger.debug(this, SEHN + " Service error while returning connection to pool" + e);
						}
						catch(LifecycleException e)
						{
							if (this.debug)
								logger.debug(this, SEHN + " Lifecyle error while returning connection to pool" + e);
						}
					}
					else
					{
						// Close connection 
						connection.close();
						connection = null;
					}
				}
			}
		}

		if (this.debug)
		{
			logger.debug(this, SEHN + " Handled ManagedSessionEvent: class=" +
				this.getClass().getName());
		}
	}

	//
	// Private methods
	//

	/**
	 * Create a LDAPConnection.
	 *
	 * @return a LDAPConnection that is ready for use.
	 * @throws LDAPException is a connection problem occurs
	 */
	private LDAPConnection getConnection()
		throws LDAPException
	{
		Logger logger = config.getContext().getLogger();

		if (this.debug)
			logger.debug(this, SEHN + " Instantiating LDAP connection");

		// Initialize the connection
		LDAPConnection connection = null;
		SSLSocketFactory sslSocketFactory = null;

		// Set the connection timeout
		LDAPConnectionOptions connectionOptions = new LDAPConnectionOptions();
		connectionOptions.setConnectTimeoutMillis(this.connectionTimeout);
		connectionOptions.setResponseTimeoutMillis(this.responseTimeout);

		try
		{
			if (useSSL)
			{
				SSLUtil sslUtil = new SSLUtil(new TrustAllTrustManager());
				sslSocketFactory = sslUtil.createSSLSocketFactory();
			}

			long timer = System.currentTimeMillis(); 

			// If not specified, fall through to default (legacy)
			String serverSetType = config.getInitParameter("serverSetType");

			// Use "SINGLE" ServerSet if configuration parameter is not 
			// supplied to support legacy integrations
			if (serverSetType == null)
				serverSetType = "SINGLE";

			if (serverSetType.equalsIgnoreCase("FAILOVER"))
			{
				if (this.debug)
					logger.debug(this, SEHN + 
						" Using 'FAILOVER' ServerSet for LDAP connection"); 

				FailoverServerSet failoverServerSet = 
					new FailoverServerSet(this.connectionHost, 
					this.connectionPort, sslSocketFactory, 
					connectionOptions);

				connection = failoverServerSet.getConnection();
			}
			else if (serverSetType.equalsIgnoreCase("ROUND_ROBIN"))
			{
				if (this.debug)
					logger.debug(this, SEHN + 
						" Using 'ROUND_ROBIN' ServerSet for LDAP connection"); 

				RoundRobinServerSet roundRobinServerSet = 
					new RoundRobinServerSet(this.connectionHost, 
					this.connectionPort, sslSocketFactory, 
					connectionOptions);

				connection = roundRobinServerSet.getConnection();
			}
			else // Fall through is "SINGLE"
			{
				if (this.debug)
					logger.debug(this, SEHN + 
						" Using 'SINGLE' ServerSet for LDAP connection"); 

				SingleServerSet singleServerSet = 
					new SingleServerSet(this.connectionHost[0], 
					this.connectionPort[0], sslSocketFactory,
					connectionOptions);

				connection = singleServerSet.getConnection();
			}

			if (this.debug)
			{
				logger.debug(this, SEHN + " Time to connect: '" +
					(System.currentTimeMillis() - timer) + "'");

				if (connection.isConnected())
				{
					logger.debug(this, SEHN + " LDAP connection succeeded to: '" + 
						connection.getConnectedAddress() + ":" +
						connection.getConnectedPort() + "'");
				}
				else
				{
					logger.debug(this, SEHN + " LDAP connection failed");
				}

				logger.debug(this, SEHN + " Returning LDAP connection");
			}
		}
		catch (GeneralSecurityException e)
		{
			logger.error(this, SEHN + " General security error", e);

			// Dump the nested Throwables to get at the root cause
			int i = 1;
			Throwable cause = e.getCause();
			do
			{
				if (cause != null)
					logger.error(this, SEHN + " LDAP error: CAUSE " + 
						i + ": " + cause.getMessage());
			}
			while ((cause = cause.getCause()) != null);
		}
		catch (LDAPException e)
		{
			logger.error(this, SEHN + " LDAP error", e);

			// Dump the nested Throwables to get at the root cause
			int i = 1;
			Throwable cause = e.getCause();
			do
			{
				if (cause != null)
					logger.error(this, SEHN + " LDAP error: CAUSE " + i + ": " +
						cause.getMessage());
			}
			while ((cause = cause.getCause()) != null);
		}

		return connection;
	}

	/**
	 * Get an LDAP server connection from a connection pool.
	 *
	 * @return a LDAPConnection.
	 * @exception LDAPException if an error occurs trying to get
	 * 		a LDAPConnection.
	 */
	private LDAPConnection getPooledConnection()
		throws ServiceException
	{
		Logger logger = config.getContext().getLogger();

		// Get the connection values to use
		LDAPConnection connection = null;

		try
		{
			connection = ulcps.getConnection();

			if (this.debug)
				logger.debug(this, SEHN +  
				" Got connection from UnboundIDLdapConnectionPoolService: " + 
				connection.getConnectionStatistics());
		}
		catch(ServiceException e)
		{
			logger.error(this, SEHN +  
				" Service error while getting UnboundIDLdapConnectionPoolService", e);
			throw new ServiceException(" Error: " + e.getMessage());
		}
		catch(LifecycleException e)
		{
			logger.error(this, SEHN +  
				" Lifecycle error while getting UnboundIDLdapConnectionPoolService", e);
			throw new ServiceException(" Error: " + e.getMessage());
		}

		return connection;
	}

	/**
	 * Parse a string for String values separated by a single input character. 
	 * For example, "foo,goober,bar" could be parsed by specifying "," as the
	 * token. It would return a String[3] with foo, goober, and bar as the values.
	 *
	 * @param s the string to evaluate
	 * @param token the token that separates the String values
	 *
	 * @return a String array
	 */
	static String[] getStringArray(String s, String token)
	{
		// Create tokenizer and initialize array based on the number of tokens
		StringTokenizer st = new StringTokenizer(s.trim(), token);
		String[] sa = new String[st.countTokens()];

		// Parse the tokens and store them in the array
		int i = 0;
		while(st.hasMoreTokens())
		{
			sa[i] = st.nextToken();
			i++;
		}

		return sa;
	}

	/**
	 * Parse a string for int values separated by a single input character. 
	 * For example, "389,5389,389" could be parsed by specifying "," as the
	 * token. It would return a int[3] with 389, 389 and 389 as the values.
	 *
	 * @param s the string to evaluate
	 * @param token the token that separates the String values
	 *
	 * @throws ConfigException if values in option are not integers
	 * 
	 * @return a int array
	 */
	static int[] getIntArray(String s, String token)
		throws ConfigException
	{
		// Create tokenizer and initialize array based on the number of tokens
		StringTokenizer st = new StringTokenizer(s.trim(), token);
		int[] ia = new int[st.countTokens()];

		// Parse the tokens and store them in the array
		int i = 0;
		while(st.hasMoreTokens())
		{
			try
			{
				ia[i] = Integer.parseInt(st.nextToken());
			}
			catch (NumberFormatException e)
			{
				throw new ConfigException("Value '" + ia[i] + 
					"' is invalid, must be an integer");
			}

			i++;
		}

		return ia;
	}
} // End of class: UnboundIDLdapUserAttributeManagedSessionEventHandler
